/*
 *    Copyright (c) 2005 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: usercopy.c,v 1.10.2.6 2005/09/06 09:06:58 addy Exp $"
/*$Source: /cvs/master/quadrics/elanmod/modsrc/usercopy.c,v $*/

#include <qsnet/kernel.h>
#include <qsnet/autoconf.h>

#include <elan/elanmod.h>
#include <elan/elanmod_linux.h>

#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/highmem.h>
#include <linux/pagemap.h>

/*
 * Access another process' address space copying directly to/from user space (current)
 *
 * Remote is the non-local process memory address, which we access using get_user_pages() and kmap()
 * For the local memory (i.e. owned by current task) we use the standard copy_[to|from]_user interfaces
 *
 * Code based on linux/kernel/ptrace.c
 */

#if defined(NO_MMPUT) || LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 9)
static size_t 
rw_process_vm (struct task_struct *tsk, unsigned long remote, void *local, size_t len, int write)
{
#warning "NO EXPORTED MMPUT - usercopy not possible"

	/* Without an exported mmput() function we cannot make this
	 * safe as the remote process may be torn down during the copy
	 * I experimented with taking a write lock on the remote mmap_sem
	 * but this seemed to lead to deadlocks when pagefaulting
	 */
	/* GNAT 7768: We have also found that some older versions of the get_task_mm() code
	 * in linux/sched.h call mmgrab() which is not exported in any 2.6.X kernel
	 */
	return 0;
}

#else
static size_t 
rw_process_vm (struct task_struct *tsk, unsigned long remote, void *local, size_t len, int write)
{
        struct mm_struct *mm;
        struct vm_area_struct *vma;
        struct page *page;
        void *old_buf = local;

	if (write)
		ELAN_DEBUG5(ELAN_DBG_USERCOPY, "%p remote write from %p to %lx len %ld tsk %p\n",
			    current, local, remote, (long)len, tsk);
	else
		ELAN_DEBUG5(ELAN_DBG_USERCOPY, "%p remote read from %lx to %p len %ld tsk %p\n",
			    current, remote, local, (long)len, tsk);

	/* This locks the task, grabs a reference to the mm and then unlocks the task */
	mm = get_task_mm(tsk);

	if (!mm)
	{
		/* GNAT 7777: Must drop lock before returning */
		ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);
		return 0;
	}

	/* Do not try and copy from ourselves! */
	if (mm == current->mm)
	{
		/* GNAT 7777: Must now drop the elanmod lock as otherwise we can create a deadlock
		 * during the mmput() due it it calling exit_mmap() for the remote process
		 */
		ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);
		mmput(mm);
		return 0;
	}
	
        down_read(&mm->mmap_sem);

        /* ignore errors, just check how much was sucessfully transfered */
        while (len) {
                size_t bytes, ret, offset;
                void *maddr;

                ret = get_user_pages(tsk, mm, remote, 1, write, 1, &page, &vma);
                if (ret <= 0)
                        break;

                bytes = len;
                offset = remote & (PAGE_SIZE-1);
                if (bytes > PAGE_SIZE-offset)
                        bytes = PAGE_SIZE-offset;

                maddr = kmap(page);
                if (write) {
                        if (copy_from_user(/* remote to */maddr + offset, /* user from */local, bytes))	{
				kunmap(page);
				page_cache_release(page);	
				break;
			}
			set_page_dirty_lock(page);
                } else {
                        if (copy_to_user(/* user to */local, /* remote from */maddr + offset, bytes)) {
				kunmap(page);
				page_cache_release(page);
				break;
			}
                }
                kunmap(page);
		page_cache_release(page);
                len -= bytes;
                local += bytes;
                remote += bytes;
        }

        up_read(&mm->mmap_sem);

	/* GNAT 7777: Must now drop the elanmod lock as otherwise can we create a deadlock
	 * during the mmput() due it it calling exit_mmap() in the remote process
	 */
	ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);
	mmput(mm);

	/* Return num bytes copied */
        return local - old_buf;
}
#endif /* !defined(NO_MMPUT) || LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 9) */

int
elan_usercopy (void *remote, void *local, size_t len, int write, int ctxId, struct elan_cap_node_struct *cap_ptr)
{
	int ret = 0;
	size_t bytes;
	
	struct task_struct *tsk;

	/* Grab a read lock on elanmod lock 
	 *
	 * This prevents any process from exiting whilst the copy is in progress
	 * as it will need to take a write lock on the elanmod lock in order to do so
	 * As exit_fs() is called before the task is destroyed this should prevent
	 * the remote tsk from being torn down during the copy
	 *
	 * It would be much easier if we could just use get_task_struct()/put_task_struct()
	 * but __put_task_struct() is not exported by the 2.6.X kernels - sigh.
	 */
	ELANMOD_RWLOCK_READ(&elan_rwlock);

	/* Get the task handle from the cap node for the supplied ctxId */
	if ((ret = elan_usercopy_handle(cap_ptr, ctxId, (void **)&tsk)) < 0)
	{
		ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);
		return ret;
	}
	
	ELAN_DEBUG6(ELAN_DBG_USERCOPY,
		    "elan_usercopy: remote %p local %p len %ld write %d ctxId %d tsk %p\n",
		    remote, local, (long) len, write, ctxId, tsk);
	
	ASSERT(tsk);

	/* The BKL - why ??? (arch/[i386|ia64]/kernel/ptrace.c seems to hold it) */
//	lock_kernel();

	bytes = rw_process_vm(tsk, (unsigned long)remote, local, len, write);
	
	if (bytes != len)
	{
		ELAN_DEBUG2(ELAN_DBG_USERCOPY, "elan_usercopy: Failed to read %ld bytes (%ld copied)\n",
			    (long)len, (long)bytes);
		ret = -EPERM;
	}

	/* The BKL - why ??? (arch/[i386|ia64]/kernel/ptrace.c seems to hold it) */
//	unlock_kernel();

	/* GNAT 7777: rw_process_vm() now drops the elanmod lock 
	 *
	 * ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);
	 */

	return ret;
}



/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
