/*
 *    Copyright (c) 2003 by Quadrics Limited.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: context_linux.c,v 1.32 2004/10/28 11:51:00 david Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/os/context_linux.c,v $*/

#include <qsnet/kernel.h>
#include <qsnet/kpte.h>

#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elanvp.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanctxt.h>
#include <elan3/elandebug.h>
#include <elan3/urom_addrs.h>
#include <elan3/thread.h>

int
LoadElanTranslation (ELAN3_CTXT *ctxt, E3_Addr addr, int len, int protFault, int writeable)
{
    ELAN3MMU      	*elan3mmu = ctxt->Elan3mmu;
    ELAN3MMU_RGN 	*rgn;
    caddr_t		mainAddr;
    int			perm;
    unsigned int        off;
    unsigned long       flags;

    ASSERT (PAGE_ALIGNED (addr) && PAGE_ALIGNED (len));

    PRINTF (ctxt, DBG_FAULT, "LoadElanTranslation: addr %08x len %08x%s%s\n", 
        addr, len, protFault ? " prot fault" : "", writeable ? " writeable" : "");

    /* Ensure there's enough elan mmu tables for us to use */
    elan3mmu_expand (elan3mmu, addr, len, PTBL_LEVEL_3, 0);

    while (len > 0) 
    {
	/*
	 * Retrieve permission region and calculate main address
	 */
	spin_lock (&elan3mmu->elan3mmu_lock);

	rgn = elan3mmu_rgnat_elan (elan3mmu, addr);
	if (rgn == NULL) {
	    PRINTF (ctxt, DBG_FAULT, "LoadElanTranslation: no permission region at %lx %p\n", 
	        (u_long) addr, rgn);
	    spin_unlock (&elan3mmu->elan3mmu_lock);
	    return (EFAULT);
	}
	mainAddr = rgn->rgn_mbase + (addr - rgn->rgn_ebase);

	ASSERT (PAGE_ALIGNED ((unsigned long)mainAddr));

	spin_unlock (&elan3mmu->elan3mmu_lock);

	/*
	 * If we're tying to load a translation to the elan command port, 
	 * then don't do it now, but mark the context to have it reloaded
	 * just before we restart any threads. We do this because we don't
	 * want to call into the segment driver since we could then block
	 * waiting for the command port to become available.
	 */
	if (mainAddr == ctxt->CommandPageMapping)
	{
	    PRINTF (ctxt, DBG_FAULT, "LoadElanTranslation: addr=%08x maps command port\n", addr);

	    spin_lock_irqsave (&ctxt->Device->IntrLock, flags);
	    UnloadCommandPageMapping (ctxt);
	    spin_unlock_irqrestore (&ctxt->Device->IntrLock, flags);
	}
	else 
	{
	    struct vm_area_struct *area;
	    struct mm_struct *mm = current->mm;
	    pte_t *ptep_ptr;
	    pte_t  ptep_value;

	    down_read (&current->mm->mmap_sem);

	    if ((area = find_vma_intersection(mm, (unsigned long)mainAddr, (unsigned long)mainAddr + PAGESIZE)) == NULL)
	    {
		PRINTF (ctxt, DBG_FAULT, "LoadElanTranslation: %p no vma\n", mainAddr);
		up_read (&current->mm->mmap_sem);
		return EFAULT;
	    }

	    if (writeable && !(area->vm_flags & VM_WRITE)) 
	    {
		PRINTF (ctxt, DBG_FAULT, "LoadElanTranslation: %p not writeable\n", mainAddr);
		up_read (&current->mm->mmap_sem);
		return EFAULT;
	    }
	    
	    spin_lock (&mm->page_table_lock);

	    /* dont deference the pointer after the unmap */
	    ptep_ptr = find_pte_map (mm, (unsigned long)mainAddr);  
	    if (ptep_ptr) {
		ptep_value = *ptep_ptr;
		pte_unmap(ptep_ptr);
	    }

	    PRINTF (ctxt, DBG_FAULT, "LoadElanTranslation: %p %s %s\n", 
		    mainAddr, writeable ? "writeable" : "readonly", 
		    !ptep_ptr ? "invalid" : pte_none(ptep_value) ? "none " : !pte_present(ptep_value) ? "swapped " : 
		    writeable && !pte_write(ptep_value) ? "COW" : "OK");
	    
	    if (!ptep_ptr || pte_none(ptep_value) || !pte_present(ptep_value) || (writeable && !pte_write(ptep_value))) 
	    {  
		spin_unlock (&mm->page_table_lock);

		get_user_pages (current, current->mm, (unsigned long) mainAddr, PAGE_SIZE, 
				(area->vm_flags & VM_WRITE), 0, NULL, NULL);

		spin_lock (&mm->page_table_lock);

		/* dont deference the pointer after the unmap */
		ptep_ptr = find_pte_map (mm, (unsigned long)mainAddr);  
		if (ptep_ptr) {
		    ptep_value = *ptep_ptr;
		    pte_unmap(ptep_ptr);
		}

		if (!ptep_ptr || pte_none(ptep_value) || !pte_present(ptep_value) || (writeable && !pte_write(ptep_value))) 
		{
		    spin_unlock (&mm->page_table_lock);
		    up_read (&current->mm->mmap_sem);
		    return EFAULT;
		}
	    } 

	    /* don't allow user write access to kernel pages if not kernel */
	    if (!pte_read(ptep_value))
	    {
		spin_unlock (&mm->page_table_lock);
		up_read (&current->mm->mmap_sem);
		return EFAULT;
	    }

	    if (writeable)
		pte_mkdirty(ptep_value);
	    pte_mkyoung (ptep_value);

	    /* now load the elan pte */
	    if (writeable)
		perm  = rgn->rgn_perm;
	    else
		perm = ELAN3_PERM_READONLY(rgn->rgn_perm & ELAN3_PTE_PERM_MASK) | (rgn->rgn_perm & ~ELAN3_PTE_PERM_MASK);

	    for (off = 0; off < PAGE_SIZE; off += ELAN3_PAGE_SIZE)
		elan3mmu_pteload (elan3mmu, PTBL_LEVEL_3, addr + off, pte_phys(ptep_value) + off, perm, PTE_LOAD | PTE_NO_SLEEP);

	    spin_unlock (&mm->page_table_lock);
	    up_read (&current->mm->mmap_sem);
	}

	len -= PAGESIZE;
	addr += PAGESIZE;
    }
    return (ESUCCESS);
}


/*
 * LoadCommandPortTranslation:
 *    explicitly load an elan translation to the command port.
 *    but only do it if the command port is accessible.
 *
 *    we call this function just after we have restarted
 *    and trapped commands,  since when a command traps
 *    the elan translation to the command port is unloaded.
 */
void
LoadCommandPortTranslation (ELAN3_CTXT *ctxt)
{
    ELAN3MMU     *elan3mmu = ctxt->Elan3mmu;
    ELAN3MMU_RGN *rgn;
    E3_Addr       addr;
    int	          perm;
    physaddr_t    phys;
    unsigned int  off;
    unsigned long flags;

    PRINTF (ctxt, DBG_FAULT, "LoadCommandPortTranslation: SegAddr=%p Status=%x\n", ctxt->CommandPageMapping, ctxt->Status);

    if (ctxt->CommandPageMapping != NULL  && !(ctxt->Status & CTXT_COMMAND_MAPPED_ELAN))
    {
	spin_lock (&elan3mmu->elan3mmu_lock);
	
	rgn = elan3mmu_rgnat_main (elan3mmu, ctxt->CommandPageMapping);
	if (rgn == (ELAN3MMU_RGN *) NULL) 
	{
	    PRINTF(ctxt, DBG_FAULT, "LoadCommandPortTranslation: no permission for command port\n");
	    spin_unlock (&elan3mmu->elan3mmu_lock);
	    return;
	}
	
	addr = rgn->rgn_ebase + (ctxt->CommandPageMapping - rgn->rgn_mbase);
	perm = rgn->rgn_perm;
	phys = kmem_to_phys((caddr_t) ctxt->CommandPage);

	spin_lock_irqsave (&ctxt->Device->IntrLock, flags);
	if (ELAN3_QUEUE_EMPTY(ctxt->CommandTrapQ) && !(ctxt->Status & CTXT_OTHERS_REASONS))
	{
	    PRINTF(ctxt, DBG_FAULT, "LoadCommandPortTranslation: load xlation addr=%08x phys=%llx perm=%d\n", 
		   addr, (unsigned long long)phys, perm);

	    ctxt->Status |= CTXT_COMMAND_MAPPED_ELAN;

	    for (off = 0; off < PAGESIZE; off += ELAN3_PAGE_SIZE)
		elan3mmu_pteload (elan3mmu, PTBL_LEVEL_3, addr + off, phys + off, perm, PTE_LOAD | PTE_NO_SLEEP);
	}
	spin_unlock_irqrestore (&ctxt->Device->IntrLock, flags);
	
	spin_unlock (&elan3mmu->elan3mmu_lock);
    }
}


/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
