/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: elan3mmu_linux.c,v 1.53 2004/12/14 10:19:38 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/vm/elan3mmu_linux.c,v $*/

#include <qsnet/kernel.h>
#include <qsnet/kpte.h>

#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elanvp.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanctxt.h>
#include <elan3/elandebug.h>
#include <elan3/urom_addrs.h>
#include <elan3/thread.h>

/*
 * Strategy for syncing main <-> elan pte's:
 * 
 *   Install callbacks for linux flush_tlb_page(), flush_tlb_range(),
 *   flush_tlb_all(), and flush_tlb_mm() so when a main PTE changes,
 *   the elan translations, if any, are invalidated.  They can then be
 *   faulted in again with the correct physical page, perms, etc., on demand. 
 * 
 *   Callbacks are stacked on the mm_struct, one per context.  We also stack
 *   a ctxt pointer so we don't have to do lookups on every call.
 *
 *   Sanity check -- we clearly want to flush the elan PTEs in these 
 *   situations, all of which are covered by tlb_flush_{page,range}()
 *
 *     1) kernel/vmscan.c::try_to_swap_out() swaps out a page
 *
 *     2) kernel/mremap.c::copy_one_pte() moves a page as a result of the 
 *     mremap system call
 * 
 *     3) kernel/mprotect.c::change_pte_range() changes the permissions of a 
 *     page as the result of the mprotect system call
 *
 * Other Notes: 
 * 
 *   Dirty a page in the mains page tables when it is faulted into the elan.
 *   This way it will not be thrown away by the swapper.
 * 
 *   Pages write protected for COW are copied by elan3mmu_main_pagefault()
 *   when a writeable translation is loaded into the elan.
 */

caddr_t	     elan3mmu_kernel_invalid_space;
ELAN3_PTE     elan3mmu_kernel_invalid_pte_val;

void
elan3mmu_init_osdep (void)
{
    pte_t *pte;

    KMEM_GETPAGES (elan3mmu_kernel_invalid_space, caddr_t, 1, TRUE);

    ASSERT(elan3mmu_kernel_invalid_space != NULL);

    pte = find_pte_kernel ((unsigned long) elan3mmu_kernel_invalid_space);

    elan3mmu_kernel_invalid_pte_val = ELAN3_PTE_64_BIT | (pte_phys(*pte) & ELAN3_PTE_PFN_MASK) | ELAN3_PERM_REMOTEREAD | ELAN3_ET_PTE;

#ifdef __alpha
    /*
     * NOTE: Elan sign-extends bit 48 of the physical address, so if we need to
     *       set any of bits 63:48, then we will set them all by setting bit 48/
     */
    if (alpha_mv.pci_dac_offset & 0xFFFF000000000000ull)
        elan3mmu_kernel_invalid_pte_val |= (1ull << 48);
    else
	elan3mmu_kernel_invalid_pte_val |= alpha_mv.pci_dac_offset;
#endif

    HAT_PRINTF(0x10, "elan3mmu_invalid_space at %p phys=%llx pte=%llx\n", elan3mmu_kernel_invalid_space, 
	       (unsigned long long) pte_phys(*pte), (unsigned long long) elan3mmu_kernel_invalid_pte_val);
}

void
elan3mmu_fini_osdep()
{
    KMEM_FREEPAGES (elan3mmu_kernel_invalid_space, 1);
}

void
elan3mmu_alloc_osdep (ELAN3MMU *elan3mmu)
{
    elan3mmu->elan3mmu_coproc_mm = current->mm;
}

/*
 * Convert physical page frame number to elan pte.
 */
ELAN3_PTE
elan3mmu_phys_to_pte (ELAN3_DEV *dev, physaddr_t paddr, int perm)
{
    ELAN3_PTE newpte;
    
    ASSERT (paddr != 0);
    
    if ((paddr & dev->SdramPhysMask) == dev->SdramPhysBase)		/* SDRAM, turn on PTE_LOCAL bit */
    {
	PRINTF(NULL, DBG_HAT, "elan3mmu_phys_to_pte: phys %llx SDRAM\n", (unsigned long long) paddr);
	
	newpte = ELAN3_PTE_LOCAL | (paddr & ELAN3_PTE_PFN_MASK & ~dev->SdramPhysMask) | perm | ELAN3_ET_PTE;
    }
#if defined(LINUX_ALPHA)
    else if ((paddr & dev->PciPhysMask) == dev->PciPhysBase)
    {
	PRINTF(NULL, DBG_HAT, "elan3mmu_phys_to_pte: phys %llx PCI\n", (unsigned long long) paddr);
	newpte = ELAN3_PTE_64_BIT | (paddr & ELAN3_PTE_PFN_MASK & ~dev->PciPhysMask) | perm | ELAN3_ET_PTE;
    }
#endif
    else 						/* main memory, must convert to PCI view */
    {
	PRINTF(NULL, DBG_HAT, "elan3mmu_phys_to_pte: phys %llx is main memory\n", (unsigned long long) paddr);

	/* main memory, just set the architecture specific PTE_BYPASS bit */
	/* This requires the Tsunami chipset being programmed to support
	 * the monster window option. This is in linux-2.4.5 and later kernels 
	 * and is also patched into the RH 7.1/2.4.3-12 Alpha kernel
	 */
	newpte = ELAN3_PTE_64_BIT | (paddr & ELAN3_PTE_PFN_MASK) | perm | ELAN3_ET_PTE;

#ifdef __alpha
	/*
	 * NOTE: Elan sign-extends bit 48 of the physical address, so if we need to
	 *       set any of bits 63:48, then we will set them all by setting bit 48/
	 */
	if (alpha_mv.pci_dac_offset & 0xFFFF000000000000ull)
            newpte |= (1ull << 48);
        else
	    newpte |= alpha_mv.pci_dac_offset;
#endif
    }

    if ( ELAN3_PERM_WRITEABLE( perm )) 
	newpte |= ( ELAN3_PTE_MOD | ELAN3_PTE_REF );
    else
	newpte |= ( ELAN3_PTE_REF ) ;

    return (newpte);
}

ELAN3_PTE
elan3mmu_kernel_invalid_pte (ELAN3MMU *elan3mmu)
{
    if (elan3mmu->elan3mmu_dev->Devinfo.dev_revision_id == PCI_REVISION_ID_ELAN3_REVB)
	return (elan3mmu_kernel_invalid_pte_val);
    return (ELAN3_INVALID_PTE);
}

/* 
 * Invalidate a range of addresses for specified context.
 */
void
elan3mmu_pte_range_unload (ELAN3MMU *elan3mmu, struct mm_struct *mm, caddr_t addr, unsigned long len)
{
    E3_Addr       eaddr;
    ELAN3MMU_RGN *rgn;
    unsigned long span;

    spin_lock (&elan3mmu->elan3mmu_lock);

    for (; len; len -= span, addr += span)
    {
    	rgn = elan3mmu_findrgn_main (elan3mmu, addr, 0);

	if (rgn == NULL || (rgn->rgn_mbase + rgn->rgn_len) < addr)
	    span = len;
    	else if (rgn->rgn_mbase > addr)
	    span = MIN(len, rgn->rgn_mbase - addr);
	else
	{
	    span  = MIN(len, (rgn->rgn_mbase + rgn->rgn_len) - addr);
	    eaddr = rgn->rgn_ebase + (addr - rgn->rgn_mbase);

            HAT_PRINTF(0x10, "  unloading eaddr %x main %p (%ld pages)\n", 
	      eaddr, addr, btopr(span));
	    elan3mmu_unload (elan3mmu, eaddr, span, PTE_UNLOAD);
	}			/* takes care of elan tlb flush also */
    }

    spin_unlock (&elan3mmu->elan3mmu_lock);
}

/*
 *
 */
void
elan3mmu_update_range (ELAN3MMU *elan3mmu, struct mm_struct *mm, caddr_t vaddr, E3_Addr eaddr, u_int len, u_int perm)
{
    u_int roperm = ELAN3_PERM_READONLY(perm & ELAN3_PTE_PERM_MASK) | (perm & ~ELAN3_PTE_PERM_MASK);
    u_int off;

    HAT_PRINTF3(1, "elan3mmu_update_range (elan3mmu %p addr %p -> %p)\n", elan3mmu, vaddr, vaddr+len-1);

    while (len > 0)
    {
	pte_t *pte_ptr;
	pte_t  pte_value;

	pte_ptr = find_pte_map(mm, (unsigned long)vaddr);
	if (pte_ptr) {
	    pte_value = *pte_ptr;
	    pte_unmap(pte_ptr);
	}

	HAT_PRINTF(0x10, "  elan3mmu_update_range %x (%p) %s\n", eaddr, vaddr, 
		!pte_ptr ? "invalid" : pte_none(pte_value) ? "none " : !pte_present(pte_value) ? "swapped " : 
		!pte_write(pte_value) ? "RO/COW" : "OK");
	
	if (pte_ptr && !pte_none(pte_value) && pte_present(pte_value))
	    for (off = 0; off < PAGE_SIZE; off += ELAN3_PAGE_SIZE)
		elan3mmu_pteload (elan3mmu, PTBL_LEVEL_3, eaddr + off, pte_phys(pte_value) + off, pte_write(pte_value) ? perm : roperm, PTE_LOAD|PTE_NO_SLEEP|PTE_NO_STEAL);
	vaddr += PAGESIZE;
	eaddr += PAGESIZE;
	len   -= PAGESIZE;
    }
}

/* 
 * Update a range of addresses for specified context.
 */
void
elan3mmu_pte_range_update (ELAN3MMU *elan3mmu, struct mm_struct *mm,caddr_t vaddr, unsigned long len)
{
    E3_Addr       eaddr;
    ELAN3MMU_RGN *rgn;
    unsigned long span;

    spin_lock (&elan3mmu->elan3mmu_lock);

    for (; len; len -= span, vaddr += span)
    {
    	rgn = elan3mmu_findrgn_main (elan3mmu, vaddr, 0);

	if (rgn == NULL || (rgn->rgn_mbase + rgn->rgn_len) < vaddr)
	    span = len;
    	else if (rgn->rgn_mbase > vaddr)
	    span = MIN(len, rgn->rgn_mbase - vaddr);
	else
	{
	    span  = MIN(len, (rgn->rgn_mbase + rgn->rgn_len) - vaddr);
	    eaddr = rgn->rgn_ebase + (vaddr - rgn->rgn_mbase);

            HAT_PRINTF(0x10, "  updating eaddr %u main %p (%ld pages)\n", 
	      eaddr, vaddr, btopr(span));
	    
	    elan3mmu_update_range(elan3mmu, mm, vaddr, eaddr, span, rgn->rgn_perm);
	}			
    }

    spin_unlock (&elan3mmu->elan3mmu_lock);
}

/* 
 * Invalidate all ptes for the given context.
 */
void
elan3mmu_pte_ctxt_unload(ELAN3MMU *elan3mmu)
{
    ELAN3_PTBL  *l1ptbl   = (elan3mmu ? elan3mmu->elan3mmu_l1ptbl : NULL);
    spinlock_t *l1mtx;
    unsigned long flags;

    if (l1ptbl && elan3mmu_lock_ptbl (l1ptbl, LK_PTBL_FAILOK, elan3mmu, (E3_Addr) 0, 1, &l1mtx, &flags) == LK_PTBL_OK)
    {
	elan3mmu_l1inval(elan3mmu, elan3mmu->elan3mmu_l1ptbl, 0);
	elan3mmu_unlock_ptbl (l1ptbl, l1mtx, flags);
    }
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
