/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: eventcookie.c,v 1.7 2003/08/13 10:03:03 fabien Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/os/eventcookie.c,v $*/

#include <qsnet/kernel.h>
#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elanvp.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanctxt.h>
#include <elan3/elandebug.h>
#include <elan3/urom_addrs.h>
#include <elan3/thread.h>
#include <elan3/vmseg.h>

static EVENT_COOKIE_TABLE *cookie_tables;
static spinlock_t	   cookie_table_lock;

/*
 * cookie_drop_entry:
 *   drop the reference to a cookie held 
 *   by the cookie table
 */
static void
cookie_drop_entry (EVENT_COOKIE_ENTRY *ent)
{
    unsigned long flags;

    spin_lock_irqsave (&ent->ent_lock, flags);
    if (--ent->ent_ref != 0)
    {
	ent->ent_fired = ent->ent_cookie;
	kcondvar_wakeupall (&ent->ent_wait, &ent->ent_lock);

	spin_unlock_irqrestore (&ent->ent_lock, flags);
    }
    else
    {
	spin_unlock_irqrestore (&ent->ent_lock, flags);

	spin_lock_destroy (&ent->ent_lock);
	kcondvar_destroy (&ent->ent_wait);

	KMEM_FREE (ent, sizeof (EVENT_COOKIE_ENTRY));
    }
}

void
cookie_init()
{
    spin_lock_init (&cookie_table_lock);
}

void
cookie_fini()
{
    spin_lock_destroy (&cookie_table_lock);
}

EVENT_COOKIE_TABLE *
cookie_alloc_table (unsigned long task, unsigned long handle)
{
    EVENT_COOKIE_TABLE *tbl, *ntbl;

    KMEM_ZALLOC (ntbl, EVENT_COOKIE_TABLE *, sizeof (EVENT_COOKIE_TABLE), TRUE);

    if (ntbl == NULL)
	return (NULL);

    spin_lock (&cookie_table_lock);
    
    for (tbl = cookie_tables; tbl; tbl = tbl->tbl_next)
	if (tbl->tbl_task == task && tbl->tbl_handle == handle)
	    break;
    
    if (tbl != NULL)
	tbl->tbl_ref++;
    else
    {
	spin_lock_init (&ntbl->tbl_lock);

	ntbl->tbl_task    = task;
	ntbl->tbl_handle  = handle;
	ntbl->tbl_ref     = 1;
	ntbl->tbl_entries = NULL;

	if ((ntbl->tbl_next = cookie_tables) != NULL)
	    cookie_tables->tbl_prev = ntbl;
	cookie_tables = ntbl;
	ntbl->tbl_prev = NULL;
    }
    spin_unlock (&cookie_table_lock);

    if (tbl == NULL)
	return (ntbl);
    else
    {
	KMEM_FREE (ntbl, sizeof (EVENT_COOKIE_TABLE));
	return (tbl);
    }    
}

void
cookie_free_table (EVENT_COOKIE_TABLE *tbl)
{
    EVENT_COOKIE_ENTRY *ent;

    spin_lock (&cookie_table_lock);
    if (tbl->tbl_ref > 1)
    {
	tbl->tbl_ref--;
	spin_unlock (&cookie_table_lock);
	return;
    }
    
    if (tbl->tbl_prev)
	tbl->tbl_prev->tbl_next = tbl->tbl_next;
    else
	cookie_tables = tbl->tbl_next;
    if (tbl->tbl_next)
	tbl->tbl_next->tbl_prev = tbl->tbl_prev;
    
    spin_unlock (&cookie_table_lock);
    
    /* NOTE - table no longer visible to other threads
     *        no need to aquire tbl_lock */
    while ((ent = tbl->tbl_entries) != NULL)
    {
	if ((tbl->tbl_entries = ent->ent_next) != NULL)
	    ent->ent_next->ent_prev = NULL;
	
	cookie_drop_entry (ent);
    }
    spin_lock_destroy (&tbl->tbl_lock);

    KMEM_FREE (tbl, sizeof (EVENT_COOKIE_TABLE));
}

int
cookie_alloc_cookie (EVENT_COOKIE_TABLE *tbl, EVENT_COOKIE cookie)
{
    EVENT_COOKIE_ENTRY *ent, *nent;
    unsigned long flags;

    KMEM_ZALLOC (nent, EVENT_COOKIE_ENTRY *, sizeof (EVENT_COOKIE_ENTRY), TRUE);
    
    spin_lock_irqsave (&tbl->tbl_lock, flags);
    for (ent = tbl->tbl_entries; ent; ent = ent->ent_next)
	if (ent->ent_cookie == cookie)
	    break;

    if (ent == NULL)
    {
	kcondvar_init (&nent->ent_wait);
	spin_lock_init (&nent->ent_lock);

	nent->ent_ref    = 1;
	nent->ent_cookie = cookie;

	if ((nent->ent_next = tbl->tbl_entries) != NULL)
	    tbl->tbl_entries->ent_prev = nent;
	tbl->tbl_entries = nent;
	nent->ent_prev = NULL;
    }
    spin_unlock_irqrestore (&tbl->tbl_lock, flags);

    if (ent == NULL)
	return (ESUCCESS);
    else
    {
	KMEM_FREE (nent, sizeof (EVENT_COOKIE_ENTRY));
	return (EINVAL);
    }
}

int
cookie_free_cookie (EVENT_COOKIE_TABLE *tbl, EVENT_COOKIE cookie)
{
    EVENT_COOKIE_ENTRY *ent;
    unsigned long flags;

    spin_lock_irqsave (&tbl->tbl_lock, flags);
    for (ent = tbl->tbl_entries; ent; ent = ent->ent_next)
	if (ent->ent_cookie == cookie)
	    break;
    
    if (ent == NULL)
    {
	spin_unlock_irqrestore (&tbl->tbl_lock, flags);
	return (EINVAL);
    }

    if (ent->ent_prev == NULL)
	tbl->tbl_entries = ent->ent_next;
    else
	ent->ent_prev->ent_next = ent->ent_next;

    if (ent->ent_next != NULL)
	ent->ent_next->ent_prev = ent->ent_prev;
    
    spin_unlock_irqrestore (&tbl->tbl_lock, flags);

    cookie_drop_entry (ent);

    return (ESUCCESS);
}

/*
 * cookie_fire_cookie:
 *    fire the cookie - this is called from the event interrupt.
 */
int
cookie_fire_cookie (EVENT_COOKIE_TABLE *tbl, EVENT_COOKIE cookie)
{
    EVENT_COOKIE_ENTRY *ent;
    unsigned long flags;

    spin_lock_irqsave (&tbl->tbl_lock, flags);
    for (ent = tbl->tbl_entries; ent; ent = ent->ent_next)
	if (ent->ent_cookie == cookie)
	    break;
    
    if (ent == NULL)
    {
	spin_unlock_irqrestore (&tbl->tbl_lock, flags);
	return (EINVAL);
    }
	    
    spin_lock (&ent->ent_lock);
    ent->ent_fired = cookie;
    kcondvar_wakeupall (&ent->ent_wait, &ent->ent_lock);
    spin_unlock (&ent->ent_lock);

    spin_unlock_irqrestore (&tbl->tbl_lock, flags);

    return (ESUCCESS);
}    

/*
 * cookie_wait_cookie:
 *    deschedule on a cookie if it has not already fired.
 *    note - if the cookie is removed from the table, then
 *           we free it off when we're woken up.
 */
int
cookie_wait_cookie (EVENT_COOKIE_TABLE *tbl, EVENT_COOKIE cookie)
{
    EVENT_COOKIE_ENTRY *ent;
    unsigned long flags;
    
    spin_lock_irqsave (&tbl->tbl_lock, flags);
    for (ent = tbl->tbl_entries; ent; ent = ent->ent_next)
	if (ent->ent_cookie == cookie)
	    break;
    
    if (ent == NULL)
    {
	spin_unlock_irqrestore (&tbl->tbl_lock, flags);
	return (EINVAL);
    }

    spin_lock (&ent->ent_lock);
    spin_unlock (&tbl->tbl_lock);

    if (ent->ent_fired != 0)
    {
	spin_unlock_irqrestore (&ent->ent_lock, flags);
	return (ESUCCESS);
    }

    ent->ent_ref++;
    kcondvar_waitsig (&ent->ent_wait, &ent->ent_lock, &flags);
    
    if (--ent->ent_ref > 0)
	spin_unlock_irqrestore (&ent->ent_lock, flags);
    else
    {
	spin_unlock_irqrestore (&ent->ent_lock, flags);
	
	spin_lock_destroy (&ent->ent_lock);
	kcondvar_destroy (&ent->ent_wait);

	KMEM_FREE (ent, sizeof (EVENT_COOKIE_ENTRY));
    }
    return (ESUCCESS);
}

int
cookie_arm_cookie (EVENT_COOKIE_TABLE *tbl, EVENT_COOKIE cookie)
{
    EVENT_COOKIE_ENTRY *ent;
    unsigned long flags;

    spin_lock_irqsave (&tbl->tbl_lock, flags);
    for (ent = tbl->tbl_entries; ent; ent = ent->ent_next)
	if (ent->ent_cookie == cookie)
	    break;
    
    if (ent == NULL)
    {
	spin_unlock_irqrestore (&tbl->tbl_lock, flags);
	return (EINVAL);
    }
	    
    spin_lock (&ent->ent_lock);
    ent->ent_fired = 0;
    spin_unlock (&ent->ent_lock);

    spin_unlock_irqrestore (&tbl->tbl_lock, flags);

    return (ESUCCESS);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
