/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: network_error.c,v 1.33 2004/10/28 11:51:00 david Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/os/network_error.c,v $*/

#include <qsnet/kernel.h>
#include <qsnet/kthread.h>

#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elanvp.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanctxt.h>
#include <elan3/elan3mmu.h>
#include <elan3/elandebug.h>

#ifdef DIGITAL_UNIX
#include <sys/cred.h>
#include <sys/mbuf.h>
#include <sys/utsname.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/in_var.h>

#include <rpc/types.h>
#include <rpc/auth.h>
#include <rpc/xdr.h>
#include <rpc/clnt.h>

typedef xdrproc_t kxdrproc_t;
#endif

#ifdef LINUX
#include <linux/sunrpc/types.h>
#include <linux/sunrpc/auth.h>
#include <linux/sunrpc/xdr.h>
#include <linux/sunrpc/clnt.h>

#include <linux/utsname.h>
#define SYS_NMLN	__NEW_UTS_LEN
#endif

#include <elan3/neterr_rpc.h>

spinlock_t	  ResolveRequestLock;
kcondvar_t	  ResolveRequestWait;

NETERR_RESOLVER  *ResolveRequestHead;
NETERR_RESOLVER **ResolveRequestTailp = &ResolveRequestHead;
int		  ResolveRequestCount;
int		  ResolveRequestThreads;
int		  ResolveRequestMaxThreads = 4;
int		  ResolveRequestTimeout = 60;

typedef struct neterr_server
{
    struct neterr_server *Next;
    struct neterr_server *Prev;
    unsigned		  ElanId;

    char	         *Name;
    int		          RefCount;
    struct sockaddr_in    Addr;
} NETERR_SERVER;

#define NETERR_HASH_ENTRIES	64
#define NETERR_HASH(elanid)	(((unsigned) elanid) % NETERR_HASH_ENTRIES)
NETERR_SERVER *NeterrServerHash[NETERR_HASH_ENTRIES];
kmutex_t       NeterrServerLock;

static NETERR_SERVER *FindNeterrServer (int elanId);
static void           DereferenceNeterrServer (NETERR_SERVER *server);
static int            CallNeterrServer (NETERR_SERVER *server, NETERR_MSG *msg);

void
InitialiseNetworkErrorResolver ()
{
    spin_lock_init (&ResolveRequestLock);
    kcondvar_init (&ResolveRequestWait);
    
    ResolveRequestHead  = NULL;
    ResolveRequestTailp = &ResolveRequestHead;

    kmutex_init (&NeterrServerLock);
}

void
FinaliseNetworkErrorResolver ()
{
    spin_lock_destroy (&ResolveRequestLock);
    kcondvar_destroy (&ResolveRequestWait);
    
    kmutex_destroy (&NeterrServerLock);
}

static NETERR_RESOLVER *
AllocateNetworkErrorResolver (void)
{
    NETERR_RESOLVER *rvp;

    KMEM_ZALLOC (rvp, NETERR_RESOLVER *, sizeof (NETERR_RESOLVER), TRUE);
    spin_lock_init (&rvp->Lock);

    return (rvp);
}

void
FreeNetworkErrorResolver (NETERR_RESOLVER *rvp)
{
    spin_lock_destroy (&rvp->Lock);
    KMEM_FREE (rvp, sizeof (NETERR_RESOLVER));
}

static void
elan3_neterr_resolver (void)
{
    NETERR_RESOLVER *rvp;
    NETERR_SERVER   *server;
    int		     status;
    unsigned long    flags;

    kernel_thread_init("elan3_neterr_resolver");
    spin_lock (&ResolveRequestLock);

    while ((rvp = ResolveRequestHead) != NULL)
    {
	if ((ResolveRequestHead = rvp->Next) == NULL)
	    ResolveRequestTailp = &ResolveRequestHead;
	
	spin_unlock (&ResolveRequestLock);

	PRINTF1 (DBG_DEVICE, DBG_NETERR, "elan3_neterr_resolver: rvp = %p\n", rvp);
	PRINTF1 (DBG_DEVICE, DBG_NETERR, "                      Rail          %d\n", rvp->Message.Rail);
	PRINTF1 (DBG_DEVICE, DBG_NETERR, "                      SrcCapability %s\n", CapabilityString (&rvp->Message.SrcCapability));
	PRINTF1 (DBG_DEVICE, DBG_NETERR, "                      DstCapability %s\n", CapabilityString (&rvp->Message.DstCapability));
	PRINTF1 (DBG_DEVICE, DBG_NETERR, "                      CookieAddr    %08x\n", rvp->Message.CookieAddr);
	PRINTF1 (DBG_DEVICE, DBG_NETERR, "                      CookieVProc   %08x\n", rvp->Message.CookieVProc);
	PRINTF1 (DBG_DEVICE, DBG_NETERR, "                      NextCookie    %08x\n", rvp->Message.NextCookie);
	PRINTF1 (DBG_DEVICE, DBG_NETERR, "                      WaitForEop    %08x\n", rvp->Message.WaitForEop);
	
	if ((server = FindNeterrServer (rvp->Location.loc_node)) == NULL)
	    status = ECONNREFUSED;
	else if (ResolveRequestTimeout && ((int)(lbolt - rvp->Timestamp)) > (ResolveRequestTimeout*HZ))
	{
	    printk ("elan_neterr: rpc to '%s' timedout - context %d killed\n", server->Name, rvp->Message.SrcCapability.cap_mycontext);
	    status = ECONNABORTED;
	}
	else
	{
	    status = CallNeterrServer (server, &rvp->Message);

	    DereferenceNeterrServer (server);
	}
	
	if ((status == EINTR || status == ETIMEDOUT) && rvp->Ctxt != NULL)
	{
	    PRINTF1 (DBG_DEVICE, DBG_NETERR, "elan3_neterr_resolver: retry rvp=%p\n", rvp);
	    spin_lock (&ResolveRequestLock);
	    rvp->Next = NULL;
	    *ResolveRequestTailp = rvp;
	    ResolveRequestTailp = &rvp->Next;
	}
	else
	{
	    rvp->Status = status;
	    
	    spin_lock (&rvp->Lock);
	    
	    if (rvp->Ctxt != NULL)
	    {
		PRINTF2 (rvp->Ctxt, DBG_NETERR, "elan3_neterr_resolver: completing rvp %p for ctxt %p\n", rvp, rvp->Ctxt);
		spin_lock_irqsave (&rvp->Ctxt->Device->IntrLock, flags);
		
		rvp->Completed = TRUE;
		
		kcondvar_wakeupall (&rvp->Ctxt->Wait, &rvp->Ctxt->Device->IntrLock);
		
		/*
		 * drop the locks out of order since the rvp can get freeed
		 * as soon as we drop the IntrLock - so cannot reference the
		 * rvp after this.
		 */
		
		spin_unlock (&rvp->Lock);
		spin_unlock_irqrestore (&rvp->Ctxt->Device->IntrLock, flags);
	    }
	    else
	    {
		PRINTF2 (DBG_DEVICE, DBG_NETERR, "elan3_neterr_resolver: completing rvp %p for deceased ctxt %p\n", rvp, rvp->Ctxt);
		spin_unlock (&rvp->Lock);
		FreeNetworkErrorResolver (rvp);
	    }
	    
	    spin_lock (&ResolveRequestLock);
	    ResolveRequestCount--;
	}
    }

    ResolveRequestThreads--;

    spin_unlock (&ResolveRequestLock);
    kernel_thread_exit();
}

int
QueueNetworkErrorResolver (ELAN3_CTXT *ctxt, INPUT_TRAP *trap, NETERR_RESOLVER **rvpp)
{
    int			   isdma   = trap->DmaIdentifyTransaction != NULL;
    E3_IprocTrapHeader_BE *hdrp    = isdma ? trap->DmaIdentifyTransaction : trap->ThreadIdentifyTransaction;
    E3_uint32              process = isdma ? (hdrp->s.TrAddr & 0xFFFF) : (hdrp->s.TrData0 & 0xFFFF);
    NETERR_RESOLVER       *rvp;

    PRINTF2 (ctxt, DBG_NETERR, "QueueNetworkErrorResolver: process = %d %s\n", process, isdma ? "(dma)" : "(thread)");

    if ((rvp = AllocateNetworkErrorResolver()) == NULL)
    {
	PRINTF0 (ctxt, DBG_NETERR, "QueueNetworkErrorResolver: cannot allocate resolver\n");
	return (ENOMEM);
    }

    rvp->Message.Rail = ctxt->Device->Devinfo.dev_rail;

    krwlock_read (&ctxt->VpLock);
    rvp->Location = ProcessToLocation (ctxt, NULL, process, &rvp->Message.SrcCapability);
    krwlock_done (&ctxt->VpLock);

    if (rvp->Location.loc_node == ELAN3_INVALID_NODE)
    {
	PRINTF0 (ctxt, DBG_NETERR, "QueueNetworkErrorResolver: invalid elan id\n");

	FreeNetworkErrorResolver (rvp);
	return (EINVAL);
    }

    rvp->Message.DstCapability = ctxt->Capability;
    rvp->Message.DstProcess    = elan3_process (ctxt);
    rvp->Message.WaitForEop    = (trap->WaitForEopTransaction != NULL);

    if (isdma)
    {
	rvp->Message.CookieAddr  = 0;
	rvp->Message.CookieVProc = hdrp->s.TrAddr;
	rvp->Message.NextCookie  = 0;
    }
    else
    {
	rvp->Message.CookieAddr  = hdrp->s.TrAddr;
	rvp->Message.CookieVProc = hdrp->s.TrData0;
	rvp->Message.NextCookie  = hdrp->s.TrData1;
    }

    rvp->Completed = FALSE;
    rvp->Ctxt      = ctxt;
    rvp->Timestamp = lbolt;

    spin_lock (&ResolveRequestLock);

    rvp->Next = NULL;
    *ResolveRequestTailp = rvp;
    ResolveRequestTailp = &rvp->Next;
    ResolveRequestCount++;

    kcondvar_wakeupone (&ResolveRequestWait, &ResolveRequestLock);

    if (ResolveRequestCount < ResolveRequestThreads || ResolveRequestThreads >= ResolveRequestMaxThreads)
	spin_unlock (&ResolveRequestLock);
    else
    {
	ResolveRequestThreads++;

	spin_unlock (&ResolveRequestLock);
	if (kernel_thread_create (elan3_neterr_resolver, NULL) == NULL)
	{
	    spin_lock (&ResolveRequestLock);
	    ResolveRequestThreads--;
	    spin_unlock (&ResolveRequestLock);
	    
	    if (ResolveRequestThreads == 0)
	    {
		PRINTF0 (ctxt, DBG_NETERR, "QueueNetworkErrorResolver: cannot thread pool\n");

		FreeNetworkErrorResolver (rvp);
		return (ENOMEM);
	    }
	}
    }

    *rvpp = rvp;
    return (ESUCCESS);
}

void
CancelNetworkErrorResolver (NETERR_RESOLVER *rvp)
{
    spin_lock (&rvp->Lock);

    PRINTF2 (rvp->Ctxt, DBG_NETERR, "CancelNetworkErrorResolver: rvp=%p %s\n", rvp, rvp->Completed ? "Completed" : "Pending");

    if (rvp->Completed)
    {
	spin_unlock (&rvp->Lock);
	FreeNetworkErrorResolver (rvp);
    }
    else
    {
	rvp->Ctxt = NULL;
	spin_unlock (&rvp->Lock);
    }
}

static NETERR_FIXUP *
AllocateNetworkErrorFixup (void)
{
    NETERR_FIXUP *nef;

    KMEM_ZALLOC (nef, NETERR_FIXUP *, sizeof (NETERR_FIXUP), TRUE);

    if (nef == (NETERR_FIXUP *) NULL)
	return (NULL);

    kcondvar_init (&nef->Wait);

    return (nef);
}

static void
FreeNetworkErrorFixup (NETERR_FIXUP *nef)
{
    kcondvar_destroy (&nef->Wait);
    KMEM_FREE (nef, sizeof (NETERR_FIXUP));
}

int
ExecuteNetworkErrorFixup (NETERR_MSG *msg)
{
    ELAN3_DEV      *dev;
    ELAN3_CTXT	  *ctxt;
    NETERR_FIXUP  *nef;
    NETERR_FIXUP **predp;
    int		   rc;
    unsigned long  flags;

    PRINTF1 (DBG_DEVICE, DBG_NETERR, "ExecuteNetworkErrorFixup: msg = %p\n", msg);
    PRINTF1 (DBG_DEVICE, DBG_NETERR, "                      Rail          %d\n", msg->Rail);
    PRINTF1 (DBG_DEVICE, DBG_NETERR, "                      SrcCapability %s\n", CapabilityString (&msg->SrcCapability));
    PRINTF1 (DBG_DEVICE, DBG_NETERR, "                      DstCapability %s\n", CapabilityString (&msg->DstCapability));
    PRINTF1 (DBG_DEVICE, DBG_NETERR, "                      CookieAddr    %08x\n", msg->CookieAddr);
    PRINTF1 (DBG_DEVICE, DBG_NETERR, "                      CookieVProc   %08x\n", msg->CookieVProc);
    PRINTF1 (DBG_DEVICE, DBG_NETERR, "                      NextCookie    %08x\n", msg->NextCookie);
    PRINTF1 (DBG_DEVICE, DBG_NETERR, "                      WaitForEop    %08x\n", msg->WaitForEop);
	
    if ((dev = elan3_device (msg->Rail)) == NULL)
	return (ESRCH);

    if ((nef = AllocateNetworkErrorFixup()) == NULL)
	return (ENOMEM);

    if (nef == (NETERR_FIXUP *) NULL)
	return (ENOMEM);
    
    bcopy (msg, &nef->Message, sizeof (NETERR_MSG));

    spin_lock_irqsave (&dev->IntrLock, flags);
    
    ctxt = ELAN3_DEV_CTX_TABLE(dev, msg->SrcCapability.cap_mycontext);

    if (ctxt == NULL)
	rc = ESRCH;
    else if (!ELAN_CAP_MATCH (&msg->SrcCapability, &ctxt->Capability))
	rc = EPERM;
    else
    {	
	if (ctxt->Status & CTXT_NO_LWPS)
	    rc = EAGAIN;
	else
	{
	    for (predp = &ctxt->NetworkErrorFixups; *predp != NULL; predp = &(*predp)->Next)
		;
	    nef->Next = NULL;
	    *predp = nef;
	    
	    kcondvar_wakeupone (&ctxt->Wait, &dev->IntrLock);

	    while (! nef->Completed)
		kcondvar_wait (&nef->Wait, &dev->IntrLock, &flags);

	    rc = nef->Status;
	}
    }
    
    spin_unlock_irqrestore (&dev->IntrLock, flags);

    FreeNetworkErrorFixup (nef);

    return (rc);
}

void
CompleteNetworkErrorFixup (ELAN3_CTXT *ctxt, NETERR_FIXUP *nef, int status)
{
    ELAN3_DEV *dev = ctxt->Device;
    unsigned long flags;

    PRINTF2 (ctxt, DBG_NETERR, "CompleteNetworkErrorFixup: %p %d\n", nef, status);

    spin_lock_irqsave (&dev->IntrLock, flags);

    nef->Status = status;
    nef->Completed = TRUE;
    kcondvar_wakeupone (&nef->Wait, &dev->IntrLock);

    spin_unlock_irqrestore (&dev->IntrLock, flags);
}


static NETERR_SERVER *
NewNeterrServer (int elanId, struct sockaddr_in *addr, char *name)
{
    NETERR_SERVER *server;

    KMEM_ZALLOC (server, NETERR_SERVER *, sizeof (NETERR_SERVER), TRUE);
    KMEM_ALLOC  (server->Name, char *, strlen (name)+1, TRUE);

    bcopy (addr, &server->Addr, sizeof (struct sockaddr_in));
    bcopy (name, server->Name, strlen (name)+1);

    server->ElanId   = elanId;
    server->RefCount = 1;
    
    return (server);
}

static void
DeleteNeterrServer (NETERR_SERVER *server)
{
    KMEM_FREE (server->Name, strlen(server->Name)+1);
    KMEM_FREE (server, sizeof (NETERR_SERVER));
}

static NETERR_SERVER *
FindNeterrServer (int elanId)
{
    NETERR_SERVER *server;
    
    kmutex_lock (&NeterrServerLock);
    
    for (server = NeterrServerHash[NETERR_HASH(elanId)]; server != NULL; server = server->Next)
	if (server->ElanId == elanId)
	    break;

    if (server != NULL)
	server->RefCount++;
    kmutex_unlock (&NeterrServerLock);

    return (server);
}

static void
DereferenceNeterrServer (NETERR_SERVER *server)
{
    kmutex_lock (&NeterrServerLock);
    if ((--server->RefCount) == 0)
	DeleteNeterrServer (server);
    kmutex_unlock  (&NeterrServerLock);
}

int
AddNeterrServer (int elanId, struct sockaddr_in *addr, char *name)
{
    NETERR_SERVER *server;
    NETERR_SERVER *old;
    int            hashval = NETERR_HASH(elanId);

    server = NewNeterrServer (elanId, addr, name);
    
    if (server == NULL)
	return (ENOMEM);
    
    kmutex_lock (&NeterrServerLock);
    for (old = NeterrServerHash[hashval]; old != NULL; old = old->Next)
	if (old->ElanId == elanId)
	    break;
    
    /* remove "old" server from hash table */
    if (old != NULL)
    {
	if (old->Prev)
	    old->Prev->Next = old->Next;
	else
	    NeterrServerHash[hashval] = old->Next;
	if (old->Next)
	    old->Next->Prev = old->Prev;
    }

    /* insert "new" server into hash table */
    if ((server->Next = NeterrServerHash[hashval]) != NULL)
	server->Next->Prev = server;
    server->Prev = NULL;
    NeterrServerHash[hashval] = server;

    kmutex_unlock (&NeterrServerLock);

    if (old != NULL)
	DereferenceNeterrServer (old);
    
    return (ESUCCESS);
}

int
AddNeterrServerSyscall (int elanId, void *addrp, void *namep, char *unused)
{
    struct sockaddr_in addr;
    char              *name;
    int                error;
    int                nob;

    /* Sanity check the supplied elanId argument */
    if (elanId < 0)
	return ( set_errno(EINVAL) );

    KMEM_ALLOC (name, caddr_t, SYS_NMLN, TRUE);
    
    if (copyin ((caddr_t) addrp, (caddr_t) &addr, sizeof (addr)) ||
	copyinstr ((caddr_t) namep, name, SYS_NMLN, &nob))
    {
	error = EFAULT;
    }
    else
    {
	PRINTF2 (DBG_DEVICE, DBG_NETERR, "AddNeterrServer: '%s' at elanid %d\n", name, elanId);

	error = AddNeterrServer (elanId, &addr, name);
    }
    KMEM_FREE (name, SYS_NMLN);

    return (error ? set_errno(error) : ESUCCESS);
}


#if defined(DIGITAL_UNIX)
static int
CallNeterrServer (NETERR_SERVER *server, NETERR_MSG *msg)
{
    cred_t	   *cr = crget();
    struct rpc_err  rpcerr;
    extern cred_t  *kcred;
    struct timeval  wait;
    enum clnt_stat  rc;
    int		    status;
    CLIENT         *clnt;
    int             error;

    PRINTF4 (DBG_DEVICE, DBG_NETRPC, "CallNeterrServer(%s) - family=%d port=%d addr=%08x\n", server->Name,
	     server->Addr.sin_family, server->Addr.sin_port, server->Addr.sin_addr.s_addr);

    if ((clnt = clntkudp_create (&server->Addr, (struct sockaddr_in *)0, NETERR_PROGRAM, NETERR_VERSION, 1, cr)) == NULL)
    {
	PRINTF1 (DBG_DEVICE, DBG_NETRPC, "CallNeterrServer(%s): clntkudp_create error\n", server->Name);

	return (ENOMEM);
    }
    
    wait.tv_sec  = NETERR_RPC_TIMEOUT;
    wait.tv_usec = 0;
    
    PRINTF2 (DBG_DEVICE, DBG_NETRPC, "CallNeterrServer(%s): CLNT_CALL timeout = %d\n", server->Name, NETERR_RPC_TIMEOUT);
   
    rc = CLNT_CALL(clnt, NETERR_FIXUP_RPC, xdr_neterr_msg, (void *)msg, xdr_int, (void *) &status, wait);

    PRINTF3 (DBG_DEVICE, DBG_NETRPC, "CallNeterrServer(%s): CLNT_CALL -> %d (%s)\n", server->Name, rc, clnt_sperrno(rc));;

    switch (rc)
    {
    case RPC_SUCCESS:
	break;

    case RPC_INTR:
	status = EINTR;
	break;

    case RPC_TIMEDOUT:
	status = ETIMEDOUT;
	break;

    default:
	printf ("CallNeterrServer(%s): %s\n", server->Name, clnt_sperrno(status));
	status = ENOENT;
	break;
    }

    CLNT_DESTROY(clnt);

    crfree(cr);
    
    ASSERT(rc == RPC_SUCCESS || status != 0);

    PRINTF2 (DBG_DEVICE, DBG_NETRPC, "CallNeterrServer(%s): status=%d\n", server->Name, status);

    return (status);
}
#endif

#if defined(LINUX)

#define xdrsize(type) ((sizeof(type) + 3) >> 2)

static int
xdr_error(struct rpc_rqst *req, u32 *p, void *dummy)
{
    return -EIO;
}

static int
xdr_decode_int(struct rpc_rqst *req, u32 *p, int *res)
{ 
    *res = ntohl(*p++);
    return 0;
}

#define XDR_capability_sz ((12 + BT_BITOUL(ELAN3_MAX_VPS)) * sizeof (u32))

static int
xdr_encode_capability(u32 *p, ELAN_CAPABILITY *cap)
{
    u32 *pp = p;

    /* basic xdr unit is u32 - for opaque types we must round up to that */
    memcpy(p, &cap->cap_userkey, sizeof(cap->cap_userkey));
    p += xdrsize(cap->cap_userkey);

    *p++ = htonl(cap->cap_version);
    ((u16 *) (p++))[1] = htons(cap->cap_type);
    *p++ = htonl(cap->cap_lowcontext);
    *p++ = htonl(cap->cap_highcontext);
    *p++ = htonl(cap->cap_mycontext);
    *p++ = htonl(cap->cap_lownode);
    *p++ = htonl(cap->cap_highnode);
    *p++ = htonl(cap->cap_railmask);

    memcpy(p, &cap->cap_bitmap[0], sizeof(cap->cap_bitmap));
    p += xdrsize(cap->cap_bitmap);

    ASSERT (((unsigned long) p - (unsigned long) pp) == XDR_capability_sz);

    return (p - pp);
}


#define XDR_neterr_sz	(((1 + 5) * sizeof (u32)) + (2*XDR_capability_sz))

static int
xdr_encode_neterr_msg(struct rpc_rqst *req, u32 *p, NETERR_MSG *msg)
{
    u32 *pp = p;

    *p++ = htonl(msg->Rail);

    p += xdr_encode_capability(p, &msg->SrcCapability);
    p += xdr_encode_capability(p, &msg->DstCapability);

    *p++ = htonl(msg->DstProcess);
    *p++ = htonl(msg->CookieAddr);
    *p++ = htonl(msg->CookieVProc);
    *p++ = htonl(msg->NextCookie);
    *p++ = htonl(msg->WaitForEop);

    ASSERT (((unsigned long) p - (unsigned long) pp) == XDR_neterr_sz);

    req->rq_slen = xdr_adjust_iovec(req->rq_svec, p);

    return 0;
}

static struct rpc_procinfo neterr_procedures[2] = 
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#	define RPC_ID_NULL 	"neterr_null"
#	define RPC_ID_FIXUP_RPC	"neterr_fixup_rpc"
#else
#	define RPC_ID_NULL	NETERR_NULL_RPC
#	define RPC_ID_FIXUP_RPC	NETERR_FIXUP_RPC
#endif
    { 	
	RPC_ID_NULL,			/* procedure name or number*/
    	(kxdrproc_t) xdr_error,		/* xdr encode fun */
        (kxdrproc_t) xdr_error, 	/* xdr decode fun */
	0, 				/* req buffer size */
	0,				/* call count */
    },
    { 	
	RPC_ID_FIXUP_RPC,
        (kxdrproc_t) xdr_encode_neterr_msg,
        (kxdrproc_t) xdr_decode_int,
	XDR_neterr_sz,
	0,			
    },
};

static struct rpc_version neterr_version1 = 
{
    1, 			/* version */
    2, 			/* number of procedures */
    neterr_procedures	/* procedures */
};

static struct rpc_version *neterr_version[] = 
{
    NULL,
    &neterr_version1,
};

static struct rpc_stat neterr_stats;

static struct rpc_program neterr_program = 
{
    NETERR_SERVICE,
    NETERR_PROGRAM,
    sizeof(neterr_version)/sizeof(neterr_version[0]),
    neterr_version,
    &neterr_stats,
};

static int
CallNeterrServer (NETERR_SERVER *server, NETERR_MSG *msg)
{
    struct rpc_xprt   *xprt;
    struct rpc_clnt   *clnt;
    struct rpc_timeout to;
    int                rc, status;
    
    PRINTF (DBG_DEVICE, DBG_NETRPC, "CallNeterrServer(%s)\n", server->Name);

    xprt_set_timeout(&to, 1, NETERR_RPC_TIMEOUT * HZ);

    if ((xprt = xprt_create_proto(IPPROTO_UDP, &server->Addr, &to)) == NULL)
    {
	PRINTF (DBG_DEVICE, DBG_NETRPC, "CallNeterrServer(%s) xprt_create_proto failed\n", server->Name);
	return EFAIL;
    }

    if ((clnt = rpc_create_client(xprt, server->Name, &neterr_program, NETERR_VERSION, RPC_AUTH_NULL)) == NULL)
    {
	PRINTF (DBG_DEVICE, DBG_NETRPC, "CallNeterrServer(%s) rpc_create_client failed\n", server->Name);
	xprt_destroy (xprt);
	
	return EFAIL;
    }

    clnt->cl_softrtry = 1;
    clnt->cl_chatty   = 0;
    clnt->cl_oneshot  = 1;
    clnt->cl_intr     = 0;

    if ((rc = rpc_call(clnt, NETERR_FIXUP_RPC, msg, &status, 0)) < 0)
    {
	/* RPC error has occured - determine whether we should retry */

	status = ETIMEDOUT;
    }

    PRINTF (DBG_DEVICE, DBG_NETRPC, "CallNeterrServer(%s): -> %d\n", server->Name, status);

    return (status);
}

#endif /* defined(LINUX) */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
