/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "$Id: route_table.c,v 1.23 2003/09/24 13:57:25 david Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/os/route_table.c,v $ */

#include <qsnet/kernel.h>

#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elanvp.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanctxt.h>
#include <elan3/elandebug.h>

static sdramaddr_t
AllocateLargeRoute (ELAN3_DEV *dev, ELAN3_ROUTE_TABLE *tbl, int ctxnum, E3_uint64 *smallRoute)
{
    int		  bit = -1;
    ELAN3_ROUTES  *rent;
    unsigned long flags;
    
    spin_lock_irqsave (&tbl->Lock, flags);
    
    for (rent = tbl->LargeRoutes; rent; rent = rent->Next)
    {
	if ((bit = bt_freebit (rent->Bitmap, NROUTES_PER_BLOCK)) != -1)
	    break;
    }
    
    if (bit == -1)						/* No spare entries in large routes */
    {								/* so allocate a new page */
	PRINTF0 (DBG_DEVICE, DBG_VP, "AllocateLargeRoute: allocate route entries\n");
	
	spin_unlock_irqrestore (&tbl->Lock, flags);

	KMEM_ZALLOC(rent, ELAN3_ROUTES *, sizeof (ELAN3_ROUTES), TRUE);
	
	if (rent == (ELAN3_ROUTES *) NULL)
	    return ((sdramaddr_t) 0);
	
	rent->Routes = elan3_sdram_alloc (dev, PAGESIZE);
	if (rent->Routes == (sdramaddr_t) 0)
	{
	    KMEM_FREE (rent, sizeof (ELAN3_ROUTES));
	    return ((sdramaddr_t) 0);
	}

	spin_lock_irqsave (&tbl->Lock, flags);

	/* Add to list of large routes */
	rent->Next       = tbl->LargeRoutes;
	tbl->LargeRoutes = rent;

	/* and use entry 0 */
	bit = 0;
    }
    
    /* Set the bit in the bitmap to mark this route as allocated */
    BT_SET (rent->Bitmap, bit);
    
    /* And generate the small route pointer and the pointer to the large routes */
    (*smallRoute) = BIG_ROUTE_PTR(rent->Routes + (bit*NBYTES_PER_LARGE_ROUTE), ctxnum);

    PRINTF4 (DBG_DEVICE, DBG_VP, "AllocateLargeRoute: rent %p using entry %d at %lx with route pointer %llx\n",
	     rent, bit, rent->Routes + (bit * NBYTES_PER_LARGE_ROUTE), (long long) (*smallRoute));

    /* Invalidate the large route */
    elan3_sdram_zeroq_sdram (dev, rent->Routes + (bit * NBYTES_PER_LARGE_ROUTE), NBYTES_PER_LARGE_ROUTE);

    spin_unlock_irqrestore (&tbl->Lock, flags);

    return (rent->Routes + (bit * NBYTES_PER_LARGE_ROUTE));
}

static void
FreeLargeRoute (ELAN3_DEV *dev, ELAN3_ROUTE_TABLE *tbl, E3_uint64 smallRoute)
{
    E3_Addr      addr = (E3_Addr) (smallRoute & ((1ULL << ROUTE_CTXT_SHIFT)-1));
    ELAN3_ROUTES *rent;

    PRINTF1 (DBG_DEVICE, DBG_VP, "FreeLargeRoute: free route %llx\n", (long long) smallRoute);

    ASSERT (SPINLOCK_HELD (&tbl->Lock));

    for (rent = tbl->LargeRoutes; rent; rent = rent->Next)
    {
	if (rent->Routes <= addr && (rent->Routes + ROUTE_BLOCK_SIZE) > addr)
	{
	    int indx = (addr - rent->Routes)/NBYTES_PER_LARGE_ROUTE;
	    
	    PRINTF2 (DBG_DEVICE, DBG_VP, "FreeLargeRoute: rent=%p indx=%d\n", rent, indx);
	    
	    BT_CLEAR(rent->Bitmap, indx);
	    return;
	}
    }

    panic ("elan: FreeLargeRoute - route not found in large route tables");
}

static void
FreeLargeRoutes (ELAN3_DEV *dev, ELAN3_ROUTE_TABLE *tbl)
{
    ELAN3_ROUTES *rent;

    while ((rent = tbl->LargeRoutes) != NULL)
    {
	PRINTF1 (DBG_DEVICE, DBG_VP, "FreeLargeRoutes: free rent %p\n", rent);

	tbl->LargeRoutes = rent->Next;

	elan3_sdram_free (dev, rent->Routes, PAGESIZE);
	
	KMEM_FREE (rent, sizeof(ELAN3_ROUTES));
    }
}

int
GetRoute (ELAN3_DEV *dev, ELAN3_ROUTE_TABLE *tbl, int process, E3_uint16 *flits)
{
    E3_uint64	routeValue;
    sdramaddr_t largeRouteOff;
  
    if (process < 0 || process >= tbl->Size)
	return (EINVAL);

    routeValue = elan3_sdram_readq (dev, tbl->Table + process * NBYTES_PER_SMALL_ROUTE);

    if (routeValue & ROUTE_PTR)
    {
	largeRouteOff = (routeValue & ROUTE_PTR_MASK);
	
	routeValue = elan3_sdram_readq (dev, largeRouteOff + 0);
	flits[0] = routeValue & 0xffff;
	flits[1] = (routeValue >> 16)  & 0xffff;
	flits[2] = (routeValue >> 32)  & 0xffff;
	flits[3] = (routeValue >> 48)  & 0xffff;
	
	routeValue = elan3_sdram_readq (dev, largeRouteOff + 8);
	flits[4] = routeValue & 0xffff;
	flits[5] = (routeValue >> 16)  & 0xffff;
	flits[6] = (routeValue >> 32)  & 0xffff;
	flits[6] = (routeValue >> 48)  & 0xffff;
    }
    else
    {
	flits[0] = routeValue & 0xffff;
	flits[1] = (routeValue >> 16)  & 0xffff;
	flits[2] = (routeValue >> 32)  & 0xffff;
    }

    return (ESUCCESS);
}

ELAN3_ROUTE_TABLE *
AllocateRouteTable (ELAN3_DEV *dev, int size)
{
    ELAN3_ROUTE_TABLE *tbl;

    KMEM_ZALLOC (tbl, ELAN3_ROUTE_TABLE *, sizeof (ELAN3_ROUTE_TABLE), TRUE);

    if (tbl == (ELAN3_ROUTE_TABLE *) NULL)
	return (NULL);
    
    tbl->Size  = size;
    tbl->Table = elan3_sdram_alloc (dev, size*NBYTES_PER_SMALL_ROUTE);

    if (tbl->Table == 0)
    {
	KMEM_FREE (tbl, sizeof (ELAN3_ROUTE_TABLE));
	return (NULL);
    }
    spin_lock_init (&tbl->Lock);

    /* zero the route table */
    elan3_sdram_zeroq_sdram (dev, tbl->Table, size*NBYTES_PER_SMALL_ROUTE);

    return (tbl);
}

void
FreeRouteTable (ELAN3_DEV *dev, ELAN3_ROUTE_TABLE *tbl)
{
    elan3_sdram_free (dev, tbl->Table, tbl->Size*NBYTES_PER_SMALL_ROUTE);

    FreeLargeRoutes (dev, tbl);

    spin_lock_destroy (&tbl->Lock);

    KMEM_FREE (tbl, sizeof (ELAN3_ROUTE_TABLE));
}

int
LoadRoute (ELAN3_DEV *dev, ELAN3_ROUTE_TABLE *tbl, int process, int ctxnum, int nflits, E3_uint16 *flits)
{
    E3_uint64	  routeValue;
    E3_uint64	  largeRouteValue;
    sdramaddr_t   largeRouteOff;
    unsigned long flags;

    if (process < 0 || process >= tbl->Size)
	return (EINVAL);

    PRINTF3 (DBG_DEVICE, DBG_VP, "LoadRoute: table %lx process %d ctxnum %x\n", tbl->Table ,process, ctxnum);

    if (nflits < 4)
    {
	spin_lock_irqsave (&tbl->Lock, flags);

	/* See if we're replacing a "large" route */
	routeValue = elan3_sdram_readq (dev, tbl->Table + process * NBYTES_PER_SMALL_ROUTE);
	if (routeValue & ROUTE_PTR)
	    FreeLargeRoute (dev, tbl, routeValue);

	routeValue = SMALL_ROUTE(flits, ctxnum);

	if ( routeValue &  ROUTE_PTR)
	    PRINTF0 (DBG_DEVICE, DBG_VP, "SHOULD BE  A SMALL ROUTE !!!!!!!\n");

	PRINTF2 (DBG_DEVICE, DBG_VP, "LoadRoute: loading small route %d  %llx\n", process, (long long) routeValue);
	elan3_sdram_writeq (dev, tbl->Table + process * NBYTES_PER_SMALL_ROUTE, routeValue);
    }
    else
    {
	E3_uint64 value0 = BIG_ROUTE0(flits);
	E3_uint64 value1 = BIG_ROUTE1(flits);

	if ((largeRouteOff = AllocateLargeRoute (dev, tbl, ctxnum, &largeRouteValue)) == (sdramaddr_t) 0)
	    return (ENOMEM);

	spin_lock_irqsave (&tbl->Lock, flags);
	    
	routeValue = elan3_sdram_readq (dev, tbl->Table + process * NBYTES_PER_SMALL_ROUTE);

	if ((routeValue & ROUTE_PTR) == 0)
	    elan3_sdram_writeq (dev, tbl->Table + process * NBYTES_PER_SMALL_ROUTE, largeRouteValue);
	else
	{
	    FreeLargeRoute (dev, tbl, largeRouteValue);

	    largeRouteOff   = (routeValue & ROUTE_PTR_MASK);
	}

	PRINTF3 (DBG_DEVICE, DBG_VP, "LoadRoute: loading large route %d - %llx %llx\n", process, 
		 (long long) value0, (long long) value1);

	elan3_sdram_writeq (dev, largeRouteOff + 0, value0);
	elan3_sdram_writeq (dev, largeRouteOff + 8, value1);
    }

    spin_unlock_irqrestore (&tbl->Lock, flags);
    return (ESUCCESS);
}
void
InvalidateRoute (ELAN3_DEV *dev, ELAN3_ROUTE_TABLE *tbl, int process)
{
    E3_uint64 routeValue;
    unsigned long flags;

    if (process < 0 || process >= tbl->Size)
	return;

    spin_lock_irqsave (&tbl->Lock, flags);

    /* unset ROUTE_VALID
     * does not matter if its short or long, will check when we re-use it
     */
    routeValue = elan3_sdram_readq (dev, tbl->Table + process * NBYTES_PER_SMALL_ROUTE);
    elan3_sdram_writeq (dev, tbl->Table + process * NBYTES_PER_SMALL_ROUTE, (routeValue & (~ROUTE_VALID)));

    spin_unlock_irqrestore (&tbl->Lock, flags);
}
void
ValidateRoute (ELAN3_DEV *dev, ELAN3_ROUTE_TABLE *tbl, int process)
{
    E3_uint64 routeValue;
    unsigned long flags;

    if (process < 0 || process >= tbl->Size)
	return;

    PRINTF2 (DBG_DEVICE, DBG_VP, "ValidateRoute: table %ld process %d  \n", tbl->Table ,process);

    spin_lock_irqsave (&tbl->Lock, flags);

    /* set ROUTE_VALID
     */
    routeValue = elan3_sdram_readq (dev, tbl->Table + process * NBYTES_PER_SMALL_ROUTE);
    elan3_sdram_writeq (dev, tbl->Table + process * NBYTES_PER_SMALL_ROUTE, (routeValue | ROUTE_VALID));

    spin_unlock_irqrestore (&tbl->Lock, flags);
}
void
ClearRoute (ELAN3_DEV *dev, ELAN3_ROUTE_TABLE *tbl, int process)
{
    E3_uint64 routeValue;
    unsigned long flags;

    if (process < 0 || process >= tbl->Size)
	return;

    spin_lock_irqsave (&tbl->Lock, flags);

    PRINTF2 (DBG_DEVICE, DBG_VP, "ClearRoute: table %ld process %d  \n", tbl->Table ,process);

    routeValue = elan3_sdram_readq (dev, tbl->Table + process * NBYTES_PER_SMALL_ROUTE);

    elan3_sdram_writeq (dev, tbl->Table + process * NBYTES_PER_SMALL_ROUTE, 0);

    if (routeValue & ROUTE_PTR)
	FreeLargeRoute (dev, tbl, routeValue);

    spin_unlock_irqrestore (&tbl->Lock, flags);
}

static int
ElanIdEqual (ELAN_POSITION *pos, int level, int ida, int idb)
{
    int l;

    for (l = pos->pos_levels-1; l >= level; l--)
    {
	ida /= pos->pos_arity[l];
	idb /= pos->pos_arity[l];
    }
	
    return (ida == idb);
}

static int
RouteDown (ELAN_POSITION *pos, int level, int elanid)
{
    int l;

    for (l = (pos->pos_levels - 1); level < pos->pos_levels - 1; level++, l--)
    {	
	if (  pos->pos_arity[l] )
	    elanid /= pos->pos_arity[l];    
    }
    elanid %= pos->pos_arity[l];

    return elanid;
}

static int
InitPackedAndFlits (u_char *packed, E3_uint16 *flits)
{
    int rb = 0;

    bzero ((caddr_t) packed, MAX_PACKED+4);
    bzero ((caddr_t) flits, MAX_FLITS * sizeof (E3_uint16));

    /* Initialise 4 bytes of packed, so that the "padding" */
    /* NEVER terminates with 00, as this is recognised as */
    /* as CRC flit */
    packed[rb++] = 0xF;
    packed[rb++] = 0xF;
    packed[rb++] = 0xF;
    packed[rb++] = 0xF;

    return (rb);
}

static int
PackThemRoutesUp (E3_uint16 *flits, u_char *packed, int rb, int timeout, int highPri)
{
    int i, nflits;

    flits[0] |= FIRST_TIMEOUT(timeout);
    if (highPri)
	flits[0] |= FIRST_HIGH_PRI;

    /* round up the number of route bytes to flits */
    /* and subtract the 4 extra we've padded out with */
    nflits = (rb-1)/4;
    
    for (i = nflits; i > 0; i--)
    {
	flits[i] = (packed[rb-1] << 12 |
		    packed[rb-2] << 8  |
		    packed[rb-3] << 4  |
		    packed[rb-4] << 0);
	rb -= 4;
    }
    
    /* Now set the position of the first packed route  */
    /* byte in the 2nd 16 bit flit, taking account of the */
    /* 4 byte padding */
    flits[0] |= FIRST_PACKED (4-rb);
    
    return (nflits+1);
}

int
GenerateRoute (ELAN_POSITION *pos, E3_uint16 *flits, int lowid, int highid, int timeout, int highPri)
{
    int     broadcast  = (lowid != highid);
    int     rb         = 0;
    int	    first      = 1;
    int     noRandom   = 0;
    int     level;
    u_char  packed[MAX_PACKED+4];
    int     numDownLinks;

    rb = InitPackedAndFlits (packed, flits);

    for (level = pos->pos_levels-1;				/* Move up out of the elan */
	 level > 0 && ! (ElanIdEqual (pos, level, pos->pos_nodeid, lowid) &&
			 ElanIdEqual (pos, level, pos->pos_nodeid, highid)); level--)
    {
	noRandom |= pos->pos_random_disabled & (1 << (pos->pos_levels-1-level));
    }
    
    for (level = pos->pos_levels-1;				/* Move up out of the elan */
	 level > 0 && ! (ElanIdEqual (pos, level, pos->pos_nodeid, lowid) &&
			 ElanIdEqual (pos, level, pos->pos_nodeid, highid)); level--)
    {
	numDownLinks = pos->pos_arity [level];
	if (first)
	{
	    if (broadcast || noRandom)
		flits[0] = FIRST_BCAST_TREE;
	    else
	    {
		if (numDownLinks == 4) 
		    flits[0] = FIRST_ADAPTIVE;
		else
		    flits[0] = FIRST_ROUTE( numDownLinks + ( lowid % (8-numDownLinks) ));
	    }
	    first = 0;
	}
	else
	{
	    if (broadcast || noRandom)
		packed[rb++] = PACKED_BCAST_TREE;
	    else 
	    {
		if (numDownLinks == 4) 
		    packed[rb++] = PACKED_ADAPTIVE;
		else
		    packed[rb++] = PACKED_ROUTE( numDownLinks + ( lowid % (8-numDownLinks) ));
	    }		    
	}
    }
    
    while (level < pos->pos_levels)
    {
	int lowRoute  = RouteDown (pos, level, lowid);
	int highRoute = RouteDown (pos, level, highid);

	if (first)
	{
	    if (broadcast)
		flits[0] = FIRST_BCAST(highRoute, lowRoute);
	    else
		flits[0] = FIRST_ROUTE(lowRoute);

	    first = 0;
	}
	else
	{
	    if (broadcast)
	    {
		packed[rb++] = PACKED_BCAST0(highRoute, lowRoute);
		packed[rb++] = PACKED_BCAST1(highRoute, lowRoute);
	    }
	    else
		packed[rb++] = PACKED_ROUTE(lowRoute);
	}
	
	level++;
    }

#ifdef ELITE_REVA_SUPPORTED
    if (broadcast && (pos->pos_levels == 3))
    {
      packed[rb++] = PACKED_BCAST0(0, 0);
      packed[rb++] = PACKED_BCAST1(0, 0);
    }
#endif

    return (PackThemRoutesUp (flits, packed, rb, timeout, highPri));
}

int
GenerateCheckRoute (ELAN_POSITION *pos, E3_uint16 *flits, int level, int adaptive)
{
    int     notfirst = 0;
    int     l, rb;
    u_char  packed[MAX_PACKED+4];

    rb = InitPackedAndFlits (packed, flits);

    for (l = pos->pos_levels-1; l > level; l--)
	if (! notfirst++)
	    flits[0] = adaptive ? FIRST_ADAPTIVE : FIRST_BCAST_TREE;
	else
	    packed[rb++] = adaptive ? PACKED_ADAPTIVE : PACKED_BCAST_TREE;

    if (! notfirst++ ) 
	flits[0] = FIRST_MYLINK;
    else
	packed[rb++] = PACKED_MYLINK;

    for (l++ /* consume mylink */; l < pos->pos_levels; l++)
	if (! notfirst++)
	    flits[0] = FIRST_ROUTE (RouteDown (pos, l, pos->pos_nodeid));
	else
	    packed[rb++] = PACKED_ROUTE (RouteDown (pos, l, pos->pos_nodeid));


    return (PackThemRoutesUp (flits, packed, rb, DEFAULT_ROUTE_TIMEOUT, HIGH_ROUTE_PRIORITY));
}


/*
 * In this case "level" is the number of levels counted from the bottom.
 */
int
GenerateProbeRoute (E3_uint16 *flits, int nodeid, int level, int *linkup, int *linkdown, int adaptive )
{
    int	    first = 1;
    int     i, rb;
    u_char  packed[MAX_PACKED+4];

    rb = InitPackedAndFlits (packed, flits);

    /* Generate "up" routes */
    for (i = 0; i < level; i++)
    {
	if (first)
	    flits[0] = linkup ? FIRST_ROUTE(linkup[i]) : adaptive ? FIRST_ADAPTIVE : FIRST_BCAST_TREE;
	else
	    packed[rb++] = linkup ? PACKED_ROUTE(linkup[i]) : adaptive ? PACKED_ADAPTIVE : PACKED_BCAST_TREE;
	first = 0;
    }

    /* Generate a "to-me" route down */
    if (first)
	flits[0] = FIRST_MYLINK;
    else
	packed[rb++] = PACKED_MYLINK;

    for (i = level-1; i >= 0; i--)
	packed[rb++] =  PACKED_ROUTE(linkdown[i]);

    return (PackThemRoutesUp (flits, packed, rb, DEFAULT_ROUTE_TIMEOUT, HIGH_ROUTE_PRIORITY));
}


/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
