/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: virtual_process.c,v 1.68 2004/06/07 13:50:10 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/os/virtual_process.c,v $*/

#include <qsnet/kernel.h>

#include <elan/elanmod.h>
#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elanvp.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanctxt.h>
#include <elan3/elandebug.h>
#include <elan3/urom_addrs.h>
#include <elan3/thread.h>
#include <elan3/vmseg.h>
#include <elan3/elansyscall.h>

static ELAN3_VPSEG *
InstallSegment (ELAN3_CTXT *ctxt, int process, int entries)
{
    ELAN3_VPSEG **prevSeg, *seg;
    int lastTop = -1;
    int	top	= process + entries-1;

    ASSERT (krwlock_is_write_locked (&ctxt->VpLock));

    for (prevSeg = &ctxt->VpSegs; (seg = (*prevSeg)) != NULL; prevSeg = &seg->Next)
    {
	int thisTop = seg->Process + seg->Entries - 1;

        if (process < seg->Process && (process <= lastTop || top >= seg->Process))
	{
	    /*
	     * Overlaps with last segment, or this one 
	     */
	    return (NULL);
	}
	if (seg->Process > process)
	    break;
	
	lastTop = thisTop;
    }
    
    KMEM_ZALLOC (seg, ELAN3_VPSEG *, sizeof (ELAN3_VPSEG), TRUE);
    
    if (seg == (ELAN3_VPSEG *) NULL)
	return (NULL);
    
    seg->Process = process;
    seg->Entries = entries;


    PRINTF2 (ctxt, DBG_VP, "InstallSegment: add seg %p before %p\n", seg, *prevSeg);

    seg->Next = *prevSeg;
    *prevSeg = seg;

    return (seg);
}

static int 
RemoveSegment (ELAN3_CTXT *ctxt, ELAN3_VPSEG *seg)
{
    ELAN3_VPSEG **prevSeg, *thisSeg;

    ASSERT (krwlock_is_write_locked (&ctxt->VpLock));

    for (prevSeg = &ctxt->VpSegs; (thisSeg = (*prevSeg)) != NULL; prevSeg = &thisSeg->Next)
    {
	if (thisSeg == seg)
	    break;
    }

    if (thisSeg == (ELAN3_VPSEG *) NULL)
	return (EINVAL);


    PRINTF2 (ctxt, DBG_VP, "RemoveSegment: remove seg %p next %p\n", thisSeg, thisSeg->Next);

    *prevSeg = thisSeg->Next;
    
    KMEM_FREE ((caddr_t) seg, sizeof (ELAN3_VPSEG));

    return (ESUCCESS);
}

static ELAN3_VPSEG *
FindSegment (ELAN3_CTXT *ctxt, int low, int high)
{
    ELAN3_VPSEG *seg;

    ASSERT(krwlock_is_locked (&ctxt->VpLock));
    
    for (seg = ctxt->VpSegs; seg; seg = seg->Next)
    {
	if (seg->Process <= low && (seg->Process + seg->Entries) > high)
	    return (seg);
    }

    return ((ELAN3_VPSEG *) NULL);
}

ELAN_LOCATION
ProcessToLocation (ELAN3_CTXT *ctxt, ELAN3_VPSEG *seg, int process, ELAN_CAPABILITY *cap)
{
    ELAN_LOCATION location;
    int           nnodes,nctxs;
    int           node,ctx,i;

    ASSERT(krwlock_is_locked (&ctxt->VpLock));

    location.loc_node    = ELAN3_INVALID_NODE;
    location.loc_context = -1;

    PRINTF3 (ctxt, DBG_VP, "ProcessToLocation: process %d seg %p cap %p\n", process, seg, cap);

    if (seg == NULL)
	seg = FindSegment (ctxt, process, process);

    if (!seg || (seg->Type != ELAN3_VPSEG_P2P))
	return (location);

    cap    = &seg->SegCapability;
    nnodes = ELAN_CAP_NUM_NODES (cap);
    nctxs  = ELAN_CAP_NUM_CONTEXTS (cap);

    switch (seg->SegCapability.cap_type & ELAN_CAP_TYPE_MASK)
    {
    case ELAN_CAP_TYPE_BLOCK:
    {
	int entries = ELAN_CAP_ENTRIES(cap);

	for (node = 0, i = 0; node < nnodes && i < entries; node++)
	{
	    for (ctx = 0; ctx < nctxs && i < entries; ctx++)
	    {
		if (( seg->SegCapability.cap_type & ELAN_CAP_TYPE_NO_BITMAP) || BT_TEST (seg->SegCapability.cap_bitmap, ctx + (node * nctxs)))
		{
		    if (i++ == (process - seg->Process))
		    { 
			location.loc_node    = seg->SegCapability.cap_lownode    + node;
			location.loc_context = seg->SegCapability.cap_lowcontext + ctx;
			goto found;
		    }
		}
	    }
	}
	break;
    }
    case ELAN_CAP_TYPE_CYCLIC:
    {
	int entries = ELAN_CAP_ENTRIES(cap);

	for (ctx = 0, i = 0; ctx < nctxs && i < entries; ctx++)
	{
	    for (node = 0; node < nnodes && i < entries; node++)
	    {
		if ((seg->SegCapability.cap_type & ELAN_CAP_TYPE_NO_BITMAP) || BT_TEST (seg->SegCapability.cap_bitmap, node + (ctx * nnodes)))
		{		    		    
		    if (i++ ==  (process - seg->Process))
		    { 
			location.loc_node    = seg->SegCapability.cap_lownode    + node;
			location.loc_context = seg->SegCapability.cap_lowcontext + ctx;
			goto found;
		    }
		}
	    }
	}
	break;	
    }
    default:
	break;
    }
       
 found:
    
    PRINTF3 (ctxt, DBG_VP, "ProcessToLocation: process %d -> Node %d Context %d\n", process, location.loc_node,  location.loc_context);

    if (cap != NULL)
    {
	bcopy ((caddr_t) &seg->SegCapability, (caddr_t) cap, sizeof (ELAN_CAPABILITY));
	cap->cap_mycontext = location.loc_context;
    }

    return (location);
}

int
LocationToProcess (ELAN3_CTXT *ctxt, ELAN3_VPSEG *seg, ELAN_LOCATION loc, ELAN_CAPABILITY *cap)
{
    int nnodes,nctxs;
    int node,ctx,i;

    if (seg == NULL)
	return ELAN3_INVALID_PROCESS;

    if (!seg || (seg->Type != ELAN3_VPSEG_P2P))
	return ELAN3_INVALID_PROCESS;

    nnodes = cap->cap_highnode - cap->cap_lownode + 1;
    nctxs  = cap->cap_highcontext - cap->cap_lowcontext + 1;

    switch (cap->cap_type & ELAN_CAP_TYPE_MASK)
    {
    case ELAN_CAP_TYPE_BLOCK:
    {
	int entries = ELAN_CAP_ENTRIES(cap);

	for (node = 0, i = 0; node < nnodes && i < entries; node++)
	{
	    for (ctx = 0; ctx < nctxs && i < entries; ctx++)
	    {
		if ((cap->cap_type & ELAN_CAP_TYPE_NO_BITMAP) || BT_TEST (cap->cap_bitmap, ctx + (node * nctxs)))
		{
		    if ((loc.loc_node    == (cap->cap_lownode + node) ) 
			&& (loc.loc_context == (cap->cap_lowcontext + ctx) ))
		    {
			return (i + seg->Process);
		    }
		    i++;
		}
	    }
	}
	break;
    }	
    case ELAN_CAP_TYPE_CYCLIC:
    {
	int entries = ELAN_CAP_ENTRIES(cap);

	for (ctx = 0, i = 0; ctx < nctxs && i < entries; ctx++)
	{
	    for (node = 0; node < nnodes && i < entries; node++)
	    {
		if ((cap->cap_type & ELAN_CAP_TYPE_NO_BITMAP) || BT_TEST (cap->cap_bitmap, node + (ctx * nnodes)))
		{
		    if ((loc.loc_node   == (cap->cap_lownode + node) ) 
			&& (loc.loc_context == (cap->cap_lowcontext + ctx) ))
		    {
			return (i + seg->Process);
		    }
		    i++;
		    
		}
	    }
	}
	break;
    }	
    default:
	break;
    }
       
    return ELAN3_INVALID_PROCESS;
}

int
elan3_addvp (ELAN3_CTXT *ctxt, int process, ELAN_CAPABILITY *cap)
{
    ELAN3_DEV	     *dev = ctxt->Device;
    ELAN_POSITION    *pos = &ctxt->Position;
    ELAN3_VPSEG       *seg;
    int		      i;
    int		      nodeOff;
    int		      ctxOff;
    int		      nnodes;
    int		      nctxs;
    E3_uint16         flits[MAX_FLITS];
    int               nflits;
    int               entries;

    PRINTF2 (ctxt, DBG_VP, "elan3_addvp: %d -> %s\n", process, CapabilityString (cap));

    entries = ELAN_CAP_ENTRIES(cap);
    if (entries <= 0 || (process + entries) > ELAN3_MAX_VPS)
	return (EINVAL);

    /*
     * Scan the virtual process segment list, to add this entry, and ensure that
     * the ranges don't overlap.
     */
    krwlock_write (&ctxt->VpLock);

    /* check cap. */
    switch (elan3_validate_cap (ctxt->Device, cap, ELAN_USER_P2P))
    {
    case ELAN_CAP_OK:
	/* nothing */
	break;

    case ELAN_CAP_RMS:
	if ( elan_validate_map(cap, cap) != ESUCCESS)
	{
	    krwlock_done (&ctxt->VpLock);
	    return (EINVAL);
	}
	break;

    default:
	krwlock_done (&ctxt->VpLock);
	return (EINVAL);
    }

    if ((seg = InstallSegment (ctxt, process, entries)) == NULL)
    {
	PRINTF0 (ctxt, DBG_VP, "elan3_addvp: failed to find a seg\n");
	krwlock_done (&ctxt->VpLock);
	return (EINVAL);
    }
    
    seg->Type                        = ELAN3_VPSEG_P2P;
    seg->SegCapability               = *cap;
    seg->SegCapability.cap_mycontext = ELAN_CAP_UNINITIALISED;

    PRINTF3 (ctxt, DBG_VP, "elan3_addvp: segment type %x  %d %d\n",
	    seg->SegCapability.cap_type, seg->Process, entries);


    nnodes = cap->cap_highnode - cap->cap_lownode + 1;
    nctxs  = cap->cap_highcontext - cap->cap_lowcontext + 1;

    /* position not determined, so cannot load any routes, the hwtest
     * process must explicitly set it's own routes */
    
    if (!(cap->cap_type & ELAN_CAP_TYPE_HWTEST) && (pos->pos_mode != ELAN_POS_UNKNOWN))
    {
	switch (cap->cap_type & ELAN_CAP_TYPE_MASK)
	{
	case ELAN_CAP_TYPE_BLOCK:
	    for (nodeOff = 0, i = 0; nodeOff < nnodes && i < entries; nodeOff++)
	    {
		for (ctxOff = 0; ctxOff < nctxs && i < entries; ctxOff++)
		{
		    if ((cap->cap_type & ELAN_CAP_TYPE_NO_BITMAP) || BT_TEST (cap->cap_bitmap, ctxOff + (nodeOff * nctxs)))
		    {
			/* Don't load a route if there's no switch and trying to talk to myself */
			if (pos->pos_mode == ELAN_POS_MODE_SWITCHED ||
			    (pos->pos_mode == ELAN_POS_MODE_LOOPBACK && cap->cap_lownode + nodeOff == pos->pos_nodeid) ||
			    (pos->pos_mode == ELAN_POS_MODE_BACKTOBACK && cap->cap_lownode + nodeOff != pos->pos_nodeid))
			{
			    PRINTF3 (ctxt, DBG_VP, "elan3_addvp: virtual process %d -> node %d context %d\n",
				     seg->Process + i, cap->cap_lownode +nodeOff, cap->cap_lowcontext +ctxOff);
			    
			    nflits = GenerateRoute (pos, flits, cap->cap_lownode + nodeOff, cap->cap_lownode + nodeOff,
						    DEFAULT_ROUTE_TIMEOUT, DEFAULT_ROUTE_PRIORITY);
			    


			    LoadRoute (dev, ctxt->RouteTable, seg->Process+i, cap->cap_lowcontext + ctxOff, nflits, flits);  
			}
			
			i++;
		    }
		}
	    }
	    break;
	    
	case ELAN_CAP_TYPE_CYCLIC:
	    for (ctxOff = 0, i = 0; ctxOff < nctxs && i < entries; ctxOff++)
	    {
		for (nodeOff = 0; nodeOff < nnodes && i < entries; nodeOff++)
		{
		    if ((cap->cap_type & ELAN_CAP_TYPE_NO_BITMAP) || BT_TEST (cap->cap_bitmap, nodeOff + (ctxOff * nnodes)))
		    {
			/* Don't load a route if there's no switch and trying to talk to myself */
			if (pos->pos_mode == ELAN_POS_MODE_SWITCHED ||
			    (pos->pos_mode == ELAN_POS_MODE_LOOPBACK && cap->cap_lownode + nodeOff == pos->pos_nodeid) ||
			    (pos->pos_mode == ELAN_POS_MODE_BACKTOBACK && cap->cap_lownode + nodeOff != pos->pos_nodeid))
			{
			    PRINTF3 (ctxt, DBG_VP, "elan3_addvp: virtual process %d -> node %d context %d\n",
				     seg->Process + i, cap->cap_lownode + nodeOff, cap->cap_lowcontext +ctxOff);
			
			    nflits = GenerateRoute (pos, flits, cap->cap_lownode + nodeOff, cap->cap_lownode + nodeOff,
						    DEFAULT_ROUTE_TIMEOUT, DEFAULT_ROUTE_PRIORITY);
			    

			    LoadRoute (dev, ctxt->RouteTable, seg->Process+i, cap->cap_lowcontext +ctxOff, nflits, flits);  
			} 
			i++;		    
		    }
		}
	    }
	    break;	
	default:
	    break;
	}
    }
  
    krwlock_done (&ctxt->VpLock);

    return (ESUCCESS);
}

int
elan3_removevp (ELAN3_CTXT *ctxt, int process)
{
    ELAN3_VPSEG *seg;
    ELAN3_VPSEG *next;
    int		i;

    krwlock_write (&ctxt->VpLock);

    PRINTF1 (ctxt, DBG_VP, "elan3_removevp: remove process %d\n", process);

    if (process == ELAN3_INVALID_PROCESS)
	seg = ctxt->VpSegs;
    else
	seg = FindSegment (ctxt, process, process);

    if (seg == (ELAN3_VPSEG *) NULL)
    {
	krwlock_done (&ctxt->VpLock);
	return (EINVAL);
    }
    
    do {
	PRINTF3 (ctxt, DBG_VP, "elan3_removevp: segment is %p [%x,%x]\n",
		 seg, seg->Process, seg->Process+seg->Entries);

	for (i = 0; i < seg->Entries; i++)
	    ClearRoute (ctxt->Device, ctxt->RouteTable, seg->Process+i);

        /* get Next pointer value before structure is free'd */
        next = seg->Next;	
	RemoveSegment (ctxt, seg);

    } while (process == ELAN3_INVALID_PROCESS && (seg = next) != NULL);
    
    krwlock_done (&ctxt->VpLock);

    return (ESUCCESS);
}

int
elan3_addbcastvp (ELAN3_CTXT *ctxt, int process, int lowProc, int highProc)
{
    ELAN_POSITION *pos = &ctxt->Position;
    ELAN3_VPSEG    *seg;
    ELAN3_VPSEG    *aseg;
    int            virtualProcess;
    E3_uint64	   routeValue;

    PRINTF3 (ctxt, DBG_VP, "elan3_addbcastvp: process %d [%d,%d]\n", process, lowProc, highProc);

    if (lowProc > highProc || pos->pos_mode != ELAN_POS_MODE_SWITCHED)
	return (EINVAL);
    
    krwlock_write (&ctxt->VpLock);

    if ((aseg = FindSegment (ctxt, lowProc, highProc)) == NULL || (aseg->Type != ELAN3_VPSEG_P2P))
    {
	PRINTF2 (ctxt, DBG_VP, "elan3_addbcastvp: process [%d,%d] does not map to p2p segment\n", lowProc, highProc);
	krwlock_done (&ctxt->VpLock);
	return (EINVAL);
    }

    /* check aseg->SegCapability */    
    switch (elan3_validate_cap (ctxt->Device, &aseg->SegCapability, ELAN_USER_BROADCAST))
    {
    case ELAN_CAP_OK:
	/* nothing */
	break;
	
    case ELAN_CAP_RMS:
	if ( elan_validate_map(&ctxt->Capability, &aseg->SegCapability) != ESUCCESS )
	{
	    krwlock_done (&ctxt->VpLock);
	    return (EINVAL);
	}
	break;

    default:
	krwlock_done (&ctxt->VpLock);
	return (EINVAL);
    }

    if ( ProcessToLocation (ctxt, aseg, lowProc,  NULL).loc_context != 
	 ProcessToLocation (ctxt, aseg, highProc, NULL).loc_context)
    {
	PRINTF2 (ctxt, DBG_VP, "elan3_addbcastvp: process [%d,%d] does not map to single context\n", lowProc, highProc);
	krwlock_done (&ctxt->VpLock);
	return (EINVAL);
    }
    
    if ((seg = InstallSegment (ctxt, process, 1)) == NULL)
    {
	krwlock_done (&ctxt->VpLock);
	return (EINVAL);
    }

    seg->Type        = ELAN3_VPSEG_BROADCAST;
    seg->SegLowProc  = lowProc;
    seg->SegHighProc = highProc;

    PRINTF4 (ctxt, DBG_VP, "elan3_addbcastvp: installed seg %p Type %d LowProc %d HighProc %d\n",
	    seg, seg->Type, seg->SegLowProc, seg->SegHighProc);

    for (virtualProcess = lowProc; virtualProcess <= highProc; virtualProcess++)
    {
	if (virtualProcess < 0 || virtualProcess >= ctxt->RouteTable->Size)
	    routeValue = 0;
	else
	    routeValue = elan3_sdram_readq ( ctxt->Device, ctxt->RouteTable->Table + virtualProcess * NBYTES_PER_SMALL_ROUTE);
	
	if (! (routeValue & ROUTE_VALID))
	{
	    PRINTF2 (ctxt, DBG_VP, "loadvp[%x]: broadcast %x not valid\n", 
		     ctxt->Capability.cap_mycontext, virtualProcess);
	    break;
	}
    }
	    
    if (virtualProcess > highProc)			/* All vps now present */
    {						/* so load up broadcast route */
	E3_uint16     flits[MAX_FLITS];
	ELAN_LOCATION low    = ProcessToLocation (ctxt, aseg, lowProc,   NULL);
	ELAN_LOCATION high   = ProcessToLocation (ctxt, aseg, highProc,  NULL);
	int           nflits = GenerateRoute (pos, flits, low.loc_node, high.loc_node, DEFAULT_ROUTE_TIMEOUT, DEFAULT_ROUTE_PRIORITY);
	
	PRINTF6 (ctxt, DBG_VP, "loadvp[%x]: broadcast %d -> %x.%x [%x.%x]\n", ctxt->Capability.cap_mycontext,
		 seg->Process, low.loc_node, high.loc_node, 
		 low.loc_context, high.loc_context);
	
	LoadRoute ( ctxt->Device, ctxt->RouteTable, seg->Process, low.loc_context, nflits, flits);
    }

    krwlock_done (&ctxt->VpLock);

    return (ESUCCESS);
}

int
elan3_process (ELAN3_CTXT *ctxt)
{
    int           res = ELAN3_INVALID_PROCESS;
    ELAN3_VPSEG   *seg;
    ELAN_LOCATION loc;

    krwlock_write (&ctxt->VpLock);

    loc.loc_node    = ctxt->Position.pos_nodeid;
    loc.loc_context = ctxt->Capability.cap_mycontext;

    for (seg = ctxt->VpSegs ; seg; seg = seg->Next)
    {
	if (seg->Type == ELAN3_VPSEG_P2P &&
	    seg->SegCapability.cap_lowcontext  <= ctxt->Capability.cap_mycontext &&
	    seg->SegCapability.cap_highcontext >= ctxt->Capability.cap_mycontext &&
	    seg->SegCapability.cap_lownode     <= ctxt->Position.pos_nodeid &&
	    seg->SegCapability.cap_highnode    >= ctxt->Position.pos_nodeid)
	{
	    if ((res=LocationToProcess (ctxt,seg,loc,&ctxt->Capability)) != ELAN3_INVALID_PROCESS)
	    {
		 krwlock_done (&ctxt->VpLock);
		 return res;
	    }
	}
    }

    krwlock_done (&ctxt->VpLock);

    return (res);
}

int
elan3_check_route (ELAN3_CTXT *ctxt, int process, E3_uint16 *flits, E3_uint32 *routeError)
{
    PRINTF5 (ctxt, DBG_VP, "elan3_check_route: vp=%d flits=%04x %04x %04x %04x\n",
	     process, flits[0], flits[1], flits[2], flits[3]);
    PRINTF4 (ctxt, DBG_VP, "                            %04x %04x %04x %04x\n",
	     flits[4], flits[5], flits[6], flits[7]);

    krwlock_read (&ctxt->VpLock);
    *routeError=elan3_route_check(ctxt,flits,ProcessToLocation (ctxt, NULL, process, NULL).loc_node);
    krwlock_done (&ctxt->VpLock);

    return (ESUCCESS); /* the call is a success tho the errorcode may be set */
}

int
elan3_load_route (ELAN3_CTXT *ctxt, int process, E3_uint16 *flits)
{
    ELAN3_VPSEG *seg;
    int		res = 0;
    int	        nflits;
    int         err;

    PRINTF5 (ctxt, DBG_VP, "elan3_load_route: vp=%d flits=%04x %04x %04x %04x\n",
	     process, flits[0], flits[1], flits[2], flits[3]);
    PRINTF4 (ctxt, DBG_VP, "                            %04x %04x %04x %04x\n",
	     flits[4], flits[5], flits[6], flits[7]);

    krwlock_write (&ctxt->VpLock);

    /* check the route is valid */
    if (!(ctxt->Capability.cap_type & ELAN_CAP_TYPE_HWTEST))
    {
	/* must have already attached to define my context number */
	if (ctxt->Capability.cap_mycontext == ELAN_CAP_UNINITIALISED)
	{
	    krwlock_done (&ctxt->VpLock);
	    return (EINVAL);
	}

	if ((err=elan3_route_check(ctxt,flits,ProcessToLocation (ctxt, NULL, process, NULL).loc_node)) != ELAN3_ROUTE_SUCCESS)
	{
	    krwlock_done (&ctxt->VpLock);
	    return (EINVAL);
	}
    }

    if ((seg = FindSegment (ctxt, process, process)) == NULL || seg->Type != ELAN3_VPSEG_P2P)
    {
	krwlock_done (&ctxt->VpLock);
	return (EINVAL);
    }

    /* Calculate number of flits in this route */
    for (nflits = 0; nflits < MAX_FLITS && flits[nflits]; nflits++)
	;
    
    res = LoadRoute (ctxt->Device, ctxt->RouteTable, process, ProcessToLocation (ctxt, seg, process, NULL).loc_context, nflits, flits);

    krwlock_done (&ctxt->VpLock);

    return (res);
}

int
elan3_get_route (ELAN3_CTXT *ctxt, int process, E3_uint16 *flits)
{
    ELAN3_VPSEG *seg;
    int		res = 0;

    PRINTF1 (ctxt, DBG_VP, "elan3_get_route: vp=%d \n",  process);

    krwlock_write (&ctxt->VpLock);

    if (ctxt->RouteTable == NULL)  /* is there a route table */
    {
	krwlock_done (&ctxt->VpLock);
	return (EINVAL);
    }

    if ((seg = FindSegment (ctxt, process, process)) != NULL && seg->Type != ELAN3_VPSEG_P2P)
    {
	krwlock_done (&ctxt->VpLock);
	return (EINVAL);
    }
    
    if (seg == NULL)
    {
	krwlock_done (&ctxt->VpLock);
	return (EINVAL);
    }
    
    res = GetRoute (ctxt->Device, ctxt->RouteTable, process, flits);
    
    krwlock_done (&ctxt->VpLock);

    return (res);
}

int
elan3_reset_route (ELAN3_CTXT *ctxt, int process)
{
    E3_uint16     flits[MAX_FLITS];

    PRINTF1 (ctxt, DBG_VP, "elan3_reset_route: vp=%d \n",  process);
 
    GenerateRoute (&ctxt->Position, flits, process, process, DEFAULT_ROUTE_TIMEOUT, DEFAULT_ROUTE_PRIORITY);
    
    return elan3_load_route(ctxt,process,flits);
}

int
ResolveVirtualProcess (ELAN3_CTXT *ctxt, int process)
{
    E3_uint16	  flits[MAX_FLITS];
    ELAN3_DEV     *dev = ctxt->Device;
    int		  res = ESUCCESS;
    ELAN3_VPSEG   *seg;
    ELAN3_VPSEG   *aseg;
    E3_uint64	  routeValue;

    krwlock_read (&ctxt->VpLock);

    PRINTF1 (ctxt, DBG_VP, "ResolveVirtualProcess: vp=%d \n",  process);

    if (ctxt->RouteTable == NULL || process < 0 || process >= ctxt->RouteTable->Size)
    {
	krwlock_done (&ctxt->VpLock);
	return (EINVAL);
    }

    if (! (seg = FindSegment (ctxt, process, process)))
    {
	PRINTF1 (ctxt, DBG_VP, "ResolveVirtualProcess: cannot find segment for virtual process %d\n", process);
	krwlock_done (&ctxt->VpLock);
	return (EINVAL);
    }
    
    /* check cap. */
    switch (elan3_validate_cap (ctxt->Device, &seg->SegCapability, ((seg->Type == ELAN3_VPSEG_P2P) ? ELAN_USER_P2P : ELAN_USER_BROADCAST)))
    {
    case ELAN_CAP_OK:
	/* nothing */
	break;

    case ELAN_CAP_RMS:
	if ( elan_validate_map(&ctxt->Capability, &seg->SegCapability) != ESUCCESS)
	{
	    krwlock_done (&ctxt->VpLock);
	    return (EINVAL);
	}
	break;

    default:
	krwlock_done (&ctxt->VpLock);
	return (EINVAL);
    }

    BumpUserStat (ctxt, LoadVirtualProcess);

    routeValue = elan3_sdram_readq (dev, ctxt->RouteTable->Table + process * NBYTES_PER_SMALL_ROUTE);
    if (routeValue & ROUTE_VALID)				/* Virtual process already */
    {								/* loaded */
	krwlock_done (&ctxt->VpLock);
	return (ESUCCESS);			
    }
    
    switch (seg->Type)
    {
    case ELAN3_VPSEG_P2P:
	switch (seg->SegCapability.cap_type & ELAN_CAP_TYPE_MASK)
	{
	case ELAN_CAP_TYPE_BLOCK:
	case ELAN_CAP_TYPE_CYCLIC:
	    if ((res = elan_validate_map (&ctxt->Capability,&seg->SegCapability)) == ESUCCESS &&
		(res = GetRoute(dev, ctxt->RouteTable ,process,  flits)) == ESUCCESS)
	    {
		if (elan3_route_check(ctxt, flits, ProcessToLocation (ctxt, seg, process, NULL).loc_node))
		    res = EINVAL;
		else
		    ValidateRoute(dev, ctxt->RouteTable, process);
	    }
	    break;
	default:
	    res = EINVAL;
	    break;
	}
	break;

    case ELAN3_VPSEG_BROADCAST:
	/* Find the segment that this broadcast range spans. */
	aseg = FindSegment (ctxt, seg->SegLowProc, seg->SegHighProc);
	
	if (aseg == NULL || (aseg->Type != ELAN3_VPSEG_P2P) || !(aseg->SegCapability.cap_type & ELAN_CAP_TYPE_BROADCASTABLE))
	{
	    PRINTF2 (ctxt, DBG_VP, "resolveVirtualProcess: %d -> EINVAL (%s)\n", process, 
		     (aseg == NULL ? "no segment" : ((seg->Type != ELAN3_VPSEG_P2P) ? "not point to point" :
						     "not broadcastable")));
	    res = EINVAL;
	    break;
	}
	
	switch (aseg->SegCapability.cap_type & ELAN_CAP_TYPE_MASK)
	{
	case ELAN_CAP_TYPE_BLOCK:
	case ELAN_CAP_TYPE_CYCLIC:
	{
	    ELAN_LOCATION lowNode  = ProcessToLocation (ctxt,aseg,seg->SegLowProc  , NULL);
	    ELAN_LOCATION highNode = ProcessToLocation (ctxt,aseg,seg->SegHighProc , NULL);


	    if ((res = elan_validate_map (&ctxt->Capability,&aseg->SegCapability)) == ESUCCESS &&
		(res=GetRoute(dev, ctxt->RouteTable ,process,  flits)) == ESUCCESS)
	    {
		if (elan3_route_broadcast_check(ctxt,flits, lowNode.loc_node , highNode.loc_node ) != ELAN3_ROUTE_SUCCESS )
		    res = EINVAL;
		else
		    ValidateRoute(dev, ctxt->RouteTable, process);
	    }
	    break;
	}

	default:
	    res = EINVAL;
	    break;
	}
    default:
	res  = EINVAL;
	break;
    }

    krwlock_done (&ctxt->VpLock);
    return (res);
}	  

void
UnloadVirtualProcess (ELAN3_CTXT *ctxt, ELAN_CAPABILITY *cap)
{
    ELAN3_DEV        *dev  = ctxt->Device;
    ELAN3_VPSEG      *seg;
    ELAN_CAPABILITY *scap;
    int              i;

    for (seg = ctxt->VpSegs; seg; seg = seg->Next)
    {
	switch (seg->Type)
	{
	case ELAN3_VPSEG_P2P:
	    scap = &seg->SegCapability;
	    
	    if (cap == NULL || ELAN_CAP_MATCH (scap, cap))
	    {
		PRINTF2 (ctxt, DBG_VP, "unloadvp: segment [%x.%x]\n", 
			 seg->Process, seg->Process + seg->Entries-1);
		
		for (i = 0; i < seg->Entries; i++)
		    InvalidateRoute (dev, ctxt->RouteTable, seg->Process+i);
	    }
	    break;

	case ELAN3_VPSEG_BROADCAST:
	    for (i = 0; i < seg->Entries; i++)
	    {
		ELAN3_VPSEG *aseg = FindSegment (ctxt, seg->SegLowProc, seg->SegHighProc);
		
		if (aseg != NULL && ELAN_CAP_MATCH(&aseg->SegCapability, cap))
		{
		    PRINTF1 (ctxt, DBG_VP, "unloadvp: broadcast vp %d\n", seg->Process);
		
		    InvalidateRoute (dev, ctxt->RouteTable, seg->Process+i);
		}
	    }
	}
    }
}

caddr_t
CapabilityString (ELAN_CAPABILITY *cap)
{
#define CAPSTR_LEN	200
#define NCAPSTRS	4
    static char       space[CAPSTR_LEN*NCAPSTRS];
    static int        bufnum;
    static spinlock_t lock;
    static int	      lockinitialised;
    int		      num;
    unsigned long     flags;

    if (! lockinitialised)
    {
	spin_lock_init (&lock);
	lockinitialised = 1;
    }

    spin_lock_irqsave (&lock, flags);
    
    if ((num = ++bufnum) == NCAPSTRS)
	num = bufnum = 0;
    spin_unlock_irqrestore (&lock, flags);

    sprintf (space + (num * CAPSTR_LEN), "%4x %4x %4x %4x %4x %4x %4x [%x.%x.%x.%x]", cap->cap_type,
	     cap->cap_lownode, cap->cap_highnode, 
	     cap->cap_lowcontext, cap->cap_mycontext, cap->cap_highcontext,  ELAN_CAP_ENTRIES(cap),
	     cap->cap_userkey.key_values[0],  cap->cap_userkey.key_values[1],
	     cap->cap_userkey.key_values[2],  cap->cap_userkey.key_values[3]);

    return (space + (num * CAPSTR_LEN));
}


/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
