/*
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: i2c.c,v 1.4 2004/01/07 13:37:45 jon Exp $"
/*      $Source: /cvs/master/quadrics/elan4mod/i2c.c,v $*/
#include <qsnet/kernel.h>

#include <elan4/sdram.h>
#include <elan4/debug.h>
#include <elan4/device.h>
#include <elan4/commands.h>

#include <elan4/i2c.h>
#include <elan4/pci.h>
#include <elan4/ioctl.h>
#include <elan4/registers.h>

#define I2C_POLL_LIMIT		8

static int
i2c_poll_busy (ELAN4_DEV *dev)
{
    int t    = 100;
    int loop = 0;
    volatile unsigned char val;

    /* wait for any led I2C operation to finish */
    while (((val = read_i2c (dev, I2cPortControl)) & I2cCntl_I2cPortBusy) && loop++ < I2C_POLL_LIMIT)
    {
	DELAY (t);
	
	if (t < 500000)
	    t <<= 1;
    }
    if (loop >= I2C_POLL_LIMIT)
    {
	printk ("elan%d: I2c has timed out waiting for I2cPortBusy to clear!\n", dev->dev_instance);
	printk ("elan%d: I2cPortControl=%x I2cLedBase=%x I2cStatus=%x\n",
		dev->dev_instance, val, read_i2c (dev, I2cLedBase), read_i2c (dev, I2cStatus));
    }

    return val;
}

static int
i2c_poll_stopped (ELAN4_DEV *dev)
{
    int t    = 100;
    int loop = 0;
    unsigned char val=0, newval;

    /* wait for any led I2C operation to finish. Must see it stopped at least twice */
    while (!(((newval = read_i2c (dev, I2cPortControl)) & I2cCntl_I2cStopped) &&
	      (val & I2cCntl_I2cStopped)) &&
	      (loop++ < I2C_POLL_LIMIT))
    {
	DELAY (t);
	
	if (t < 500000)
	    t <<= 1;
	val = newval;
    }

    return val;
}

int
i2c_disable_auto_led_update (ELAN4_DEV *dev)
{
    spin_lock (&dev->dev_i2c_lock);

    if (dev->dev_i2c_led_disabled++ == 0)
    {
	write_i2c (dev, I2cLedBase, read_i2c (dev, I2cLedBase) & ~I2cCntl_I2cUpdatingLedReg);

	if (! (i2c_poll_stopped (dev) & I2cCntl_I2cStopped))
	{
	    write_i2c (dev, I2cLedBase, read_i2c (dev, I2cLedBase) | I2cCntl_I2cUpdatingLedReg);
	    
	    spin_unlock (&dev->dev_i2c_lock);
	    
	    return -EAGAIN;
	}
	
	write_i2c (dev, I2cStatus, read_i2c (dev, I2cStatus) & ~I2cCntl_SampleNewLedValues);
    }

    spin_unlock (&dev->dev_i2c_lock);

    return 0;
}

void
i2c_enable_auto_led_update (ELAN4_DEV *dev)
{
    spin_lock (&dev->dev_i2c_lock);
    if (--dev->dev_i2c_led_disabled == 0)
    {
	write_i2c (dev, I2cLedBase, read_i2c (dev, I2cLedBase) | I2cCntl_I2cUpdatingLedReg);
	write_i2c (dev, I2cStatus, read_i2c (dev, I2cStatus) | I2cCntl_SampleNewLedValues);
    }

    spin_unlock (&dev->dev_i2c_lock);
}

int
i2c_write (ELAN4_DEV *dev, unsigned int address, unsigned int count, unsigned char *data)
{
    int i;

    if (! (i2c_poll_busy (dev) & I2cCntl_I2cStopped))
	return -EAGAIN;
    
    write_i2c (dev, I2cWrData,      I2C_WRITE_ADDR(address));
    write_i2c (dev, I2cPortControl, I2cCntl_I2cPortWrite);
    
    if (i2c_poll_busy (dev) & I2cCntl_I2cPortAccFailed)
	return -ENXIO;

    for (i = 0; i < count; i++)
    {
	write_i2c (dev, I2cWrData, data[i]);
	write_i2c (dev, I2cPortControl, I2cCntl_I2cPortWrite | (i == (count-1) ? I2cCntl_I2cPortGenStopBit : 0));
    }

    return 0;
}

int
i2c_read (ELAN4_DEV *dev, unsigned int address, unsigned int count, unsigned char *data)
{
    int i;

    if (! (i2c_poll_busy (dev) & I2cCntl_I2cStopped))
	return -EAGAIN; /* not idle */ 

    write_i2c (dev, I2cWrData,      I2C_READ_ADDR(address));
    write_i2c (dev, I2cPortControl, I2cCntl_I2cPortWrite);

    if (i2c_poll_busy (dev) & I2cCntl_I2cPortAccFailed)
	return -ENXIO;
    
    for (i = 0; i < count; i++)
    {
	write_i2c (dev, I2cWrData, 0xff);
	write_i2c (dev, I2cPortControl, I2cCntl_I2cPortRead | ((i == count-1) ? I2cCntl_I2cPortGenStopBit : 0));

	i2c_poll_busy (dev);

	data[i] = read_i2c (dev, I2cRdData);
    }

    return 0;
}

int
i2c_writereg (ELAN4_DEV *dev, unsigned int address, unsigned int reg, unsigned int count, unsigned char *data)
{
    int i;

    if (! (i2c_poll_busy (dev) & I2cCntl_I2cStopped))
	return -EAGAIN; /* not idle */ 

    write_i2c (dev, I2cWrData,      I2C_WRITE_ADDR(address));
    write_i2c (dev, I2cPortControl, I2cCntl_I2cPortWrite);

    if (i2c_poll_busy (dev) & I2cCntl_I2cPortAccFailed)
	return -ENXIO;
    
    write_i2c (dev, I2cWrData,      reg);
    write_i2c (dev, I2cPortControl, I2cCntl_I2cPortWrite);

    if (i2c_poll_busy (dev) & I2cCntl_I2cPortAccFailed)
	return -ENXIO;
    
    for (i = 0; i < count; i++)
    {
	write_i2c (dev, I2cWrData, data[i]);
	write_i2c (dev, I2cPortControl, I2cCntl_I2cPortWrite | ((i == count-1) ? I2cCntl_I2cPortGenStopBit : 0));

	if (i2c_poll_busy (dev) & I2cCntl_I2cPortAccFailed)
	    printk (" i2c_writereg: off %d failed\n", i);
    }

    return 0;
}

int
i2c_readreg (ELAN4_DEV *dev, unsigned int address, unsigned int reg, unsigned int count, unsigned char *data)
{
    if (! (i2c_poll_busy (dev) & I2cCntl_I2cStopped))
	return -EAGAIN; /* not idle */ 

    write_i2c (dev, I2cWrData,      I2C_WRITE_ADDR(address));
    write_i2c (dev, I2cPortControl, I2cCntl_I2cPortWrite);

    if (i2c_poll_busy (dev) & I2cCntl_I2cPortAccFailed)
	return -ENXIO;
    
    write_i2c (dev, I2cWrData,      reg);
    write_i2c (dev, I2cPortControl, I2cCntl_I2cPortWrite | I2cCntl_I2cPortGenStopBit);

    if (i2c_poll_busy (dev) & I2cCntl_I2cPortAccFailed)
	return -ENXIO;

    return i2c_read (dev, address, count, data);
}

int
i2c_read_rom (ELAN4_DEV *dev, unsigned int addr, unsigned int len, unsigned char *data)
{
    unsigned int top = addr + len;
    int res;

    if ((res = i2c_disable_auto_led_update (dev)) == 0)
    {
	/* read the rom in chunks that don't span the block boundary */
	while (addr < top)
	{
	    unsigned int thisnob  = top - addr;
	    unsigned int blocknob = I2C_24LC16B_BLOCKSIZE - I2C_24LC16B_BLOCKOFFSET(addr);
	    
	    if (thisnob > blocknob)
		thisnob = blocknob;

	    if ((res = i2c_readreg (dev, I2C_EEPROM_ADDR + I2C_24LC16B_BLOCKADDR(addr),
				    I2C_24LC16B_BLOCKOFFSET(addr), thisnob, data)) < 0)
		break;
	    
	    addr += thisnob;
	    data += thisnob;
	}

	i2c_enable_auto_led_update (dev);
    }
    return res;
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
