/*
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: intcookie.c,v 1.15 2005/02/03 16:24:44 addy Exp $"
/*      $Source: /cvs/master/quadrics/elan4mod/intcookie.c,v $*/

#include <qsnet/kernel.h>

#include <elan4/debug.h>
#include <elan4/types.h>
#include <elan/capability.h>
#include <elan4/intcookie.h>

static INTCOOKIE_TABLE *intcookie_tables;
static spinlock_t	intcookie_table_lock;

/*
 * intcookie_drop_entry:
 *   drop the reference to a cookie held 
 *   by the cookie table
 */
static void
intcookie_drop_entry (INTCOOKIE_ENTRY *ent)
{
    unsigned long flags;

    spin_lock_irqsave (&ent->ent_lock, flags);
    if (--ent->ent_ref != 0)
    {
	ent->ent_fired = ent->ent_cookie;
	kcondvar_wakeupall (&ent->ent_wait, &ent->ent_lock);

	spin_unlock_irqrestore (&ent->ent_lock, flags);
    }
    else
    {
	spin_unlock_irqrestore (&ent->ent_lock, flags);

	spin_lock_destroy (&ent->ent_lock);
	kcondvar_destroy (&ent->ent_wait);

	KMEM_FREE (ent, sizeof (INTCOOKIE_ENTRY));
    }
}

void
intcookie_init()
{
    spin_lock_init (&intcookie_table_lock);
}

void
intcookie_fini()
{
    spin_lock_destroy (&intcookie_table_lock);
}

INTCOOKIE_TABLE *
intcookie_alloc_table (ELAN_CAPABILITY *cap)
{
    INTCOOKIE_TABLE *tbl, *ntbl;
    ELAN_CAPABILITY *ncap;
    
    KMEM_ZALLOC (ntbl, INTCOOKIE_TABLE *, sizeof (INTCOOKIE_TABLE), 1);

    if (ntbl == NULL)
	return (NULL);

    KMEM_ALLOC (ncap, ELAN_CAPABILITY *, ELAN_CAP_SIZE(cap), 1);

    if (ncap == NULL)
    {
	KMEM_FREE (ntbl, sizeof (INTCOOKIE_TABLE));
	return (NULL);
    }

    spin_lock (&intcookie_table_lock);
    
    for (tbl = intcookie_tables; tbl; tbl = tbl->tbl_next)
	if (ELAN_CAP_MATCH (tbl->tbl_cap, cap) && tbl->tbl_cap->cap_mycontext == cap->cap_mycontext)
	    break;
    
    if (tbl != NULL)
	tbl->tbl_ref++;
    else
    {
	spin_lock_init (&ntbl->tbl_lock);

	ntbl->tbl_cap     = ncap;
	ntbl->tbl_ref     = 1;
	ntbl->tbl_entries = NULL;
	
	/* Save supplied cap */
	memcpy (ncap, cap, ELAN_CAP_SIZE(cap));

	if ((ntbl->tbl_next = intcookie_tables) != NULL)
	    intcookie_tables->tbl_prev = ntbl;
	intcookie_tables = ntbl;
	ntbl->tbl_prev = NULL;
    }
    spin_unlock (&intcookie_table_lock);

    if (tbl == NULL)
	return (ntbl);
    else
    {
	KMEM_FREE (ntbl, sizeof (INTCOOKIE_TABLE));
	KMEM_FREE (ncap, ELAN_CAP_SIZE(cap));
	return (tbl);
    }    
}

void
intcookie_free_table (INTCOOKIE_TABLE *tbl)
{
    INTCOOKIE_ENTRY *ent;

    spin_lock (&intcookie_table_lock);
    if (tbl->tbl_ref > 1)
    {
	tbl->tbl_ref--;
	spin_unlock (&intcookie_table_lock);
	return;
    }
    
    if (tbl->tbl_prev)
	tbl->tbl_prev->tbl_next = tbl->tbl_next;
    else
	intcookie_tables = tbl->tbl_next;
    if (tbl->tbl_next)
	tbl->tbl_next->tbl_prev = tbl->tbl_prev;
    
    spin_unlock (&intcookie_table_lock);
    
    /* NOTE - table no longer visible to other threads
     *        no need to aquire tbl_lock */
    while ((ent = tbl->tbl_entries) != NULL)
    {
	if ((tbl->tbl_entries = ent->ent_next) != NULL)
	    ent->ent_next->ent_prev = NULL;
	
	intcookie_drop_entry (ent);
    }
    spin_lock_destroy (&tbl->tbl_lock);

    KMEM_FREE (tbl->tbl_cap, ELAN_CAP_SIZE(tbl->tbl_cap));
    KMEM_FREE (tbl, sizeof (INTCOOKIE_TABLE));
}

int
intcookie_alloc (INTCOOKIE_TABLE *tbl, ELAN4_INTCOOKIE cookie)
{
    INTCOOKIE_ENTRY *ent, *nent;
    unsigned long flags;

    KMEM_ZALLOC (nent, INTCOOKIE_ENTRY *, sizeof (INTCOOKIE_ENTRY), 1);

    if (nent == NULL)
	return (-ENOMEM);
    
    spin_lock_irqsave (&tbl->tbl_lock, flags);
    for (ent = tbl->tbl_entries; ent; ent = ent->ent_next)
	if (ent->ent_cookie == cookie)
	    break;

    if (ent == NULL)
    {
	kcondvar_init (&nent->ent_wait);
	spin_lock_init (&nent->ent_lock);

	nent->ent_ref    = 1;
	nent->ent_cookie = cookie;

	if ((nent->ent_next = tbl->tbl_entries) != NULL)
	    tbl->tbl_entries->ent_prev = nent;
	tbl->tbl_entries = nent;
	nent->ent_prev = NULL;
    }
    spin_unlock_irqrestore (&tbl->tbl_lock, flags);

    if (ent == NULL)
	return (0);
    else
    {
	KMEM_FREE (nent, sizeof (INTCOOKIE_ENTRY));
	return (-EINVAL);
    }
}

int
intcookie_free (INTCOOKIE_TABLE *tbl, ELAN4_INTCOOKIE cookie)
{
    INTCOOKIE_ENTRY *ent;
    unsigned long flags;

    spin_lock_irqsave (&tbl->tbl_lock, flags);
    for (ent = tbl->tbl_entries; ent; ent = ent->ent_next)
	if (ent->ent_cookie == cookie)
	    break;
    
    if (ent == NULL)
    {
	spin_unlock_irqrestore (&tbl->tbl_lock, flags);
	return (-EINVAL);
    }

    if (ent->ent_prev == NULL)
	tbl->tbl_entries = ent->ent_next;
    else
	ent->ent_prev->ent_next = ent->ent_next;

    if (ent->ent_next != NULL)
	ent->ent_next->ent_prev = ent->ent_prev;
    
    spin_unlock_irqrestore (&tbl->tbl_lock, flags);

    intcookie_drop_entry (ent);

    return (0);
}

/*
 * intcookie_fire_cookie:
 *    fire the cookie - this is called from the event interrupt.
 */
int
intcookie_fire (INTCOOKIE_TABLE *tbl, ELAN4_INTCOOKIE cookie)
{
    INTCOOKIE_ENTRY *ent;
    unsigned long flags;

    spin_lock_irqsave (&tbl->tbl_lock, flags);
    for (ent = tbl->tbl_entries; ent; ent = ent->ent_next)
	if (ent->ent_cookie == cookie)
	    break;
    
    if (ent == NULL)
    {
	spin_unlock_irqrestore (&tbl->tbl_lock, flags);
	return (-EINVAL);
    }
	    
    spin_lock (&ent->ent_lock);
    ent->ent_fired = cookie;
    kcondvar_wakeupall (&ent->ent_wait, &ent->ent_lock);
    spin_unlock (&ent->ent_lock);

    spin_unlock_irqrestore (&tbl->tbl_lock, flags);

    return (0);
}    

int
intcookie_fire_cap (ELAN_CAPABILITY *cap, ELAN4_INTCOOKIE cookie)
{
    int res;
    INTCOOKIE_TABLE *tbl;

    spin_lock (&intcookie_table_lock);
    
    for (tbl = intcookie_tables; tbl; tbl = tbl->tbl_next)
	if (ELAN_CAP_MATCH (tbl->tbl_cap, cap) && tbl->tbl_cap->cap_mycontext == cap->cap_mycontext)
	    break;
    
    if (tbl != NULL)
	tbl->tbl_ref++;

    spin_unlock (&intcookie_table_lock);

    /* No matching table found */
    if (tbl == NULL)
	return (-EINVAL);

    /* Fire the correct cookie */
    res = intcookie_fire (tbl, cookie);

    /* Decrement reference count (and free if necessary) */
    intcookie_free_table (tbl);

    return (res);
}

/*
 * intcookie_wait_cookie:
 *    deschedule on a cookie if it has not already fired.
 *    note - if the cookie is removed from the table, then
 *           we free it off when we're woken up.
 */
int
intcookie_wait (INTCOOKIE_TABLE *tbl, ELAN4_INTCOOKIE cookie)
{
    INTCOOKIE_ENTRY *ent;
    unsigned long flags;
    int res;
    
    spin_lock_irqsave (&tbl->tbl_lock, flags);
    for (ent = tbl->tbl_entries; ent; ent = ent->ent_next)
	if (ent->ent_cookie == cookie)
	    break;
    
    if (ent == NULL)
    {
	spin_unlock_irqrestore (&tbl->tbl_lock, flags);
	return (-EINVAL);
    }

    spin_lock (&ent->ent_lock);
    spin_unlock (&tbl->tbl_lock);

    if (ent->ent_fired != 0)
    {
	spin_unlock_irqrestore (&ent->ent_lock, flags);
	return (0);
    }

    ent->ent_ref++;
    kcondvar_waitsig (&ent->ent_wait, &ent->ent_lock, &flags);
    
    res = ent->ent_fired ? 0 : -EINTR;

    if (--ent->ent_ref > 0)
	spin_unlock_irqrestore (&ent->ent_lock, flags);
    else
    {
	spin_unlock_irqrestore (&ent->ent_lock, flags);
	
	spin_lock_destroy (&ent->ent_lock);
	kcondvar_destroy (&ent->ent_wait);

	KMEM_FREE (ent, sizeof (INTCOOKIE_ENTRY));
    }

    return (res);
}

int
intcookie_arm (INTCOOKIE_TABLE *tbl, ELAN4_INTCOOKIE cookie)
{
    INTCOOKIE_ENTRY *ent;
    unsigned long flags;

    spin_lock_irqsave (&tbl->tbl_lock, flags);
    for (ent = tbl->tbl_entries; ent; ent = ent->ent_next)
	if (ent->ent_cookie == cookie)
	    break;
    
    if (ent == NULL)
    {
	spin_unlock_irqrestore (&tbl->tbl_lock, flags);
	return (-EINVAL);
    }
	    
    spin_lock (&ent->ent_lock);
    ent->ent_fired = 0;
    spin_unlock (&ent->ent_lock);

    spin_unlock_irqrestore (&tbl->tbl_lock, flags);

    return (0);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
