/*
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: routetable.c,v 1.17 2005/04/15 12:38:22 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan4mod/routetable.c,v $*/

#include <qsnet/kernel.h>

#include <elan4/sdram.h>
#include <elan4/debug.h>
#include <elan4/device.h>

ELAN4_ROUTE_TABLE *
elan4_alloc_routetable (ELAN4_DEV *dev, unsigned size)
{
    ELAN4_ROUTE_TABLE *tbl;

    KMEM_ZALLOC (tbl, ELAN4_ROUTE_TABLE *, sizeof (ELAN4_ROUTE_TABLE), 1);

    if (tbl == (ELAN4_ROUTE_TABLE *) NULL)
	return (NULL);
    
    tbl->tbl_size    = (size & E4_VPT_SIZE_MASK);
    tbl->tbl_entries = elan4_sdram_alloc (dev, (E4_VPT_MIN_ENTRIES << tbl->tbl_size) * sizeof (E4_VirtualProcessEntry));

    if (tbl->tbl_entries == 0)
    {
	KMEM_FREE (tbl, sizeof (ELAN4_ROUTE_TABLE));
	return ((ELAN4_ROUTE_TABLE *) NULL);
    }

    spin_lock_init (&tbl->tbl_lock);

    /* zero the route table */
    elan4_sdram_zeroq_sdram (dev, tbl->tbl_entries, (E4_VPT_MIN_ENTRIES << tbl->tbl_size) * sizeof (E4_VirtualProcessEntry));

    return (tbl);
}

void
elan4_free_routetable (ELAN4_DEV *dev, ELAN4_ROUTE_TABLE *tbl)
{
    elan4_sdram_free (dev, tbl->tbl_entries, (E4_VPT_MIN_ENTRIES << tbl->tbl_size) * sizeof (E4_VirtualProcessEntry));
    
    spin_lock_destroy (&tbl->tbl_lock);

    KMEM_FREE (tbl, sizeof (ELAN4_ROUTE_TABLE));
}

void
elan4_write_route (ELAN4_DEV *dev, ELAN4_ROUTE_TABLE *tbl, unsigned vp, E4_VirtualProcessEntry *entry)
{
    ASSERT (vp < (E4_VPT_MIN_ENTRIES << tbl->tbl_size));
    
    elan4_sdram_writeq (dev, tbl->tbl_entries + (vp * sizeof (E4_VirtualProcessEntry)) + offsetof (E4_VirtualProcessEntry, Values[1]), entry->Values[1]);
    elan4_sdram_writeq (dev, tbl->tbl_entries + (vp * sizeof (E4_VirtualProcessEntry)) + offsetof (E4_VirtualProcessEntry, Values[0]), entry->Values[0]);
    pioflush_sdram (dev);
}

void
elan4_read_route (ELAN4_DEV *dev, ELAN4_ROUTE_TABLE *tbl, unsigned vp, E4_VirtualProcessEntry *entry)
{
    ASSERT (vp < (E4_VPT_MIN_ENTRIES << tbl->tbl_size));
    
    entry->Values[0] = elan4_sdram_readq (dev, tbl->tbl_entries + (vp * sizeof (E4_VirtualProcessEntry)) + offsetof (E4_VirtualProcessEntry, Values[0]));
    entry->Values[1] = elan4_sdram_readq (dev, tbl->tbl_entries + (vp * sizeof (E4_VirtualProcessEntry)) + offsetof (E4_VirtualProcessEntry, Values[1]));
}

void
elan4_invalidate_route (ELAN4_DEV *dev, ELAN4_ROUTE_TABLE *tbl, unsigned vp)
{
    ASSERT (vp < (E4_VPT_MIN_ENTRIES << tbl->tbl_size));

    elan4_sdram_writeq (dev, tbl->tbl_entries + (vp * sizeof (E4_VirtualProcessEntry)) + offsetof (E4_VirtualProcessEntry, Values[0]), 0);
    elan4_sdram_writeq (dev, tbl->tbl_entries + (vp * sizeof (E4_VirtualProcessEntry)) + offsetof (E4_VirtualProcessEntry, Values[1]), 0);
    pioflush_sdram (dev);
}

static void
pack_them_routes (E4_VirtualProcessEntry *entry, E4_uint16 first, E4_uint8 *packed, unsigned ctx)
{
    E4_uint64 value0 = first;
    E4_uint64 value1 = ROUTE_CTXT_VALUE(ctx);
    E4_uint32 ThirdRouteBCastVal;
    register int i;

    for (i = 0; i < (ROUTE_NUM_PACKED >> 1); i++)
    {
	value0 |= ((E4_uint64) packed[i]) << ((i << 2) + ROUTE_PACKED_OFFSET);
	value1 |= ((E4_uint64) packed[i+(ROUTE_NUM_PACKED >> 1)]) << ((i << 2));
    }

    /* DMA fix for large broadcast route values that fall into the double issue of route value 3 bug. */
    /* NOTE - this is only required when the link is running in Mod45 mode, it could be automatically
     *        disabled when Mod44 is detected */
    
    /* First seach for the alignment type. The bug is only sensitive to an odd bcast aligment on the 3rd word. */
    for (i=4;i<16;i++)
	if (((value0 >> (i*4)) & 0xc) == 4)
	    i++;
    
    if (i == 17)
    {
	ThirdRouteBCastVal = value1 & 0xcccccccc;
	if      (((value1 & 0xfffff0000000ULL) == 0ULL) && (ThirdRouteBCastVal == 0x04444444))
	    value1 |= 0x140000000ULL;
	else if (((value1 & 0xfffffff00000ULL) == 0ULL) && (ThirdRouteBCastVal == 0x00044444))
	    value1 |= 0x1400000ULL;
	else if (((value1 & 0xfffffffff000ULL) == 0ULL) && (ThirdRouteBCastVal == 0x00000444))
	    value1 |= 0x14000ULL;
	else if (((value1 & 0xfffffffffff0ULL) == 0ULL) && (ThirdRouteBCastVal == 0x00000004))
	    value1 |= 0x140ULL;
    }
    
    entry->Values[0] = value0;
    entry->Values[1] = value1;
}

int
elan4_generate_route (ELAN_POSITION *pos, E4_VirtualProcessEntry *route, unsigned ctx, unsigned lowid, unsigned highid, unsigned options)
{
    unsigned int broadcast    = (lowid != highid);
    unsigned int noadaptive   = 0;
    int          padbcast     = 0;
    E4_uint16    first;
    int	         rb;
    E4_uint8     packed[ROUTE_NUM_PACKED];
    int	         level, llink, hlink;

    /* sanity check on lowid highid */
    if (highid < lowid)           return (-EINVAL);
    if (lowid  < 0)               return (-EINVAL);
    if (highid >= pos->pos_nodes) return (-EINVAL);

 regenerate_routes:
    first = 0;
    rb    = 0;

    switch (pos->pos_mode)
    {
    case ELAN_POS_MODE_LOOPBACK:
	if (lowid != highid || lowid != pos->pos_nodeid)
	    return (-EINVAL);
	
	route->Values[0] = FIRST_MYLINK;
	route->Values[1] = ROUTE_CTXT_VALUE (ctx);
	return (0);

    case ELAN_POS_MODE_BACKTOBACK:
	if (lowid != highid || lowid == pos->pos_nodeid)
	    return (-EINVAL);
	
	route->Values[0] = FIRST_MYLINK;
	route->Values[1] = ROUTE_CTXT_VALUE (ctx);
	return (0);

    case ELAN_POS_MODE_SWITCHED:
    {
	unsigned char *arityp  = &pos->pos_arity[pos->pos_levels - 1];
	unsigned int   spanned = *arityp;
	unsigned int   broadcasting = 0;
	
	bzero (packed, sizeof (packed));

	/* XXXX compute noadaptive ? */

	for (level = 0; 
	     level < pos->pos_levels && ! ((pos->pos_nodeid / spanned) == (lowid / spanned) &&
					   (pos->pos_nodeid / spanned) ==  (highid / spanned)); 
	     level++, spanned *= *(--arityp))
	{
	    if (first == 0)
		first = (broadcast || noadaptive) ? FIRST_BCAST_TREE : FIRST_ADAPTIVE;
	    else if (broadcast && padbcast)
	    {
		padbcast = 0;
		packed[rb++] = PACKED_BCAST0(4, 4);
		packed[rb++] = PACKED_BCAST1(4, 4);
	    }
	    else
		packed[rb++] = (broadcast || noadaptive) ? PACKED_BCAST_TREE : PACKED_ADAPTIVE;    
	}

	while (level >= 0)
	{
	    spanned /= *arityp;
	    
	    llink = (lowid  / spanned) % *arityp;
	    hlink = (highid / spanned) % *arityp;
	    
	    if (llink != hlink || broadcasting)
	    {
		broadcasting = 1;
		
		if (first == 0)
		    first = FIRST_BCAST (hlink, llink);
		else
		{
		    packed[rb++] = PACKED_BCAST0(hlink, llink);
		    
		    if ((rb % 4) == 0 && PACKED_BCAST1(hlink, llink) == 0)
		    {
			padbcast = 1;
			goto regenerate_routes;
		    }
		    
		    packed[rb++] = PACKED_BCAST1(hlink, llink);
		}
	    }
	    else
	    {
		if (first == 0)
		    first = FIRST_ROUTE(llink);
		else
		    packed[rb++] = PACKED_ROUTE(llink);
	    }
	    
	    level--;
	    arityp++;
	}

	pack_them_routes (route, first | (options & FIRST_OPTIONS_MASK), packed, ctx);
	return (0);
    }
    }

    return (-EINVAL);
}

int
elan4_check_route (ELAN_POSITION *postiion, ELAN_LOCATION location, E4_VirtualProcessEntry *route, unsigned flags)
{
    /* XXXX - TBD */
    return (0);
}

EXPORT_SYMBOL(elan4_alloc_routetable);
EXPORT_SYMBOL(elan4_free_routetable);
EXPORT_SYMBOL(elan4_write_route);
EXPORT_SYMBOL(elan4_read_route);
EXPORT_SYMBOL(elan4_invalidate_route);
EXPORT_SYMBOL(elan4_generate_route);
EXPORT_SYMBOL(elan4_check_route);

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
