/*
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: user.c,v 1.89.2.2 2005/07/19 13:45:36 daniel Exp $"
/*      $Source: /cvs/master/quadrics/elan4mod/user.c,v $*/

#include <qsnet/kernel.h>
#include <qsnet/kpte.h>

#include <elan/elanmod.h>
#include <elan4/debug.h>
#include <elan4/device.h>
#include <elan4/user.h>

#include <elan4/trtype.h>
#include <elan4/commands.h>

#include <stdarg.h>

/* allow this code to compile against an Eagle elanmod */
#ifdef __ELANMOD_DEVICE_H
#define elan_attach_cap(cap,rnum,args,func)	elanmod_attach_cap(cap,args,func)
#define elan_detach_cap(cap,rnum)		elanmod_detach_cap(cap)
#endif

#define NETERR_MSGS	16

int user_p2p_route_options   = FIRST_TIMEOUT(3);
int user_bcast_route_options = FIRST_TIMEOUT(3);
int user_dproc_retry_count   = 15;
int user_cproc_retry_count   = 2;
int user_ioproc_enabled      = 1;
int user_pagefault_enabled   = 1;

int num_fault_save           = 30;
int min_fault_pages          = 1;
int max_fault_pages          = 128;

static int
user_validate_cap (USER_CTXT *uctx, ELAN_CAPABILITY *cap, unsigned use)
{
    /* Don't allow a user process to attach to system context */
    if (ELAN4_SYSTEM_CONTEXT (cap->cap_lowcontext) || ELAN4_SYSTEM_CONTEXT (cap->cap_highcontext))
    {
	PRINTF3 (DBG_DEVICE, DBG_VP,"user_validate_cap: lctx %x hctx %x high %x\n", cap->cap_lowcontext, cap->cap_highcontext, ELAN4_KCOMM_BASE_CONTEXT_NUM);
	PRINTF0 (DBG_DEVICE, DBG_VP,"user_validate_cap: user process cant attach to system cap\n");
	return (EINVAL);
    }
    
    return elanmod_classify_cap(&uctx->uctx_position, cap, use);
}

static __inline__ void
__user_signal_trap (USER_CTXT *uctx)
{
    switch (uctx->uctx_trap_state)
    {
    case UCTX_TRAP_IDLE:
	PRINTF (uctx, DBG_TRAP, "user_signal_trap: deliver signal %d to pid %d\n", uctx->uctx_trap_signo, uctx->uctx_trap_pid);

	if (uctx->uctx_trap_signo)
	    kill_proc (uctx->uctx_trap_pid, uctx->uctx_trap_signo, 1);
	break;

    case UCTX_TRAP_SLEEPING:
	PRINTF (uctx, DBG_TRAP, "user_signal_trap: wakeup sleeping trap handler\n");

	kcondvar_wakeupone (&uctx->uctx_wait, &uctx->uctx_spinlock);
	break;
    }
    uctx->uctx_trap_state = UCTX_TRAP_SIGNALLED;
}

static void
user_signal_timer (unsigned long arg)
{
    USER_CTXT    *uctx = (USER_CTXT *) arg;
    unsigned long flags;

    PRINTF (uctx, DBG_TRAP, "user_signal_timer: state=%d pid=%d signal=%d (now %d start %d)\n",
	    uctx->uctx_trap_state, uctx->uctx_trap_pid, uctx->uctx_trap_signo, jiffies,
	    uctx->uctx_int_start);

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    __user_signal_trap (uctx);
    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
}

#define MAX_INTS_PER_TICK	50
#define MIN_INTS_PER_TICK	20

static void
user_shuffle_signal_trap (USER_CTXT *uctx)
{
    ASSERT (SPINLOCK_HELD (&uctx->uctx_spinlock));

    PRINTF (uctx, DBG_TRAP, "user_shuffle_signal_trap: signal=%d%s\n", 
	    uctx->uctx_trap_signo, timer_pending(&uctx->uctx_shuffle_timer) ? " (timer-pending)" : "");

    if (timer_pending (&uctx->uctx_shuffle_timer))
	return;

    uctx->uctx_shuffle_timer.expires =  jiffies + (HZ*2);

    add_timer (&uctx->uctx_shuffle_timer);
}

static void
user_signal_trap (USER_CTXT *uctx)
{
    ASSERT (SPINLOCK_HELD (&uctx->uctx_spinlock));

    PRINTF (uctx, DBG_TRAP, "user_signal_trap: state=%d pid=%d signal=%d%s\n", uctx->uctx_trap_state,
	    uctx->uctx_trap_pid, uctx->uctx_trap_signo, timer_pending(&uctx->uctx_int_timer) ? " (timer-pending)" : "");

    uctx->uctx_int_count++;

    if (timer_pending (&uctx->uctx_int_timer))
	return;

    if (uctx->uctx_int_count > ((int)(jiffies - uctx->uctx_int_start) * MAX_INTS_PER_TICK))
    {
	PRINTF (uctx, DBG_TRAP, "user_signal_trap: deferring signal for %d ticks (count %d ticks %d -> %d)\n", 
		uctx->uctx_int_delay + 1, uctx->uctx_int_count, (int) (jiffies - uctx->uctx_int_start),
		((int)(jiffies - uctx->uctx_int_start) * MAX_INTS_PER_TICK));

	/* We're interrupting too fast, so defer this signal */
	uctx->uctx_int_timer.expires = jiffies + (++uctx->uctx_int_delay);

	add_timer (&uctx->uctx_int_timer);
    }
    else
    {
	__user_signal_trap (uctx);

	PRINTF (uctx, DBG_TRAP, "user_signal_trap: check signal for %d ticks (count %d ticks %d -> %d)\n", 
		uctx->uctx_int_delay + 1, uctx->uctx_int_count, (int) (jiffies - uctx->uctx_int_start),
		(int)(jiffies - uctx->uctx_int_start) * MIN_INTS_PER_TICK);
	    
	if (uctx->uctx_int_count < ((int) (jiffies - uctx->uctx_int_start)) * MIN_INTS_PER_TICK)
	{
	    PRINTF (uctx, DBG_TRAP, "user_signal_trap: reset interrupt throttle (count %d ticks %d)\n", 
		    uctx->uctx_int_count, (int) (jiffies - uctx->uctx_int_start));

	    uctx->uctx_int_start = jiffies;
	    uctx->uctx_int_count = 0;
	    uctx->uctx_int_delay = 0;
	}
    }
}

static void
user_neterr_timer (unsigned long arg)
{
    USER_CTXT *uctx = (USER_CTXT *) arg;
    unsigned long flags;
    
    spin_lock_irqsave (&uctx->uctx_spinlock, flags);

    uctx->uctx_status |= UCTX_NETERR_TIMER;
    
    user_signal_trap (uctx);

    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
}

static void
user_flush_dma_runqueue (ELAN4_DEV *dev, USER_CTXT *uctx, int qfull)
{
    E4_uint64          qptrs = read_reg64 (dev, DProcLowPriPtrs);
    E4_uint32          qsize = E4_QueueSize (E4_QueueSizeValue (qptrs));
    E4_uint32          qfptr = E4_QueueFrontPointer (qptrs);
    E4_uint32          qbptr = E4_QueueBackPointer (qptrs);
    E4_DProcQueueEntry qentry;

    while ((qfptr != qbptr) || qfull)
    {
	E4_uint64 typeSize = elan4_sdram_readq (dev, qfptr + offsetof (E4_DProcQueueEntry, Desc.dma_typeSize));

	if (DMA_Context (typeSize) == uctx->uctx_ctxt.ctxt_num)
	{
	    elan4_sdram_copyq_from_sdram (dev, qfptr, &qentry, sizeof (E4_DProcQueueEntry));

	    PRINTF4 (uctx, DBG_SWAP, "user_flush_dma_runqueue: %016llx %016llx %016llx %016llx\n", qentry.Desc.dma_typeSize, 
		     qentry.Desc.dma_cookie, qentry.Desc.dma_vproc, qentry.Desc.dma_srcAddr);
	    PRINTF3 (uctx, DBG_SWAP, "                         %016llx %016llx %016llx\n", qentry.Desc.dma_dstAddr, 
		     qentry.Desc.dma_srcEvent, qentry.Desc.dma_dstEvent);

	    if (RING_QUEUE_REALLY_FULL (uctx->uctx_dmaQ))
	    {
		PRINTF (uctx, DBG_SWAP, "user_flush_dma_runqueue: queue overflow\n");
		uctx->uctx_status |= UCTX_DPROC_QUEUE_OVERFLOW;
	    }
	    else
	    {
		*RING_QUEUE_BACK (uctx->uctx_dmaQ, uctx->uctx_dmas) = qentry.Desc;
		(void) RING_QUEUE_ADD (uctx->uctx_dmaQ);
	    }
	    
	    qentry.Desc.dma_typeSize = DMA_ShMemWrite | dev->dev_ctxt.ctxt_num;
	    qentry.Desc.dma_cookie   = 0;
	    qentry.Desc.dma_vproc    = 0;
	    qentry.Desc.dma_srcAddr  = 0;
	    qentry.Desc.dma_dstAddr  = 0;
	    qentry.Desc.dma_srcEvent = 0;
	    qentry.Desc.dma_dstEvent = 0;

	    elan4_sdram_copyq_to_sdram (dev, &qentry, qfptr, sizeof (E4_DProcQueueEntry));
	}

	qfptr = (qfptr & ~(qsize-1)) | ((qfptr + sizeof (E4_DProcQueueEntry)) & (qsize-1));
	qfull = 0;
    }
}

static void
user_flush_thread_runqueue (ELAN4_DEV *dev, USER_CTXT *uctx, int qfull)
{
    E4_uint64          qptrs = read_reg64 (dev, TProcLowPriPtrs);
    E4_uint32          qsize = E4_QueueSize (E4_QueueSizeValue (qptrs));
    E4_uint32          qfptr = E4_QueueFrontPointer (qptrs);
    E4_uint32          qbptr = E4_QueueBackPointer (qptrs);
    E4_TProcQueueEntry qentry;

    while ((qfptr != qbptr) || qfull)
    {
	E4_uint64 context = elan4_sdram_readq (dev, qfptr + offsetof (E4_TProcQueueEntry, Context));

	if (TPROC_Context (context) == uctx->uctx_ctxt.ctxt_num)
	{
	    elan4_sdram_copyq_from_sdram (dev, qfptr, &qentry, sizeof (E4_TProcQueueEntry));

	    PRINTF (uctx, DBG_SWAP, "user_flush_thread_runqueue: %016llx %016llx %016llx %016llx\n", qentry.Regs.Registers[0],
		    qentry.Regs.Registers[1], qentry.Regs.Registers[2], qentry.Regs.Registers[3]);
	    PRINTF (uctx, DBG_SWAP, "                            %016llx %016llx %016llx\n", 
		    qentry.Regs.Registers[4], qentry.Regs.Registers[5], qentry.Regs.Registers[6]);

	    if (RING_QUEUE_REALLY_FULL (uctx->uctx_threadQ))
		uctx->uctx_status |= UCTX_TPROC_QUEUE_OVERFLOW;
	    else
	    {
		*RING_QUEUE_BACK (uctx->uctx_threadQ, uctx->uctx_threads) = qentry.Regs;
		(void) RING_QUEUE_ADD (uctx->uctx_threadQ);
	    }
	    
	    /* change the thread to execute the suspend sequence */
	    qentry.Regs.Registers[0] = dev->dev_tproc_suspend;
	    qentry.Regs.Registers[1] = dev->dev_tproc_space;
	    qentry.Context           = dev->dev_ctxt.ctxt_num;

	    elan4_sdram_copyq_to_sdram (dev, &qentry, qfptr, sizeof (E4_TProcQueueEntry));
	}
	
	qfptr = (qfptr & ~(qsize-1)) | ((qfptr + sizeof (E4_TProcQueueEntry)) & (qsize-1));
	qfull = 0;
    }
}

static void
user_flush_dmas (ELAN4_DEV *dev, void *arg, int qfull)
{
    USER_CTXT        *uctx = (USER_CTXT *) arg;
    unsigned long     flags;
    
    ASSERT ((read_reg32 (dev, InterruptReg) & INT_DProcHalted) != 0);

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);

    if ((uctx->uctx_status & (UCTX_SWAPPED_REASONS|UCTX_STOPPED_REASONS)) == 0)
    {
	PRINTF1 (uctx, DBG_SWAP, "user_flush_dmas: status %x - no more reasons\n", uctx->uctx_status);

	uctx->uctx_status &= ~UCTX_STOPPING;

	user_signal_trap (uctx);
    }
    else
    {
	user_flush_dma_runqueue (dev, uctx, qfull);

	uctx->uctx_status = (uctx->uctx_status | UCTX_STOPPED) & ~UCTX_STOPPING;
    
	PRINTF1 (uctx, DBG_SWAP, "user_flush_dmas: statux %x - stopped\n", uctx->uctx_status);

	kcondvar_wakeupall (&uctx->uctx_wait, &uctx->uctx_spinlock);
    }

    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
}

static void
user_flush (ELAN4_DEV *dev, void *arg)
{
    USER_CTXT        *uctx = (USER_CTXT *) arg;
    struct list_head *entry;
    unsigned long     flags;

    ASSERT ((read_reg32 (dev, InterruptReg) & (INT_Halted|INT_Discarding)) == (INT_Halted|INT_Discarding));

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);

    if ((uctx->uctx_status & (UCTX_SWAPPED_REASONS|UCTX_STOPPED_REASONS)) == 0)
    {
	PRINTF1 (uctx, DBG_SWAP, "user_flush: status %x - no more reasons\n", uctx->uctx_status);

	uctx->uctx_status &= ~UCTX_STOPPING;

	user_signal_trap (uctx);
    }
    else
    {
	PRINTF1 (uctx, DBG_SWAP, "user_flush: status %x - flushing context\n", uctx->uctx_status);

	list_for_each (entry, &uctx->uctx_cqlist) {
	    USER_CQ *ucq = list_entry (entry, USER_CQ, ucq_link);

	    if (ucq->ucq_state == UCQ_RUNNING)
	    {
		/* NOTE: since the inserter can still be running we modify the permissions
		 *       to zero then when the extractor starts up again it will trap */
		PRINTF1 (uctx, DBG_SWAP, "user_flush: stopping cq indx=%d\n", elan4_cq2idx(ucq->ucq_cq));

		elan4_updatecq (dev, ucq->ucq_cq, 0, 0);
	    }
	}
	
	user_flush_thread_runqueue (dev, uctx, TPROC_LowRunQueueFull(read_reg64 (dev, TProcStatus)));

	/* since we can't determine whether the dma run queue is full or empty, we use a dma
	 * halt operation to do the flushing - as the reason for halting the dma processor 
	 * will be released when we return, we keep it halted until the flush has completed */
	elan4_queue_dma_flushop (dev, &uctx->uctx_dma_flushop, 0);

	if (uctx->uctx_status & UCTX_EXITING)
	    elan4_flush_icache_halted (&uctx->uctx_ctxt);
    }

    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
}

static void
user_set_filter (USER_CTXT *uctx, E4_uint32 state)
{
    struct list_head *entry;

    ASSERT (SPINLOCK_HELD (&uctx->uctx_spinlock));

    list_for_each (entry, &uctx->uctx_cent_list) {
	USER_CTXT_ENTRY *cent = list_entry (entry, USER_CTXT_ENTRY, cent_link);

	elan4_set_filter (&uctx->uctx_ctxt, cent->cent_cap->cap_mycontext, state);
    }
}

static void
user_start_nacking (USER_CTXT *uctx, unsigned reason)
{
    PRINTF2 (uctx, DBG_SWAP, "user_start_nacking: status %x reason %x\n", uctx->uctx_status, reason);

    ASSERT (SPINLOCK_HELD (&uctx->uctx_spinlock));

    if (UCTX_NACKING(uctx))
	uctx->uctx_status |= reason;
    else
    {
	uctx->uctx_status |= reason;

	user_set_filter (uctx, E4_FILTER_STATS | E4_FILTER_DISCARD_ALL);
    }
}

static void
user_stop_nacking (USER_CTXT *uctx, unsigned reason)
{
    PRINTF2 (uctx, DBG_SWAP, "user_stop_nacking: status %x reason %x\n", uctx->uctx_status, reason);
    
    ASSERT (SPINLOCK_HELD (&uctx->uctx_spinlock));
    
    uctx->uctx_status &= ~reason;
    
    if (! UCTX_NACKING (uctx))
	user_set_filter (uctx, E4_FILTER_STATS);
}

static void
user_start_stopping (USER_CTXT *uctx, unsigned reason)
{
    ELAN4_DEV *dev =uctx->uctx_ctxt.ctxt_dev;

    PRINTF2 (uctx, DBG_SWAP, "user_start_stopping: status %x reason %x\n", uctx->uctx_status, reason);

    ASSERT (! (uctx->uctx_status & UCTX_STOPPED));

    user_start_nacking (uctx, reason);
    
    if ((uctx->uctx_status & UCTX_STOPPING) != 0)
	return;
    
    uctx->uctx_status |= UCTX_STOPPING;

    /* queue the halt operation to  remove all threads/dmas/cqs from the run queues */
    /*    and also flush through the context filter change */
    elan4_queue_haltop (dev, &uctx->uctx_haltop);
}

static void
user_stop_stopping (USER_CTXT *uctx, unsigned reason)
{
    PRINTF2 (uctx, DBG_SWAP, "user_stop_stopping: status %x reason %x\n", uctx->uctx_status, reason);
    
    user_stop_nacking (uctx, reason);

    if (UCTX_RUNNABLE (uctx))
    {
	uctx->uctx_status &= ~UCTX_STOPPED;

	PRINTF1 (uctx, DBG_SWAP, "user_stop_stopping: no more reasons => %x\n", uctx->uctx_status);

	user_signal_trap (uctx);
    }
}

void
user_swapout (USER_CTXT *uctx, unsigned reason)
{
    ELAN4_DEV *dev = uctx->uctx_ctxt.ctxt_dev;
    unsigned long flags;
    
    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    
    PRINTF2 (uctx, DBG_SWAP, "user_swapout: status %x reason %x\n", uctx->uctx_status, reason);
    
    user_start_nacking (uctx, reason);
    
    while (uctx->uctx_status & (UCTX_SWAPPING|UCTX_STOPPING) && 		/* wait for someone else to finish */
	   uctx->uctx_trap_count > 0)						/* and for trap handlers to notice */
    {										/* and exit */
	PRINTF1 (uctx, DBG_SWAP, "user_swapout: waiting for %d trap handlers to exit/previous swapout\n", uctx->uctx_trap_count);

	kcondvar_wakeupall (&uctx->uctx_wait, &uctx->uctx_spinlock);
	kcondvar_wait (&uctx->uctx_wait, &uctx->uctx_spinlock, &flags);
    }

    if (uctx->uctx_status & UCTX_SWAPPED)					/* already swapped out */
    {
	spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
	return;
    }
    
    uctx->uctx_status |= (UCTX_SWAPPING|UCTX_STOPPING);				/* mark the context as swapping & stopping */
    
    /* queue the halt operation to  remove all threads/dmas/cqs from the run queues */
    /*    and also flush through the context filter change */
    elan4_queue_haltop (dev, &uctx->uctx_haltop);
    
    while (! (uctx->uctx_status & UCTX_STOPPED))
	kcondvar_wait (&uctx->uctx_wait, &uctx->uctx_spinlock, &flags);

    /* all state has been removed from the elan - we can now "tidy" it up */

    PRINTF0 (uctx, DBG_SWAP, "user_swapout: swapped out\n");
    
    uctx->uctx_status = (uctx->uctx_status & ~UCTX_SWAPPING) | UCTX_SWAPPED;
    
    kcondvar_wakeupall (&uctx->uctx_wait, &uctx->uctx_spinlock);

    PRINTF1 (uctx, DBG_SWAP, "user_swapout: all done - status %x\n", uctx->uctx_status);

    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
}

void
user_swapin (USER_CTXT *uctx, unsigned reason)
{
    unsigned long flags;

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);

    ASSERT (uctx->uctx_status & UCTX_SWAPPED_REASONS);

    PRINTF2 (uctx, DBG_SWAP, "user_swapin: status %x reason %x\n", uctx->uctx_status, reason);

    while (uctx->uctx_status & (UCTX_SWAPPING|UCTX_STOPPING))			/* wait until other threads have */
	kcondvar_wait (&uctx->uctx_wait, &uctx->uctx_spinlock, &flags);		/* completed their swap operation */

    ASSERT (uctx->uctx_status & (UCTX_SWAPPED | UCTX_STOPPED));

    user_stop_nacking (uctx, reason);

    if (! (uctx->uctx_status & UCTX_SWAPPED_REASONS))
    {
	uctx->uctx_status &= ~UCTX_SWAPPED;

	/* no longer swapped out - wakeup anyone sleeping waiting for swapin */
	kcondvar_wakeupall (&uctx->uctx_wait, &uctx->uctx_spinlock);

	if (! (uctx->uctx_status & UCTX_STOPPED_REASONS))
	{
	    uctx->uctx_status &= ~UCTX_STOPPED;
	    user_signal_trap (uctx);
	}
    }

    PRINTF1 (uctx, DBG_SWAP, "user_swapin: all done - status %x\n", uctx->uctx_status);

    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
}

void
user_destroy_callback (void *arg, ELAN_CAPABILITY *cap, ELAN_CAPABILITY *map)
{
    USER_CTXT *uctx = (USER_CTXT *) arg;

    PRINTF (uctx, DBG_VP, "user_destroy_callback: %s\n", map == NULL ? "cap destoyed" : "map destroyed");
}

int
user_attach (USER_CTXT *uctx, ELAN_CAPABILITY *cap)
{
    ELAN4_DEV       *dev = uctx->uctx_ctxt.ctxt_dev;
    USER_CTXT_ENTRY *cent;
    unsigned long flags;
    int ctype, res;
    
    if ((ctype = user_validate_cap (uctx, cap, ELAN_USER_ATTACH)) < 0)
	return ctype;

    if ((ctype == ELAN_CAP_RMS) && (res = elan_attach_cap (cap, dev->dev_devinfo.dev_rail, uctx, user_destroy_callback)) != 0)
    {
	/* NOTE: elan_attach_cap returns +ve errnos */
	return -res;
    }

    KMEM_ALLOC (cent, USER_CTXT_ENTRY *, sizeof (USER_CTXT_ENTRY), 1);
    if (cent == NULL)
    {
	if (ctype == ELAN_CAP_RMS)
	    elan_detach_cap (cap, dev->dev_devinfo.dev_rail);

	return -ENOMEM;
    }

    KMEM_ALLOC (cent->cent_cap, ELAN_CAPABILITY *, ELAN_CAP_SIZE(cap), 1);
    if (cent->cent_cap == NULL)
    {
	if (ctype == ELAN_CAP_RMS)
	    elan_detach_cap (cap, dev->dev_devinfo.dev_rail);

	KMEM_FREE (cent, sizeof (USER_CTXT_ENTRY));
	return -ENOMEM;
    }

    memcpy (cent->cent_cap, cap, ELAN_CAP_SIZE(cap));

    if ((res = elan4_attach_filter (&uctx->uctx_ctxt, cap->cap_mycontext)) != 0)
    {
	if (ctype == ELAN_CAP_RMS)
	    elan_detach_cap (cap, dev->dev_devinfo.dev_rail);
	
	KMEM_FREE (cent->cent_cap, ELAN_CAP_SIZE (cap));
	KMEM_FREE (cent, sizeof (USER_CTXT_ENTRY));

	return res;
    }

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);

    list_add_tail (&cent->cent_link, &uctx->uctx_cent_list);

    if (! UCTX_NACKING (uctx))
	user_set_filter (uctx, E4_FILTER_STATS);

    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);

    return (0);
    
}

void
user_detach (USER_CTXT *uctx, ELAN_CAPABILITY *cap)
{
    ELAN4_DEV         *dev = uctx->uctx_ctxt.ctxt_dev;
    struct list_head  *entry;
    struct list_head  *next;
    struct list_head   list;
    unsigned long      flags;

    INIT_LIST_HEAD (&list);

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    
    PRINTF (uctx, DBG_NETWORK_CTX, cap ? "user_detach: network context %d\n" : "user_detach: all network contexts\n", cap ? cap->cap_mycontext : 0);

    list_for_each_safe (entry, next, &uctx->uctx_cent_list) {
	USER_CTXT_ENTRY *cent = list_entry (entry, USER_CTXT_ENTRY, cent_link);

	if (cap == NULL || ELAN_CAP_MATCH (cap, cent->cent_cap))
	{
	    PRINTF1 (uctx, DBG_NETWORK_CTX, "user_detach: detach from network context %d\n", cent->cent_cap->cap_mycontext);
	    
	    elan4_detach_filter (&uctx->uctx_ctxt, cent->cent_cap->cap_mycontext);

	    list_del (&cent->cent_link);
	    list_add_tail (&cent->cent_link, &list);
	}
    }

    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);

    while (! list_empty (&list))
    {
	USER_CTXT_ENTRY *cent = list_entry (list.next, USER_CTXT_ENTRY, cent_link);

	list_del (&cent->cent_link);

	if (user_validate_cap (uctx, cent->cent_cap, ELAN_USER_DETACH) == ELAN_CAP_RMS)
	    elan_detach_cap (cent->cent_cap, dev->dev_devinfo.dev_rail); 
	
	KMEM_FREE (cent->cent_cap, ELAN_CAP_SIZE (cent->cent_cap));
	KMEM_FREE (cent, sizeof (USER_CTXT_ENTRY));
    }
}

void
user_block_inputter (USER_CTXT *uctx, unsigned blocked)
{
    unsigned long flags;
    int isblocked;

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    
    isblocked = (uctx->uctx_status & UCTX_USER_FILTERING);

    if (blocked && !isblocked)
	user_start_nacking (uctx, UCTX_USER_FILTERING);

    if (!blocked && isblocked)
	user_stop_nacking (uctx, UCTX_USER_FILTERING);

    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
}

static USER_VPSEG *
user_install_vpseg (USER_CTXT *uctx, unsigned process, unsigned entries)
{
    struct list_head *entry;
    USER_VPSEG       *seg;

    if ((process + entries) > (E4_VPT_MIN_ENTRIES << uctx->uctx_routetable->tbl_size))
	return (NULL);

    ASSERT (kmutex_is_locked (&uctx->uctx_vpseg_lock));

    list_for_each (entry, &uctx->uctx_vpseg_list) {
	seg = list_entry (entry, USER_VPSEG, vps_link);

	if (process <= (seg->vps_process + seg->vps_entries-1) && 
	    (process + entries - 1) >= seg->vps_process)
	    return ((USER_VPSEG *) NULL);
    }

    KMEM_ZALLOC (seg, USER_VPSEG *, sizeof (USER_VPSEG), 1);
    
    if (seg == (USER_VPSEG *) NULL)
	return ((USER_VPSEG *) NULL);

    seg->vps_process = process;
    seg->vps_entries = entries;

    list_add_tail (&seg->vps_link, &uctx->uctx_vpseg_list);

    return (seg);
}

static void
user_remove_vpseg (USER_CTXT *uctx, USER_VPSEG *seg)
{
    ASSERT (kmutex_is_locked (&uctx->uctx_vpseg_lock));

    list_del (&seg->vps_link);
    
    switch (seg->vps_type)
    {
    case USER_VPSEG_P2P:
	/* These pointers (union) are only valid for P2P segs */
	if (seg->vps_p2p_routes)
	    KMEM_FREE (seg->vps_p2p_routes, sizeof (E4_VirtualProcessEntry) * seg->vps_entries);
	
	if (seg->vps_p2p_cap)
	    KMEM_FREE (seg->vps_p2p_cap, ELAN_CAP_SIZE(seg->vps_p2p_cap));

	break;
	
    case USER_VPSEG_BCAST:
	;
    }

    KMEM_FREE (seg, sizeof (USER_VPSEG));
}

static USER_VPSEG *
user_find_vpseg (USER_CTXT *uctx, unsigned low, unsigned high)
{
    struct list_head *entry;

    ASSERT (kmutex_is_locked (&uctx->uctx_vpseg_lock));

    list_for_each (entry, &uctx->uctx_vpseg_list) {
	USER_VPSEG *seg = list_entry (entry, USER_VPSEG, vps_link);

	if (seg->vps_process <= low && (seg->vps_process + seg->vps_entries) > high)
	    return (seg);
    }

    return ((USER_VPSEG *) NULL);
}

static ELAN_LOCATION 
user_process2location (USER_CTXT *uctx, USER_VPSEG *seg, unsigned process)
{
    ELAN_LOCATION location;
    int           nnodes, nctxs;
    int           nodeOff, ctxOff, vpOff;

    location.loc_node    = ELAN_INVALID_NODE;
    location.loc_context = -1;

    if (seg == NULL)
	seg = user_find_vpseg (uctx, process, process);

    if (seg == NULL || (seg->vps_type != USER_VPSEG_P2P))
	return (location);

    nnodes = ELAN_CAP_NUM_NODES (seg->vps_p2p_cap);
    nctxs  = ELAN_CAP_NUM_CONTEXTS (seg->vps_p2p_cap);

    switch (seg->vps_p2p_cap->cap_type & ELAN_CAP_TYPE_MASK)
    {
    case ELAN_CAP_TYPE_BLOCK:
	for (nodeOff = 0, vpOff = 0; nodeOff < nnodes; nodeOff++)
	{
	    for (ctxOff = 0; ctxOff < nctxs; ctxOff++)
	    {
		if ((seg->vps_p2p_cap->cap_type & ELAN_CAP_TYPE_NO_BITMAP) || BT_TEST (seg->vps_p2p_cap->cap_bitmap, ctxOff + (nodeOff * nctxs)))
		{
		    if (vpOff++ == (process - seg->vps_process))
		    { 
			location.loc_node    = seg->vps_p2p_cap->cap_lownode + nodeOff;
			location.loc_context = seg->vps_p2p_cap->cap_lowcontext + ctxOff;
			goto found;
		    }
		}
	    }
	}
	break;
	
    case ELAN_CAP_TYPE_CYCLIC:
	for (ctxOff = 0, vpOff = 0; ctxOff < nctxs; ctxOff++)
	{
	    for (nodeOff = 0; nodeOff < nnodes; nodeOff++)
	    {
		if ((seg->vps_p2p_cap->cap_type & ELAN_CAP_TYPE_NO_BITMAP) || BT_TEST (seg->vps_p2p_cap->cap_bitmap, nodeOff + (ctxOff * nnodes)))
		{		    		    
		    if (vpOff++ ==  (process - seg->vps_process))
		    { 
			location.loc_node    = seg->vps_p2p_cap->cap_lownode + nodeOff;
			location.loc_context = seg->vps_p2p_cap->cap_lowcontext + ctxOff;
			goto found;
		    }
		}
	    }
	}
	break;	
    }
       
 found:
    return (location);
}

static unsigned 
user_location2process (USER_CTXT *uctx, ELAN_LOCATION location)
{
    unsigned int      process = ELAN_INVALID_PROCESS;
    struct list_head *entry;
    int               nnodes, nctxs;
    int               nodeOff, ctxOff, vpOff;

    kmutex_lock (&uctx->uctx_vpseg_lock);
    list_for_each (entry, &uctx->uctx_vpseg_list) {
	USER_VPSEG *seg = list_entry (entry, USER_VPSEG, vps_link);

	if (seg->vps_type != USER_VPSEG_P2P)
	    continue;

	if (location.loc_node >= seg->vps_p2p_cap->cap_lownode && location.loc_node <= seg->vps_p2p_cap->cap_highnode &&
	    location.loc_context >= seg->vps_p2p_cap->cap_lowcontext && location.loc_context <= seg->vps_p2p_cap->cap_highcontext)
	{
	    nnodes = ELAN_CAP_NUM_NODES (seg->vps_p2p_cap);
	    nctxs  = ELAN_CAP_NUM_CONTEXTS (seg->vps_p2p_cap);

	    switch (seg->vps_p2p_cap->cap_type & ELAN_CAP_TYPE_MASK)
	    {
	    case ELAN_CAP_TYPE_BLOCK:
		for (nodeOff = 0, vpOff = 0; nodeOff < nnodes; nodeOff++)
		{
		    for (ctxOff = 0; ctxOff < nctxs; ctxOff++)
		    {
			if ((seg->vps_p2p_cap->cap_type & ELAN_CAP_TYPE_NO_BITMAP) || BT_TEST (seg->vps_p2p_cap->cap_bitmap, ctxOff + (nodeOff * nctxs)))
			{
			    if (location.loc_node == seg->vps_p2p_cap->cap_lownode + nodeOff &&
				location.loc_context == seg->vps_p2p_cap->cap_lowcontext + ctxOff)
			    {
				process = seg->vps_process + vpOff;
				goto found;
			    }
			    vpOff++;
			}
		    }
		}
		break;
	
	    case ELAN_CAP_TYPE_CYCLIC:
		for (ctxOff = 0, vpOff = 0; ctxOff < nctxs; ctxOff++)
		{
		    for (nodeOff = 0; nodeOff < nnodes; nodeOff++)
		    {
			if ((seg->vps_p2p_cap->cap_type & ELAN_CAP_TYPE_NO_BITMAP) || BT_TEST (seg->vps_p2p_cap->cap_bitmap, nodeOff + (ctxOff * nnodes)))
			{
			    if (location.loc_node == seg->vps_p2p_cap->cap_lownode + nodeOff &&
				location.loc_context == seg->vps_p2p_cap->cap_lowcontext + ctxOff)
			    {
				process = seg->vps_process + vpOff;
				goto found;
			    }
			    vpOff++;
			}
		    }
		}
		break;
	    }
	}
    }
 found:
    kmutex_unlock (&uctx->uctx_vpseg_lock);

    return (process);
}

static void
user_loadroute_vpseg (USER_CTXT *uctx, USER_VPSEG *seg, ELAN_POSITION *pos)
{
    ELAN4_DEV             *dev    = uctx->uctx_ctxt.ctxt_dev;
    ELAN_CAPABILITY       *cap    = seg->vps_p2p_cap;
    unsigned               nnodes = ELAN_CAP_NUM_NODES (cap);
    unsigned               nctxs  = ELAN_CAP_NUM_CONTEXTS (cap);
    E4_VirtualProcessEntry route;
    unsigned	           nodeOff;
    unsigned	           ctxOff;
    unsigned	           vpOff;

    switch (cap->cap_type & ELAN_CAP_TYPE_MASK)
    {
    case ELAN_CAP_TYPE_BLOCK:
	for (nodeOff = 0, vpOff = 0; nodeOff < nnodes; nodeOff++)
	{
	    for (ctxOff = 0; ctxOff < nctxs; ctxOff++)
	    {
		if ((cap->cap_type & ELAN_CAP_TYPE_NO_BITMAP) || BT_TEST (cap->cap_bitmap, ctxOff + (nodeOff * nctxs)))
		{
		    if (seg->vps_p2p_routes != NULL)
			route = seg->vps_p2p_routes[vpOff];
		    else if (elan4_generate_route (&uctx->uctx_position, &route, cap->cap_lowcontext + ctxOff,
						   cap->cap_lownode + nodeOff, cap->cap_lownode + nodeOff, user_p2p_route_options) < 0)
		    {
			vpOff++;
			continue;
		    }

		    PRINTF5 (uctx, DBG_VP, "user_loadroute_vpseg: virtual process %d -> node %d context %d [%016llx.%016llx]\n",
			     seg->vps_process + vpOff, cap->cap_lownode + nodeOff, cap->cap_lowcontext + ctxOff,
			     route.Values[0], route.Values[1]);
		    
		    elan4_write_route (dev, uctx->uctx_routetable, seg->vps_process + vpOff, &route);
					      
		    vpOff++;
		}
	    }
	}
	break;

    case ELAN_CAP_TYPE_CYCLIC:
	for (ctxOff = 0, vpOff = 0; ctxOff < nctxs; ctxOff++)
	{
	    for (nodeOff = 0; nodeOff < nnodes; nodeOff++)
	    {
		if ((cap->cap_type & ELAN_CAP_TYPE_NO_BITMAP) || BT_TEST (cap->cap_bitmap, nodeOff + (ctxOff * nnodes)))
		{
		    if (seg->vps_p2p_routes != NULL)
			route = seg->vps_p2p_routes[vpOff];
		    else if (elan4_generate_route (&uctx->uctx_position, &route, cap->cap_lowcontext + ctxOff,
						   cap->cap_lownode + nodeOff, cap->cap_lownode + nodeOff, user_p2p_route_options) < 0)
		    {
			vpOff++;
			continue;
		    }

		    PRINTF5 (uctx, DBG_VP, "user_loadroute_vpseg: virtual process %d -> node %d context %d [%016llx.%016llx]\n",
			     seg->vps_process + vpOff, cap->cap_lownode + nodeOff, cap->cap_lowcontext + ctxOff,
			     route.Values[0], route.Values[1]);
		    
		    elan4_write_route (dev, uctx->uctx_routetable, seg->vps_process + vpOff, &route);
					      
		    vpOff++;
		}
	    }
	}
	break;
    }
}

static int
user_loadroute_bcast (USER_CTXT *uctx, USER_VPSEG *seg)
{
    ELAN4_DEV             *dev = uctx->uctx_ctxt.ctxt_dev;
    ELAN_POSITION         *pos = &uctx->uctx_position;
    E4_VirtualProcessEntry route;
    USER_VPSEG            *aseg;
    int                    res;
    ELAN_LOCATION          low;
    ELAN_LOCATION          high;

    if ((aseg = user_find_vpseg (uctx, seg->vps_bcast_lowvp, seg->vps_bcast_highvp)) == NULL || aseg->vps_type != USER_VPSEG_P2P)
	return (-EINVAL);
    
#ifdef use_elanmod
    if ((res = user_validate_cap (dev, aseg->vps_p2p_cap, ELAN_USER_BROADCAST)) < 0)
	return (res);
#endif
    
    low  = user_process2location (uctx, aseg, seg->vps_bcast_lowvp);
    high = user_process2location (uctx, aseg, seg->vps_bcast_highvp);

    if (low.loc_context != high.loc_context)
	return (-EINVAL);

    /* NOTE: if loopback can only broadcast to ourself - 
     *       if back-to-back can only broadcast to other node */
    if ((pos->pos_mode == ELAN_POS_MODE_LOOPBACK   && low.loc_node != high.loc_node && low.loc_node != pos->pos_nodeid) ||
	(pos->pos_mode == ELAN_POS_MODE_BACKTOBACK && low.loc_node != high.loc_node && low.loc_node == pos->pos_nodeid))
    {
	return (-EINVAL);
    }
    
    if ((res = elan4_generate_route (pos, &route, low.loc_context, low.loc_node, high.loc_node, user_bcast_route_options)) < 0)
	return (res);

    PRINTF (uctx, DBG_VP, "user_loadroute_bcast: virtual process %d -> nodes %d.%d context %d [%016llx.%016llx]\n",
	    seg->vps_process, low.loc_node, high.loc_node, low.loc_context, route.Values[0], route.Values[1]);
    
    elan4_write_route (dev, uctx->uctx_routetable, seg->vps_process, &route);
    return (0);
}

int
user_add_p2pvp (USER_CTXT *uctx, unsigned process, ELAN_CAPABILITY *cap)
{
    USER_VPSEG      *seg;
    ELAN_CAPABILITY *ncap;
    unsigned         entries;

    if ((cap->cap_type & ELAN_CAP_TYPE_NO_BITMAP) == 0)
	entries = bt_nbits (cap->cap_bitmap , ELAN_CAP_BITMAPSIZE(cap));
    else
	entries = ELAN_CAP_BITMAPSIZE(cap);
    
    if ((process + entries) > (E4_VPT_MIN_ENTRIES << uctx->uctx_routetable->tbl_size))
	return (-EINVAL);

    KMEM_ALLOC (ncap, ELAN_CAPABILITY *, ELAN_CAP_SIZE (cap), 1);

    if (ncap == NULL)
	return (-ENOMEM);
    
    memcpy (ncap, cap, ELAN_CAP_SIZE (cap));

    kmutex_lock (&uctx->uctx_vpseg_lock);

    if ((seg = user_install_vpseg (uctx, process, entries)) == NULL)
    {
	kmutex_unlock (&uctx->uctx_vpseg_lock);
	return (-EINVAL);
    }
    
    seg->vps_type       = USER_VPSEG_P2P;
    seg->vps_p2p_cap    = ncap;
    seg->vps_p2p_routes = NULL;

    user_loadroute_vpseg (uctx, seg, &uctx->uctx_position);
    
    kmutex_unlock (&uctx->uctx_vpseg_lock);

    return (0);
}

int
user_add_bcastvp (USER_CTXT *uctx, unsigned process, unsigned lowvp, unsigned highvp)
{
    USER_VPSEG *seg;
    int         res;

    if (lowvp > highvp || process >= (E4_VPT_MIN_ENTRIES << uctx->uctx_routetable->tbl_size))
	return (-EINVAL);

    kmutex_lock (&uctx->uctx_vpseg_lock);

    if ((seg = user_install_vpseg (uctx, process, 1)) == NULL)
    {
	kmutex_unlock (&uctx->uctx_vpseg_lock);
	return (-EINVAL);
    }

    seg->vps_type         = USER_VPSEG_BCAST;
    seg->vps_bcast_lowvp  = lowvp;
    seg->vps_bcast_highvp = highvp;

    if ((res = user_loadroute_bcast (uctx, seg)) < 0)
	user_remove_vpseg (uctx, seg);

    kmutex_unlock (&uctx->uctx_vpseg_lock);
    return (res);
}

int
user_removevp (USER_CTXT *uctx, unsigned process)
{
    USER_VPSEG *seg;

    kmutex_lock (&uctx->uctx_vpseg_lock);
    
    if (process == ELAN_INVALID_PROCESS)
	seg = list_entry (uctx->uctx_vpseg_list.next, USER_VPSEG, vps_link);
    else
	seg = user_find_vpseg (uctx, process, process);

    if (seg == NULL)
    {
	kmutex_unlock (&uctx->uctx_vpseg_lock);
	return (-EINVAL);
    }

    do {
	ELAN4_DEV    *dev = uctx->uctx_ctxt.ctxt_dev;
	int i;

	for (i = 0; i < seg->vps_entries; i++)
	    elan4_invalidate_route (dev, uctx->uctx_routetable, seg->vps_process + i);

	user_remove_vpseg (uctx, seg);

    } while (process == ELAN_INVALID_PROCESS && (seg = list_entry (uctx->uctx_vpseg_list.next, USER_VPSEG, vps_link)) != NULL);

    kmutex_unlock (&uctx->uctx_vpseg_lock);

    return (0);
}

int
user_set_route (USER_CTXT *uctx, unsigned process, E4_VirtualProcessEntry *route)
{
    ELAN4_DEV    *dev = uctx->uctx_ctxt.ctxt_dev;
    USER_VPSEG   *seg;
    ELAN_LOCATION location;

    kmutex_lock (&uctx->uctx_vpseg_lock);

    if ((seg = user_find_vpseg (uctx, process, process)) == NULL || seg->vps_type != USER_VPSEG_P2P)
    {
	kmutex_unlock (&uctx->uctx_vpseg_lock);
	return (-EINVAL);
    }

    /* check that the route supplied is valid and goes to the correct place */
    location = user_process2location (uctx, seg, process);

    if (elan4_check_route (&uctx->uctx_position, location, route, 0) != 0)
    {
	kmutex_unlock (&uctx->uctx_vpseg_lock);
	return (-EINVAL);
    }

    if (seg->vps_p2p_routes == NULL)
	KMEM_ZALLOC (seg->vps_p2p_routes, E4_VirtualProcessEntry *, sizeof (E4_VirtualProcessEntry) * seg->vps_entries, 1);
    
    if (seg->vps_p2p_routes == NULL)
    {
	kmutex_unlock (&uctx->uctx_vpseg_lock);
	return (-ENOMEM);
    }
    
    seg->vps_p2p_routes[process - seg->vps_process].Values[0] = route->Values[0];
    seg->vps_p2p_routes[process - seg->vps_process].Values[1] = ROUTE_CTXT_VALUE(location.loc_context) | (route->Values[1] & ~ROUTE_CTXT_MASK);
    
    PRINTF (uctx, DBG_ROUTE, "user_set_route: vp=%d -> %016llx%016llx\n", process, 
	    seg->vps_p2p_routes[process - seg->vps_process].Values[1], seg->vps_p2p_routes[process - seg->vps_process].Values[0]);

    elan4_write_route (dev, uctx->uctx_routetable, process, &seg->vps_p2p_routes[process - seg->vps_process]);

    kmutex_unlock (&uctx->uctx_vpseg_lock);

    return (0);
}

int
user_reset_route (USER_CTXT *uctx, unsigned process)
{
    ELAN4_DEV             *dev = uctx->uctx_ctxt.ctxt_dev;
    E4_VirtualProcessEntry route;
    ELAN_LOCATION          location;
    USER_VPSEG            *seg;

    kmutex_lock (&uctx->uctx_vpseg_lock);

    if ((seg = user_find_vpseg (uctx, process, process)) == NULL || seg->vps_type != USER_VPSEG_P2P)
    {
	kmutex_unlock (&uctx->uctx_vpseg_lock);
	return (-EINVAL);
    }

    if (seg->vps_p2p_routes != NULL)
    {
	seg->vps_p2p_routes[process - seg->vps_process].Values[0] = 0;
	seg->vps_p2p_routes[process - seg->vps_process].Values[1] = 0;
    }
    
    /* generate the default route to this location */
    location = user_process2location (uctx, seg, process);

    PRINTF (uctx, DBG_ROUTE, "user_reset_route: vp=%d\n", process);

    if (elan4_generate_route (&uctx->uctx_position, &route, location.loc_context, location.loc_node, location.loc_node, 0) < 0)
	elan4_invalidate_route (dev, uctx->uctx_routetable, process);
    else
	elan4_write_route (dev, uctx->uctx_routetable, process, &route);

    kmutex_unlock (&uctx->uctx_vpseg_lock);

    return (0);
}

int
user_get_route (USER_CTXT *uctx, unsigned process, E4_VirtualProcessEntry *route)
{
    ELAN4_DEV  *dev = uctx->uctx_ctxt.ctxt_dev;
    USER_VPSEG   *seg;
    
    kmutex_lock (&uctx->uctx_vpseg_lock);

    if ((seg = user_find_vpseg (uctx, process, process)) == NULL || seg->vps_type != USER_VPSEG_P2P)
    {
	kmutex_unlock (&uctx->uctx_vpseg_lock);
	return (-EINVAL);
    }

    elan4_read_route (dev, uctx->uctx_routetable, process, route);

    kmutex_unlock (&uctx->uctx_vpseg_lock);
    return (0);
}

int
user_check_route (USER_CTXT *uctx, unsigned process, E4_VirtualProcessEntry *route, unsigned *error)
{
    ELAN4_DEV  *dev = uctx->uctx_ctxt.ctxt_dev;
    USER_VPSEG *seg;
    
    kmutex_lock (&uctx->uctx_vpseg_lock);

    if ((seg = user_find_vpseg (uctx, process, process)) == NULL || seg->vps_type != USER_VPSEG_P2P)
    {
	kmutex_unlock (&uctx->uctx_vpseg_lock);
	return (-EINVAL);
    }

    elan4_read_route (dev, uctx->uctx_routetable, process, route);

    *error = elan4_check_route (&uctx->uctx_position, user_process2location (uctx, seg, process), route, 0);

    kmutex_unlock (&uctx->uctx_vpseg_lock);
    return (0);
}

int
user_send_neterr_msg (USER_CTXT *uctx, unsigned int vp, unsigned int nctx, unsigned int retries, ELAN4_NETERR_MSG *msg)
{
    USER_VPSEG   *seg;
    ELAN_LOCATION location;
    unsigned long flags;
    int		  res, found = 0;
    struct list_head *el;

    kmutex_lock (&uctx->uctx_vpseg_lock);
    /* determine the location of the virtual process */
    if ((seg = user_find_vpseg (uctx, vp, vp)) == NULL)
    {
	PRINTF (uctx, DBG_NETERR, "user_send_neterr_msg: vp=%d has no vpseg\n", vp);

	kmutex_unlock (&uctx->uctx_vpseg_lock);
	return -EINVAL;
    }

    switch (seg->vps_type)
    {
    case USER_VPSEG_P2P:
	location = user_process2location (uctx, seg, vp);
	break;

    case USER_VPSEG_BCAST:
	PRINTF (uctx, DBG_NETERR, "user_send_neterr_msg: vp=%d is a bcast vp\n", vp);
	kmutex_unlock (&uctx->uctx_vpseg_lock);
	return -EINVAL;
    }

    /*  check that we're attached to the network context */
    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    list_for_each (el , &uctx->uctx_cent_list) {
	USER_CTXT_ENTRY *cent = list_entry (el, USER_CTXT_ENTRY, cent_link);
	
	if (cent->cent_cap->cap_mycontext == nctx)
	    found++;
    }
    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
    
    if (! found)
    {
	PRINTF (uctx, DBG_NETERR, "user_send_neterr_msg: nctx=%d not attached\n", nctx);
	kmutex_unlock (&uctx->uctx_vpseg_lock);

	return -EINVAL;
    }

    /* Update the fields which the user might have "faked" */
    msg->msg_context            = location.loc_context;
    msg->msg_sender.loc_node    = uctx->uctx_position.pos_nodeid;
    msg->msg_sender.loc_context = nctx;

    res = elan4_neterr_sendmsg (uctx->uctx_ctxt.ctxt_dev, location.loc_node, retries, msg);

    kmutex_unlock (&uctx->uctx_vpseg_lock);

    return (res);
}


static int
user_resolvevp (USER_CTXT *uctx, unsigned process)
{
    int                    res = 0;
    USER_VPSEG            *seg;
    ELAN_LOCATION          location;
    E4_VirtualProcessEntry route;

    PRINTF1 (uctx, DBG_VP, "user_resolvevp: process=%d\n", process);

    kmutex_lock (&uctx->uctx_vpseg_lock);

    if ((seg = user_find_vpseg (uctx, process, process)) == NULL)
    {
	kmutex_unlock (&uctx->uctx_vpseg_lock);
	return (-EINVAL);
    }

    switch (seg->vps_type)
    {
    case USER_VPSEG_P2P:
#ifdef use_elanmod
	if ((res = user_validate_cap (uctx, seg->vps_p2p_cap, ELAN_USER_P2P)) != 0)
	    break;
#endif

	location = user_process2location (uctx, seg, process);

	PRINTF (uctx, DBG_VP, "user_resolvevp: vp=%d -> node=%d ctx=%d\n", process, location.loc_node, location.loc_context);
	
	if (seg->vps_p2p_routes != NULL && seg->vps_p2p_routes[process - seg->vps_process].Values[0] != 0)
	    route = seg->vps_p2p_routes[process - seg->vps_process];
	else if ((res = elan4_generate_route (&uctx->uctx_position, &route, location.loc_context, location.loc_node, location.loc_node, user_p2p_route_options)) < 0)
	    break;;
	
	elan4_write_route (uctx->uctx_ctxt.ctxt_dev, uctx->uctx_routetable, process, &route);
	break;

    case USER_VPSEG_BCAST:
	res = user_loadroute_bcast (uctx, seg);
	break;
	
    default:
	res = -EINVAL;
	break;
    }

    kmutex_unlock (&uctx->uctx_vpseg_lock);
    return (res);
}

static void
user_eproc_trap (ELAN4_CTXT *ctxt, E4_uint64 status)
{
    USER_CTXT    *uctx = (USER_CTXT *) ctxt;
    unsigned long flags;

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);

    if (RING_QUEUE_REALLY_FULL (uctx->uctx_eprocTrapQ) || (uctx->uctx_status & UCTX_STOPPED))
    {
	PRINTF (uctx, DBG_EPROC, "user_eproc_trap: %s\n", (uctx->uctx_status & UCTX_STOPPED) ? "context stopped" : "trap queue overflow");

	uctx->uctx_status |= UCTX_EPROC_QUEUE_ERROR;
    }
    else
    {
	elan4_extract_eproc_trap (ctxt->ctxt_dev, status, RING_QUEUE_BACK (uctx->uctx_eprocTrapQ, uctx->uctx_eprocTraps), 0);
	
	DBGCMD (ctxt, DBG_EPROC, elan4_display_eproc_trap (ctxt, DBG_EPROC, "user_eproc_trap", RING_QUEUE_BACK(uctx->uctx_eprocTrapQ, uctx->uctx_eprocTraps)));
	
	if (RING_QUEUE_ADD (uctx->uctx_eprocTrapQ))
	    user_start_stopping (uctx, UCTX_EPROC_QUEUE_FULL);
    }

    user_signal_trap (uctx);

    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
}

static void
user_cproc_trap (ELAN4_CTXT *ctxt, E4_uint64 status, unsigned cqnum)
{
    USER_CTXT        *uctx = (USER_CTXT *) ctxt;
    USER_CQ          *ucq  = NULL;
    struct list_head *entry;
    unsigned long     flags;

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    
    list_for_each (entry, &uctx->uctx_cqlist) {
	ucq = list_entry (entry, USER_CQ, ucq_link);

	if (elan4_cq2num(ucq->ucq_cq) == cqnum)
	    break;
    }

    ASSERT (ucq != NULL);

    if (ucq->ucq_state != UCQ_RUNNING && CPROC_TrapType (status) == CommandProcInserterError)
    {
	PRINTF (ctxt, DBG_TRAP, "user_cproc_trap CommandProcInserterError\n");
	ucq->ucq_errored++;
    }
    else
    {
	ASSERT (ucq->ucq_state == UCQ_RUNNING);

	elan4_extract_cproc_trap (ctxt->ctxt_dev, status, &ucq->ucq_trap, cqnum);

	DBGCMD (ctxt, DBG_CPROC, elan4_display_cproc_trap (ctxt, DBG_CPROC, "user_cproc_trap", &ucq->ucq_trap));

	ucq->ucq_state = UCQ_TRAPPED;
	
    }

    user_signal_trap (uctx);
	
    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
}

static void
user_dproc_trap (ELAN4_CTXT *ctxt, E4_uint64 status, unsigned unit)
{
    USER_CTXT *uctx = (USER_CTXT *) ctxt;
    unsigned long flags;

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);

    if (RING_QUEUE_REALLY_FULL (uctx->uctx_dprocTrapQ) || (uctx->uctx_status & UCTX_STOPPED))
    {
	PRINTF (uctx, DBG_DPROC, "user_dproc_trap: %s\n", (uctx->uctx_status & UCTX_STOPPED) ? "context stopped" : "trap queue overflow");

	uctx->uctx_status |= UCTX_DPROC_QUEUE_ERROR;
    }
    else
    {
	ELAN4_DPROC_TRAP *trap = RING_QUEUE_BACK (uctx->uctx_dprocTrapQ, uctx->uctx_dprocTraps);
	
	elan4_extract_dproc_trap (ctxt->ctxt_dev, status, trap, unit);

	DBGCMD (ctxt, DBG_DPROC, elan4_display_dproc_trap (ctxt, DBG_DPROC, "user_dproc_trap", trap));

	if (!DPROC_PrefetcherFault (status) && DPROC_TrapType(status) == DmaProcFailCountError && !RING_QUEUE_FULL (uctx->uctx_dmaQ))
	{
	    trap->tr_desc.dma_typeSize |= DMA_FailCount (user_dproc_retry_count);

	    *RING_QUEUE_BACK (uctx->uctx_dmaQ, uctx->uctx_dmas) = trap->tr_desc;
    
	    (void) RING_QUEUE_ADD (uctx->uctx_dmaQ);
	}
	else
	{
	    if (RING_QUEUE_ADD (uctx->uctx_dprocTrapQ))
		user_start_stopping (uctx, UCTX_DPROC_QUEUE_FULL);
	}
    }

    user_signal_trap (uctx);

    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
}

static void
user_tproc_trap (ELAN4_CTXT *ctxt, E4_uint64 status)
{
    USER_CTXT *uctx = (USER_CTXT *) ctxt;
    unsigned long flags;

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);

    if (RING_QUEUE_REALLY_FULL (uctx->uctx_tprocTrapQ) || (uctx->uctx_status & UCTX_STOPPED))
    {
	PRINTF (uctx, DBG_TPROC, "user_tproc_trap: %s\n", (uctx->uctx_status & UCTX_STOPPED) ? "context stopped" : "trap queue overflow");

	uctx->uctx_status |= UCTX_TPROC_QUEUE_ERROR;
    }
    else
    {
	elan4_extract_tproc_trap (ctxt->ctxt_dev, status, RING_QUEUE_BACK (uctx->uctx_tprocTrapQ, uctx->uctx_tprocTraps));
	
	DBGCMD (ctxt, DBG_TPROC, elan4_display_tproc_trap (ctxt, DBG_TPROC, "user_tproc_trap", RING_QUEUE_BACK (uctx->uctx_tprocTrapQ, uctx->uctx_tprocTraps)));
	
	if (RING_QUEUE_ADD (uctx->uctx_tprocTrapQ))
	    user_start_stopping (uctx, UCTX_TPROC_QUEUE_FULL);
    }
    user_signal_trap (uctx);

    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
}

static void
user_iproc_trap (ELAN4_CTXT *ctxt, E4_uint64 status, unsigned unit)
{
    USER_CTXT       *uctx  = (USER_CTXT *) ctxt;
    USER_IPROC_TRAP *utrap = &uctx->uctx_iprocTrap[unit & 1];
    unsigned long    flags;

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);

    ASSERT (utrap->ut_state == UTS_IPROC_RUNNING);

    elan4_extract_iproc_trap (ctxt->ctxt_dev, status, &utrap->ut_trap, unit);
    DBGCMD (ctxt, DBG_IPROC, elan4_display_iproc_trap (ctxt, DBG_IPROC, "user_iproc_trap", &utrap->ut_trap));

    utrap->ut_state = UTS_IPROC_TRAPPED;

    user_start_nacking (uctx, unit ? UCTX_IPROC_CH0_TRAPPED : UCTX_IPROC_CH1_TRAPPED);

    user_signal_trap (uctx);

    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
}

static void
user_interrupt (ELAN4_CTXT *ctxt, E4_uint64 cookie)
{
    USER_CTXT *uctx = (USER_CTXT *) ctxt;
    unsigned long flags;

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);

    PRINTF1 (uctx, DBG_TRAP, "user_interrupt: cookie=%llx\n", cookie);

    switch (cookie)
    {
    case ELAN4_INT_COOKIE_DDCQ:
	uctx->uctx_ddcq_intr--;

	user_signal_trap (uctx);
	break;

    default:
	if (uctx->uctx_intcookie_table == NULL || intcookie_fire (uctx->uctx_intcookie_table, cookie) != 0)
	{
	    PRINTF2 (uctx, DBG_TRAP, "user_interrupt: cookie=%llx %s\n", cookie, uctx->uctx_intcookie_table ? "not found" : "no table");
	    uctx->uctx_status |= UCTX_EPROC_QUEUE_ERROR;
	    user_signal_trap (uctx);
	}
	break;
    }

    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
}
static void
user_needs_shuffle (ELAN4_CTXT *ctxt, int tbl, int hashidx)
{
    USER_CTXT     *uctx = (USER_CTXT *) ctxt;
    unsigned long  flags;

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);

    elan4mmu_set_shuffle(ctxt, tbl, hashidx);

    if (ctxt->shuffle_needed[tbl]) 
	user_shuffle_signal_trap (uctx);

    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);

}
static void
user_neterrmsg (ELAN4_CTXT *ctxt, ELAN4_NETERR_MSG *msg)
{
    USER_CTXT *uctx = (USER_CTXT *) ctxt;
    unsigned long flags;

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    
    if (! RING_QUEUE_FULL (uctx->uctx_msgQ))
    {
	memcpy (RING_QUEUE_BACK (uctx->uctx_msgQ, uctx->uctx_msgs), msg, sizeof (ELAN4_NETERR_MSG));

	(void) RING_QUEUE_ADD (uctx->uctx_msgQ);
    
	user_signal_trap (uctx);
    }
    
    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
}

ELAN4_TRAP_OPS user_trap_ops = 
{
    user_eproc_trap,
    user_cproc_trap,
    user_dproc_trap,
    user_tproc_trap,
    user_iproc_trap,
    user_interrupt,
    user_neterrmsg,
    user_needs_shuffle,
};

static int
deliver_trap (ELAN4_USER_TRAP *utrapp, int type, unsigned proc, void *trap, ...)
{
    register int i, len;
    va_list ap;

    PRINTF (NULL, DBG_TRAP, "deliver_trap: type=%d proc=%d\n", type, proc);

    switch (proc)
    {
    case UTS_CPROC:      len = sizeof (ELAN4_CPROC_TRAP); break;
    case UTS_DPROC:      len = sizeof (ELAN4_DPROC_TRAP); break;
    case UTS_EPROC:      len = sizeof (ELAN4_EPROC_TRAP); break;
    case UTS_IPROC:      len = sizeof (ELAN4_IPROC_TRAP); break;
    case UTS_TPROC:      len = sizeof (ELAN4_TPROC_TRAP); break;
    case UTS_NETERR_MSG: len = sizeof (ELAN4_NETERR_MSG); break;
    default:             len = 0; break;
    }

    if (put_user (type, &utrapp->ut_type) || put_user (proc, &utrapp->ut_proc) || copy_to_user (&utrapp->ut_trap, trap, len))
	return (UTS_EFAULT);

    va_start (ap, trap);
    for (i = 0; i < sizeof (utrapp->ut_args)/sizeof (utrapp->ut_args[0]); i++)
	if (put_user (va_arg (ap, unsigned long), &utrapp->ut_args[i]))
	    return (UTS_EFAULT);
    va_end (ap);

    return (type);
}

static int
user_pagefault (USER_CTXT *uctx, E4_FaultSave *farea)
{
    E4_Addr      addr = farea->FaultAddress;
    E4_uint32    fsr  = FaultSaveFSR(farea->FSRAndFaultContext);
    FAULT_SAVE  *entry;
    FAULT_SAVE **predp;
    int count;

    PRINTF2 (uctx, DBG_FAULT, "user_pagefault: addr=%llx fsr %x\n", (unsigned long long) addr, fsr);
    
    if ((fsr & FSR_FaultForBadData) != 0)			/* Memory ECC error during walk */
    {
	PRINTF0 (uctx, DBG_FAULT, "user_pagefault: ECC error during walk\n");
	return (-EFAULT);
    }
    
    if ((fsr & FSR_FaultForMaxChainCount) != 0)			/* Have walked a chain of 1024 items */
    {
	PRINTF0 (uctx, DBG_FAULT, "user_pagefault: pte chain too long\n");
	return (-EFAULT);
    }
    
    if (! user_pagefault_enabled)
	return (-EFAULT);

    if (uctx->uctx_num_fault_save)
    {
        spin_lock (&uctx->uctx_fault_lock);
        for( predp = &uctx->uctx_fault_list; (entry = *predp)->next != NULL; predp = &entry->next)
        {
	    if (entry->addr == (addr & ~((E4_Addr) PAGE_SIZE-1)))
	        break;
        }

        *predp = entry->next;
        entry->next = uctx->uctx_fault_list;
        uctx->uctx_fault_list = entry;

        if (entry->addr == (addr & ~((E4_Addr) PAGE_SIZE-1)))
        {
	    if ((entry->count <<= 1) > max_fault_pages)
	        entry->count = max_fault_pages;
        }
        else
	    entry->count = min_fault_pages;

        entry->addr = (addr & ~((E4_Addr) PAGE_SIZE-1))+(entry->count * PAGE_SIZE);
        count = entry->count;
        spin_unlock (&uctx->uctx_fault_lock);

        if (user_load_range (uctx, addr & ~((E4_Addr) PAGE_SIZE-1), count * PAGESIZE, fsr) == 0)
	    return 0;

	/* else pre-faulting has failed, try just this page */
    }

    return (user_load_range (uctx, addr & ~((E4_Addr) PAGE_SIZE-1), PAGE_SIZE, fsr));

}

static int
queue_dma_for_retry (USER_CTXT *uctx, ELAN4_USER_TRAP *utrapp, E4_DMA *dma)
{
    unsigned long flags;

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);

    if (RING_QUEUE_FULL (uctx->uctx_dmaQ))
    {
	spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
	
	PRINTF (uctx, DBG_DPROC, "queue_dma_for_retry: overflow\n");
	
	return (deliver_trap (utrapp, UTS_QUEUE_OVERFLOW, UTS_NOPROC, NULL, UCTX_DPROC_QUEUE_OVERFLOW));
    }

    *RING_QUEUE_BACK (uctx->uctx_dmaQ, uctx->uctx_dmas) = *dma;
    
    (void) RING_QUEUE_ADD (uctx->uctx_dmaQ);
    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);

    return (UTS_FINISHED);
}

static int
queue_thread_for_retry (USER_CTXT *uctx, ELAN4_USER_TRAP *utrapp, E4_ThreadRegs *regs)
{
    unsigned long flags;
    
    spin_lock_irqsave (&uctx->uctx_spinlock, flags);

    if (RING_QUEUE_FULL (uctx->uctx_threadQ))
    {
	spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);

	PRINTF (uctx, DBG_TPROC, "queue_thread_for_retry: overflow\n");

	return (deliver_trap (utrapp, UTS_QUEUE_OVERFLOW, UTS_NOPROC, NULL, UCTX_TPROC_QUEUE_OVERFLOW));
    }

    *RING_QUEUE_BACK (uctx->uctx_threadQ, uctx->uctx_threads) = *regs;
    (void) RING_QUEUE_ADD (uctx->uctx_threadQ);
    
    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);

    return (UTS_FINISHED);
}

static int
fixup_eproc_trap (USER_CTXT *uctx, ELAN4_EPROC_TRAP *trap, int waitevent)
{
    E4_FaultSave *farea = &trap->tr_faultarea;
    E4_uint32     fsr   = FaultSaveFSR(farea->FSRAndFaultContext);
    E4_uint64     CountAndType;
    E4_uint64     CopySource;
    E4_uint64     CopyDest;

    /*
     * Event processor can trap as follows :
     *   1) Event location read		(faddr == event location & Event Permission)
     *   2) Event location write	(faddr == event location & Event Permission)
     *   3) Copy Source read		Read Access
     *   4) Copy/Write dest write	other
     *
     *  NOTE - it is possible to see both 3) and 4) together - but only with physical errors.
     */
    if (AT_Perm(fsr) == AT_PermLocalDataRead || AT_Perm(fsr) == AT_PermLocalDataWrite)
    {
	/* 
	 * We complete the copy/write by issuing a waitevent 0 of the approriate type.
	 *   - NB mask off bottom bits of EventAddr in case of partial setevent
	 */
	E4_uint64 EventAddr = trap->tr_eventaddr & ~((E4_uint64) E4_EVENT_ALIGN-1);

	if (! user_ddcq_check (uctx, 4))
	    return (0);
	
	if ((trap->tr_event.ev_CountAndType & E4_EVENT_COPY_TYPE_MASK) == E4_EVENT_WRITE)
	{
	    /* case 4) faulted on write word to destination */

	    CountAndType = trap->tr_event.ev_CountAndType & E4_EVENT_TYPE_MASK;
	    
	    PRINTF (uctx, DBG_TRAP, "fixup_eproc_trap: write Event=%llx CountAndType=%llx\n", EventAddr, CountAndType);
	    PRINTF (uctx, DBG_TRAP, "                  WritePtr=%llx WriteValue=%llx\n", 
		    trap->tr_event.ev_WritePtr, trap->tr_event.ev_WriteValue);

	    user_ddcq_waitevent (uctx, EventAddr, CountAndType, trap->tr_event.ev_WritePtr, trap->tr_event.ev_WriteValue);
	}
	else
	{
	    /* case 3) or case 4) faulted on read/write of copy */
	    if (AT_Perm (fsr) == AT_PermLocalDataRead)
	    {
		CountAndType = (trap->tr_event.ev_CountAndType & E4_EVENT_DATA_TYPE_MASK) | EPROC_CopySize(trap->tr_status);
		CopySource   = trap->tr_event.ev_CopySource - EVENT_COPY_BLOCK_SIZE;
		CopyDest     = trap->tr_event.ev_CopyDest;
	    }
	    else
	    {
		CountAndType = ((trap->tr_event.ev_CountAndType & E4_EVENT_DATA_TYPE_MASK) | 
				((EPROC_CopySize(trap->tr_status) + EVENT_COPY_NDWORDS) & E4_EVENT_COPY_SIZE_MASK));
		CopySource   = trap->tr_event.ev_CopySource - EVENT_COPY_BLOCK_SIZE;
		CopyDest     = trap->tr_event.ev_CopyDest - EVENT_COPY_BLOCK_SIZE;
	    }
	    
	    PRINTF (uctx, DBG_TRAP, "fixup_eproc_trap: copy Event=%llx CountAndType=%llx\n", EventAddr, CountAndType);
	    PRINTF (uctx, DBG_TRAP, "                  CopySource=%llx CopyDest=%llx\n", CopySource, CopyDest);

	    user_ddcq_waitevent (uctx, EventAddr, CountAndType, CopySource, CopyDest);
	}
    }
    else
    {
	E4_uint64 EventAddr  = trap->tr_eventaddr & ~((E4_uint64) E4_EVENT_ALIGN-1);
	E4_uint32 EventCount = trap->tr_eventaddr & (E4_EVENT_ALIGN-1);

	/* case 1) or 2) - just reissue the event */
	if (! waitevent)
	    PRINTF (uctx, DBG_TRAP, "fixup_eproc_trap: setevent EventAddr=%llx EventCount=%x\n", EventAddr, EventCount);
	else
	{
	    PRINTF (uctx, DBG_TRAP, "fixup_eproc_trap: waitevent Event=%llx CountAndType=%llx\n", EventAddr, trap->tr_event.ev_CountAndType);
	    PRINTF (uctx, DBG_TRAP, "                  Param[0]=%llx Param[1]=%llx\n",
		     trap->tr_event.ev_Params[0], trap->tr_event.ev_Params[1]);
	}

	if (! user_ddcq_check (uctx, waitevent ? 4 : 2))
	    return (0);
	
	if (waitevent)
	    user_ddcq_waitevent (uctx, EventAddr, trap->tr_event.ev_CountAndType, 
				  trap->tr_event.ev_Params[0], trap->tr_event.ev_Params[1]);
	else
	    user_ddcq_seteventn (uctx, EventAddr, EventCount);
    }

    return (1);
}


static int
resolve_eproc_trap (USER_CTXT *uctx, ELAN4_USER_TRAP *utrapp, ELAN4_EPROC_TRAP *trap)
{
    switch (EPROC_TrapType (trap->tr_status))
    {
    case EventProcNoFault:
	PRINTF (uctx, DBG_TRAP, "resolve_eproc_trap: EventProcNoFault\n");

	return (UTS_FINISHED);
	
    case EventProcAddressAlignment:
	return (deliver_trap (utrapp, UTS_ALIGNMENT_ERROR, UTS_EPROC, trap));

    case EventProcMemoryFault:
	PRINTF (uctx, DBG_TRAP, "resolve_eproc_trap: EventProcMemoryFault @ %llx\n", trap->tr_faultarea.FaultAddress);

	if (user_pagefault (uctx, &trap->tr_faultarea) != 0)
	    return (deliver_trap (utrapp, UTS_INVALID_ADDR, UTS_EPROC, trap));

	return (UTS_FINISHED);
	
    case EventProcCountWrapError:
	return (deliver_trap (utrapp, UTS_BAD_TRAP, UTS_EPROC, trap));

    default:
	printk ("resolve_eproc_trap: bad trap type %d\n", EPROC_TrapType (trap->tr_status));
	BUG();
    }

    return (UTS_FINISHED);
}

static int
resolve_cproc_trap (USER_CTXT *uctx, ELAN4_USER_TRAP *utrapp, USER_CQ *ucq)
{
    ELAN4_DEV        *dev    = uctx->uctx_ctxt.ctxt_dev;
    ELAN4_CPROC_TRAP *trap   = &ucq->ucq_trap;
    E4_uint64         command;
    int               res;
    int               chan;

    ELAN_LOCATION location;
    int vp, node;

    PRINTF2 (uctx, DBG_CPROC, "resolve_cproc_trap: cq %p is trapped - Status %lx\n", ucq, trap->tr_status);
    
    switch (CPROC_TrapType (trap->tr_status))
    {
    case CommandProcDmaQueueOverflow:
	PRINTF (uctx, DBG_CPROC, "resolve_cproc_trap: CommandProcDmaQueueOverflow\n");
	/*
	 * XXXX: should wait for the queue to become emptier if we're 
	 *       responsible for it being very full
	 */
	ucq->ucq_state = UCQ_NEEDS_RESTART;
	break;

    case CommandProcInterruptQueueOverflow:
	PRINTF (uctx, DBG_CPROC, "resolve_cproc_trap: CommandProcInterruptQueueOverflow\n");
	/*
	 * XXXX: should wait for the queue to become emptier if we're
	 *       responsible for it being very full
	 */
	ucq->ucq_state = UCQ_NEEDS_RESTART;
	break;
	
    case CommandProcWaitTrap:
	PRINTF0 (uctx, DBG_CPROC, "resolve_cproc_trap: CommandProcWaitTrap\n");
	
	if ((res = resolve_eproc_trap (uctx, utrapp, &trap->tr_eventtrap)) != UTS_FINISHED)
	{
	    ucq->ucq_state = UCQ_STOPPED;

	    return (res);
	}
	
	if (fixup_eproc_trap (uctx, &trap->tr_eventtrap, 1) == 0)
	    return UTS_RESCHEDULE;

	ucq->ucq_state = UCQ_NEEDS_RESTART;
	break;
	
    case CommandProcMemoryFault:
	PRINTF1 (uctx, DBG_CPROC, "resolve_cproc_trap: CommandProcMemoryFault at %llx\n", trap->tr_faultarea.FaultAddress);
	if (user_pagefault (uctx, &trap->tr_faultarea) != 0)
	{
	    ucq->ucq_state = UCQ_STOPPED;

	    return (deliver_trap (utrapp, UTS_INVALID_ADDR, UTS_CPROC, trap, elan4_cq2idx(ucq->ucq_cq)));
	}
	
	ucq->ucq_state = UCQ_NEEDS_RESTART;
	break;
	
    case CommandProcRouteFetchFault:
	command = elan4_trapped_open_command (dev, ucq->ucq_cq);
	
	PRINTF1 (uctx, DBG_CPROC, "resolve_cproc_trap: CommandProcRouteFetchFault to vp %d\n", (int) (command >> 32));
	
	if (user_resolvevp (uctx, (unsigned) (command >> 32)) != 0)
	{
	    ucq->ucq_state = UCQ_STOPPED;

	    return (deliver_trap (utrapp, UTS_INVALID_VPROC, UTS_CPROC, trap, elan4_cq2idx(ucq->ucq_cq), (long) (command >> 32)));
	}

	ucq->ucq_state = UCQ_NEEDS_RESTART;
	break;
	
    case CommandProcFailCountZero:
	PRINTF0 (uctx, DBG_CPROC, "resolve_cproc_trap: CommandProcFailCountZero - reset failcount\n");
	
	/* Update CPROC timeout route statistics */
	for (chan = 0; chan <= 1; chan++)
	{
	    /* Was there a timeout on this channel ? */
	    if (PackValue(trap->tr_qdesc.CQ_AckBuffers, chan) == PackTimeout)
	    {
		/* Find the last open command for that channel to extract the relevant vp */
		if ((vp = cproc_open_extract_vp(uctx->uctx_ctxt.ctxt_dev, ucq->ucq_cq, chan)) != -1)
		{
		    E4_VirtualProcessEntry route;
		    
		    kmutex_lock (&uctx->uctx_vpseg_lock);
		    location = user_process2location(uctx, NULL, vp);
		    elan4_read_route (uctx->uctx_ctxt.ctxt_dev, uctx->uctx_routetable, vp, &route);
		    kmutex_unlock (&uctx->uctx_vpseg_lock);
		    node = location.loc_node;
		    
		    kmutex_lock(&uctx->uctx_ctxt.ctxt_dev->dev_lock);
		    
		    if ((node >= 0) && (node < uctx->uctx_ctxt.ctxt_dev->dev_position.pos_nodes))
		    {
			uctx->uctx_ctxt.ctxt_dev->dev_cproc_timeout[node]++;
			
			elan4_ringbuf_store(&uctx->uctx_ctxt.ctxt_dev->dev_cproc_timeout_routes,
					    &route, uctx->uctx_ctxt.ctxt_dev);
		    }
		    
		    kmutex_unlock(&uctx->uctx_ctxt.ctxt_dev->dev_lock);
		}
	    }
	}
	    
	/* NOTE - we must not modify the ChannelNotCompleted bits - so modify */
	/*        the restart count with a part-word store */
	elan4_updatecq (dev, ucq->ucq_cq, ucq->ucq_cq->cq_perm, user_cproc_retry_count);

	ucq->ucq_state = UCQ_NEEDS_RESTART;
	break;

    case CommandProcAddressAlignment:
	ucq->ucq_state = UCQ_STOPPED;

	return (deliver_trap (utrapp, UTS_ALIGNMENT_ERROR, UTS_CPROC, trap, elan4_cq2idx(ucq->ucq_cq)));

    case CommandProcPermissionTrap:
    {
	sdramaddr_t cqdesc = dev->dev_cqaddr + (elan4_cq2num(ucq->ucq_cq) * sizeof (E4_CommandQueueDesc));
	E4_uint64   control = elan4_sdram_readq (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_Control));

	PRINTF (uctx, DBG_CPROC, "resolve_cproc_trap: CommandProcPermissionTrap - %s\n",
		(control & CQ_PermissionMask) != ucq->ucq_cq->cq_perm ? "resume from stop" : "permission denied");
	
	if ((control & CQ_PermissionMask) == ucq->ucq_cq->cq_perm)
	    return (deliver_trap (utrapp, UTS_PERMISSION_DENIED, UTS_CPROC, trap, elan4_cq2idx(ucq->ucq_cq)));

	elan4_updatecq (dev, ucq->ucq_cq, ucq->ucq_cq->cq_perm, 0);

	ucq->ucq_state = UCQ_NEEDS_RESTART;
	break;
    }
    
    case CommandProcBadData:
	ucq->ucq_state = UCQ_STOPPED;

	return (deliver_trap (utrapp, UTS_INVALID_COMMAND, UTS_CPROC, trap, elan4_cq2idx(ucq->ucq_cq)));

    default:
	ucq->ucq_state = UCQ_STOPPED;

	return (deliver_trap (utrapp, UTS_BAD_TRAP, UTS_CPROC, trap, elan4_cq2idx(ucq->ucq_cq)));
    }

    return (UTS_FINISHED);
}

static int
resolve_dproc_trap (USER_CTXT *uctx, ELAN4_USER_TRAP *utrapp, ELAN4_DPROC_TRAP *trap)
{
    ELAN_LOCATION location;
    int node;
    E4_VirtualProcessEntry route;

    if (DPROC_PrefetcherFault (trap->tr_status))
    {
	PRINTF (uctx, DBG_DPROC, "resolve_dproc_trap: PrefetcherFault at %llx\n", trap->tr_prefetchFault.FaultAddress);

	if (user_pagefault (uctx, &trap->tr_prefetchFault) != 0)
	    return (deliver_trap (utrapp, UTS_INVALID_ADDR, UTS_DPROC, trap));
	
	return (queue_dma_for_retry (uctx, utrapp, &trap->tr_desc));
    }
    
    switch (DPROC_TrapType (trap->tr_status))
    {
    case DmaProcRouteFetchFault:
	PRINTF (uctx, DBG_DPROC, "resolve_dproc_trap: DmaProcRouteFetchFault vp %d\n", trap->tr_desc.dma_vproc);

	if (user_resolvevp (uctx, trap->tr_desc.dma_vproc) != 0)
	    return (deliver_trap (utrapp, UTS_INVALID_VPROC, UTS_DPROC, trap, trap->tr_desc.dma_vproc));
	
	return (queue_dma_for_retry (uctx, utrapp, &trap->tr_desc /* immediate */));
	
    case DmaProcFailCountError:
	PRINTF (uctx, DBG_DPROC, "resolve_dproc_trap: DmaProcFailCountError - vp %d cookie %llx\n",
		trap->tr_desc.dma_vproc, trap->tr_desc.dma_cookie);

	trap->tr_desc.dma_typeSize |= DMA_FailCount (user_dproc_retry_count);

	return (queue_dma_for_retry (uctx, utrapp, &trap->tr_desc /* XXXX - backoff for some time later */));

    case DmaProcPacketAckError:
	PRINTF (uctx, DBG_DPROC, "resolve_dproc_trap: DmaProcPacketAckError - %d%s\n", DPROC_PacketAckValue (trap->tr_status), 
		DPROC_PacketTimeout (trap->tr_status) ? " timeout" : "");

	kmutex_lock (&uctx->uctx_vpseg_lock);
	location = user_process2location(uctx, NULL, trap->tr_desc.dma_vproc);
	elan4_read_route(uctx->uctx_ctxt.ctxt_dev, uctx->uctx_routetable, trap->tr_desc.dma_vproc, &route);
	kmutex_unlock (&uctx->uctx_vpseg_lock);
	node = location.loc_node;

	/* Update dproc route timeout statistics */
	if ((node >= 0) && (node < uctx->uctx_ctxt.ctxt_dev->dev_position.pos_nodes))
	{
	    kmutex_lock(&uctx->uctx_ctxt.ctxt_dev->dev_lock);
	    
	    if ((route.Values[0] != 0) || (route.Values[1] != 0))
	    {
		if (DPROC_PacketTimeout (trap->tr_status))
		{
		    uctx->uctx_ctxt.ctxt_dev->dev_dproc_timeout[node]++;
		    elan4_ringbuf_store(&uctx->uctx_ctxt.ctxt_dev->dev_dproc_timeout_routes,
					&route, uctx->uctx_ctxt.ctxt_dev);
		}
		else
		{
		    uctx->uctx_ctxt.ctxt_dev->dev_ack_errors[node]++;
		    elan4_ringbuf_store(&uctx->uctx_ctxt.ctxt_dev->dev_ack_error_routes,
					&route, uctx->uctx_ctxt.ctxt_dev);
		}
	    }
	    
	    kmutex_unlock(&uctx->uctx_ctxt.ctxt_dev->dev_lock);
	}

	return (queue_dma_for_retry (uctx, utrapp, &trap->tr_desc /* XXXX - backoff for some time later */));

    case DmaProcQueueOverflow:
	PRINTF (uctx, DBG_DPROC, "resolve_dproc_trap: DmaProcQueueOverflow\n");
	return (queue_dma_for_retry (uctx, utrapp, &trap->tr_desc /* XXXX - backoff for some time later */));
	
    case DmaProcRunQueueReadFault:
	return (deliver_trap (utrapp, UTS_BAD_TRAP, UTS_DPROC, trap));
	
    default:
	printk ("resolve_dproc_trap: unknown trap type : %d\n", DPROC_TrapType(trap->tr_status));
	BUG();
    }
    return UTS_FINISHED;
}

int
resolve_tproc_trap (USER_CTXT *uctx, ELAN4_USER_TRAP *utrapp, ELAN4_TPROC_TRAP *trap)
{
    PRINTF (uctx, DBG_TPROC, "resolve_tproc_trap: trap state = %lx\n", trap->tr_state);

    if (trap->tr_state & TS_TrapForTooManyInstructions)
	return (deliver_trap (utrapp, UTS_BAD_TRAP, UTS_TPROC, trap));
    
    if (trap->tr_state & TS_Unimplemented)
	return (deliver_trap (utrapp, UTS_UNIMP_INSTR, UTS_TPROC, trap));
    
    if (trap->tr_state & TS_DataAlignmentError)
	return (deliver_trap (utrapp, UTS_ALIGNMENT_ERROR, UTS_TPROC, trap));
    
    if ((trap->tr_state & TS_InstAccessException) && user_pagefault (uctx, &trap->tr_instFault) != 0)
	return (deliver_trap (utrapp, UTS_INVALID_ADDR, UTS_TPROC, trap));
    
    if ((trap->tr_state & TS_DataAccessException) && user_pagefault (uctx, &trap->tr_dataFault) != 0)
	return (deliver_trap (utrapp, UTS_INVALID_ADDR, UTS_TPROC, trap));
    
    /* If we're restarting from trap - then just need to re-issue it */
    if (trap->tr_pc == uctx->uctx_trestart_addr || (trap->tr_state & TS_TrappedFlag))
    {
	PRINTF (uctx, DBG_TPROC, "resolve_tproc_trap: trapped in trap code PC=%llx SP=%llx\n", trap->tr_pc, trap->tr_regs[1]);

	trap->tr_regs[0] = uctx->uctx_trestart_addr;
    }
    else
    {
	E4_uint64 *sp = (E4_uint64 *) user_elan2main (uctx, trap->tr_regs[1]);
	int        i, reload;

	/* need to store the register on the stack see */
	/*  lib_tproc_trampoline_elan4_thread.S for stack layout */
#define TS_STACK_OFF(REG)	((((REG)&7)) - (((REG)>>3)*8) - 8)
	for (reload = 0, i = 0; i < 64; i++)
	{
	    if (trap->tr_dirty & ((E4_uint64) 1 << i))
	    {
		PRINTF (uctx, DBG_TPROC, "resolve_tproc_trap: %%r%d  [%016llx] -> %p\n", i, trap->tr_regs[i], &sp[TS_STACK_OFF(i)]);

		sulonglong ((u64 *) &sp[TS_STACK_OFF(i)], trap->tr_regs[i]);
		
		reload |= (1 << (i >> 3));
	    }
	}
#undef TS_STACK_OFF

	PRINTF (uctx, DBG_TPROC, "resolve_tproc_trap: pc %llx npc %llx\n", trap->tr_pc, trap->tr_npc);
	PRINTF (uctx, DBG_TPROC, "resolve_tproc_trap: CC %x reload %x\n", (int) (trap->tr_state >> TS_XCCshift), reload);

	trap->tr_regs[0] = uctx->uctx_trestart_addr;
	trap->tr_regs[2] = trap->tr_pc;
	trap->tr_regs[3] = trap->tr_npc;
	trap->tr_regs[4] = (trap->tr_state >> TS_XCCshift) & TS_XCCmask;
	trap->tr_regs[5] = reload;
    }

    return (queue_thread_for_retry (uctx, utrapp, (E4_ThreadRegs *) trap->tr_regs));
}

static int
resolve_iproc_trap (USER_CTXT *uctx, ELAN4_USER_TRAP *utrapp, int channel)
{
    USER_IPROC_TRAP  *utrap = &uctx->uctx_iprocTrap[channel];
    ELAN4_IPROC_TRAP *trap  = &utrap->ut_trap;
    unsigned long     flags;

    elan4_inspect_iproc_trap (trap);

    if (trap->tr_flags & TR_FLAG_TOOMANY_TRANS)
	return (deliver_trap (utrapp, UTS_INVALID_TRANS, UTS_IPROC, trap, channel));

    ASSERT (trap->tr_trappedTrans >= 0 && trap->tr_trappedTrans < trap->tr_numTransactions);

    switch (IPROC_TrapValue (trap->tr_transactions[trap->tr_trappedTrans].IProcStatusCntxAndTrType))
    {
    case InputMemoryFault:
	if (user_pagefault (uctx, &trap->tr_faultarea) != 0)
	{
	    utrap->ut_state = UTS_IPROC_STOPPED;
	    
	    return (deliver_trap (utrapp, UTS_INVALID_ADDR, UTS_IPROC, trap, channel));
	}
	break;

    case InputDmaQueueOverflow:
    case InputEventEngineTrapped:
	/* nothing to do for these 2 - restarting will simulate the transactions */
	break;

    case InputEopErrorOnWaitForEop:
    case InputEopErrorTrap:
	break;

    case InputCrcErrorAfterPAckOk:
	PRINTF (DBG_DEVICE, DBG_IPROC, "InputCrcErrorAfterPAckOk: flags %x\n", trap->tr_flags);

	ASSERT ((trap->tr_flags & TR_FLAG_ACK_SENT) && ((trap->tr_flags & (TR_FLAG_DMA_PACKET|TR_FLAG_BAD_TRANS)) ||
							((trap->tr_flags & TR_FLAG_EOP_ERROR) && trap->tr_identifyTrans == TR_TRANS_INVALID)));
	break;

    case InputDiscardAfterAckOk:
	return (deliver_trap (utrapp, UTS_INVALID_TRANS, UTS_IPROC, trap, channel));

    case InputAddressAlignment:
	return (deliver_trap (utrapp, UTS_ALIGNMENT_ERROR, UTS_IPROC, trap, channel));

    case InputInvalidTransType:
	return (deliver_trap (utrapp, UTS_INVALID_TRANS, UTS_IPROC, trap, channel));

    default:
	printk ("resolve_iproc_trap: unknown trap type %d\n", IPROC_TrapValue (trap->tr_transactions[trap->tr_trappedTrans].IProcStatusCntxAndTrType));
	BUG();
	/* NOTREACHED */
    }

    if (! (trap->tr_flags & TR_FLAG_ACK_SENT) || (trap->tr_flags & TR_FLAG_EOP_BAD))
    {
	spin_lock_irqsave (&uctx->uctx_spinlock, flags);

	utrap->ut_state = UTS_IPROC_RUNNING;

	user_stop_nacking (uctx, channel ? UCTX_IPROC_CH0_TRAPPED : UCTX_IPROC_CH1_TRAPPED);
	
	spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
    }
    else if ((trap->tr_flags & (TR_FLAG_DMA_PACKET | TR_FLAG_BAD_TRANS)) || ((trap->tr_flags & TR_FLAG_EOP_ERROR) && (trap->tr_identifyTrans == TR_TRANS_INVALID)))
    {
	/* 
	 * TR_FLAG_DMA_PACKET	means a DMA packet has faulted.
	 *
	 * TR_FLAG_BAD_TRANS	means we have a transaction with a bad CRC after the transaction
	 *			which sent the ack - this implies it's an overlapped ack DMA packet
	 *
	 * TR_FLAG_EOP_ERROR	means we've received an EOP reset - if we hadn't seen an identify
	 *			transaction then it's a DMA packet.
	 *
	 * To ensure that the DMA processor works correctly the next packet must be NACKed to 
	 * cause it to resend this one.
	 */
	PRINTF (uctx, DBG_IPROC, "resolve_iproc_trap: %s during DMA packet\n",
		(trap->tr_flags & TR_FLAG_BAD_TRANS) ? "BadTransaction" : (trap->tr_flags & TR_FLAG_EOP_ERROR) ? "EopError" : "trap");

	spin_lock_irqsave (&uctx->uctx_spinlock, flags);

	if (trap->tr_flags & TR_FLAG_DMA_PACKET)
	{
	    if (! (trap->tr_flags & TR_FLAG_BAD_TRANS))
		utrap->ut_state = UTS_IPROC_EXECUTE_PACKET;
	    else
	    {
		kcondvar_t waithere;

		/* We must ensure that the next packet is always nacked, so
		 * we wait here for an output timeout before dropping the 
		 * context filter - we just pause here for 4 mS */
		kcondvar_init (&waithere);
		kcondvar_timedwait (&waithere, &uctx->uctx_spinlock, &flags, lbolt + (HZ/250) + 1);;
		kcondvar_destroy (&waithere);

		utrap->ut_state = UTS_IPROC_RUNNING;
		
		user_stop_nacking (uctx, channel ? UCTX_IPROC_CH0_TRAPPED : UCTX_IPROC_CH1_TRAPPED);
	    }
	}
	else
	{
	    utrap->ut_state = UTS_IPROC_RUNNING;

	    user_stop_nacking (uctx, channel ? UCTX_IPROC_CH0_TRAPPED : UCTX_IPROC_CH1_TRAPPED);
	}

	spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
    }
    else if (trap->tr_flags & TR_FLAG_EOP_ERROR)
    {
	PRINTF (uctx, DBG_IPROC, "resolve_iproc_trap: EopError with identify\n");

	utrap->ut_state = UTS_IPROC_NETWORK_ERROR;
    }
    else
    {
	PRINTF (uctx, DBG_IPROC, "resolve_iproc_trap: execute packet\n");

	utrap->ut_state = UTS_IPROC_EXECUTE_PACKET;
    }

    return UTS_FINISHED;
}


static int
resolve_cproc_traps (USER_CTXT *uctx, ELAN4_USER_TRAP *utrapp)
{
    struct list_head *entry;
    int res = UTS_FINISHED;

    kmutex_lock (&uctx->uctx_cqlock);
    list_for_each (entry, &uctx->uctx_cqlist) {
	USER_CQ *ucq = list_entry (entry, USER_CQ, ucq_link);

	if (ucq->ucq_state == UCQ_TRAPPED)
	{
	    res = resolve_cproc_trap (uctx, utrapp, ucq);

	    if (res != UTS_FINISHED)
		break;
	}

	if (ucq->ucq_errored)
	{
	    ucq->ucq_errored = 0;
	    res = deliver_trap (utrapp, UTS_CPROC_ERROR, UTS_CPROC, &ucq->ucq_trap, elan4_cq2idx(ucq->ucq_cq));
	    break;
	}
    }
    kmutex_unlock (&uctx->uctx_cqlock);

    return (res);
}

static int
resolve_eproc_traps (USER_CTXT *uctx, ELAN4_USER_TRAP *utrapp)
{
    unsigned long flags;
    int res;

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    while (! RING_QUEUE_EMPTY (uctx->uctx_eprocTrapQ))
    {
	ELAN4_EPROC_TRAP trap = *RING_QUEUE_FRONT (uctx->uctx_eprocTrapQ, uctx->uctx_eprocTraps);

	(void) RING_QUEUE_REMOVE (uctx->uctx_eprocTrapQ);

	spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);

	if ((res = resolve_eproc_trap (uctx, utrapp, &trap)) != UTS_FINISHED)
	    return (res);

	if (fixup_eproc_trap (uctx, &trap, 0) == 0)
	{
	    PRINTF (uctx, DBG_EPROC, "resolve_eproc_trap: could not fixup eproc trap - requeue it\n");

	    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
	    if (RING_QUEUE_REALLY_FULL(uctx->uctx_eprocTrapQ))
	    {
		PRINTF (uctx, DBG_EPROC, "resolve_eproc_trap: queue overflow\n");
		uctx->uctx_status |= UCTX_EPROC_QUEUE_OVERFLOW;
	    }
	    else
	    {
		/* Requeue at front to preserve setevent ordering */
		/* GNAT 7504: Must move fptr before writing over it */
		(void) RING_QUEUE_ADD_FRONT(uctx->uctx_eprocTrapQ);
		
		*RING_QUEUE_FRONT(uctx->uctx_eprocTrapQ, uctx->uctx_eprocTraps) = trap;
	    }
	    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);

	    return UTS_RESCHEDULE;
	}
	
	spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    }

    if (uctx->uctx_status & UCTX_EPROC_QUEUE_FULL)
	user_stop_stopping (uctx, UCTX_EPROC_QUEUE_FULL);

    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
    return (UTS_FINISHED);
}
	    
static int
resolve_dproc_traps (USER_CTXT *uctx, ELAN4_USER_TRAP *utrapp)
{
    unsigned long flags;
    int res;
    
    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    while (! RING_QUEUE_EMPTY (uctx->uctx_dprocTrapQ))
    {
	ELAN4_DPROC_TRAP trap = *RING_QUEUE_FRONT(uctx->uctx_dprocTrapQ, uctx->uctx_dprocTraps);
	
	(void) RING_QUEUE_REMOVE (uctx->uctx_dprocTrapQ);

	spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);

	if ((res = resolve_dproc_trap (uctx, utrapp, &trap)) != UTS_FINISHED)
	    return (res);
	
	spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    }

    if (uctx->uctx_status & UCTX_DPROC_QUEUE_FULL)
	user_stop_stopping (uctx, UCTX_DPROC_QUEUE_FULL);
    
    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
    return (UTS_FINISHED);
}

static int
resolve_tproc_traps (USER_CTXT *uctx, ELAN4_USER_TRAP *utrapp)
{
    unsigned long flags;
    int res;
    
    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    while (! RING_QUEUE_EMPTY (uctx->uctx_tprocTrapQ))
    {
	ELAN4_TPROC_TRAP trap = *RING_QUEUE_FRONT(uctx->uctx_tprocTrapQ, uctx->uctx_tprocTraps);
	
	(void) RING_QUEUE_REMOVE (uctx->uctx_tprocTrapQ);

	spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);

	if ((res = resolve_tproc_trap (uctx, utrapp, &trap)) != UTS_FINISHED)
	    return (res);
	
	spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    }

    if (uctx->uctx_status & UCTX_TPROC_QUEUE_FULL)
	user_stop_stopping (uctx, UCTX_TPROC_QUEUE_FULL);

    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
    return (UTS_FINISHED);
}

static int
resolve_iproc_traps (USER_CTXT *uctx, ELAN4_USER_TRAP *utrapp)
{
    unsigned long flags;
    int i, res;

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    for (i = 0; i < 2; i++)
	if (uctx->uctx_iprocTrap[i].ut_state == UTS_IPROC_TRAPPED)
	{
	    uctx->uctx_iprocTrap[i].ut_state = UTS_IPROC_RESOLVING;
	    spin_unlock_irqrestore(&uctx->uctx_spinlock, flags);
	    
	    if ((res = resolve_iproc_trap (uctx, utrapp, i)) != UTS_FINISHED)
		return (res);
	    
	    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
	}
    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
    
    return (UTS_FINISHED);
}

static int
resolve_all_traps (USER_CTXT *uctx, ELAN4_USER_TRAP *utrapp)
{
    int res;

    if ((res = resolve_iproc_traps (uctx, utrapp)) != UTS_FINISHED ||
	(res = resolve_cproc_traps (uctx, utrapp)) != UTS_FINISHED ||
	(res = resolve_eproc_traps (uctx, utrapp)) != UTS_FINISHED ||
	(res = resolve_dproc_traps (uctx, utrapp)) != UTS_FINISHED ||
	(res = resolve_tproc_traps (uctx, utrapp)) != UTS_FINISHED)
	return (res);

    if (uctx->uctx_status & UCTX_OVERFLOW_REASONS)
    {
	PRINTF (uctx, DBG_TRAP, "resolve_all_traps: overflow reasons %x\n", uctx->uctx_status);

	return (deliver_trap (utrapp, UTS_QUEUE_OVERFLOW, UTS_NOPROC, NULL, uctx->uctx_status));
    }

    if (uctx->uctx_status & UCTX_ERROR_REASONS)
	return (deliver_trap (utrapp, UTS_QUEUE_ERROR, UTS_NOPROC, NULL, uctx->uctx_status));

    return (UTS_FINISHED);
}

static int
execute_iproc_traps (USER_CTXT *uctx, ELAN4_USER_TRAP *utrapp)
{
    unsigned long flags;
    int i;

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    for (i = 0; i < 2; i++)
	switch (uctx->uctx_iprocTrap[i].ut_state)
	{
	case UTS_IPROC_EXECUTE_PACKET:
	    uctx->uctx_iprocTrap[i].ut_state = UTS_IPROC_EXECUTING;
	    spin_unlock_irqrestore(&uctx->uctx_spinlock, flags);
	    
	    return (deliver_trap (utrapp, UTS_EXECUTE_PACKET, UTS_IPROC, &uctx->uctx_iprocTrap[i].ut_trap, i));

	case UTS_IPROC_NETWORK_ERROR:
	    spin_unlock_irqrestore(&uctx->uctx_spinlock, flags);
	    
	    return (deliver_trap (utrapp, UTS_NETWORK_ERROR_TRAP, UTS_IPROC, &uctx->uctx_iprocTrap[i].ut_trap, i));
	}
    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
    
    return (UTS_FINISHED);
}

static int
progress_neterr (USER_CTXT *uctx, ELAN4_USER_TRAP *utrapp)
{
    unsigned long flags;

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    if (! RING_QUEUE_EMPTY (uctx->uctx_msgQ))
    {
	ELAN4_NETERR_MSG msg = *RING_QUEUE_FRONT (uctx->uctx_msgQ, uctx->uctx_msgs);
	
	(void) RING_QUEUE_REMOVE (uctx->uctx_msgQ);
	spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
	
	return deliver_trap (utrapp, UTS_NETWORK_ERROR_MSG, UTS_NETERR_MSG, &msg, user_location2process (uctx, msg.msg_sender));
    }
    
    if (uctx->uctx_status & UCTX_NETERR_TIMER)
    {
	uctx->uctx_status &= ~UCTX_NETERR_TIMER;

	spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
	
	return deliver_trap (utrapp, UTS_NETWORK_ERROR_TIMER, UTS_NOPROC, NULL);
    }
    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
    
    return (UTS_FINISHED);
}

static void
restart_command_queues (USER_CTXT *uctx)
{
    struct list_head *entry;

    ASSERT (SPINLOCK_HELD (&uctx->uctx_spinlock));
    
    list_for_each (entry, &uctx->uctx_cqlist) {
	USER_CQ *ucq = list_entry (entry, USER_CQ, ucq_link);
	
	if (ucq->ucq_state == UCQ_NEEDS_RESTART)
	{
	    ucq->ucq_state = UCQ_RUNNING;
	    
	    elan4_restartcq (uctx->uctx_ctxt.ctxt_dev, ucq->ucq_cq);
	}
    }
}

static int
restart_dmas (USER_CTXT *uctx)
{
    PRINTF (uctx, DBG_TRAP, "restart_dmas: back=%d front=%d\n", uctx->uctx_dmaQ.q_back, uctx->uctx_dmaQ.q_front);

    while (! RING_QUEUE_EMPTY (uctx->uctx_dmaQ))
    {
	if (! user_ddcq_check (uctx, 7))
	    return (0);

	user_ddcq_run_dma (uctx, RING_QUEUE_FRONT(uctx->uctx_dmaQ, uctx->uctx_dmas));
	
	(void) RING_QUEUE_REMOVE (uctx->uctx_dmaQ);
    }

    return (1);
}

static int
restart_threads (USER_CTXT *uctx)
{
    PRINTF (uctx, DBG_TRAP, "restart_threads: back=%d front=%d\n", uctx->uctx_threadQ.q_back, uctx->uctx_threadQ.q_front);

    while (! RING_QUEUE_EMPTY (uctx->uctx_threadQ))
    {
	if (! user_ddcq_check (uctx, 7))
	    return (0);

	user_ddcq_run_thread (uctx, RING_QUEUE_FRONT(uctx->uctx_threadQ, uctx->uctx_threads));
	
	(void) RING_QUEUE_REMOVE (uctx->uctx_threadQ);
    }

    return (1);
}

int
user_resume_eproc_trap (USER_CTXT *uctx, E4_Addr addr)
{
    PRINTF2 (uctx, DBG_RESUME, "user_resume_eproc_trap: addr=%llx -> %s\n", addr, user_ddcq_check(uctx, 2) ? "success" : "EAGAIN");

    if (! user_ddcq_check (uctx, 2))
	return (-EAGAIN);

    user_ddcq_setevent (uctx, addr);

    return (0);
}

int
user_resume_cproc_trap (USER_CTXT *uctx, unsigned indx)
{
    struct list_head *entry;
    unsigned long flags;

    PRINTF1 (uctx, DBG_RESUME, "user_resume_cproc_trap: indx=%d\n", indx);

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);

    list_for_each (entry, &uctx->uctx_cqlist) {
	USER_CQ *ucq = list_entry (entry, USER_CQ, ucq_link);
	
	if (elan4_cq2idx(ucq->ucq_cq) == indx && ucq->ucq_state == UCQ_STOPPED && !(ucq->ucq_flags & UCQ_SYSTEM))
	{
	    ucq->ucq_state = UCQ_NEEDS_RESTART;
	    
	    user_signal_trap (uctx);

	    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
	    return (0);
	}
    }
    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);

    return (-EINVAL);
}

int
user_resume_dproc_trap (USER_CTXT *uctx, E4_DMA *dma)
{
    unsigned long flags;
    int res = 0;

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    if (RING_QUEUE_FULL (uctx->uctx_dmaQ))
	res = -ENOMEM;
    else
    {
	*RING_QUEUE_BACK (uctx->uctx_dmaQ, uctx->uctx_dmas) = *dma;
	(void) RING_QUEUE_ADD (uctx->uctx_dmaQ);

	user_signal_trap (uctx);
    }
    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
    return (res);
}

int
user_resume_tproc_trap (USER_CTXT *uctx, E4_ThreadRegs *regs)
{
    unsigned long flags;
    int res = 0;

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    if (RING_QUEUE_FULL (uctx->uctx_threadQ))
	res = -ENOMEM;
    else
    {
	*RING_QUEUE_BACK (uctx->uctx_threadQ, uctx->uctx_threads) = *regs;
	(void) RING_QUEUE_ADD (uctx->uctx_threadQ);

	user_signal_trap (uctx);
    }
    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
    return (res);
}

int
user_resume_iproc_trap (USER_CTXT *uctx, unsigned channel, unsigned trans,
			E4_IprocTrapHeader *hdrp, E4_IprocTrapData *datap)
{
    unsigned long flags;
    int res = 0;

    if (channel >= 2)
	return (-EINVAL);

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    if (uctx->uctx_iprocTrap[channel].ut_state != UTS_IPROC_STOPPED &&
	uctx->uctx_iprocTrap[channel].ut_state != UTS_IPROC_EXECUTING &&
	uctx->uctx_iprocTrap[channel].ut_state != UTS_IPROC_NETWORK_ERROR)
	res = -EINVAL;
    else
    {
	ELAN4_IPROC_TRAP *trap = &uctx->uctx_iprocTrap[channel].ut_trap;

	if (trans < trap->tr_numTransactions)
	{
	    PRINTF1 (uctx, DBG_RESUME, "user_resume_iproc_trap: trans=%d -> execute\n", trans);

	    uctx->uctx_iprocTrap[channel].ut_state = UTS_IPROC_EXECUTE_PACKET;
	    trap->tr_trappedTrans                  = trans;
	    trap->tr_transactions[trans]           = *hdrp;
	    trap->tr_dataBuffers[trans]            = *datap;
	}
	else
	{
	    PRINTF1 (uctx, DBG_RESUME, "user_resume_iproc_trap: trans=%d -> running\n", trans);

	    uctx->uctx_iprocTrap[channel].ut_state = UTS_IPROC_RUNNING;
	
	    user_stop_nacking (uctx, channel ? UCTX_IPROC_CH0_TRAPPED : UCTX_IPROC_CH1_TRAPPED);
	}
    }
    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
    
    return (res);
}

int
__categorise_command (E4_uint64 command, int *cmdSize)
{
    switch (command & 0x3)
    {
    case RUN_THREAD_CMD: *cmdSize = 7; break;
	
    default:
	switch (command & 0x7)
	{
	case WRITE_DWORD_CMD: *cmdSize = 2; break;
	case ADD_DWORD_CMD:   *cmdSize = 2; break;
	    
	default:
	    switch (command & 0xF)
	    {
	    case OPEN_STEN_PKT_CMD:
		*cmdSize = 1;
		return 1;
		
	    case COPY64_CMD:    *cmdSize = 2; break;
	    case GUARD_CMD:     *cmdSize = 1; break;
	    case INTERRUPT_CMD: *cmdSize = 1; break;
	    case RUN_DMA_CMD:   *cmdSize = 7; break;
		
	    default:
		switch (command & 0x1f)
		{
		case SEND_TRANS_CMD:
		    *cmdSize = 2 + (((command >> 16) & TR_SIZE_MASK) >> TR_SIZE_SHIFT);
		    return 2;
		    
		case SET_EVENT_CMD:    *cmdSize = 1; break;
		case SET_EVENTN_CMD:   *cmdSize = 2; break;
		case WAIT_EVENT_CMD:   *cmdSize = 4; break;

		default:
		    switch (command & 0x3f)
		    {
		    case NOP_CMD: 	     *cmdSize = 1; break;
		    case MAKE_EXT_CLEAN_CMD: *cmdSize = 1; break;
		    default:
			return 3;
		    }
		    break;
		}
	    }
	}
    }

    return 0;
}

int
__whole_command (sdramaddr_t *commandPtr, sdramaddr_t insertPtr, unsigned int cqSize, unsigned int cmdSize)
{
    /* Move onto next command */
    while (cmdSize-- && (*commandPtr) != insertPtr)
	*commandPtr = ((*commandPtr) & ~(cqSize-1)) | (((*commandPtr) + sizeof (E4_uint64)) & (cqSize-1));

    return cmdSize == -1;
}

int
user_neterr_sten (USER_CTXT *uctx, unsigned int vp, E4_uint64 cookie, int waitforeop)
{
    ELAN4_DEV        *dev   = uctx->uctx_ctxt.ctxt_dev;
    int		      found = 0;
    struct list_head *el;

    user_swapout (uctx, UCTX_NETERR_FIXUP);

    kmutex_lock (&uctx->uctx_cqlock);
    list_for_each (el, &uctx->uctx_cqlist) {
	USER_CQ *ucq = list_entry (el, USER_CQ, ucq_link);
	
	if ((ucq->ucq_cq->cq_perm & CQ_STENEnableBit) != 0)
	{
	    sdramaddr_t   cqdesc       = dev->dev_cqaddr + (elan4_cq2num(ucq->ucq_cq) * sizeof (E4_CommandQueueDesc));
	    E4_uint64	  queuePtrs    = elan4_sdram_readq (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_QueuePtrs));
	    sdramaddr_t   insertPtr    = (queuePtrs & CQ_PtrMask);
	    sdramaddr_t   commandPtr   = CQ_CompletedPtr (queuePtrs);
	    unsigned int  cqSize       = CQ_Size ((queuePtrs >> CQ_SizeShift) & CQ_SizeMask);
	    E4_uint64     openCommand  = 0;

	    if (dev->dev_devinfo.dev_revision_id != PCI_REVISION_ID_ELAN4_REVA && (queuePtrs & CQ_RevB_ReorderingQueue))
	    {
		E4_uint32 oooMask = elan4_sdram_readl (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_HoldingValue));
		
		for (; (oooMask & 1) != 0; oooMask >>= 1)
		    insertPtr = (insertPtr & ~(cqSize-1)) | ((insertPtr + sizeof (E4_uint64)) & (cqSize-1));
	    }

	    while (commandPtr != insertPtr)
	    {
		E4_uint64    command = elan4_sdram_readq (dev, commandPtr);
		sdramaddr_t  identifyPtr;
		unsigned int cmdSize;
		
		switch (__categorise_command (command, &cmdSize))
		{
		case 0:
		    (void) __whole_command (&commandPtr, insertPtr, cqSize, cmdSize);
		    break;
		    
		case 1:
		    PRINTF (uctx, DBG_NETERR, "user_neterr_sten: cq=%d OPEN %llx\n", elan4_cq2num (ucq->ucq_cq), command);
		    
		    if ((command >> 32) == vp)
			openCommand = command;

		    (void) __whole_command (&commandPtr, insertPtr, cqSize, cmdSize);
		    break;
		    
		case 2:
		    PRINTF (uctx, DBG_NETERR, "user_neterr_sten: cq=%d SENDTRANS %llx\n", elan4_cq2num (ucq->ucq_cq), command);
		    
		    if (openCommand == 0)
			(void) __whole_command (&commandPtr, insertPtr, cqSize, cmdSize);
		    else
		    {
			switch ((command >> 16) & (TR_OPCODE_MASK | TR_SIZE_MASK))
			{
			case TR_IDENTIFY  & (TR_OPCODE_MASK | TR_SIZE_MASK):
			case TR_REMOTEDMA & (TR_OPCODE_MASK | TR_SIZE_MASK):
			    PRINTF (uctx, DBG_NETERR, "user_neterr_sten: TR_IDENTIFY/TR_REMOTEDMA\n");
			    identifyPtr = (commandPtr & ~(cqSize-1)) | ((commandPtr + sizeof (E4_uint64)) & (cqSize-1));
			    break;
			    
			case TR_SETEVENT_IDENTIFY & (TR_OPCODE_MASK | TR_SIZE_MASK):
			case TR_INPUT_Q_COMMIT    & (TR_OPCODE_MASK | TR_SIZE_MASK):
			    PRINTF (uctx, DBG_NETERR, "user_neterr_sten: TR_SETEVENT_IDENTIFY/TR_INPUT_Q_COMMIT\n");
			    identifyPtr = (commandPtr & ~(cqSize-1)) | ((commandPtr + 2*sizeof (E4_uint64)) & (cqSize-1));
			    break;
			    
			case TR_ADDWORD & (TR_OPCODE_MASK | TR_SIZE_MASK):
			    PRINTF (uctx, DBG_NETERR, "user_neterr_sten: TR_ADDWORD\n");
			    identifyPtr = (commandPtr & ~(cqSize-1)) | ((commandPtr + 3*sizeof (E4_uint64)) & (cqSize-1));
			    break;
			    
			case TR_TESTANDWRITE & (TR_OPCODE_MASK | TR_SIZE_MASK):
			    PRINTF (uctx, DBG_NETERR, "user_neterr_sten: TR_TESTANDWRITE\n");
			    identifyPtr = (commandPtr & ~(cqSize-1)) | ((commandPtr + 4*sizeof (E4_uint64)) & (cqSize-1));
			    break;
			    
			default:
			    identifyPtr = 0;
			}
			
			if (! __whole_command (&commandPtr, insertPtr, cqSize, cmdSize))
			{
			    PRINTF (uctx, DBG_NETERR, "user_neterr_sten: not whole command\n");
			    openCommand = 0;
			}

			else if (identifyPtr)
			{
			    E4_uint64 tcookie = elan4_sdram_readq (dev, identifyPtr);
			    
			    PRINTF (uctx, DBG_NETERR, "user_neterr_sten: cookie=%llx [%llx]\n", tcookie, cookie);
			    
			    if (tcookie == cookie)
			    {
				unsigned int vchan = (openCommand >> 4) & 0x1f;
				
				PRINTF (uctx, DBG_NETERR, "user_neterr_sten: cookie matches - vchan=%d\n", vchan);
				
				if (! waitforeop)
				{
				    /* Alter the CQ_AckBuffer for this channel to indicate an
				     * ack was received */
				    E4_uint64 value  = elan4_sdram_readq (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_AckBuffers));
				    E4_uint64 nvalue = ((value & ~((E4_uint64)0xf << ((vchan & 0xf) << 2))) |
							((E4_uint64) PackOk << ((vchan & 0xf) << 2)));
				    
				    PRINTF (uctx, DBG_NETERR, "user_neterr_sten: CQ_AckBuffers %llx -> %llx\n", value, nvalue);
				    
				    elan4_sdram_writeq (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_AckBuffers), nvalue);
				    pioflush_sdram (dev);
				}
				
				found++;
			    }
			    openCommand = 0;
			}
			
			if ((command >> 16) & TR_LAST_AND_SEND_ACK)
			    openCommand = 0;
		    }
		    break;
		    
		case 3:
		    PRINTF (uctx, DBG_NETERR, "user_neterr_sten: invalid command %llx\n", command);
		    kmutex_unlock (&uctx->uctx_cqlock);
		    return -EINVAL;
		}
		
	    }
	}
    }
    kmutex_unlock (&uctx->uctx_cqlock);

    user_swapin (uctx, UCTX_NETERR_FIXUP);

    return found;
}

int
user_neterr_dma (USER_CTXT *uctx, unsigned int vp, E4_uint64 cookie, int waitforeop)
{
    unsigned long flags;
    int found = 0;
    int idx;

    user_swapout (uctx, UCTX_NETERR_FIXUP);

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    RING_QUEUE_ITERATE (uctx->uctx_dmaQ, idx) {
	E4_DMA *dma = &uctx->uctx_dmas[idx];

	if (dma->dma_vproc == vp && dma->dma_cookie == cookie)
	{
	    PRINTF (uctx, DBG_NETERR, "user_neterr_dma: dmaQ matches %s\n", waitforeop ? "waitforeop" : "remove remoteness");

	    if (! waitforeop) 
	    {
		dma->dma_dstEvent = 0;
		dma->dma_typeSize = DMA_ShMemWrite | DMA_Context (dma->dma_typeSize);
	    }
	    found++;
	}
    }

    RING_QUEUE_ITERATE (uctx->uctx_dprocTrapQ, idx) {
	ELAN4_DPROC_TRAP *trap = &uctx->uctx_dprocTraps[idx];

	if (trap->tr_desc.dma_vproc == vp && trap->tr_desc.dma_cookie == cookie)
	{
	    PRINTF (uctx, DBG_NETERR, "user_neterr_dma: dmaTrapQ matches %s\n", waitforeop ? "waitforeop" : "remove remoteness");

	    if (! waitforeop) 
	    {
		trap->tr_desc.dma_dstEvent = 0;
		trap->tr_desc.dma_typeSize = DMA_ShMemWrite | DMA_Context (trap->tr_desc.dma_typeSize);
	    }
	    found++;
	}
    }

    /* The device driver command queue should be empty at this point ! */
    if (user_ddcq_flush (uctx) == 0)
	found = -EAGAIN;

    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);

    /* The device driver command queue should be empty at this point ! */
    if (user_ddcq_flush (uctx) == 0)
	found = -EAGAIN;
    
    user_swapin (uctx, UCTX_NETERR_FIXUP);

    return found;
}

int
user_trap_handler (USER_CTXT *uctx, ELAN4_USER_TRAP *utrapp, int nticks)
{
    unsigned long entered = jiffies;
    unsigned int  need_reenter = 0;
    unsigned long flags;
    int           res;
    int           tbl;

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);

    PRINTF1 (uctx, DBG_TRAP, "user_trap_handler: entered state=%d\n", uctx->uctx_trap_state);

    uctx->uctx_trap_count++;
    
    for (;;)
    {
	if (uctx->uctx_status & UCTX_SWAPPED_REASONS)
	{
	    PRINTF0 (uctx, DBG_TRAP, "user_trap_handler: exiting on swapped reasons\n");
	    
	    res = UTS_FINISHED;
	    goto no_more_to_do;
	}

	if ((long) (jiffies - entered) > HZ)
	{
	    PRINTF0 (uctx, DBG_TRAP, "user_trap_handler: exiting for reschedule\n");
	    res = UTS_RESCHEDULE;
	    goto no_more_to_do;
	}
	
	switch (uctx->uctx_trap_state)
	{
	case UCTX_TRAP_ACTIVE:
	    uctx->uctx_trap_state = UCTX_TRAP_SLEEPING;
	    
	    if (nticks == 0 || need_reenter || kcondvar_timedwaitsig (&uctx->uctx_wait, &uctx->uctx_spinlock, &flags, lbolt + nticks) != CV_RET_NORMAL)
	    {
		PRINTF0 (uctx, DBG_TRAP, "user_trap_handler: exiting by kcondvar_timedwaitsig\n");

		res = UTS_FINISHED;
		goto no_more_to_do;
	    }

	    /* Have slept above, so resample entered */
	    entered = jiffies;
	    
	    uctx->uctx_trap_state = UCTX_TRAP_SIGNALLED;
	    continue;

	case UCTX_TRAP_IDLE:
	case UCTX_TRAP_SIGNALLED:
	    uctx->uctx_trap_state = UCTX_TRAP_ACTIVE;
	    break;
	}
	spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);

	PRINTF2 (uctx, DBG_TRAP, "user_trap_handler: resolve traps - state=%d status=%x\n", uctx->uctx_trap_state, uctx->uctx_status);

	switch ((res = resolve_all_traps (uctx, utrapp)))
	{
	case UTS_FINISHED:
	    break;
	    
	case UTS_RESCHEDULE:
	    need_reenter++;
	    break;

	default:
	    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
	    goto no_more_to_do;
	}

	spin_lock_irqsave (&uctx->uctx_spinlock, flags);
	if (! user_ddcq_flush (uctx))
	{
	    PRINTF0 (uctx, DBG_TRAP, "user_trap_handler: ddcq not flushed - re-enter\n");
	    need_reenter++;
	    continue;
	}
	spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);

	for(tbl=0; tbl < NUM_HASH_TABLES; tbl++)
	    if (uctx->uctx_ctxt.shuffle_needed[tbl]) 
		elan4mmu_do_shuffle (&uctx->uctx_ctxt, tbl);

	if ((res = progress_neterr (uctx, utrapp)) != UTS_FINISHED)
	{
	    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
	    goto no_more_to_do;
	}

	if ((res = execute_iproc_traps (uctx, utrapp)) != UTS_FINISHED)
	{
	    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
	    goto no_more_to_do;
	}

	PRINTF2 (uctx, DBG_TRAP, "user_trap_handler: restart items - state=%d status=%x\n", uctx->uctx_trap_state, uctx->uctx_status);

	spin_lock_irqsave (&uctx->uctx_spinlock, flags);
	if (UCTX_RUNNABLE (uctx))
	{
	    restart_command_queues (uctx);

	    if (! restart_threads (uctx) || ! restart_dmas (uctx))
	    {
		PRINTF0 (uctx, DBG_TRAP, "user_trap_handler: ddcq full - re-enter\n");
		need_reenter++;
	    }
	}
    }
 no_more_to_do:
    uctx->uctx_trap_state = UCTX_TRAP_IDLE;

    /*
     * Always ensure that the command queue is flushed with a flow control
     * write, so that on the next trap we (hopefully) find it empty and so
     * can immediately restart the context.   Also if we need to be re-enter
     * the trap handler and don't have an interrupt outstanding, then issue
     * one now.
     */
    user_ddcq_flush (uctx);
    if (need_reenter && uctx->uctx_ddcq_intr == 0)
    {
	uctx->uctx_ddcq_intr++;
	user_ddcq_intr (uctx);
    }

    if (--uctx->uctx_trap_count == 0 && (uctx->uctx_status & UCTX_SWAPPING))
	kcondvar_wakeupall (&uctx->uctx_wait, &uctx->uctx_spinlock);

    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);

    /* Should update the user trap area in this case as deliver_trap()
     * has not been called 
     */
    if (res == UTS_RESCHEDULE)
	put_user (res, &utrapp->ut_type);

    PRINTF2 (uctx, DBG_TRAP, "user_trap_handler: finished state=%d res=%d\n", uctx->uctx_trap_state, res);

    return (res == UTS_EFAULT ? -EFAULT : 0);
}

USER_CQ *
user_alloccq (USER_CTXT *uctx, unsigned cqsize, unsigned perm, unsigned uflags)
{
    USER_CQ      *ucq;
    unsigned long flags;

    KMEM_ZALLOC (ucq, USER_CQ *, sizeof (USER_CQ), 1);

    if (ucq == (USER_CQ *) NULL)
	return ERR_PTR(-ENOMEM);
    
    /* NOTE - do not allow the user to create high-priority queues as we only flush through the low-priority run queues */
    if ((ucq->ucq_cq = elan4_alloccq (&uctx->uctx_ctxt, cqsize, perm, (uflags & UCQ_REORDER) ? CQ_Reorder : 0)) == NULL)
    {
	KMEM_FREE (ucq, sizeof (USER_CQ));
	
	PRINTF2 (uctx, DBG_CQ, "user_alloccq: failed elan4_allocq cqsize %d uflags %x\n", cqsize, uflags);

	return ERR_PTR(-ENOMEM);
    }
    
    atomic_set (&ucq->ucq_ref, 1);

    ucq->ucq_state = UCQ_RUNNING;
    ucq->ucq_flags = uflags;
    
    PRINTF3 (uctx, DBG_CQ, "user_alloccq: ucq=%p idx=%d cqnum=%d\n", ucq, elan4_cq2idx (ucq->ucq_cq), elan4_cq2num(ucq->ucq_cq));

    /* chain it onto the context */
    kmutex_lock (&uctx->uctx_cqlock);
    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    list_add (&ucq->ucq_link, &uctx->uctx_cqlist);
    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
    kmutex_unlock (&uctx->uctx_cqlock);

    return (ucq);
}

USER_CQ *
user_findcq (USER_CTXT *uctx, unsigned idx)
{
    struct list_head *entry;
    unsigned long flags;

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    list_for_each (entry, &uctx->uctx_cqlist) {
	USER_CQ *ucq = list_entry (entry, USER_CQ, ucq_link);

	if (elan4_cq2idx(ucq->ucq_cq) == idx)
	{
	    atomic_inc (&ucq->ucq_ref);
	    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
	    return (ucq);
	}
    }
    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);

    return (NULL);
}

void
user_dropcq (USER_CTXT *uctx, USER_CQ *ucq)
{
    unsigned long flags;

    PRINTF2 (uctx, DBG_CQ, "user_dropcq: ucq=%p ref=%d\n", ucq, atomic_read (&ucq->ucq_ref));

    kmutex_lock (&uctx->uctx_cqlock);
    if (! atomic_dec_and_test (&ucq->ucq_ref))
    {
	kmutex_unlock (&uctx->uctx_cqlock);
	return;
    }

    spin_lock_irqsave (&uctx->uctx_spinlock, flags);
    list_del (&ucq->ucq_link);
    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);

    kmutex_unlock (&uctx->uctx_cqlock);

    elan4_freecq (&uctx->uctx_ctxt, ucq->ucq_cq);

    KMEM_FREE (ucq, sizeof (USER_CQ));
}

int
user_alloc_trap_queues (USER_CTXT *uctx, unsigned ndproc_traps, unsigned neproc_traps, 
			unsigned ntproc_traps, unsigned nthreads, unsigned ndmas)
{
    ELAN4_DPROC_TRAP *dprocs;
    ELAN4_EPROC_TRAP *eprocs;
    ELAN4_TPROC_TRAP *tprocs;
    E4_DMA           *dmas;
    E4_ThreadRegs    *threads;
    ELAN4_NETERR_MSG *msgs;
    unsigned long     flags;

    int nmsgs = NETERR_MSGS;

    /* bounds check the values that have been passed in */
    if (ndproc_traps < 2 || ndproc_traps > 10000 ||
	ntproc_traps < 1 || ntproc_traps > 100   ||
	neproc_traps < 6 || neproc_traps > 10000 ||
	nthreads     < 2 || nthreads     > 10000 ||
	ndmas        < 2 || ndmas        > 10000)
	return -EINVAL;

    if (uctx->uctx_dmas != NULL)
	return -EBUSY;

    KMEM_ZALLOC (dprocs, ELAN4_DPROC_TRAP *, ndproc_traps * sizeof (ELAN4_DPROC_TRAP), 1);
    KMEM_ZALLOC (eprocs, ELAN4_EPROC_TRAP *, neproc_traps * sizeof (ELAN4_EPROC_TRAP), 1);
    KMEM_ZALLOC (tprocs, ELAN4_TPROC_TRAP *, ntproc_traps * sizeof (ELAN4_TPROC_TRAP), 1);
    KMEM_ZALLOC (threads, E4_ThreadRegs *, nthreads * sizeof (E4_ThreadRegs), 1);
    KMEM_ZALLOC (dmas, E4_DMA *, ndmas * sizeof (E4_DMA), 1);
    KMEM_ZALLOC (msgs, ELAN4_NETERR_MSG *, nmsgs * sizeof (ELAN4_NETERR_MSG), 1);

    if (dprocs == NULL || eprocs == NULL || tprocs == NULL || dmas == NULL || threads == NULL || msgs == NULL)
    {
	if (dprocs != NULL) KMEM_FREE (dprocs, ndproc_traps * sizeof (ELAN4_DPROC_TRAP));
	if (eprocs != NULL) KMEM_FREE (eprocs, neproc_traps * sizeof (ELAN4_EPROC_TRAP));
	if (tprocs != NULL) KMEM_FREE (tprocs, ntproc_traps * sizeof (ELAN4_TPROC_TRAP));
	if (threads != NULL) KMEM_FREE (threads, nthreads * sizeof (E4_ThreadRegs));
	if (dmas != NULL) KMEM_FREE (dmas, ndmas * sizeof (E4_DMA));
	if (msgs != NULL) KMEM_FREE (msgs, nmsgs * sizeof (ELAN4_NETERR_MSG));
	
	return -ENOMEM;
    }
    
    spin_lock_irqsave (&uctx->uctx_spinlock, flags);

    uctx->uctx_dprocTraps = dprocs;
    uctx->uctx_eprocTraps = eprocs;
    uctx->uctx_tprocTraps = tprocs;
    uctx->uctx_threads    = threads;
    uctx->uctx_dmas       = dmas;
    uctx->uctx_msgs       = msgs;

    RING_QUEUE_INIT (uctx->uctx_dprocTrapQ, ndproc_traps, 1 /* 1 for 2nd dma */);
    RING_QUEUE_INIT (uctx->uctx_tprocTrapQ, ntproc_traps, 0);
    RING_QUEUE_INIT (uctx->uctx_eprocTrapQ, neproc_traps, 5 /* 1 for command, 2 for dma, 2 for inputter */);
    RING_QUEUE_INIT (uctx->uctx_threadQ,    nthreads,     1);
    RING_QUEUE_INIT (uctx->uctx_dmaQ,       ndmas,        1);
    RING_QUEUE_INIT (uctx->uctx_msgQ,       nmsgs,        0);

    spin_unlock_irqrestore (&uctx->uctx_spinlock, flags);
    
    return 0;
}

USER_CTXT *
user_alloc (ELAN4_DEV *dev)
{
    USER_CTXT *uctx;
    int res;
    int i;

    /* Allocate and initialise the context private data */
    KMEM_ZALLOC (uctx, USER_CTXT *, sizeof  (USER_CTXT), 1);

    if (uctx == NULL)
	return ERR_PTR(-ENOMEM);

    if (elan4_get_position (dev, &uctx->uctx_position) == ELAN_POS_UNKNOWN)
    {
	KMEM_FREE (uctx, sizeof (USER_CTXT));
	return ERR_PTR(-EAGAIN);
    }
    
    if ((res = elan4_insertctxt (dev, &uctx->uctx_ctxt, &user_trap_ops)) != 0)
    {
	KMEM_FREE (uctx, sizeof (USER_CTXT));
	return ERR_PTR(res);
    }

    if (! user_ioproc_enabled)
	uctx->uctx_ctxt.ctxt_features |= ELAN4_FEATURE_NO_IOPROC | ELAN4_FEATURE_PIN_DOWN;
    if (! user_pagefault_enabled)
	uctx->uctx_ctxt.ctxt_features |= ELAN4_FEATURE_NO_PAGEFAULT;

    KMEM_GETPAGES (uctx->uctx_upage, ELAN4_USER_PAGE *, btopr (sizeof (ELAN4_USER_PAGE)), 1);
    if (uctx->uctx_upage == NULL)
    {
	elan4_removectxt (dev, &uctx->uctx_ctxt);
	KMEM_FREE (uctx, sizeof (USER_CTXT));
	return ERR_PTR(-ENOMEM);
    }
    
    if ((uctx->uctx_trampoline = elan4_sdram_alloc (dev, SDRAM_PAGE_SIZE)) == 0)
    {
	KMEM_FREEPAGES (uctx->uctx_upage, btopr (sizeof (ELAN4_USER_PAGE)));
	elan4_removectxt (dev, &uctx->uctx_ctxt);

	KMEM_FREE (uctx, sizeof (USER_CTXT));
	return ERR_PTR(-ENOMEM);
    }
    
    if ((uctx->uctx_routetable = elan4_alloc_routetable (dev, 4 /* 512 << 4 == 8192 entries */)) == NULL)
    {
	elan4_sdram_free (dev, uctx->uctx_trampoline, SDRAM_PAGE_SIZE);
	KMEM_FREEPAGES (uctx->uctx_upage, btopr (sizeof (ELAN4_USER_PAGE)));
	elan4_removectxt (dev, &uctx->uctx_ctxt);

	KMEM_FREE (uctx, sizeof (USER_CTXT));
	return ERR_PTR(-ENOMEM);
    }

    elan4_set_routetable (&uctx->uctx_ctxt, uctx->uctx_routetable);

    /* initialise the trap and swap queues to be really full */
    RING_QUEUE_INIT (uctx->uctx_dprocTrapQ, 0, 1);
    RING_QUEUE_INIT (uctx->uctx_tprocTrapQ, 0, 1);
    RING_QUEUE_INIT (uctx->uctx_eprocTrapQ, 0, 1);
    RING_QUEUE_INIT (uctx->uctx_threadQ, 0, 1);
    RING_QUEUE_INIT (uctx->uctx_dmaQ, 0, 1);

    INIT_LIST_HEAD (&uctx->uctx_cent_list);
    INIT_LIST_HEAD (&uctx->uctx_vpseg_list);
    INIT_LIST_HEAD (&uctx->uctx_cqlist);

    uctx->uctx_haltop.op_function = user_flush;
    uctx->uctx_haltop.op_arg      = uctx;
    uctx->uctx_haltop.op_mask     = INT_Halted|INT_Discarding;

    uctx->uctx_dma_flushop.op_function = user_flush_dmas;
    uctx->uctx_dma_flushop.op_arg      = uctx;

    kmutex_init (&uctx->uctx_vpseg_lock);
    kmutex_init (&uctx->uctx_cqlock);
    kmutex_init (&uctx->uctx_rgnmutex);

    spin_lock_init (&uctx->uctx_spinlock);
    spin_lock_init (&uctx->uctx_rgnlock);
    spin_lock_init (&uctx->uctx_fault_lock);

    kcondvar_init (&uctx->uctx_wait);

    if ((uctx->uctx_ddcq = user_alloccq (uctx, CQ_Size1K, CQ_EnableAllBits, UCQ_SYSTEM)) == NULL)
    {
	user_free (uctx);
	return ERR_PTR(-ENOMEM);
    }

    uctx->uctx_trap_count = 0;
    uctx->uctx_trap_state = UCTX_TRAP_IDLE;
    uctx->uctx_status     = 0 /* UCTX_DETACHED | UCTX_SWAPPED | UCTX_STOPPED */;

    init_timer (&uctx->uctx_int_timer);

    uctx->uctx_int_timer.function = user_signal_timer;
    uctx->uctx_int_timer.data     = (unsigned long) uctx;
    uctx->uctx_int_start          = jiffies;
    uctx->uctx_int_count          = 0;
    uctx->uctx_int_delay          = 0;

    init_timer (&uctx->uctx_shuffle_timer);

    uctx->uctx_shuffle_timer.function = user_signal_timer;
    uctx->uctx_shuffle_timer.data     = (unsigned long) uctx;


    init_timer (&uctx->uctx_neterr_timer);
    uctx->uctx_neterr_timer.function = user_neterr_timer;
    uctx->uctx_neterr_timer.data     = (unsigned long) uctx;

    uctx->uctx_upage->upage_ddcq_completed = 0;
    uctx->uctx_ddcq_completed              = 0;
    uctx->uctx_ddcq_insertcnt              = 0;

    uctx->uctx_num_fault_save = num_fault_save;
    if (uctx->uctx_num_fault_save) 
    {  
	KMEM_ZALLOC (uctx->uctx_faults, FAULT_SAVE *, (sizeof(FAULT_SAVE) * uctx->uctx_num_fault_save), 1);
        if ( uctx->uctx_faults == NULL) 
	{
	    user_free (uctx);
	    return ERR_PTR(-ENOMEM);
        }
    
        for (i = 0; i < uctx->uctx_num_fault_save; i++)
	    uctx->uctx_faults[i].next = (i == (uctx->uctx_num_fault_save-1) ? NULL : &uctx->uctx_faults[i+1]);

    }
    uctx->uctx_fault_list = uctx->uctx_faults;

    return (uctx);
}

void
user_free (USER_CTXT *uctx)
{
    ELAN4_DEV *dev = uctx->uctx_ctxt.ctxt_dev;

    user_swapout (uctx, UCTX_EXITING);

    /* Detach from all input contexts */
    user_detach (uctx, NULL);

    /* since we're single threaded here - (called from close()) */
    /* we don't need to hold the lock to drop the command queues */
    /* since they cannot be mapped into user space */
    while (! list_empty (&uctx->uctx_cqlist))
	user_dropcq (uctx, list_entry (uctx->uctx_cqlist.next, USER_CQ, ucq_link));

    /* Free off all of vpseg_list */
    kmutex_lock (&uctx->uctx_vpseg_lock);
    while (! list_empty (&uctx->uctx_vpseg_list))
	user_remove_vpseg(uctx, list_entry (uctx->uctx_vpseg_list.next, USER_VPSEG, vps_link));
    kmutex_unlock (&uctx->uctx_vpseg_lock);
    
    if (timer_pending (&uctx->uctx_int_timer))
	del_timer_sync (&uctx->uctx_int_timer);
    
    if (timer_pending (&uctx->uctx_shuffle_timer))
	del_timer_sync (&uctx->uctx_shuffle_timer);

    if (timer_pending (&uctx->uctx_neterr_timer))
	del_timer_sync (&uctx->uctx_neterr_timer);

    if (uctx->uctx_dprocTraps)
	KMEM_FREE (uctx->uctx_dprocTraps, uctx->uctx_dprocTrapQ.q_size * sizeof (ELAN4_DPROC_TRAP));
    if (uctx->uctx_tprocTraps)
	KMEM_FREE (uctx->uctx_tprocTraps, uctx->uctx_tprocTrapQ.q_size * sizeof (ELAN4_TPROC_TRAP));
    if (uctx->uctx_eprocTraps)
	KMEM_FREE (uctx->uctx_eprocTraps, uctx->uctx_eprocTrapQ.q_size * sizeof (ELAN4_EPROC_TRAP));
    if (uctx->uctx_dmas)
	KMEM_FREE (uctx->uctx_dmas, uctx->uctx_dmaQ.q_size * sizeof (E4_DMA));
    if (uctx->uctx_msgs)
	KMEM_FREE (uctx->uctx_msgs, NETERR_MSGS * sizeof (ELAN4_NETERR_MSG));
    if (uctx->uctx_threads)
	KMEM_FREE (uctx->uctx_threads, uctx->uctx_threadQ.q_size * sizeof (E4_ThreadRegs));
    if (uctx->uctx_faults)
	KMEM_FREE (uctx->uctx_faults, (sizeof(FAULT_SAVE) * uctx->uctx_num_fault_save));

    if (uctx->uctx_intcookie_table)
	intcookie_free_table (uctx->uctx_intcookie_table);

    elan4_set_routetable (&uctx->uctx_ctxt, NULL);
    elan4_free_routetable (dev, uctx->uctx_routetable);

    /* Free off all USER_RGNs */
    user_freergns(uctx);

    elan4_sdram_free (dev, uctx->uctx_trampoline, SDRAM_PAGE_SIZE);

    /* Clear the PG_Reserved bit before free to avoid a memory leak */
    ClearPageReserved(pte_page(*find_pte_kernel((unsigned long) uctx->uctx_upage)));
    KMEM_FREEPAGES (uctx->uctx_upage, btopr (sizeof (ELAN4_USER_PAGE)));

    elan4_removectxt (dev, &uctx->uctx_ctxt);

    kcondvar_destroy (&uctx->uctx_wait);

    spin_lock_destroy (&uctx->uctx_rgnlock);
    spin_lock_destroy (&uctx->uctx_spinlock);

    kmutex_destroy (&uctx->uctx_rgnmutex);
    kmutex_destroy (&uctx->uctx_cqlock);
    kmutex_destroy (&uctx->uctx_vpseg_lock);

    KMEM_FREE (uctx, sizeof (USER_CTXT));
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
