/*
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: user_ddcq.c,v 1.16.2.1 2005/07/20 11:35:36 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan4mod/user_ddcq.c,v $*/

#include <qsnet/kernel.h>

#include <elan4/debug.h>
#include <elan4/device.h>
#include <elan4/user.h>
#include <elan4/commands.h>

#if PAGE_SIZE < CQ_CommandMappingSize
#  define ELAN4_COMMAND_QUEUE_MAPPING	PAGE_SIZE
#else
#  define ELAN4_COMMAND_QUEUE_MAPPING	CQ_CommandMappingSize
#endif

/* The user device driver command queue is used for re-issuing 
 * trapped items.  It is allocated as a 1K command queue, and
 * we insert command flow writes event 256 bytes (32 dwords).
 */
#define USER_CTRLFLOW_COUNT	32

/* Flow control of the device driver command queue is handled by periodically 
 * inserting dword writes into the command stream.   When you need to know 
 * that the queue has been flushed, then you insert an extra contorl flow
 * write into the command queue.   Should the queue not be flushed, but the
 * trap handler be returning to user space, then it will also insert and 
 * extra interrupt command to ensure that it is re-entered after the queue
 * has been flushed.
 *
 * Note - we account the space for the interrupt command on each control 
 * flow write so that we do not overflow the queue even if we end up 
 * inserting an interrupt for every command flow write.  In general only
 * a single interrupt should get inserted....
 */

#define user_ddcq_command_write(value,off) do { \
    PRINTF(uctx, DBG_DDCQ, "user_ddcq_command_write: cmdptr=%x off=%d value=%llx\n", cmdptr, off, value);\
    writeq(value, (void *)(cmdptr + (off << 3))); \
} while (0)

#define user_ddcq_command_space(uctx)	\
    ((CQ_Size (uctx->uctx_ddcq->ucq_cq->cq_size)>>3) - ((uctx)->uctx_ddcq_insertcnt - (uctx)->uctx_upage->upage_ddcq_completed))

#define user_ddcq_command_flow_write(uctx) do { \
   E4_uint64 iptr   = (uctx)->uctx_ddcq_insertcnt; \
   ioaddr_t  cmdptr = (uctx)->uctx_ddcq->ucq_cq->cq_mapping + ((iptr<<3) & ((ELAN4_COMMAND_QUEUE_MAPPING >> 1)-1));\
\
    (uctx)->uctx_ddcq_completed = ((uctx)->uctx_ddcq_insertcnt += 3);\
\
    PRINTF (uctx, DBG_DDCQ, "user_ddcq_command_flow_write: completed=%llx [%llx] addr=%llx\n", (uctx)->uctx_ddcq_completed, \
	    (uctx)->uctx_upage->upage_ddcq_completed, (uctx)->uctx_upage_addr); \
    user_ddcq_command_write (GUARD_CMD       | GUARD_ALL_CHANNELS,      0);\
    user_ddcq_command_write (WRITE_DWORD_CMD | (uctx)->uctx_upage_addr, 1);\
    user_ddcq_command_write ((uctx)->uctx_ddcq_completed,               2);\
} while (0)

#define user_ddcq_command_flow_intr(uctx) do { \
   E4_uint64 iptr   = (uctx)->uctx_ddcq_insertcnt; \
   ioaddr_t  cmdptr = (uctx)->uctx_ddcq->ucq_cq->cq_mapping + ((iptr<<3) & ((ELAN4_COMMAND_QUEUE_MAPPING >> 1)-1));\
\
    PRINTF (uctx, DBG_DDCQ, "user_ddcq_command_flow_intr: completed=%llx [%llx] addr=%llx\n", (uctx)->uctx_ddcq_completed, \
	    (uctx)->uctx_upage->upage_ddcq_completed, (uctx)->uctx_upage_addr); \
    user_ddcq_command_write (INTERRUPT_CMD   | ELAN4_INT_COOKIE_DDCQ,   3);\
} while (0)

#define user_ddcq_command_prologue(uctx, count) do { \
   E4_uint64 iptr   = (uctx)->uctx_ddcq_insertcnt; \
   ioaddr_t  cmdptr = (uctx)->uctx_ddcq->ucq_cq->cq_mapping + ((iptr<<3) & ((ELAN4_COMMAND_QUEUE_MAPPING >> 1)-1));\
   PRINTF(uctx, DBG_DDCQ, "user_ddcq_command_prologue: iptr=%llx cmdptr=%x\n", iptr, cmdptr);

#define user_ddcq_command_epilogue(uctx, count, extra) \
   (uctx)->uctx_ddcq_insertcnt = iptr + (count);\
\
   PRINTF(uctx, DBG_DDCQ, "user_ddcq_command_epilogue: iptr=%llx + %x + %x - completed %llx\n", iptr, count, extra, (uctx)->uctx_ddcq_completed);\
   if (((iptr) + (count) + (extra)) > ((uctx)->uctx_ddcq_completed + USER_CTRLFLOW_COUNT))\
       user_ddcq_command_flow_write(uctx); \
} while (0)

int
user_ddcq_check (USER_CTXT *uctx, unsigned num)
{
    PRINTF (uctx, DBG_DDCQ, "user_ddcq_check: insert=%llx completed=%llx [%llx] space=%d num=%d\n", 
	    uctx->uctx_ddcq_insertcnt, uctx->uctx_ddcq_completed,
	    uctx->uctx_upage->upage_ddcq_completed,
	    user_ddcq_command_space (uctx),
	    num
	);

    /* Ensure that there is enough space for the command we want to issue,
     * PLUS the guard/writeword for the control flow flush.
     * PLUS the interrupt command for rescheduling */
    if (user_ddcq_command_space (uctx) > (num + 4))
    {
	PRINTF (uctx, DBG_DDCQ, "user_ddcq_check: loads of space\n");

	return (1);
    }
    
    PRINTF (uctx, DBG_DDCQ, "user_ddcq_check: not enough space - reschedule\n");

    uctx->uctx_trap_state = UCTX_TRAP_SIGNALLED;
    return (0);
}

int
user_ddcq_flush (USER_CTXT *uctx)
{
    ELAN4_DEV *dev = uctx->uctx_ctxt.ctxt_dev;
    USER_CQ   *ucq = uctx->uctx_ddcq;

    switch (ucq->ucq_state)
    {
    case UCQ_TRAPPED:
	PRINTF (uctx, DBG_DDCQ, "user_ddcq_flush: command queue is trapped\n");
	return (0);
	
    case UCQ_NEEDS_RESTART:
	PRINTF (uctx, DBG_DDCQ, "user_ddcq_flush: restarting command queue\n");

	if (UCTX_RUNNABLE (uctx))
	{
	    ucq->ucq_state = UCQ_RUNNING;
	    elan4_restartcq (dev, ucq->ucq_cq);
	}
	break;
    }

    PRINTF (uctx, DBG_DDCQ, "user_ddcq_flush: insertcnt=%llx completed=%llx [%llx]\n", 
	    uctx->uctx_ddcq_insertcnt, uctx->uctx_ddcq_completed, uctx->uctx_upage->upage_ddcq_completed);

    if (uctx->uctx_ddcq_completed != uctx->uctx_ddcq_insertcnt)
	user_ddcq_command_flow_write (uctx);

    return (uctx->uctx_ddcq_completed == uctx->uctx_upage->upage_ddcq_completed);
}

void
user_ddcq_intr (USER_CTXT *uctx)
{
    user_ddcq_command_flow_intr (uctx);
}

void
user_ddcq_run_dma (USER_CTXT *uctx, E4_DMA *dma)
{
    PRINTF (uctx, DBG_DDCQ, "user_ddcq_run_dma: cookie=%llx vproc=%llx\n",  dma->dma_cookie, dma->dma_vproc);

    user_ddcq_command_prologue(uctx, 7) {

	user_ddcq_command_write ((dma->dma_typeSize & ~DMA_ContextMask) | RUN_DMA_CMD, 0);
	user_ddcq_command_write (dma->dma_cookie,                                      1);
	user_ddcq_command_write (dma->dma_vproc,                                       2);
	user_ddcq_command_write (dma->dma_srcAddr,                                     3);
	user_ddcq_command_write (dma->dma_dstAddr,                                     4);
	user_ddcq_command_write (dma->dma_srcEvent,                                    5);
	user_ddcq_command_write (dma->dma_dstEvent,                                    6);

    } user_ddcq_command_epilogue (uctx, 7, 0);
}

void
user_ddcq_run_thread (USER_CTXT *uctx, E4_ThreadRegs *regs)
{
    PRINTF (uctx, DBG_DDCQ, "user_ddcq_run_thread: PC=%llx SP=%llx\n", regs->Registers[0], regs->Registers[1]);

    user_ddcq_command_prologue(uctx, 7) {

	user_ddcq_command_write (regs->Registers[0] | RUN_THREAD_CMD, 0);
	user_ddcq_command_write (regs->Registers[1],                  1);
	user_ddcq_command_write (regs->Registers[2],                  2);
	user_ddcq_command_write (regs->Registers[3],                  3);
	user_ddcq_command_write (regs->Registers[4],                  4);
	user_ddcq_command_write (regs->Registers[5],                  5);
	user_ddcq_command_write (regs->Registers[6],                  6);
	
    } user_ddcq_command_epilogue (uctx, 7, 0);
}

void
user_ddcq_setevent (USER_CTXT *uctx, E4_Addr addr)
{
    user_ddcq_command_prologue (uctx, 1) {

	user_ddcq_command_write (SET_EVENT_CMD | addr, 0);
    
    } user_ddcq_command_epilogue (uctx, 1, 0);
}

void
user_ddcq_seteventn (USER_CTXT *uctx, E4_Addr addr, E4_uint32 count)
{
    PRINTF (uctx, DBG_DDCQ, "user_ddcq_seteventn: addr=%llx count=%lx\n", addr, count);

    user_ddcq_command_prologue (uctx, 2) {

	user_ddcq_command_write (SET_EVENTN_CMD, 0);
	user_ddcq_command_write (addr | count,   1);

    } user_ddcq_command_epilogue (uctx, 2, 0);
}

void
user_ddcq_waitevent (USER_CTXT *uctx, E4_Addr addr, E4_uint64 CountAndType, E4_uint64 Param0, E4_uint64 Param1)
{
    PRINTF (uctx, DBG_DDCQ, "user_ddcq_waitevent: addr=%llx CountAndType=%llx Param=%llx,%llx\n", addr, CountAndType, Param0, Param1);

    user_ddcq_command_prologue (uctx, 4) {

	user_ddcq_command_write (WAIT_EVENT_CMD | addr, 0);
	user_ddcq_command_write (CountAndType,          1);
	user_ddcq_command_write (Param0,                2);
	user_ddcq_command_write (Param1,                3);

    } user_ddcq_command_epilogue (uctx, 4, 0);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
