/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: cm.c,v 1.90 2005/05/24 09:19:12 david Exp $"
/*      $Source: /cvs/master/quadrics/epmod/cm.c,v $ */

#include <qsnet/kernel.h>

#include <elan/kcomm.h>

#include "kcomm_vp.h"
#include "debug.h"
#include "cm.h"
#include <elan/epsvc.h>

#include <qsnet/procfs_linux.h>

#if defined(LINUX)
#include "conf_linux.h"
#endif

int BranchingRatios[CM_MAX_LEVELS];

int MachineId      = -1;
int BrokenLevel    = -1;			/* Simulates Broken Network */
int RejoinCheck    = 1;
int RejoinPanic    = 0;

static int
SegmentNo (CM_RAIL *cmRail, u_int nodeid, u_int lvl)
{
    int i;

    ASSERT (lvl < cmRail->NumLevels);
    
    for (i = 0; i < lvl; i++)
	nodeid /= cmRail->Levels[i].NumSegs;
    
    return (nodeid % cmRail->Levels[lvl].NumSegs);
}

static int
ClusterIds (CM_RAIL *cmRail, int clvl, int *clmin, int *clmax)
{
    int clid  = cmRail->Rail->Position.pos_nodeid - cmRail->Levels[clvl].MinNodeId;

    if (clvl == 0)
	*clmin = *clmax = clid;
    else
    {
	*clmin = cmRail->Levels[clvl - 1].MinNodeId - cmRail->Levels[clvl].MinNodeId;
	*clmax = *clmin + cmRail->Levels[clvl - 1].NumNodes - 1;
    }
    return (clid);
}

static void
__Schedule_Timer (CM_RAIL *cmRail, long tick)
{
    if (! timer_pending (&cmRail->HeartbeatTimer) || AFTER (cmRail->NextRunTime, tick))
    {
	cmRail->NextRunTime = tick;

	mod_timer (&cmRail->HeartbeatTimer, tick);
    }
}

static void
__Schedule_Discovery (CM_RAIL *cmRail)		/* we urgently need to schedule discovery */
{
    __Schedule_Timer (cmRail, cmRail->NextDiscoverTime = lbolt);
}

static int
MsgBusy (CM_RAIL *cmRail, int msgNumber)
{
    switch (ep_outputq_state (cmRail->Rail, cmRail->MsgQueue, msgNumber))
    {
    case EP_OUTPUTQ_BUSY:			/* still busy */
	return 1;
	
    case EP_OUTPUTQ_FAILED:			/* NACKed */
    {
#if defined(DEBUG_PRINTF)
	CM_MSG  *msg  = ep_outputq_msg (cmRail->Rail, cmRail->MsgQueue, msgNumber);
	uint8_t  type  = msg->Hdr.Type;
	uint16_t nmaps = msg->Hdr.NumMaps;
	int16_t  off   = msg->Payload.Statemaps[CM_MSG_MAP(0)].offset;
	
	CPRINTF4 (((type == CM_MSG_TYPE_DISCOVER_LEADER) || (type == CM_MSG_TYPE_DISCOVER_SUBORDINATE))  ? 6 : 3, /* we expect broadcasts to be NACKed */
		  "%s: msg %d type %d  failed%s\n", cmRail->Rail->Name, msgNumber, type, 
		  (type != CM_MSG_TYPE_HEARTBEAT) ? "" : nmaps == 0 ? ": null heartbeat" :
		  off == STATEMAP_RESET ? ": heartbeat with R statemaps" : ": heartbeat with statemaps");
#endif
	return 0;
    }
    
    case EP_OUTPUTQ_FINISHED:
	return 0;

    default:
	panic ("MsgBusy - bad return code from ep_outputq_state\n");
	/* NOTREACHED */
    }
    return 0;
}

static void
LaunchMessage (CM_RAIL *cmRail, int msgNumber, int vp, int qnum, int retries, int type, int lvl, int nmaps)
{
    CM_MSG *msg = ep_outputq_msg (cmRail->Rail, cmRail->MsgQueue, msgNumber);
    CM_HDR *hdr = &msg->Hdr;

    ASSERT (nmaps >= 0 && nmaps <= CM_MSG_MAXMAPS);
    ASSERT (SPINLOCK_HELD (&cmRail->Lock));

    hdr->Version   = CM_MSG_VERSION;
    hdr->ParamHash = cmRail->ParamHash;
    hdr->Timestamp = cmRail->Timestamp;
    hdr->Checksum  = 0;
    hdr->NodeId    = cmRail->Rail->Position.pos_nodeid;
    hdr->MachineId = MachineId;
    hdr->NumMaps   = nmaps;
    hdr->Level     = lvl;
    hdr->Type      = type;
    hdr->Checksum  = CheckSum ((char *)msg + CM_MSG_BASE(nmaps), CM_MSG_SIZE(nmaps));

    if (BrokenLevel != -1 && (lvl >= ((BrokenLevel >> (cmRail->Rail->Number*4)) & 0xf)))			/* Simulate broken network? */
       return;

    if (ep_outputq_send (cmRail->Rail, cmRail->MsgQueue, msgNumber, 
			 CM_MSG_SIZE(nmaps), vp, qnum, retries));
	IncrStat (cmRail, LaunchMessageFail);
}

static int
SendMessage (CM_RAIL *cmRail, int nodeId, int lvl, int type)
{
    int	msgNumber = CM_NUM_NODE_MSG_BUFFERS + cmRail->NextSpareMsg;
    int n         = CM_NUM_SPARE_MSG_BUFFERS;
    int retries;

    ASSERT (type == CM_MSG_TYPE_IMCOMING ||	/* other types must use SendToSgmt */
	    type == CM_MSG_TYPE_REJOIN);
   
    while (n-- > 0 && MsgBusy (cmRail, msgNumber))	/* search for idle "spare" buffer */
    {
	if (++(cmRail->NextSpareMsg) == CM_NUM_SPARE_MSG_BUFFERS)
	    cmRail->NextSpareMsg = 0;
      
	msgNumber = CM_NUM_NODE_MSG_BUFFERS + cmRail->NextSpareMsg;
    }

    if (n == 0)					/* all "spare" message buffers busy */
    {
	CPRINTF3 (3, "%s: all spare message buffers busy: trying to send type %d to %d\n",
		  cmRail->Rail->Name, type, nodeId);
	return (0);
    }

    /* NB IMCOMING may be echoed by MANY nodes, so we don't (and musn't) have any retries */
    retries = (type == CM_MSG_TYPE_IMCOMING) ? 0 : CM_P2P_DMA_RETRIES;
   
    LaunchMessage (cmRail, msgNumber, EP_VP_NODE (nodeId), EP_SYSTEMQ_INTR, /* eager receive */
		   retries, type, lvl, 0);
   
    if (++(cmRail->NextSpareMsg) == CM_NUM_SPARE_MSG_BUFFERS) /* check this one last next time */
	cmRail->NextSpareMsg = 0;

    return (1);
}

static int
SendToSgmt (CM_RAIL *cmRail, CM_SGMT *sgmt, int type)
{    
    bitmap_t         seg;
    int              offset;
    int              nmaps;
    int              sidx;
    int              clvl;
    
    ASSERT (sgmt->Level <= cmRail->TopLevel);

    if (MsgBusy (cmRail, sgmt->MsgNumber))		/* previous message still busy */
    {
	CPRINTF3 (3, "%s: node message buffer busy: trying to send type %d to %d\n",
		  cmRail->Rail->Name, type, sgmt->NodeId);
      
	return (0);
    }

    switch (type)
    {
    case CM_MSG_TYPE_RESOLVE_LEADER:
    case CM_MSG_TYPE_DISCOVER_LEADER:
	ASSERT (sgmt->State == CM_SGMT_ABSENT);
	ASSERT (sgmt->Level == ((cmRail->Role == CM_ROLE_LEADER_CANDIDATE) ? cmRail->TopLevel : cmRail->TopLevel - 1));
	ASSERT (sgmt->Level < cmRail->NumLevels);
	ASSERT (sgmt->Sgmt == cmRail->Levels[sgmt->Level].MySgmt);
      
	/* broadcast to me and all my peers at this level (== my segment in the level above) */
	sidx = (sgmt->Level == cmRail->NumLevels - 1) ? 0 : cmRail->Levels[sgmt->Level + 1].MySgmt;

	LaunchMessage (cmRail, sgmt->MsgNumber, EP_VP_BCAST (sgmt->Level + 1, sidx), 
		       EP_SYSTEMQ_INTR, 0,		/* eager rx; no retries */
		       type, sgmt->Level, 0);
	return (1);
      
    case CM_MSG_TYPE_DISCOVER_SUBORDINATE:
	ASSERT (sgmt->Sgmt != cmRail->Levels[sgmt->Level].MySgmt);
	ASSERT (sgmt->State == CM_SGMT_WAITING);
	ASSERT (sgmt->Level > 0);			/* broadcasting just to subtree */
      
	LaunchMessage (cmRail, sgmt->MsgNumber, EP_VP_BCAST (sgmt->Level, sgmt->Sgmt), 
		       EP_SYSTEMQ_INTR, 0,		/* eager rx; no retries */
		       CM_MSG_TYPE_DISCOVER_SUBORDINATE, sgmt->Level, 0);
	return (1);
      
    case CM_MSG_TYPE_NOTIFY:
	ASSERT (sgmt->State == CM_SGMT_PRESENT);
      
	LaunchMessage (cmRail, sgmt->MsgNumber, EP_VP_NODE (sgmt->NodeId),
		       EP_SYSTEMQ_INTR, CM_P2P_DMA_RETRIES, /* eager rx; lots of retries */
		       CM_MSG_TYPE_NOTIFY, sgmt->Level, 0);
	return (1);
      
    case CM_MSG_TYPE_HEARTBEAT:
    {
	CM_MSG *msg = ep_outputq_msg (cmRail->Rail, cmRail->MsgQueue, sgmt->MsgNumber);
	CM_HDR *hdr = &msg->Hdr;

	ASSERT (sgmt->State == CM_SGMT_PRESENT);
	
	hdr->AckSeq = sgmt->AckSeq;
   
	if (!sgmt->MsgAcked)			/* Current message not acknowledged */
	{
	    /* must have been something significant to require an ack */
	    ASSERT (sgmt->SendMaps);
	    ASSERT (sgmt->NumMaps > 0);
	    
	    CPRINTF3 (3, "%s: retrying heartbeat to %d (%d entries)\n", cmRail->Rail->Name, sgmt->NodeId, sgmt->NumMaps);

	    IncrStat (cmRail, RetryHeartbeat);

	    nmaps = sgmt->NumMaps;
	}
	else
	{
	    nmaps = 0;
      
	    if (sgmt->SendMaps)			/* can send maps */
	    {
		for (clvl = sgmt->Level; clvl < cmRail->NumLevels; clvl++)
		{
		    if (!sgmt->Maps[clvl].OutputMapValid)
			continue;
		    
		    while ((offset = statemap_findchange (sgmt->Maps[clvl].OutputMap, &seg, 1)) >= 0)
		    {
			CM_STATEMAP_ENTRY *map = &msg->Payload.Statemaps[CM_MSG_MAP(nmaps)];

			sgmt->Maps[clvl].SentChanges = 1;
			
			map->level  = clvl;
			map->offset = offset;
			map->seg[0] = seg & 0xffff;
			map->seg[1] = (seg >> 16) & 0xffff;
#if (BT_ULSHIFT == 6)
			map->seg[2] = (seg >> 32) & 0xffff;
			map->seg[3] = (seg >> 48) & 0xffff;
#elif (BT_ULSHIFT != 5)
#error "Bad value for BT_ULSHIFT"
#endif
			if (++nmaps == CM_MSG_MAXMAPS)
			    goto msg_full;
		    }

		    if (sgmt->Maps[clvl].SentChanges)
		    {
			CM_STATEMAP_ENTRY *map = &msg->Payload.Statemaps[CM_MSG_MAP(nmaps)];

			sgmt->Maps[clvl].SentChanges = 0;

			map->level  = clvl;
			map->offset = STATEMAP_NOMORECHANGES;
			
			if (++nmaps == CM_MSG_MAXMAPS)
			    goto msg_full;
		    }
		}
	    }
	    
	    ASSERT (nmaps < CM_MSG_MAXMAPS);

	msg_full:
	    sgmt->NumMaps = nmaps;		/* remember how many incase we retry */

	    if (nmaps == 0)			/* no changes to send */
		hdr->Seq = sgmt->MsgSeq;	/* this one can be dropped */
	    else
	    {
		hdr->Seq = ++(sgmt->MsgSeq);	/* on to next message number */
		sgmt->MsgAcked = 0;		/* need this one to be acked before I can send another */

		IncrStat (cmRail, MapChangesSent);
	    }
	}

	LaunchMessage (cmRail, sgmt->MsgNumber, EP_VP_NODE (sgmt->NodeId), 
		       EP_SYSTEMQ_POLLED,  CM_P2P_DMA_RETRIES, /* polled receive, lots of retries */
		       CM_MSG_TYPE_HEARTBEAT, sgmt->Level, nmaps);

	IncrStat (cmRail, HeartbeatsSent);

	return (1);
    }

    default:					/* other types must use SendMessage */
	printk ("SendToSgmt: invalid type %d\n", type);
	ASSERT (0);

	return (1);
    }
}

static char *
GlobalStatusString (statemap_t *map, int idx)
{
    char *strings[] = {"....", "S...", "C...", "R...", 
		       ".s..", "Ss..", "Cs..", "Rs..", 
		       "..r.", "S.r.", "C.r.", "R.r.", 
		       ".sr.", "Ssr.", "Csr.", "Rsr.", 
		       "...R", "S..R", "C..R", "R..R", 
		       ".s.R", "Ss.R", "Cs.R", "Rs.R", 
		       "..rR", "S.rR", "C.rR", "R.rR", 
		       ".srR", "SsrR", "CsrR", "RsrR"};
    
    return (strings[statemap_getbits (map, idx * CM_GSTATUS_BITS, CM_GSTATUS_BITS)]);
}

static char *
MapString (char *name, statemap_t *map, int nnodes, char *trailer)
{
    static char *space;
    int          i;

    if (space == NULL)
	KMEM_ALLOC (space, char *, EP_MAX_NODES*(CM_GSTATUS_BITS+1), 0);

    if (space == NULL)
	return ("<cannot allocate memory>");
    else
    {
	char *ptr = space;

	sprintf (space, "%s ", name); ptr += strlen (ptr);
	for (i = 0; i < nnodes; i++, ptr += strlen (ptr))
	    sprintf (ptr, "%s%s", i == 0 ? "" : ",", GlobalStatusString (map, i));
	sprintf (ptr, " %s", trailer);
	return (space);
    }
}

void
DisplayMap (DisplayInfo *di, CM_RAIL *cmRail, char *name, statemap_t *map, int nnodes, char *trailer)
{
    char  linebuf[256];
    char *ptr = linebuf;
    int   i;

#define NODES_PER_LINE	32
    for (i = 0; i < nnodes; i++)
    {
	if (ptr == linebuf)
	{
	    sprintf (ptr, "%4d", i);
	    ptr += strlen (ptr);
	}
	
	sprintf (ptr, ",%s", GlobalStatusString (map, i));
	ptr += strlen (ptr);

	if ((i % NODES_PER_LINE) == (NODES_PER_LINE-1) || (i == (nnodes-1)))
	{
	    (di->func)(di->arg, "%s: %s %s %s\n", cmRail->Rail->Name, name, linebuf, trailer);
	    ptr = linebuf;
	}
    }
#undef NODES_PER_LINE
}

void
DisplayNodeMaps (DisplayInfo *di, CM_RAIL *cmRail)
{
    int   lvl;
    int   clvl;
    char  mapname[128];
    
    (di->func)(di->arg, "%s: Node %d maps...\n", cmRail->Rail->Name, cmRail->Rail->Position.pos_nodeid);

    for (clvl = 0; clvl < cmRail->NumLevels; clvl++)
    {
	int nnodes = cmRail->Levels[clvl].NumNodes;

	(di->func)(di->arg, "%s: Cluster level %d: Connected %ld - %s%s\n", 
		   cmRail->Rail->Name, clvl, cmRail->Levels[clvl].Connected,
		   cmRail->Levels[clvl].Online     ? "Online" : "Offline",
		   cmRail->Levels[clvl].Restarting ? ", Restarting" : "");

	for (lvl = 0; lvl < cmRail->TopLevel && lvl <= clvl; lvl++)
	{
	    CM_LEVEL *level = &cmRail->Levels[lvl];

	    sprintf (mapname, "%10s%2d", "Level", lvl);
	    DisplayMap (di, cmRail, mapname, level->SubordinateMap[clvl], nnodes,
			level->SubordinateMapValid[clvl] ? "" : "(invalid)");
	}

	sprintf (mapname, "%12s", "Local");
	DisplayMap (di, cmRail, mapname, cmRail->Levels[clvl].LocalMap, nnodes, "");

	sprintf (mapname, "%12s", "Subtree");
	DisplayMap (di, cmRail, mapname, cmRail->Levels[clvl].SubTreeMap, nnodes, 
		    cmRail->Levels[clvl].SubTreeMapValid ? "" : "(invalid)");

	sprintf (mapname, "%12s", "Global");
	DisplayMap (di, cmRail, mapname, cmRail->Levels[clvl].GlobalMap, nnodes, 
		    cmRail->Levels[clvl].GlobalMapValid ? "" : "(invalid)");

	sprintf (mapname, "%12s", "LastGlobal");
	DisplayMap (di, cmRail, mapname, cmRail->Levels[clvl].LastGlobalMap, nnodes, "");
    }
}

void
DisplayNodeSgmts (DisplayInfo *di, CM_RAIL *cmRail)
{
    int   lvl;
    int   sidx;
    
    (di->func)(di->arg, "%s: Node %d segments...\n", cmRail->Rail->Name, cmRail->NodeId);
    
    for (lvl = 0; lvl <= cmRail->TopLevel && lvl < cmRail->NumLevels; lvl++)
    {
	(di->func)(di->arg, "   level %d: ", lvl);
	
	for (sidx = 0; sidx < ((lvl == cmRail->TopLevel) ? 1 : cmRail->Levels[lvl].NumSegs); sidx++)
	{
	    CM_SGMT *sgmt = &cmRail->Levels[lvl].Sgmts[sidx];
		
	    if (sgmt->State == CM_SGMT_PRESENT)
		(di->func)(di->arg, "[%d, in: %d out: %d %s%s]", 
			sgmt->NodeId,
			sgmt->AckSeq,
			sgmt->MsgSeq,
			sgmt->MsgAcked ? "A" : "-",
			sgmt->SendMaps ? "!" : "-");
	    else
		(di->func)(di->arg, "[%s]", (sgmt->State == CM_SGMT_ABSENT ? "absent" :
				 sgmt->State == CM_SGMT_WAITING ? "waiting" :
				 sgmt->State == CM_SGMT_COMING ? "coming" : "UNKNOWN"));
	}
	(di->func)(di->arg, "\n");
    }
}


static void
StartConnecting (CM_RAIL *cmRail, CM_SGMT *sgmt, int NodeId, int Timestamp)
{
    int clvl;

    CPRINTF4 (2, "%s: lvl %d subtree %d node %d -> connecting\n", cmRail->Rail->Name, sgmt->Level, sgmt->Sgmt, NodeId);

    /* Only reconnect the same guy if he was reborn */
    ASSERT (sgmt->State != CM_SGMT_PRESENT ||
	    (sgmt->NodeId == NodeId && sgmt->Timestamp != Timestamp));
   
    /* After we've connected to a new peer, we wait to receive
     * STATEMAP_RESET before we accumulate changes and we wait for a
     * complete map to be received before we propagate changes to other
     * nodes.
     *
     * If I'm the subordinate, I can start sending maps right away, since
     * the leader is ready for them already.  If I'm the leader, I hold off
     * sending maps until I've seen the subordinate's first heartbeat,
     * because the subordinate might miss my NOTIFY message, still think
     * she's a leader candidate and ignore my heartbeats.
     */
    sgmt->SendMaps = (sgmt->Level == cmRail->TopLevel); /* I can send maps to my leader (she NOTIFIED me) */

    for (clvl = sgmt->Level; clvl < cmRail->NumLevels; clvl++)
    {
	statemap_reset (sgmt->Maps[clvl].CurrentInputMap);
	statemap_reset (sgmt->Maps[clvl].InputMap);
	statemap_reset (sgmt->Maps[clvl].OutputMap);
	
	sgmt->Maps[clvl].InputMapValid = 0;
	sgmt->Maps[clvl].OutputMapValid = 0;
	sgmt->Maps[clvl].SentChanges = 0;

	if (sgmt->Level == cmRail->TopLevel)	/* connection to leader */
	{
	    ASSERT (sgmt->Sgmt == 0);
	    ASSERT (cmRail->Role == CM_ROLE_SUBORDINATE);

	    if (cmRail->Levels[clvl].SubTreeMapValid) /* already got a subtree map to send up */
	    {
		statemap_setmap (sgmt->Maps[clvl].OutputMap, cmRail->Levels[clvl].SubTreeMap);
		sgmt->Maps[clvl].OutputMapValid = 1;

		statemap_clearchanges (cmRail->Levels[clvl].SubTreeMap);
	    }
	}
	else					/* connection to subordinate */
	{
	    ASSERT (sgmt->Sgmt != cmRail->Levels[sgmt->Level].MySgmt);

	    if (cmRail->Levels[clvl].GlobalMapValid) /* already got a global map to broadcast */
	    {
		statemap_setmap (sgmt->Maps[clvl].OutputMap, cmRail->Levels[clvl].GlobalMap);
		sgmt->Maps[clvl].OutputMapValid = 1;
	    }
	}
    }
    
    /* Initialise sequence counters */
    sgmt->MsgSeq = sgmt->AckSeq = 0;
    sgmt->MsgAcked = 1;			/* ready to send a new sequenced message */
   
    sgmt->State      = CM_SGMT_PRESENT;
    sgmt->NodeId     = NodeId;
    sgmt->UpdateTick = lbolt;
    sgmt->Timestamp  = Timestamp;
}

static void
StartSubTreeDiscovery (CM_RAIL *cmRail, CM_SGMT *sgmt)
{
    sgmt->State = CM_SGMT_WAITING;
    sgmt->UpdateTick = lbolt;
    sgmt->WaitingTick = lbolt;

    if (sgmt->Level > 0)
	__Schedule_Discovery (cmRail);
}

void
StartSubordinateDiscovery (CM_RAIL *cmRail)
{
    int       i;
    int       lvl = cmRail->TopLevel - 1;
    CM_LEVEL *level = &cmRail->Levels[lvl];

    ASSERT (lvl >= 0 && lvl < cmRail->NumLevels);

    for (i = 0; i < level->NumSegs; i++)
    {
        CM_SGMT *sgmt = &level->Sgmts[i];
	
	if (i != level->MySgmt)		/* No-one should connect here */
	    StartSubTreeDiscovery (cmRail, sgmt);
    }
}

void
StartLeaderDiscovery (CM_RAIL *cmRail)
{
    int       i;
    int       clvl;
    CM_LEVEL *level = &cmRail->Levels[cmRail->TopLevel];

    ASSERT (cmRail->TopLevel < cmRail->NumLevels);

    for (clvl = cmRail->TopLevel; clvl < cmRail->NumLevels; clvl++)
    {
        cmRail->Levels[clvl].GlobalMapValid = 0;
	cmRail->Levels[clvl].SubTreeMapValid = 0;
        level->SubordinateMapValid[clvl] = 0;
    }

    for (i = 0; i < level->NumSegs; i++)
    {
        CM_SGMT *sgmt = &level->Sgmts[i];
	
	sgmt->State = CM_SGMT_ABSENT;
    }

    cmRail->DiscoverStartTick = lbolt;
    cmRail->Role = CM_ROLE_LEADER_CANDIDATE;
   
    __Schedule_Discovery (cmRail);
}

static void
RaiseTopLevel (CM_RAIL *cmRail)
{
    ASSERT (cmRail->NumLevels != 0);
    ASSERT (cmRail->TopLevel < cmRail->NumLevels);

    CPRINTF2 (2, "%s: RaiseTopLevel %d\n", cmRail->Rail->Name, cmRail->TopLevel + 1);

    if (++cmRail->TopLevel == cmRail->NumLevels)	/* whole machine leader? */
       cmRail->Role = CM_ROLE_LEADER;
    else
       StartLeaderDiscovery (cmRail);		/* look for my leader */

    StartSubordinateDiscovery (cmRail);		/* and any direct subordinates */
}

static void
LowerTopLevel (CM_RAIL *cmRail, int lvl)
{
    ASSERT (cmRail->NumLevels != 0);
    ASSERT (lvl < cmRail->NumLevels);

    CPRINTF2 (2, "%s: LowerTopLevel %d\n", cmRail->Rail->Name, lvl);

    if (lvl == 0)
	cmRail->Timestamp = lbolt;

    cmRail->TopLevel = lvl;

    StartLeaderDiscovery (cmRail);		/* look for my leader */
}

static int
IShouldLead (CM_RAIL *cmRail, CM_MSG *msg)
{
    /* NB, this function MUST be consistently calculated on any nodes, just
     * from the info supplied in the message.  Otherwise leadership
     * arbitration during concurrent discovery will fail.
     */
    return (cmRail->NodeId < msg->Hdr.NodeId);
}

static int
SumCheck (CM_MSG *msg)
{
    CM_HDR   *hdr   = &msg->Hdr;
    uint16_t  sum   = hdr->Checksum;
    uint16_t  nmaps = hdr->NumMaps;

    if (nmaps > CM_MSG_MAXMAPS) {
	printk ("SumCheck: nmaps %d > CM_MSG_MAXMAPS\n", nmaps);
	return 0;
    }
    
    if ((hdr->Type != CM_MSG_TYPE_HEARTBEAT) && nmaps != 0) {
	printk ("SumCheck: type(%d) not HEARTBEAT and nmaps(%d) != 0\n", hdr->Type, nmaps);
	return 0;
    }

    hdr->Checksum = 0;
    
    if (CheckSum ((char *)msg + CM_MSG_BASE(nmaps), CM_MSG_SIZE(nmaps)) != sum) {
	printk ("SumCheck: checksum failed %x %x\n", CheckSum ((char *)msg + CM_MSG_BASE(nmaps), CM_MSG_SIZE(nmaps)), sum);

	return 0;
    }
	
    return 1;
}

static void
ProcessMessage (EP_RAIL *rail, void *arg, void *msgbuf)
{
    CM_RAIL 	   *cmRail = (CM_RAIL *) arg;
    CM_MSG         *msg    = (CM_MSG *) msgbuf;
    CM_HDR         *hdr    = &msg->Hdr;
    int             lvl;
    int             sidx;
    CM_LEVEL       *level;
    CM_SGMT        *sgmt;
    bitmap_t        seg;
    int             i;
    int            delay;
    static long    tlast;
    static int     count;

    /* Poll the message Version field until the message has completely
     * arrived in main memory. */
    for (delay = 1; hdr->Version == EP_SYSTEMQ_UNRECEIVED && delay < EP_SYSTEMQ_UNRECEIVED_TLIMIT; delay <<= 1)
	DELAY (delay);

    /* Display a message every 60 seconds if we see an "old" format message */
    if (hdr->Version == EP_SYSTEMQ_UNRECEIVED && (((lbolt - tlast) > 60*HZ) ? (count = 0) : ++count) < 1)
    {
	printk ("%s: received old protocol message (type %d from node %d)\n", cmRail->Rail->Name, 
		((uint8_t *) msg)[20], ((uint16_t *) msg)[4]);

	tlast = lbolt;
	goto finished;
    }

    if (hdr->Version != CM_MSG_VERSION || hdr->ParamHash != cmRail->ParamHash || hdr->MachineId != MachineId)
    {
	CPRINTF8 (1, "%s: invalid message : Version %08x (%08x) ParamHash %08x (%08x) MachineId %04x (%04x) Nodeid %d\n", cmRail->Rail->Name,
		  hdr->Version, CM_MSG_VERSION, hdr->ParamHash, cmRail->ParamHash, hdr->MachineId, MachineId, hdr->NodeId);
	goto finished;
    }

    if (!SumCheck (msg))
    {
	printk ("%s: checksum failed on msg from %d?\n", cmRail->Rail->Name, hdr->NodeId);
	goto finished;
    }
    
    if (hdr->NodeId == cmRail->NodeId)		/* ignore my own broadcast */       
    {
	CPRINTF3 (6, "%s: node %d type %d: ignored (MESSAGE FROM ME)\n", 
		  cmRail->Rail->Name, hdr->NodeId, hdr->Type);

	if (hdr->Type != CM_MSG_TYPE_DISCOVER_LEADER && hdr->Type != CM_MSG_TYPE_RESOLVE_LEADER)
	    printk ("%s: node %d type %d: ignored (MESSAGE FROM ME)\n", 
		    cmRail->Rail->Name, hdr->NodeId, hdr->Type);
	goto finished;
    }

    lvl = hdr->Level;
    level = &cmRail->Levels[lvl];

    if (BrokenLevel != -1 && (lvl >= ((BrokenLevel >> (cmRail->Rail->Number*4)) & 0xf)))			/* Simulate broken network? */
       goto finished;
    
    if (lvl >= cmRail->NumLevels ||		/* from outer space  */
	hdr->NodeId < level->MinNodeId ||	/* from outside this level's subtree */
	hdr->NodeId >= level->MinNodeId + level->NumNodes)
    {
	printk ("%s: lvl %d node %d type %d: ignored (%s)\n", 
		cmRail->Rail->Name, lvl, hdr->NodeId, hdr->Type, 
		lvl >= cmRail->NumLevels ? "level too big for machine" : "outside subtree");
	goto finished;
    }

    sidx = SegmentNo (cmRail, hdr->NodeId, lvl);
    sgmt = &level->Sgmts[sidx];
    
    switch (hdr->Type)
    {
    case CM_MSG_TYPE_RESOLVE_LEADER:
	if (lvl >= cmRail->TopLevel)
	{
	    CPRINTF4 (6, "%s: lvl %d sidx %d node %d RESOLVE_LEADER: ignored (above my level)\n", 
		      cmRail->Rail->Name, lvl, sidx, hdr->NodeId);
	    break;
	}

	/* someone else thinks they lead at the same level as me */
	CPRINTF4 (1, "%s: lvl %d sidx %d node %d RESOLVE_LEADER: !REJOIN (putsch)\n", 
		  cmRail->Rail->Name, lvl, sidx, hdr->NodeId);
	
	printk ("%s: lvl %d sidx %d node %d RESOLVE_LEADER: !REJOIN (putsch)\n", 
		cmRail->Rail->Name, lvl, sidx, hdr->NodeId);
	

	SendMessage (cmRail, hdr->NodeId, lvl, CM_MSG_TYPE_REJOIN);
	break;
	
    case CM_MSG_TYPE_DISCOVER_LEADER:
	if (lvl > cmRail->TopLevel)
	{
	    CPRINTF4 (6, "%s: lvl %d sidx %d node %d DISCOVER_LEADER: ignored (above my level)\n", 
		      cmRail->Rail->Name, lvl, sidx, hdr->NodeId);
	    break;
	}

	if (sidx == level->MySgmt)		/* someone I led thinks they lead some of my subtrees */
	{
	    CPRINTF4 (1, "%s: lvl %d sidx %d node %d DISCOVER_LEADER: !REJOIN (putsch)\n", 
		      cmRail->Rail->Name, lvl, sidx, hdr->NodeId);

	    printk ("%s: lvl %d sidx %d node %d DISCOVER_LEADER: !REJOIN (putsch)\n", 
		    cmRail->Rail->Name, lvl, sidx, hdr->NodeId);

	    SendMessage (cmRail, hdr->NodeId, hdr->Level, CM_MSG_TYPE_REJOIN);
	    break;
	}       

	if (lvl < cmRail->TopLevel)			/* I'm the leader of this level */
	{
	    if (sgmt->State == CM_SGMT_PRESENT &&	/* someone thinks someone I lead is dead */
		sgmt->NodeId != hdr->NodeId)
	    {
		/* My subordinate's death could be noticed by one of her peers
		 * before I do.  If she _is_ dead, I'll notice before long and
		 * NOTIFY this discover.  If this discover completes before I
		 * detect my subordinate's death, the discovering node will
		 * try to take over from me, and then I'll RESET her.
		 */
		CPRINTF4 (6, "%s: lvl %d sidx %d node %d DISCOVER_LEADER: ignored (got established subordinate)\n", 
			  cmRail->Rail->Name, lvl, sidx, hdr->NodeId);
		return;
	    }

	    if (sgmt->State != CM_SGMT_PRESENT || /* New connection */
		sgmt->Timestamp != hdr->Timestamp) /* new incarnation */
		StartConnecting (cmRail, sgmt, hdr->NodeId, hdr->Timestamp);
	  
	    CPRINTF4 (2, "%s: lvl %d sidx %d node %d DISCOVER_LEADER: !NOTIFY)\n", 
		      cmRail->Rail->Name, lvl, sidx, hdr->NodeId);
	  
	    SendToSgmt (cmRail, sgmt, CM_MSG_TYPE_NOTIFY);
	    break;
	}

	ASSERT (lvl == cmRail->TopLevel);

	if (cmRail->Role == CM_ROLE_SUBORDINATE)
	{
	    /* I think my leader is alive, in which case she'll NOTIFY this
	     * DISCOVER.  If she's dead, I'll start to become a leader
	     * candidate and handle this appropriately.
	     */
	    CPRINTF3 (6, "%s: lvl %d node %d DISCOVER: ignored (I'm a subordinate)\n", 
		      cmRail->Rail->Name, lvl, hdr->NodeId);
	    break;
	}
       
	ASSERT (cmRail->Role == CM_ROLE_LEADER_CANDIDATE);

	/* A peer at this level is bidding for leadership along with me */
	if (IShouldLead (cmRail, msg))
	{
	    CPRINTF3 (6, "%s: lvl %d node %d DISCOVER: but I should lead\n", 
		      cmRail->Rail->Name, lvl, hdr->NodeId);

	    /* So there _is_ someone there; She'll be seeing my DISCOVER
	     * messages and extending her discovery period, so that when I
	     * become leader, I'll NOTIFY her.  In the meantime I'll flag her
	     * activity, so she remains WAITING.
	     */
	    sgmt->UpdateTick = lbolt;
	    break;
	}
       
	/* Defer to sender... */
	CPRINTF3 (6, "%s: lvl %d node %d DISCOVER: delaying me becoming leader\n", 
		  cmRail->Rail->Name, lvl, hdr->NodeId);
       
	StartLeaderDiscovery (cmRail);
	break;

    case CM_MSG_TYPE_DISCOVER_SUBORDINATE:
	if (lvl <= cmRail->TopLevel)
	{
	    CPRINTF3 (6, "%s: lvl %d node %d DISCOVER_SUBORDINATE: ignored (from my subtree)\n", 
		      cmRail->Rail->Name, lvl, hdr->NodeId);
	    break;
	}
       
	if (cmRail->Role != CM_ROLE_LEADER_CANDIDATE)
	{
	    CPRINTF3 (6, "%s: lvl %d node %d DISCOVER_SUBORDINATE: ignored (I'm not looking for a leader)\n", 
		      cmRail->Rail->Name, lvl, hdr->NodeId);
	    break;
	}
       
	if (hdr->Level > cmRail->BroadcastLevel && AFTER (lbolt, cmRail->BroadcastLevelTick + EP_WITHDRAW_TIMEOUT))
	{
	    CPRINTF3 (6, "%s: lvl %d node %d DISCOVER_SUBORDINATE: ignored (broadcast level too low)\n",
		      cmRail->Rail->Name, lvl, hdr->NodeId);
	    break;
	}

	CPRINTF3 (2, "%s: lvl %d node %d DISCOVER_SUBORDINATE: !IMCOMING\n", 
		  cmRail->Rail->Name, lvl, hdr->NodeId);
       
	SendMessage (cmRail, hdr->NodeId, hdr->Level, CM_MSG_TYPE_IMCOMING);
	break;

    case CM_MSG_TYPE_IMCOMING:
	if (lvl > cmRail->TopLevel ||		/* from peer or node above me */
	    sgmt->State == CM_SGMT_PRESENT ||	/* already got a subtree */
	    sgmt->State == CM_SGMT_ABSENT)	/* already written off this subtree */
	{
	    CPRINTF4 (2, "%s: lvl %d sidx %d node %d IMCOMING: ignored\n", cmRail->Rail->Name, lvl, sidx, hdr->NodeId);
	    break;
	}

	CPRINTF4 (2, "%s: lvl %d sidx %d node %d IMCOMING: waiting...\n", cmRail->Rail->Name, lvl, sidx, hdr->NodeId);

	sgmt->State = CM_SGMT_COMING;
	sgmt->UpdateTick = lbolt;
	break;
       
    case CM_MSG_TYPE_NOTIFY:
	if (cmRail->Role != CM_ROLE_LEADER_CANDIDATE || /* I'm not looking for a leader */
	    lvl != cmRail->TopLevel)		/* at this level */
	{
	    /* If this person really should be my leader, my existing leader
	     * will time out, and I'll discover this one. */
	    CPRINTF4 (2, "%s: lvl %d node %d NOTIFY: ignored (%s)\n", 
		      cmRail->Rail->Name, lvl, hdr->NodeId,
		      lvl < cmRail->TopLevel ? "already leader" : 
		      lvl > cmRail->TopLevel ? "lvl too high" : "already subordinate");
	    break;
	}

	CPRINTF3 (2, "%s: lvl %d node %d NOTIFY: becoming subordinate\n", 
		  cmRail->Rail->Name, lvl, hdr->NodeId);
       
	cmRail->Role = CM_ROLE_SUBORDINATE;		/* Now I've found my level */
	StartConnecting (cmRail, &level->Sgmts[0], hdr->NodeId, hdr->Timestamp);
	break;

    case CM_MSG_TYPE_HEARTBEAT:
	if (lvl > cmRail->TopLevel)
	{
	    CPRINTF3 (2, "%s: lvl %d node %d H/BEAT: ignored (lvl too high)\n", 
		      cmRail->Rail->Name, lvl, hdr->NodeId);
	    break;
	}

	if (lvl == cmRail->TopLevel)			/* heartbeat from my leader */
	{
	    if (cmRail->Role == CM_ROLE_LEADER_CANDIDATE) /* but I've not got one */
	    {
		/* I'm probably a new incarnation of myself; I'll keep doing
		 * discovery until my previous existence's leader NOTIFY's me.
		 * If I was this node's leader, she'll time me out (I'm not
		 * sending heartbeats to her) and we'll fight it out for
		 * leadership. */
		CPRINTF3 (2, "%s: lvl %d node %d H/BEAT ignored (no leader)\n", 
			  cmRail->Rail->Name, lvl, hdr->NodeId);
		break;
	    }
	    sidx = 0;
	    sgmt = &level->Sgmts[0];
	}
      
	if (sgmt->State != CM_SGMT_PRESENT ||	/* not fully connected with this guy */
	    sgmt->NodeId != hdr->NodeId ||	/* someone else impersonating my peer */
	    sgmt->Timestamp != hdr->Timestamp)	/* new incarnation of my peer */
	{
	    CPRINTF4 (1, "%s: lvl %d sidx %d node %d H/BEAT: !REJOIN\n", 
		      cmRail->Rail->Name, lvl, sidx, hdr->NodeId);
	  
	    printk ("%s: lvl %d sidx %d node %d H/BEAT: !REJOIN %s\n",
		    cmRail->Rail->Name, lvl, sidx, hdr->NodeId,
		    sgmt->State != CM_SGMT_PRESENT ? "not present" :
		    sgmt->NodeId != hdr->NodeId ? "someone else" : "new incarnation");
	    
	    SendMessage (cmRail, hdr->NodeId, hdr->Level, CM_MSG_TYPE_REJOIN);
	    break;
	}

	if (!((hdr->Seq == sgmt->AckSeq) ||	/* NOT duplicate message or */
	      (hdr->Seq == (CM_SEQ)(sgmt->AckSeq + 1))) || /* expected message */
	    !((hdr->AckSeq == sgmt->MsgSeq) ||	/* NOT expected ack or */
	      (hdr->AckSeq == (CM_SEQ)(sgmt->MsgSeq - 1)))) /* duplicate ack */
	{
	    CPRINTF9 (1, "%s: lvl %d sidx %d node %d type %d: H/BEAT !REJOIN (out-of-seq) M(%d,a%d) S%d,A%d\n", 
		      cmRail->Rail->Name, lvl, sidx, hdr->NodeId, hdr->Type, 
		      (int)hdr->Seq, (int)hdr->AckSeq, (int)sgmt->MsgSeq, (int)sgmt->AckSeq);
	 
	    printk ("%s: lvl %d sidx %d node %d type %d: H/BEAT !REJOIN (out-of-seq) M(%d,a%d) S%d,A%d\n", 
		    cmRail->Rail->Name, lvl, sidx, hdr->NodeId, hdr->Type, 
		    (int)hdr->Seq, (int)hdr->AckSeq, (int)sgmt->MsgSeq, (int)sgmt->AckSeq);
	 
	    SendMessage (cmRail, hdr->NodeId, hdr->Level, CM_MSG_TYPE_REJOIN);
	    break;
	}

	IncrStat (cmRail, HeartbeatsRcvd);

	sgmt->UpdateTick = lbolt;
	sgmt->SendMaps = 1;

	if (sgmt->MsgSeq == hdr->AckSeq)		/* acking current message */
	    sgmt->MsgAcked = 1;			/* can send the next one */

	if (hdr->Seq == sgmt->AckSeq)		/* discard duplicate (or NULL heartbeat) */
	{
	    CPRINTF6 (6, "%s: lvl %d sidx %d node %d type %d: %s H/BEAT\n", 
		      cmRail->Rail->Name, lvl, sidx, hdr->NodeId, hdr->Type,
		      hdr->NumMaps == 0 ? "null" : "duplicate");
	    break;
	}

	CPRINTF7 (6, "%s: lvl %d sidx %d node %d type %d: seq %d maps %d H/BEAT\n", 
		  cmRail->Rail->Name, lvl, sidx, hdr->NodeId, hdr->Type, hdr->Seq, hdr->NumMaps);

	sgmt->AckSeq = hdr->Seq;			/* ready to receive next one */
       
	for (i = 0; i < hdr->NumMaps; i++)
	{
	    CM_STATEMAP_ENTRY *map  = &msg->Payload.Statemaps[CM_MSG_MAP(i)];
	    int		       clvl = map->level;
	    
	    if (clvl < 0)			/* end of message */
		break;

	    if (clvl < sgmt->Level)		/* bad level */
	    {
		CPRINTF6 (1, "%s: lvl %d sidx %d node %d type %d: H/BEAT !REJOIN (bad clevel %d)\n", 
			  cmRail->Rail->Name, lvl, sidx, hdr->NodeId, hdr->Type, clvl);

		SendMessage (cmRail, hdr->NodeId, hdr->Level, CM_MSG_TYPE_REJOIN);
		goto finished;
	    }

	    if (map->offset == STATEMAP_NOMORECHANGES) /* end of atomic changes */
	    {
		if (!sgmt->Maps[clvl].InputMapValid || /* not set InputMap yet */
		    statemap_changed (sgmt->Maps[clvl].CurrentInputMap)) /* previously applied changes */
		{
		    CPRINTF3 (4, "%s: received new clvl %d map from %d\n", cmRail->Rail->Name, clvl, sgmt->NodeId);

		    statemap_setmap (sgmt->Maps[clvl].InputMap, sgmt->Maps[clvl].CurrentInputMap);
		    sgmt->Maps[clvl].InputMapValid = 1;

		    statemap_clearchanges (sgmt->Maps[clvl].CurrentInputMap);
		}
		continue;
	    }
	    
	    seg = ((bitmap_t)map->seg[0])
		| (((bitmap_t)map->seg[1]) << 16)
#if (BT_ULSHIFT == 6)
		| (((bitmap_t)map->seg[2]) << 32)
		| (((bitmap_t)map->seg[3]) << 48)
#elif (BT_ULSHIFT != 5)
#error "Bad value for BT_ULSHIFT"
#endif
		;
	    statemap_setseg (sgmt->Maps[clvl].CurrentInputMap, map->offset, seg);
	}
	break;

    case CM_MSG_TYPE_REJOIN:
	CPRINTF5 (1, "%s: lvl %d sidx %d node %d type %d: REJOIN\n",
		  cmRail->Rail->Name, lvl, sidx, hdr->NodeId, hdr->Type);
	printk ("%s: lvl %d sidx %d node %d type %d: REJOIN\n", 
		cmRail->Rail->Name, lvl, sidx, hdr->NodeId, hdr->Type);

	LowerTopLevel (cmRail, 0);

	IncrStat (cmRail, RejoinRequest);
	break;

    default:
	printk ("%s: lvl=%d unknown message type %d\n", cmRail->Rail->Name, lvl, hdr->Type);
	break;
    }
 finished:
    hdr->Version = EP_SYSTEMQ_UNRECEIVED;
}

static void
PollInputQueues (CM_RAIL *cmRail)
{
    ep_poll_inputq (cmRail->Rail, cmRail->IntrQueue, 0, ProcessMessage, cmRail);
    ep_poll_inputq (cmRail->Rail, cmRail->PolledQueue, 0, ProcessMessage, cmRail);
}

static void
IntrQueueCallback (EP_RAIL *rail, void *arg)
{
    CM_RAIL *cmRail = (CM_RAIL *) arg;
    unsigned long flags;

    /* If the lock is held, then don't bother spinning for it, 
     * since the messages will be received at this, or the
     * next heartbeat */
    local_irq_save (flags);
    if (spin_trylock (&cmRail->Lock))
    {
	if (AFTER (lbolt, cmRail->NextRunTime + MSEC2TICKS(CM_TIMER_SCHEDULE_TIMEOUT)))
	    printk ("%s: heartbeat timer stuck - scheduled\n", cmRail->Rail->Name);
	else
	    ep_poll_inputq (rail, cmRail->IntrQueue, 0, ProcessMessage, cmRail);
	spin_unlock (&cmRail->Lock);
    }
    local_irq_restore (flags);
}

char *
sprintClPeers (char *str, CM_RAIL *cmRail, int clvl)
{
   int clLo     = cmRail->Levels[clvl].MinNodeId;
   int clHi     = clLo + cmRail->Levels[clvl].NumNodes - 1;
   int subClLo  = (clvl == 0) ? cmRail->NodeId : cmRail->Levels[clvl - 1].MinNodeId;
   int subClHi  = subClLo + ((clvl == 0) ? 0 : cmRail->Levels[clvl - 1].NumNodes - 1);
   
   if (subClHi == clHi)
      sprintf (str, "[%d-%d]", clLo, subClLo - 1);
   else if (subClLo == clLo)
      sprintf (str, "[%d-%d]", subClHi + 1, clHi);
   else
      sprintf (str, "[%d-%d][%d-%d]", clLo, subClLo - 1, subClHi + 1, clHi);

   return (str);
}

static void
RestartComms (CM_RAIL *cmRail, int clvl)
{
    int             base;
    int             nodeId;
    int             lstat;
    int             numClNodes;
    int             subClMin;
    int             subClMax;
    int             myClId;
    int             thisClId;
    
    myClId     = ClusterIds (cmRail, clvl, &subClMin, &subClMax);
    base       = myClId * CM_GSTATUS_BITS;
    numClNodes = cmRail->Levels[clvl].NumNodes;

    statemap_setbits (cmRail->Levels[clvl].LocalMap, base, 
		      CM_GSTATUS_CLOSING | CM_GSTATUS_MAY_START | CM_GSTATUS_RESTART, CM_GSTATUS_BITS);
    cmRail->Levels[clvl].Restarting = 1;

    if (cmRail->Levels[clvl].Online)
    {
	cmRail->Levels[clvl].Online = 0;
	
	for (thisClId = 0; thisClId < numClNodes; thisClId++)
	{
	    if (thisClId == subClMin)	/* skip sub-cluster; it's just someone in this cluster */
	    {				/* that wants me to restart */
		thisClId = subClMax;
		continue;
	    }
	    
	    nodeId = cmRail->Levels[clvl].MinNodeId + thisClId;
	    base   = thisClId * CM_GSTATUS_BITS;
	    lstat  = statemap_getbits (cmRail->Levels[clvl].LocalMap, base, CM_GSTATUS_BITS);
	    
	    if ((lstat & CM_GSTATUS_ACK_MASK) == CM_GSTATUS_MAY_RUN)
	    {
		switch (ep_disconnect_node (cmRail->Rail, nodeId))
		{
		case EP_NODE_CONNECTING:
		    /* gstat must == RUNNING */
		    cmRail->Levels[clvl].Connected--;
		    break;
		case EP_NODE_DISCONNECTED:
		    /* CLOSING || STARTING || (lstat & RESTART) */
		    break;
		}
	    }
	}
    }
}

static void
UpdateGlobalStatus (CM_RAIL *cmRail)
{
    char            clNodeStr[32];				/* [%d-%d][%d-%d] */
    int             nodeId;
    int             offset;
    int             base;
    bitmap_t        gstat;
    bitmap_t        lgstat;
    bitmap_t        lstat;
    int             clvl;
    int             numClNodes;
    int             subClMin;
    int             subClMax;
    int             myClId;
    int             thisClId;
    int             lastClId;

    for (clvl = 0; clvl < cmRail->NumLevels; clvl++)
    {
	if (!cmRail->Levels[clvl].GlobalMapValid || /* not got the global map yet */
	    !statemap_changed (cmRail->Levels[clvl].GlobalMap)) /* no changes to respond to */
	{
	    CPRINTF2 (6, "%s: Got invalid or unchanged clvl %d global map\n", cmRail->Rail->Name, clvl);
	    continue;
	}
	
	CPRINTF2 (5, "%s: Got valid changed clvl %d global map\n", cmRail->Rail->Name, clvl);
	
	lastClId = -1;
	myClId = ClusterIds (cmRail, clvl, &subClMin, &subClMax);
	numClNodes = cmRail->Levels[clvl].NumNodes;
	
	while ((offset = statemap_findchange (cmRail->Levels[clvl].GlobalMap, &gstat, 1)) >= 0)
	{
	    /*
	     * Check every node that this segment covers - however
	     * if the last node we checked in the previous segmemt
	     * is also the first node in this segment, then skip
	     * it.
	     */
	    if ((thisClId = (offset/CM_GSTATUS_BITS)) == lastClId)
		thisClId++;
	    lastClId = (offset + BT_NBIPUL - 1)/CM_GSTATUS_BITS;
	    
	    /* check each node that might have changed */
	    for ( ; thisClId <= lastClId && thisClId < numClNodes; thisClId++)
	    {
		base = thisClId * CM_GSTATUS_BITS;
		nodeId = cmRail->Levels[clvl].MinNodeId + thisClId;

		if (thisClId >= subClMin && thisClId <= subClMax) /* skip sub-cluster */
		    continue;

		/* This isn't me; I need to sense what this node is driving
		 * (just the starting and running bits) and respond
		 * appropriately...
		 */
		lgstat = statemap_getbits (cmRail->Levels[clvl].LastGlobalMap, base, CM_GSTATUS_BITS) & CM_GSTATUS_STATUS_MASK;
		gstat  = statemap_getbits (cmRail->Levels[clvl].GlobalMap,     base, CM_GSTATUS_BITS) & CM_GSTATUS_STATUS_MASK;

		if (lgstat == gstat)		/* no change in peer state */
		    continue;

		CPRINTF5 (3, "%s: Node %d: lgstat %s, gstat %s, lstat %s\n", cmRail->Rail->Name, nodeId,
			  GlobalStatusString (cmRail->Levels[clvl].LastGlobalMap, thisClId),
			  GlobalStatusString (cmRail->Levels[clvl].GlobalMap, thisClId),
			  GlobalStatusString (cmRail->Levels[clvl].LocalMap, thisClId));

		/* What I'm currently driving as my acknowledgement */
		lstat = statemap_getbits (cmRail->Levels[clvl].LocalMap, base, CM_GSTATUS_BITS);

		switch (gstat)
		{
		case CM_GSTATUS_STARTING:
		    if ((lgstat == CM_GSTATUS_ABSENT || lgstat == CM_GSTATUS_CLOSING) && lstat == CM_GSTATUS_MAY_START)
		    {
			CPRINTF2 (1, "%s: ===================node %d STARTING\n", cmRail->Rail->Name, nodeId);
			
			ASSERT (cmRail->Rail->Nodes[nodeId].State == EP_NODE_DISCONNECTED);

			statemap_setbits (cmRail->Levels[clvl].LocalMap, base, CM_GSTATUS_MAY_RUN, CM_GSTATUS_BITS);
			continue;
		    }
		    break;
		    
		case CM_GSTATUS_RUNNING:
		    if ((lgstat == CM_GSTATUS_ABSENT   && lstat == CM_GSTATUS_MAY_START) ||
			(lgstat == CM_GSTATUS_STARTING && lstat == CM_GSTATUS_MAY_RUN))
		    {
			CPRINTF3 (1, "%s: ===================node %d%s RUNNING\n", cmRail->Rail->Name, nodeId,
				  lgstat == CM_GSTATUS_ABSENT ? " Already" : "");

			ASSERT (cmRail->Rail->Nodes[nodeId].State == EP_NODE_DISCONNECTED);

			if (cmRail->Levels[clvl].Online)
			{
			    ep_connect_node (cmRail->Rail, nodeId);

			    cmRail->Levels[clvl].Connected++;
			}

			statemap_setbits (cmRail->Levels[clvl].LocalMap, base, CM_GSTATUS_MAY_RUN, CM_GSTATUS_BITS);
			continue;
		    }
		    break;
		    
		case CM_GSTATUS_CLOSING:
		    CPRINTF4 (1, "%s: ===================node %d CLOSING%s%s\n", cmRail->Rail->Name, nodeId,
			      (lstat & CM_GSTATUS_RESTART) ? " for Restart" : "",
			      cmRail->Levels[clvl].Online ? "" : " (offline)");

		    if ((lstat & CM_GSTATUS_ACK_MASK) == CM_GSTATUS_MAY_RUN)
		    {
			switch (ep_disconnect_node (cmRail->Rail, nodeId))
			{
			case EP_NODE_CONNECTING:
			    cmRail->Levels[clvl].Connected--;
			    /* DROPTHROUGH */
			case EP_NODE_DISCONNECTED:
			    lstat = CM_GSTATUS_MAY_START;
			    break;
			}
		    }

		    if ((lstat & CM_GSTATUS_ACK_MASK) == CM_GSTATUS_MAY_START) /* clear restart if we've disconnected */
			statemap_setbits (cmRail->Levels[clvl].LocalMap, base, CM_GSTATUS_MAY_START, CM_GSTATUS_BITS);
		    continue;
		    
		default:
		    break;
		}

		/* "unexpected" state change forces me to ask her to restart */
		if (! (lstat & CM_GSTATUS_RESTART))		/* not requesting restart already */
		{
		    CPRINTF5 (1, "%s: ===================node %d %s, old %s new %s\n", cmRail->Rail->Name, nodeId,
			      (gstat == CM_GSTATUS_ABSENT)  ? "ABSENT" : "REQUEST RESTART", 
			      GlobalStatusString (cmRail->Levels[clvl].LastGlobalMap, thisClId),
			      GlobalStatusString (cmRail->Levels[clvl].GlobalMap, thisClId));

		    /* request restart */
		    if (cmRail->Levels[clvl].Online && lstat == CM_GSTATUS_MAY_RUN)
		    {
			switch (ep_disconnect_node (cmRail->Rail, nodeId))
			{
			case EP_NODE_CONNECTING:
			    cmRail->Levels[clvl].Connected--;
			    /* DROPTHROUGH */
			case EP_NODE_DISCONNECTED:
			    lstat = CM_GSTATUS_MAY_START;
			    break;
			}
		    }

		    statemap_setbits (cmRail->Levels[clvl].LocalMap, base, lstat | CM_GSTATUS_RESTART, CM_GSTATUS_BITS);
		    continue;
		}

		continue;
	    }
	}
    
	/* Now check myself - see what everyone else thinks I'm doing */
	base   = myClId * CM_GSTATUS_BITS;
	lstat  = statemap_getbits (cmRail->Levels[clvl].LocalMap,  base, CM_GSTATUS_BITS);
	gstat  = statemap_getbits (cmRail->Levels[clvl].GlobalMap, base, CM_GSTATUS_BITS);
	lgstat = statemap_getbits (cmRail->Levels[clvl].LastGlobalMap, base, CM_GSTATUS_BITS);

	if (lgstat == gstat)			/* my state in this cluster hasn't changed */
	{
	    CPRINTF3 (6, "%s: my clvl %d global status unchanged from %s\n", cmRail->Rail->Name,
		      clvl, GlobalStatusString (cmRail->Levels[clvl].GlobalMap, myClId));
	    goto all_done;
	}

	if ((gstat & CM_GSTATUS_RESTART) != 0)	/* someone wants me to restart */
	{
	    if ((lstat & CM_GSTATUS_STATUS_MASK) == CM_GSTATUS_CLOSING) /* I'm already restarting */
		goto all_done;
	    
	    CPRINTF2 (1, "%s: ===================RESTART REQUEST from %s\n", cmRail->Rail->Name,
		      sprintClPeers (clNodeStr, cmRail, clvl));
	    
	    printk ("%s: Restart Request from %s\n", cmRail->Rail->Name,
		    sprintClPeers (clNodeStr, cmRail, clvl));
	    
	    RestartComms (cmRail, clvl);
	    goto all_done;
	}
	
	CPRINTF6 (5, "%s: clvl %d: lgstat %s gstat %s, lstat %s%s\n", cmRail->Rail->Name, clvl,
		  GlobalStatusString (cmRail->Levels[clvl].LastGlobalMap, myClId),
		  GlobalStatusString (cmRail->Levels[clvl].GlobalMap, myClId),
		  GlobalStatusString (cmRail->Levels[clvl].LocalMap, myClId),
		  (gstat != lstat) ? " (IGNORED)" : "");
			
	if (gstat != lstat)			/* not everyone agrees with me */
	    goto all_done;

	switch (lstat)
	{
	default:
	    ASSERT (0);				/* I never drive this */
	    
	case CM_GSTATUS_CLOSING | CM_GSTATUS_MAY_START: /* I can restart now (have seen restart go away) */
	    ASSERT (!cmRail->Levels[clvl].Online);
	    
	    CPRINTF2 (1,"%s: ===================NODES %s AGREE I MAY START\n", cmRail->Rail->Name,
		      sprintClPeers (clNodeStr, cmRail, clvl));
	    printk ("%s: ===================NODES %s AGREE I MAY START\n", cmRail->Rail->Name,
		    sprintClPeers (clNodeStr, cmRail, clvl));
	    
	    statemap_setbits (cmRail->Levels[clvl].LocalMap, base, 
			      CM_GSTATUS_STARTING | CM_GSTATUS_MAY_RUN, CM_GSTATUS_BITS);
	    goto all_done;
	    
	case CM_GSTATUS_STARTING | CM_GSTATUS_MAY_RUN:
	    ASSERT (!cmRail->Levels[clvl].Online);
	    
	    CPRINTF2 (1, "%s: ===================NODES %s AGREE I MAY RUN\n", cmRail->Rail->Name,
		      sprintClPeers (clNodeStr, cmRail, clvl));
	    printk ("%s: ===================NODES %s AGREE I MAY RUN\n", cmRail->Rail->Name,
		    sprintClPeers (clNodeStr, cmRail, clvl));
	    
	    statemap_setbits (cmRail->Levels[clvl].LocalMap, base, 
			      CM_GSTATUS_RUNNING | CM_GSTATUS_MAY_RUN, CM_GSTATUS_BITS);
	    goto all_done;
	    
	case CM_GSTATUS_RUNNING | CM_GSTATUS_MAY_RUN:
	    if (! cmRail->Levels[clvl].Online)
	    {
		CPRINTF2 (1, "%s: ===================NODES %s AGREE I'M RUNNING\n", cmRail->Rail->Name,
			  sprintClPeers (clNodeStr, cmRail, clvl));
		printk ("%s: ===================NODES %s AGREE I'M RUNNING\n", cmRail->Rail->Name,
			sprintClPeers (clNodeStr, cmRail, clvl));
		
		cmRail->Levels[clvl].Online = 1;
		
		for (thisClId = 0; thisClId < numClNodes; thisClId++)
		{
		    if (thisClId == subClMin)	/* skip sub-cluster */
		    {
			thisClId = subClMax;
			continue;
		    }
		    
		    nodeId = cmRail->Levels[clvl].MinNodeId + thisClId;
		    
		    base  = thisClId * CM_GSTATUS_BITS;
		    lstat = statemap_getbits (cmRail->Levels[clvl].LocalMap, base, CM_GSTATUS_BITS);
		    gstat = statemap_getbits (cmRail->Levels[clvl].GlobalMap, base, CM_GSTATUS_BITS) & CM_GSTATUS_STATUS_MASK;
		    
		    /* Only connect to her if I see her as running and I'm not requesting her 
		     * to restart - this means that I was offline when I saw her transition
		     * to running and haven't seen her in a "bad" state since. */
		    if (gstat == CM_GSTATUS_RUNNING && ! (lstat & CM_GSTATUS_RESTART))
		    {
			CPRINTF5 (1, "%s: node %d lgstat %s gstat %s, lstat %s -> CONNECT\n", cmRail->Rail->Name, nodeId,
				  GlobalStatusString (cmRail->Levels[clvl].LastGlobalMap, thisClId),
				  GlobalStatusString (cmRail->Levels[clvl].GlobalMap, thisClId),
				  GlobalStatusString (cmRail->Levels[clvl].LocalMap, thisClId));
			
			if (lstat == CM_GSTATUS_MAY_START)
			    statemap_setbits (cmRail->Levels[clvl].LocalMap, base, CM_GSTATUS_MAY_RUN, CM_GSTATUS_BITS);

			ep_connect_node (cmRail->Rail, nodeId);

			cmRail->Levels[clvl].Connected++;
		    }
		}
	    }
	    goto all_done;
	}

    all_done:
	statemap_setmap (cmRail->Levels[clvl].LastGlobalMap, cmRail->Levels[clvl].GlobalMap);
    }
}

static void
ReduceGlobalMap (CM_RAIL *cmRail, int clvl)
{
    int       lvl;
    int       sidx;
    int       recompute;
    CM_LEVEL *level;
    int       cTopLevel;
    int       cRole;

    if (clvl < cmRail->TopLevel)
    {
	cTopLevel = clvl + 1;
	cRole = CM_ROLE_LEADER;
    }
    else
    {
	cTopLevel = cmRail->TopLevel;
	cRole = cmRail->Role;
    }
    
    /* Update cmRail->Levels[*].SubordinateMap[clvl] for all subordinate levels */
    for (lvl = 0; lvl < cTopLevel; lvl++)
    {
	level = &cmRail->Levels[lvl];

	/* We need to recompute this level's statemap if...
	 * . Previous level's statemap has changes to propagate OR
	 * . This level's statemap has not been computed yet OR
	 * . A subordinate at this level has sent me a change.
	 * Note that we can only do this if all subordinates from this
	 * level down are present with valid statemaps, or absent (i.e. not
	 * timing out).
	 */

	ASSERT (lvl == 0 || cmRail->Levels[lvl - 1].SubordinateMapValid[clvl]);

	recompute = !level->SubordinateMapValid[clvl] ||
		    (lvl > 0 && statemap_changed (cmRail->Levels[lvl - 1].SubordinateMap[clvl]));
	  
	for (sidx = 0; sidx < level->NumSegs; sidx++)
	{
	    CM_SGMT *sgmt = &level->Sgmts[sidx];

	    if (!(sgmt->State == CM_SGMT_ABSENT || /* absent nodes contribute zeros */
		  (sgmt->State == CM_SGMT_PRESENT && /* present nodes MUST have received a map to contribute */
		   sgmt->Maps[clvl].InputMapValid)))
	    {
		CPRINTF5 (5, "%s: waiting for clvl %d lvl %d seg %d node %d\n", cmRail->Rail->Name,
			  clvl, lvl, sidx, sgmt->NodeId);

		/* Gotta wait for this guy, so we can't compute this level,
		 * or any higher levels. */
		return;
	    }

	    if (statemap_changed (sgmt->Maps[clvl].InputMap))
	    {
		ASSERT (sgmt->Maps[clvl].InputMapValid);

		recompute = 1;

		CPRINTF7 (5, "%s: %s clvl %d map from @ %d %d (%d) - %s\n",
			  cmRail->Rail->Name, sgmt->State == CM_SGMT_ABSENT ? "newly absent" : "got new",
			  clvl, lvl, sidx, sgmt->NodeId, 
			  MapString ("Input", sgmt->Maps[clvl].InputMap, cmRail->Levels[clvl].NumNodes, ""));
	    }
	}

	if (recompute)
	{
	    if (lvl == 0)
		statemap_reset (cmRail->Levels[clvl].TmpMap);
	    else
	    {
		ASSERT (cmRail->Levels[lvl - 1].SubordinateMapValid[clvl]);
		
		statemap_copy (cmRail->Levels[clvl].TmpMap, cmRail->Levels[lvl - 1].SubordinateMap[clvl]);
		statemap_clearchanges (cmRail->Levels[lvl - 1].SubordinateMap[clvl]);
	    }
	 
	    for (sidx = 0; sidx < level->NumSegs; sidx++)
	    {
		CM_SGMT *sgmt = &level->Sgmts[sidx];
		
		if (sgmt->State != CM_SGMT_ABSENT)	/* absent nodes contribute zeroes */
		{
		    ASSERT (sgmt->State == CM_SGMT_PRESENT);
		    ASSERT (sgmt->Maps[clvl].InputMapValid);
		    statemap_ormap (cmRail->Levels[clvl].TmpMap, sgmt->Maps[clvl].InputMap);
		}
		statemap_clearchanges (sgmt->Maps[clvl].InputMap);
	    }

	    statemap_setmap (level->SubordinateMap[clvl], cmRail->Levels[clvl].TmpMap);
	    level->SubordinateMapValid[clvl] = 1;

	    CPRINTF4 (5, "%s: recompute clvl %d level %d statemap - %s\n", cmRail->Rail->Name, clvl, lvl,
		      MapString ("level", level->SubordinateMap[clvl], cmRail->Levels[clvl].NumNodes, ""));
	}
    }

    if (cRole == CM_ROLE_LEADER_CANDIDATE)	/* don't know this cluster's leader yet */
	return;

    ASSERT (cTopLevel == 0 || cmRail->Levels[cTopLevel - 1].SubordinateMapValid[clvl]);

    /* Update SubTreeMap */
    
    if (!cmRail->Levels[clvl].SubTreeMapValid || 
	statemap_changed (cmRail->Levels[clvl].LocalMap) ||
	(cTopLevel > 0 && statemap_changed (cmRail->Levels[cTopLevel - 1].SubordinateMap[clvl])))
    {
	statemap_copy (cmRail->Levels[clvl].TmpMap, cmRail->Levels[clvl].LocalMap);
	statemap_clearchanges (cmRail->Levels[clvl].LocalMap);

	if (cTopLevel > 0)
	{
	    statemap_ormap (cmRail->Levels[clvl].TmpMap, cmRail->Levels[cTopLevel - 1].SubordinateMap[clvl]);
	    statemap_clearchanges (cmRail->Levels[cTopLevel - 1].SubordinateMap[clvl]);
	}

	statemap_setmap (cmRail->Levels[clvl].SubTreeMap, cmRail->Levels[clvl].TmpMap);
	cmRail->Levels[clvl].SubTreeMapValid = 1;

	CPRINTF3 (5, "%s: recompute clvl %d subtree map - %s\n", cmRail->Rail->Name, clvl,
		  MapString ("subtree", cmRail->Levels[clvl].SubTreeMap, cmRail->Levels[clvl].NumNodes, ""));
    }

    if (cRole == CM_ROLE_SUBORDINATE)		/* got a leader (Not me) */
    {						/* => send SubTreeMap to her */
	CM_SGMT *leader = &cmRail->Levels[cmRail->TopLevel].Sgmts[0];

	ASSERT (leader->State == CM_SGMT_PRESENT);
	ASSERT (cmRail->Levels[clvl].SubTreeMapValid);

	if (!leader->Maps[clvl].OutputMapValid ||
	    statemap_changed (cmRail->Levels[clvl].SubTreeMap))
	{
	    statemap_setmap (leader->Maps[clvl].OutputMap, cmRail->Levels[clvl].SubTreeMap);
	    leader->Maps[clvl].OutputMapValid = 1;

	    statemap_clearchanges (cmRail->Levels[clvl].SubTreeMap);

	    CPRINTF3 (5, "%s: sending clvl %d subtree map to leader (%d)\n", cmRail->Rail->Name, clvl, leader->NodeId);
	}
    }
}

void
BroadcastGlobalMap (CM_RAIL *cmRail, int clvl)
{
    int       lvl;
    int       sidx;
    CM_LEVEL *level;
    CM_SGMT  *leader;
    int       cTopLevel;
    int       cRole;

    if (clvl < cmRail->TopLevel)
    {
	cTopLevel = clvl + 1;
	cRole = CM_ROLE_LEADER;
    }
    else
    {
	cTopLevel = cmRail->TopLevel;
	cRole = cmRail->Role;
    }

    switch (cRole)
    {
    default:
	ASSERT (0);
	
    case CM_ROLE_LEADER_CANDIDATE:		/* don't know this cluster's leader yet */
	return;

    case CM_ROLE_LEADER:			/* cluster leader: */
	ASSERT (clvl < cmRail->TopLevel);		/* set GlobalMap from SubTreeMap */
	
	if (!cmRail->Levels[clvl].SubTreeMapValid)	/* can't set global map */
	    return;

	if (cmRail->Levels[clvl].GlobalMapValid &&	/* already set global map */
	    !statemap_changed (cmRail->Levels[clvl].SubTreeMap)) /* no changes to propagate */
	    return;
	
	statemap_setmap (cmRail->Levels[clvl].GlobalMap, cmRail->Levels[clvl].SubTreeMap);
	cmRail->Levels[clvl].GlobalMapValid = 1;
	statemap_clearchanges (cmRail->Levels[clvl].SubTreeMap);

	CPRINTF2 (5, "%s: whole cluster %d leader setting global map\n", cmRail->Rail->Name, clvl);

	UpdateGlobalStatus (cmRail);
	break;
	
    case CM_ROLE_SUBORDINATE:			/* cluster subordinate: */
	ASSERT (clvl >= cmRail->TopLevel);		/* receive GlobalMap from leader */
	ASSERT (cmRail->TopLevel < cmRail->NumLevels);
	
	leader = &cmRail->Levels[cmRail->TopLevel].Sgmts[0];
	ASSERT (leader->State == CM_SGMT_PRESENT);

	if (!leader->Maps[clvl].InputMapValid)	/* can't set global map */
	    return;
	
	if (cmRail->Levels[clvl].GlobalMapValid &&	/* already set global map */
	    !statemap_changed (leader->Maps[clvl].InputMap)) /* no changes to propagate */
	    return;

	statemap_setmap (cmRail->Levels[clvl].GlobalMap, leader->Maps[clvl].InputMap);
	cmRail->Levels[clvl].GlobalMapValid = 1;
	statemap_clearchanges (leader->Maps[clvl].InputMap);

	CPRINTF3 (5, "%s: getting clvl %d global map from leader (%d)\n", cmRail->Rail->Name, clvl, leader->NodeId);

	UpdateGlobalStatus (cmRail);
	break;
    }

    CPRINTF3 (5, "%s: clvl %d %s\n", cmRail->Rail->Name, clvl,
	      MapString ("global", cmRail->Levels[clvl].GlobalMap, cmRail->Levels[clvl].NumNodes, ""));
    
    /* Broadcast global map to all subordinates */
    for (lvl = 0; lvl < cTopLevel; lvl++)
    {
	level = &cmRail->Levels[lvl];
	
	for (sidx = 0; sidx < level->NumSegs; sidx++)
	{
	    CM_SGMT *sgmt = &level->Sgmts[sidx];
	    
	    if (sgmt->State == CM_SGMT_PRESENT)
	    {
		statemap_setmap (sgmt->Maps[clvl].OutputMap, cmRail->Levels[clvl].GlobalMap);
		sgmt->Maps[clvl].OutputMapValid = 1;
		
		CPRINTF5 (5, "%s: sending clvl %d global map to subordinate %d %d (%d)\n", 
			  cmRail->Rail->Name, clvl, lvl, sidx, sgmt->NodeId);
	    }
	}
    }
}

static void
CheckPeerPulse (CM_RAIL *cmRail, CM_SGMT *sgmt)
{
    int clvl, sendRejoin;
    
    switch (sgmt->State)
    {
    case CM_SGMT_ABSENT:
	break;

    case CM_SGMT_WAITING:			/* waiting for a subtree */
	if (!AFTER (lbolt, sgmt->UpdateTick + MSEC2TICKS(CM_DISCOVER_TIMEOUT)))
	    break;
      
	CPRINTF3 (2, "%s: lvl %d subtree %d contains no live nodes\n", cmRail->Rail->Name, 
		  sgmt->Level, (int) (sgmt - &cmRail->Levels[sgmt->Level].Sgmts[0]));

	sgmt->State = CM_SGMT_ABSENT;
	for (clvl = sgmt->Level; clvl < cmRail->NumLevels; clvl++)
	{
	    statemap_zero (sgmt->Maps[clvl].InputMap); 		/* need to start propagating zeros (flags change) */
	    sgmt->Maps[clvl].InputMapValid = 1;		/* and must indicate that the map is now valid */
	}
	break;

    case CM_SGMT_COMING:				/* lost/waiting subtree sent me IMCOMING */
	ASSERT (sgmt->Level > 0);			/* we only do subtree discovery below our own level */

	if (AFTER (lbolt, sgmt->WaitingTick + MSEC2TICKS(CM_WAITING_TIMEOUT)))
	{
	    CPRINTF3 (1, "%s: lvl %d subtree %d waiting too long\n", cmRail->Rail->Name,
		      sgmt->Level, (int) (sgmt - &cmRail->Levels[sgmt->Level].Sgmts[0]));
	    printk ("%s: lvl %d subtree %d waiting too long\n", cmRail->Rail->Name,
		    sgmt->Level, (int) (sgmt - &cmRail->Levels[sgmt->Level].Sgmts[0]));

	    sgmt->State = CM_SGMT_ABSENT;
	    for (clvl = sgmt->Level; clvl < cmRail->NumLevels; clvl++)
	    {
		statemap_zero (sgmt->Maps[clvl].InputMap); 		/* need to start propagating zeros (flags change) */
		sgmt->Maps[clvl].InputMapValid = 1;		/* and must indicate that the map is now valid */
	    }
	    break;
	}

	if (!AFTER (lbolt, sgmt->UpdateTick + MSEC2TICKS(CM_DISCOVER_TIMEOUT)))
	    break;

	CPRINTF3 (2, "%s: lvl %d subtree %d hasn't connected yet\n", cmRail->Rail->Name,
		  sgmt->Level, (int) (sgmt - &cmRail->Levels[sgmt->Level].Sgmts[0]));

	sgmt->State = CM_SGMT_WAITING;
	sgmt->UpdateTick = lbolt;

	if (sgmt->Level > 0)
	    __Schedule_Discovery (cmRail);
	break;
      
    case CM_SGMT_PRESENT:
	if (!AFTER (lbolt, sgmt->UpdateTick + MSEC2TICKS(CM_HEARTBEAT_TIMEOUT)))
	    break;

	if (sgmt->Level == cmRail->TopLevel)		/* leader died */
	{
	    sendRejoin = (sgmt->State == CM_SGMT_PRESENT && sgmt->AckSeq == 0);

	    CPRINTF4 (1, "%s: leader (%d) node %d JUST DIED%s\n", 
		      cmRail->Rail->Name, sgmt->Level, sgmt->NodeId,
		      sendRejoin ? ": !REJOIN" : "");
	    
	    printk ("%s: lvl %d leader (%d) JUST DIED%s\n", 
		    cmRail->Rail->Name, sgmt->Level, sgmt->NodeId,
		    sendRejoin ? ": !REJOIN" : "");
	
	    if (sendRejoin)
	    {
		/* she's not sent us any heartbeats even though she responded to a discover
		 * so tell her to rejoin the tree at the bottom, this will mean that she 
		 * has to run the heartbeat timer before being able to rejoin the tree. */
		SendMessage (cmRail, sgmt->NodeId, sgmt->Level, CM_MSG_TYPE_REJOIN);
	    }

	    StartLeaderDiscovery (cmRail);
	    break;
	}

	sendRejoin = (sgmt->State == CM_SGMT_PRESENT && sgmt->AckSeq == 0);

	CPRINTF5 (2, "%s: lvl %d subordinate %d (%d) JUST DIED%s\n", cmRail->Rail->Name, 
		  sgmt->Level, (int) (sgmt - &cmRail->Levels[sgmt->Level].Sgmts[0]), sgmt->NodeId,
		  sendRejoin ? ": !REJOIN" : "");
	printk ("%s: lvl %d subordinate %d (%d) JUST DIED%s\n", cmRail->Rail->Name, 
		sgmt->Level, (int) (sgmt - &cmRail->Levels[sgmt->Level].Sgmts[0]), sgmt->NodeId,
		sendRejoin ? ": !REJOIN" : "");

	if (sendRejoin)
	{
	    /* she's not sent us any heartbeats even though she responded to a discover
	     * so tell her to rejoin the tree at the bottom, this will mean that she 
	     * has to run the heartbeat timer before being able to rejoin the tree. */
	    SendMessage (cmRail, sgmt->NodeId, sgmt->Level, CM_MSG_TYPE_REJOIN);
	}

	StartSubTreeDiscovery (cmRail, sgmt);
	break;
	 
    default:
	ASSERT (0);
    }
}

static void
CheckPeerPulses (CM_RAIL *cmRail)
{
    int lvl;
    int sidx;
   
    /* check children are alive */
    for (lvl = 0; lvl < cmRail->TopLevel; lvl++)
	for (sidx = 0; sidx < cmRail->Levels[lvl].NumSegs; sidx++)
	    CheckPeerPulse (cmRail, &cmRail->Levels[lvl].Sgmts[sidx]);

    /* check leader is alive */
    if (cmRail->Role == CM_ROLE_SUBORDINATE)
    {
	ASSERT (cmRail->TopLevel < cmRail->NumLevels);
	ASSERT (cmRail->Levels[cmRail->TopLevel].Sgmts[0].State == CM_SGMT_PRESENT);
      
	CheckPeerPulse (cmRail, &cmRail->Levels[cmRail->TopLevel].Sgmts[0]);
    }
}

static void
SendHeartbeats (CM_RAIL *cmRail)
{
    int lvl;

    /* Send heartbeats to my children */
    for (lvl = 0; lvl < cmRail->TopLevel; lvl++)
    {
	CM_LEVEL *level = &cmRail->Levels[lvl];
	int       sidx;
       
	for (sidx = 0; sidx < level->NumSegs; sidx++)
	{
	    CM_SGMT *sgmt = &cmRail->Levels[lvl].Sgmts[sidx];

	    if (sgmt->State == CM_SGMT_PRESENT)
		SendToSgmt (cmRail, sgmt, CM_MSG_TYPE_HEARTBEAT);
	}
    }

    /* Send heartbeat to my leader */
    if (cmRail->Role == CM_ROLE_SUBORDINATE)
    {
	ASSERT (cmRail->TopLevel < cmRail->NumLevels);
	SendToSgmt (cmRail, &cmRail->Levels[cmRail->TopLevel].Sgmts[0], CM_MSG_TYPE_HEARTBEAT);
    }
}

static int
BroadcastDiscover (CM_RAIL *cmRail)
{
    int       sidx;
    int	      lvl;
    int       msgType;
    CM_LEVEL *level;
    int       urgent;

    ASSERT (cmRail->TopLevel <= cmRail->NumLevels);
    ASSERT ((cmRail->Role == CM_ROLE_LEADER) ? (cmRail->TopLevel == cmRail->NumLevels) :
	    (cmRail->Role == CM_ROLE_SUBORDINATE) ? (cmRail->Levels[cmRail->TopLevel].Sgmts[0].State == CM_SGMT_PRESENT) :
	    (cmRail->Role == CM_ROLE_LEADER_CANDIDATE));

    if (cmRail->Role != CM_ROLE_LEADER_CANDIDATE)	/* got a leader/lead whole machine */
    {
	urgent = 0;				/* non-urgent leader discovery */
	lvl = cmRail->TopLevel - 1;		/* on nodes I lead (resolves leader conflicts) */
	msgType = CM_MSG_TYPE_RESOLVE_LEADER;
    }
    else
    {
	urgent = 1;				/* urgent leader discovery */
	lvl = cmRail->TopLevel;			/* on nodes I'd like to lead */
	msgType = CM_MSG_TYPE_DISCOVER_LEADER;
    }

    if (lvl >= 0)
    {
	if (lvl > cmRail->BroadcastLevel)
	{
	    /* Unable to broadcast at this level in the spanning tree, so we 
	     * just continue doing discovery until we are able to broadcast */
	    CPRINTF4 (6, "%s: broadcast level %d too low to discover %d at level %d\n",
		      cmRail->Rail->Name, cmRail->BroadcastLevel, msgType, lvl);

	    cmRail->DiscoverStartTick = lbolt;
	}
	else
	{
	    level = &cmRail->Levels[lvl];
	    SendToSgmt (cmRail, &level->Sgmts[level->MySgmt], msgType);
	}
    }
    
    while (lvl > 0)
    {
	level = &cmRail->Levels[lvl];
      
	for (sidx = 0; sidx < level->NumSegs; sidx++)
	{
	    CM_SGMT *sgmt = &level->Sgmts[sidx];
	 
	    if (sgmt->State == CM_SGMT_WAITING)
	    {
		ASSERT (sidx != level->MySgmt);
		/* Do subordinate discovery.  Existing subordinates will
		 * ignore it, but leader candidates will send IMCOMING.
		 * This is always urgent since we'll assume a subtree is
		 * absent if I don't get IMCOMING within the timeout.
		 */
		SendToSgmt (cmRail, sgmt, CM_MSG_TYPE_DISCOVER_SUBORDINATE);
		urgent = 1;
	    }
	}
	lvl--;
    }
   
    return (urgent);
}

static void
CheckBroadcast (CM_RAIL *cmRail)
{
    int  clvl;

    for (clvl = cmRail->NumLevels-1; clvl >= 0 && cmRail->Rail->SwitchBroadcastLevel < cmRail->Levels[clvl].SwitchLevel; clvl--)
	;

    if (cmRail->OfflineReasons || cmRail->Rail->System->Shutdown)
	clvl = -1;

    /* if the level at which we can broadcast drops, then we must rejoin the
     * spanning tree at the highest level for which broadcast is good. */
    if (cmRail->BroadcastLevel > clvl && clvl < (int)(cmRail->Role == CM_ROLE_LEADER ? cmRail->TopLevel - 1 : cmRail->TopLevel))
    {
	printk ("%s: REJOINING at level %d because %s\n", cmRail->Rail->Name, clvl+1, 
		(cmRail->OfflineReasons & CM_OFFLINE_MANAGER) ? "of manager thread" :
		(cmRail->OfflineReasons & CM_OFFLINE_PROCFS)  ? "force offline"  : 
		cmRail->Rail->System->Shutdown ? "system shutdown" : "broadcast level changed");
	LowerTopLevel (cmRail, clvl+1);
    }
    
    if (cmRail->BroadcastLevel != clvl)
    {
	cmRail->BroadcastLevel     = clvl;
	cmRail->BroadcastLevelTick = lbolt;
    }

    /* schedule the update thread, to withdraw from comms with 
     * nodes "outside" of the valid broadcastable range. */
    for (clvl = 0; clvl < cmRail->NumLevels; clvl++)
    {
	if (cmRail->BroadcastLevel < clvl)
	{
	    if (AFTER (lbolt, cmRail->BroadcastLevelTick + EP_WITHDRAW_TIMEOUT) && 
		!(cmRail->Levels[clvl].OfflineReasons & CM_OFFLINE_BROADCAST))
	    {
		printk ("%s: Withdraw at Level %d\n", cmRail->Rail->Name, clvl);
		cmRail->Levels[clvl].OfflineReasons |= CM_OFFLINE_BROADCAST;
	    }
	}
	else
	{
	    if (cmRail->Levels[clvl].OfflineReasons & CM_OFFLINE_BROADCAST)
	    {
		printk ("%s: Rejoin at Level %d\n", cmRail->Rail->Name, clvl);
		cmRail->Levels[clvl].OfflineReasons &= ~CM_OFFLINE_BROADCAST;
	    }
	}
    }
	
}

static void
CheckManager (CM_RAIL *cmRail)
{
    long time,  state = ep_kthread_state (&cmRail->Rail->System->ManagerThread, &time);

    if (state == KT_STATE_RUNNING && BEFORE (lbolt, time + MSEC2TICKS(CM_THREAD_RUNNING_TIMEOUT)))
	state = KT_STATE_SLEEPING;
    if (state != KT_STATE_SLEEPING && BEFORE (lbolt, time + MSEC2TICKS(CM_THREAD_SCHEDULE_TIMEOUT)))
	state = KT_STATE_SLEEPING;

    if ((cmRail->OfflineReasons & CM_OFFLINE_MANAGER) && state == KT_STATE_SLEEPING)
    {
	printk ("%s: manager thread unstuck\n", cmRail->Rail->Name);

	cmRail->OfflineReasons &= ~CM_OFFLINE_MANAGER;
    }

    if (!(cmRail->OfflineReasons & CM_OFFLINE_MANAGER) && state != KT_STATE_SLEEPING)
    {
	printk ("%s: manager thread stuck - %s\n", cmRail->Rail->Name,
		state == KT_STATE_SCHEDULED ? "scheduled" : 
		state == KT_STATE_RUNNING ? "running" : 
		state == KT_STATE_STALLED ? "stalled" : "unknown");

	cmRail->OfflineReasons |= CM_OFFLINE_MANAGER;
    }
}

static void
CheckOfflineReasons (CM_RAIL *cmRail, int clvl)
{
    int subClMin, subClMax, myClId;
    char clNodeStr[32];				/* [%d-%d][%d-%d] */

    if (cmRail->Levels[clvl].OfflineReasons)
    {
	if (cmRail->Levels[clvl].Online)
	{
	    printk ("%s: Withdraw from %s\n", cmRail->Rail->Name, sprintClPeers (clNodeStr, cmRail, clvl));
	    
	    RestartComms (cmRail, clvl);
	}
    }
    else
    {
	if (cmRail->Levels[clvl].Restarting && cmRail->Levels[clvl].Connected == 0)
	{
	    printk ("%s: Rejoin with %s\n", cmRail->Rail->Name, sprintClPeers (clNodeStr, cmRail, clvl));

	    myClId = ClusterIds (cmRail, clvl, &subClMin, &subClMax);
	    
	    ASSERT (statemap_getbits (cmRail->Levels[clvl].LocalMap, myClId * CM_GSTATUS_BITS, CM_GSTATUS_BITS) == 
		    (CM_GSTATUS_CLOSING | CM_GSTATUS_MAY_START | CM_GSTATUS_RESTART));
    
	    statemap_setbits (cmRail->Levels[clvl].LocalMap, myClId * CM_GSTATUS_BITS,
			      CM_GSTATUS_CLOSING | CM_GSTATUS_MAY_START, CM_GSTATUS_BITS);

	    cmRail->Levels[clvl].Restarting = 0;
	}
    }
}

void
DoHeartbeatWork (CM_RAIL *cmRail)
{
    long now = lbolt;
    int  clvl;

    if ((RejoinCheck || RejoinPanic) &&
	AFTER (now, cmRail->NextRunTime + MSEC2TICKS (CM_TIMER_SCHEDULE_TIMEOUT))) /* If I've been unresponsive for too long */
    {
	/* I'd better reconnect to the network because I've not been playing the game */
	CPRINTF4 (1, "%s: REJOINING because I was too slow (heartbeat) [%ld,%ld,(%ld)]\n", cmRail->Rail->Name, now,  cmRail->NextRunTime, (long int)MSEC2TICKS (CM_TIMER_SCHEDULE_TIMEOUT));
	printk ("%s: REJOINING because I was too slow (heartbeat) [%ld,%ld,(%ld)]\n", cmRail->Rail->Name, now,  cmRail->NextRunTime, (long int)MSEC2TICKS (CM_TIMER_SCHEDULE_TIMEOUT));
	
	LowerTopLevel (cmRail, 0);
	
	IncrStat (cmRail, RejoinTooSlow);
	
	if (RejoinPanic)
	    panic ("ep: REJOINING because I was too slow (heartbeat)\n");
    }
    
    PollInputQueues (cmRail);
    
    if (! BEFORE (now, cmRail->NextDiscoverTime))
    {
	if (BroadcastDiscover (cmRail))		/* urgent discovery required? */
	    cmRail->NextDiscoverTime = now + MSEC2TICKS (CM_URGENT_DISCOVER_INTERVAL);
	else
	    cmRail->NextDiscoverTime = now + MSEC2TICKS (CM_PERIODIC_DISCOVER_INTERVAL);
	
	if (cmRail->Role == CM_ROLE_LEADER_CANDIDATE && AFTER (now, cmRail->DiscoverStartTick + MSEC2TICKS (CM_DISCOVER_TIMEOUT)))
	    RaiseTopLevel (cmRail);
    }
    
    if (! BEFORE (now, cmRail->NextHeartbeatTime))
    {
	CheckPosition (cmRail->Rail);
	CheckPeerPulses (cmRail);
	CheckBroadcast (cmRail);
	CheckManager (cmRail);
	
	for (clvl = 0; clvl < cmRail->NumLevels; clvl++)
	{
	    CheckOfflineReasons (cmRail, clvl);
	    ReduceGlobalMap (cmRail, clvl);
	    BroadcastGlobalMap (cmRail, clvl);
	}
	
	SendHeartbeats (cmRail);
	
	/* Compute the next heartbeat time, but "drift" it towards the last
	 * periodic discovery time we saw from the whole machine leader */
	cmRail->NextHeartbeatTime = now + MSEC2TICKS (CM_HEARTBEAT_INTERVAL);
    }

    if (AFTER (cmRail->NextHeartbeatTime, cmRail->NextDiscoverTime))
	cmRail->NextRunTime = cmRail->NextDiscoverTime;
    else 
	cmRail->NextRunTime = cmRail->NextHeartbeatTime;
}

#define CM_SVC_INDICATOR_OFFSET(CMRAIL,CLVL,IND,NODEID)     ( ( CMRAIL->Levels[CLVL].NumNodes * CM_GSTATUS_BITS ) \
								+ ( CMRAIL->Levels[CLVL].NumNodes * IND ) \
								+ ( NODEID - CMRAIL->Levels[CLVL].MinNodeId ) )
int
cm_svc_indicator_set (EP_RAIL *rail, int svc_indicator)
{
    CM_RAIL *cmRail = rail->ClusterRail;
    unsigned long flags;
    int           clvl;

    EPRINTF2 (DBG_SVC,"cm_svc_indicator_set: rail %p ind %d\n", rail, svc_indicator);

    if (svc_indicator < 0 || svc_indicator >= EP_SVC_NUM_INDICATORS)
    {
	EPRINTF1 (DBG_SVC,"cm_svc_indicator_set: service indicator %d not registered\n", svc_indicator);
	return (-1);
    }

    if (rail->State == EP_RAIL_STATE_UNINITIALISED) 
	return (-2);
    
    spin_lock_irqsave (&cmRail->Lock, flags);
    for (clvl = 0; clvl < cmRail->NumLevels; clvl++)  {
	statemap_setbits (cmRail->Levels[clvl].LocalMap, CM_SVC_INDICATOR_OFFSET (cmRail, clvl, svc_indicator, cmRail->NodeId), 1, 1); 
	EPRINTF3 (DBG_SVC,"cm_svc_indicator_set: clvl %d nodeId %d offset %d\n", clvl, cmRail->NodeId, CM_SVC_INDICATOR_OFFSET (cmRail, clvl, svc_indicator, cmRail->NodeId));
    }
    spin_unlock_irqrestore (&cmRail->Lock, flags);

    return (0);
}

int
cm_svc_indicator_clear (EP_RAIL *rail, int svc_indicator)
{
    CM_RAIL *cmRail = rail->ClusterRail;
    unsigned long flags;
    int           clvl;

    EPRINTF2 (DBG_SVC, "cm_svc_indicator_clear: rail %p ind %d\n", rail, svc_indicator);

    if (svc_indicator < 0 || svc_indicator >= EP_SVC_NUM_INDICATORS)
    {
	EPRINTF1 (DBG_SVC, "cm_svc_indicator_clear: service indicator %d not registered\n", svc_indicator);
	return (-1);
    }

    if (rail->State == EP_RAIL_STATE_UNINITIALISED) 
	return (-2);

    spin_lock_irqsave (&cmRail->Lock, flags);
    for (clvl = 0; clvl < cmRail->NumLevels; clvl++)  {
	statemap_setbits (cmRail->Levels[clvl].LocalMap, CM_SVC_INDICATOR_OFFSET (cmRail, clvl, svc_indicator, cmRail->NodeId), 0, 1); 
	EPRINTF3 (DBG_SVC, "cm_svc_indicator_clear: clvl %d nodeId %d offset %d\n", clvl, cmRail->NodeId, CM_SVC_INDICATOR_OFFSET (cmRail, clvl, svc_indicator, cmRail->NodeId));
    }
    spin_unlock_irqrestore (&cmRail->Lock, flags);

    return (0);
}

int
cm_svc_indicator_is_set (EP_RAIL *rail, int svc_indicator, int nodeId)
{
    CM_RAIL *cmRail = rail->ClusterRail;
    unsigned long flags;
    int           clvl;
    bitmap_t      bits;

    EPRINTF4 (DBG_SVC, "cm_svc_indicator_is_set: rail %p ind %d nodeId %d (me=%d)\n", rail, svc_indicator, nodeId, cmRail->NodeId);

    if (svc_indicator < 0 || svc_indicator > EP_SVC_NUM_INDICATORS)
    {
	EPRINTF1 (DBG_SVC, "cm_svc_indicator_is_set: service indicator %d not registered\n", svc_indicator);
	return (0);
    }

    if (rail->State == EP_RAIL_STATE_UNINITIALISED) 
	return (0);
    
    spin_lock_irqsave (&cmRail->Lock, flags);
    for (clvl = 0; clvl < cmRail->NumLevels; clvl++)
	if (nodeId >= cmRail->Levels[clvl].MinNodeId && nodeId < (cmRail->Levels[clvl].MinNodeId + cmRail->Levels[clvl].NumNodes))
	    break;

    if ( clvl == cmRail->NumLevels) { 
	EPRINTF1 (DBG_SVC, "cm_svc_indicator_is_set: node out of range %d \n", nodeId); 
	spin_unlock_irqrestore (&cmRail->Lock, flags);
	return (0);
    }

    if ( cmRail->NodeId == nodeId ) 
	bits = statemap_getbits (cmRail->Levels[clvl].LocalMap, CM_SVC_INDICATOR_OFFSET (cmRail, clvl, svc_indicator, nodeId), 1);
    else
	bits = statemap_getbits (cmRail->Levels[clvl].GlobalMap, CM_SVC_INDICATOR_OFFSET (cmRail, clvl, svc_indicator, nodeId), 1);

    EPRINTF4 (DBG_SVC, "cm_svc_indicator_is_set: clvl %d nodeId %d offset %d %x\n", clvl, nodeId, CM_SVC_INDICATOR_OFFSET (cmRail, clvl, svc_indicator, nodeId), bits);

    spin_unlock_irqrestore (&cmRail->Lock, flags);

    return  ( (bits == 0) ? (0) : (1) );
}

int
cm_svc_indicator_bitmap (EP_RAIL *rail, int svc_indicator, bitmap_t * bitmap, int low, int nnodes)
{
    /* or in the bit map */  
    CM_RAIL      *cmRail = rail->ClusterRail;
    int           nodeId, clvl;
    bitmap_t      bits;
    unsigned long flags;
    int           clip_out_low, clip_out_high;
    int           curr_low, curr_high;
    int           check_low, check_high;

    EPRINTF4 (DBG_SVC, "cm_svc_indicator_bitmap: rail %p ind %d low %d high %d\n", rail, svc_indicator, low, (low + nnodes));

    if (svc_indicator < 0 || svc_indicator >= EP_SVC_NUM_INDICATORS)
    {
	EPRINTF1 (DBG_SVC, "cm_svc_indicator_bitmap: service indicator %d not registered\n", svc_indicator);
	return (-1);
    }

    if (rail->State != EP_RAIL_STATE_RUNNING) 
	return (-2);

    spin_lock_irqsave (&cmRail->Lock, flags);
    
    clip_out_low = clip_out_high = -1; /* all in */
    for (clvl = 0; clvl < cmRail->NumLevels; clvl++) {

	/* curr_high/low is the range of the current lvl */
	curr_low  = cmRail->Levels[clvl].MinNodeId;
	curr_high = cmRail->Levels[clvl].MinNodeId + cmRail->Levels[clvl].NumNodes;

	/* find out how much of low high is in this range and only check that part */
	check_low  = ( low  < curr_low)  ? curr_low  : low; 
	check_high = ( (low + nnodes) > curr_high) ? curr_high : (low + nnodes);

	EPRINTF6 (DBG_SVC, "cm_svc_indicator_bitmap: curr(%d,%d) check(%d,%d) clip(%d,%d)\n", curr_low, curr_high, check_low, check_high, clip_out_low, clip_out_high);

	for(nodeId = check_low; nodeId < check_high; nodeId++) {

	    if (  (clip_out_low <= nodeId) && (nodeId <= clip_out_high))
		nodeId = clip_out_high; /* step over the cliped out section */
	    else {

		if ( cmRail->NodeId == nodeId ) 
		    bits = statemap_getbits (cmRail->Levels[clvl].LocalMap, CM_SVC_INDICATOR_OFFSET (cmRail, clvl, svc_indicator, nodeId), 1);
		else
		    bits = statemap_getbits (cmRail->Levels[clvl].GlobalMap, CM_SVC_INDICATOR_OFFSET (cmRail, clvl, svc_indicator, nodeId), 1);
		
		if ( bits ) {
		    EPRINTF2 (DBG_SVC, "cm_svc_indicator_bitmap: its set nodeId %d (clvl %d)\n", nodeId, clvl);
		    BT_SET ( bitmap , nodeId - low );
		}
	    }
	}

	/* widen the clip out range */
	clip_out_low  = curr_low;
	clip_out_high = curr_high -1; 
    }

    spin_unlock_irqrestore (&cmRail->Lock, flags);      

    return (0);
}


static void
cm_heartbeat_timer (unsigned long arg)
{
    CM_RAIL *cmRail = (CM_RAIL *) arg;
    unsigned long flags;

    spin_lock_irqsave (&cmRail->Lock, flags);

    ASSERT (cmRail->Rail->State == EP_RAIL_STATE_RUNNING);

    DoHeartbeatWork (cmRail);
    
    __Schedule_Timer (cmRail, cmRail->NextRunTime);

    spin_unlock_irqrestore (&cmRail->Lock, flags);
}

void
DisplayRailDo (DisplayInfo *di, EP_RAIL *rail)
{
    CM_RAIL *cmRail = rail->ClusterRail;
    unsigned long flags;
    int  i, j;

    if (rail->State != EP_RAIL_STATE_RUNNING)
	return;

    spin_lock_irqsave (&cmRail->Lock, flags);

    (di->func)(di->arg, "NodeId=%d NodeLevel=%d NumLevels=%d NumNodes=%d\n", 
	    cmRail->NodeId, cmRail->TopLevel, cmRail->NumLevels, cmRail->Rail->Position.pos_nodes);
    
    (di->func)(di->arg, "[");

    for (i = 0; i < cmRail->NumLevels; i++)
    {
	if (i > 0)
	    (di->func)(di->arg, ",");
       
	if (i < cmRail->TopLevel)
	{
	    (di->func)(di->arg, "L ");
	  
	    for (j = 0; j < cmRail->Levels[i].NumSegs; j++)
		switch (cmRail->Levels[i].Sgmts[j].State)
		{
		case CM_SGMT_PRESENT: (di->func)(di->arg, "p%-4d", cmRail->Levels[i].Sgmts[j].NodeId); break;
		case CM_SGMT_WAITING: (di->func)(di->arg, "w%4s", ""); break;
		case CM_SGMT_COMING:  (di->func)(di->arg, "c%4s", ""); break;
		case CM_SGMT_ABSENT:  (di->func)(di->arg, ".%4s", ""); break;
		default:              (di->func)(di->arg, "?%4s", ""); break;
		}
	}
	else
	    switch (cmRail->Role)
	    {
	    case CM_ROLE_LEADER_CANDIDATE:      
		(di->func)(di->arg,"l "); 
		for (j = 0; j < cmRail->Levels[i].NumSegs; j++)
		    (di->func)(di->arg,"     ");
		break;
	  
	    case CM_ROLE_SUBORDINATE: 	    
		switch (cmRail->Levels[i].Sgmts[0].State)
		{
		case CM_SGMT_PRESENT: (di->func)(di->arg, "p%-4d", cmRail->Levels[i].Sgmts[0].NodeId); break;
		case CM_SGMT_WAITING: (di->func)(di->arg, "w%4s", ""); break;
		case CM_SGMT_COMING:  (di->func)(di->arg, "c%4s", ""); break;
		case CM_SGMT_ABSENT:  (di->func)(di->arg, ".%4s", ""); break;
		default:              (di->func)(di->arg, "?%4s", ""); break;
		}
		for (j = 1; j < cmRail->Levels[i].NumSegs; j++)
		    (di->func)(di->arg, "     ");
		break;
	  
	    default:
		(di->func)(di->arg, "####");
		break;
	    }
    }
    (di->func)(di->arg, "]\n");

    spin_unlock_irqrestore (&cmRail->Lock, flags);
}

void
DisplayRail (EP_RAIL *rail) 
{
    if (rail->State == EP_RAIL_STATE_RUNNING)
	DisplayRailDo (&di_ep_debug, rail);
}

void
DisplayStatus (EP_RAIL *rail)
{
    if (rail->State == EP_RAIL_STATE_RUNNING)
    {
	CM_RAIL *cmRail = rail->ClusterRail;
	unsigned long flags;
	
	spin_lock_irqsave (&cmRail->Lock, flags);
	
	DisplayNodeMaps (&di_ep_debug, cmRail);
	
	spin_unlock_irqrestore (&cmRail->Lock, flags);
    }
}

void
DisplaySegs (EP_RAIL *rail)
{
    if (rail->State == EP_RAIL_STATE_RUNNING)
    {
	CM_RAIL *cmRail = rail->ClusterRail;
	unsigned long flags;
	
	spin_lock_irqsave (&cmRail->Lock, flags);
	
	DisplayNodeSgmts (&di_ep_debug, cmRail);
	
	spin_unlock_irqrestore (&cmRail->Lock, flags);
    }
}

static void
LoadBroadcastRoute (CM_RAIL *cmRail, int lvl, int sidx)
{
    EP_RAIL *rail  = cmRail->Rail;
    int      nsegs = cmRail->Levels[0].NumSegs;
    int      vp    = EP_VP_BCAST(lvl, sidx);
    int      nodes = 1;
    int      baseNode;
    int      i;

    ASSERT (lvl > 0 && lvl <= cmRail->NumLevels);
    ASSERT (sidx == 0 || lvl < cmRail->NumLevels);

    ASSERT (vp >= EP_VP_BCAST_BASE && vp < EP_VP_BCAST_BASE + EP_VP_BCAST_COUNT);

    for (i = 1; i <= lvl; i++)
    {
	nodes *= nsegs;
	nsegs = (i == cmRail->NumLevels) ? 1 : cmRail->Levels[i].NumSegs;
    }

    baseNode = ((cmRail->NodeId / (nodes * nsegs)) * nsegs + sidx) * nodes;

    CPRINTF5 (2, "%s: broadcast vp lvl %d sidx %d [%d,%d]\n", 
	      cmRail->Rail->Name, lvl, sidx, baseNode, baseNode + nodes - 1);
    
    rail->Operations.LoadSystemRoute (rail, vp, baseNode, baseNode + nodes - 1);
}

static void
LoadRouteTable (CM_RAIL *cmRail)
{
    EP_RAIL *rail = cmRail->Rail;
    int      i, j;
   
   if (cmRail->NumNodes > EP_MAX_NODES)
   {
       printk ("More nodes (%d) than point-to-point virtual process table entries (%d)\n", cmRail->NumNodes, EP_MAX_NODES);
       panic ("LoadRouteTable\n");
   }

   for (i = 0; i < cmRail->NumNodes; i++)
       rail->Operations.LoadSystemRoute (rail, EP_VP_NODE(i), i, i);

   /* Generate broadcast routes for subtrees */
   for (i = 1; i < cmRail->NumLevels; i++)
      for (j = 0; j < cmRail->Levels[i].NumSegs; j++)
	  LoadBroadcastRoute (cmRail, i, j);

   /* Generate broadcast route for whole machine */
   LoadBroadcastRoute (cmRail, cmRail->NumLevels, 0);

   /* Finally invalidate all the data routes */
   for (i = 0; i < cmRail->NumNodes; i++)
       rail->Operations.UnloadNodeRoute (cmRail->Rail, i);
}

void
cm_node_disconnected (EP_RAIL *rail, unsigned nodeId)
{
    CM_RAIL *cmRail = rail->ClusterRail;
    int      base, lstat, lgstat;
    int	     clvl, subClMin, subClMax;
    int      thisClId, myClId;
    unsigned long flags;

    ASSERT (nodeId != cmRail->NodeId);

    spin_lock_irqsave (&cmRail->Lock, flags);
    for (clvl = 0; clvl < cmRail->NumLevels; clvl++)
	if (nodeId >= cmRail->Levels[clvl].MinNodeId && nodeId < (cmRail->Levels[clvl].MinNodeId + cmRail->Levels[clvl].NumNodes))
	    break;

    myClId   = ClusterIds (cmRail, clvl, &subClMin, &subClMax);
    thisClId = nodeId - cmRail->Levels[clvl].MinNodeId;
    base     = thisClId * CM_GSTATUS_BITS;
    lstat    = statemap_getbits (cmRail->Levels[clvl].LocalMap,  base, CM_GSTATUS_BITS);
    lgstat   = statemap_getbits (cmRail->Levels[clvl].LastGlobalMap, base, CM_GSTATUS_BITS) & CM_GSTATUS_STATUS_MASK;

    ASSERT ((lstat & CM_GSTATUS_ACK_MASK) == CM_GSTATUS_MAY_RUN);

    CPRINTF7 (2, "%s: cm_node_disconnected: Node %d: clvl %d, lgstat %s, gstat %s, lstat %s -> %sMAY_START\n",
	      cmRail->Rail->Name, nodeId, clvl,
	      GlobalStatusString (cmRail->Levels[clvl].LastGlobalMap, thisClId),
	      GlobalStatusString (cmRail->Levels[clvl].GlobalMap, thisClId),
	      GlobalStatusString (cmRail->Levels[clvl].LocalMap, thisClId),
	      ((lgstat != CM_GSTATUS_CLOSING) && (lstat & CM_GSTATUS_RESTART)) ? "RESTART|" : "");
    
    switch (lgstat)
    {
    case CM_GSTATUS_CLOSING:
	/* delayed ack of closing - set MAY_START and clear RESTART */
	statemap_setbits (cmRail->Levels[clvl].LocalMap, base, CM_GSTATUS_MAY_START, CM_GSTATUS_BITS);
	break;
    case CM_GSTATUS_STARTING:
    case CM_GSTATUS_RUNNING:
	IASSERT (! cmRail->Levels[clvl].Online || lstat & CM_GSTATUS_RESTART);
	break;
    case CM_GSTATUS_ABSENT:
	IASSERT (lstat & CM_GSTATUS_RESTART);
    }

    cmRail->Levels[clvl].Connected--;

    spin_unlock_irqrestore (&cmRail->Lock, flags);
}

void
cm_restart_node (EP_RAIL *rail, unsigned nodeId)
{
    CM_RAIL *cmRail = rail->ClusterRail;
    int      base, lstat, lgstat;
    int	     clvl, subClMin, subClMax;
    int      thisClId, myClId;
    unsigned long flags;

    spin_lock_irqsave (&cmRail->Lock, flags);
    if (nodeId == rail->Position.pos_nodeid)
    {
	for (clvl = 0; clvl < cmRail->NumLevels; clvl++)
	    RestartComms (cmRail, clvl);
    }
    else
    {
	for (clvl = 0; clvl < cmRail->NumLevels; clvl++)
	    if (nodeId >= cmRail->Levels[clvl].MinNodeId && nodeId < (cmRail->Levels[clvl].MinNodeId + cmRail->Levels[clvl].NumNodes))
		break;
	
	myClId   = ClusterIds (cmRail, clvl, &subClMin, &subClMax);
	thisClId = nodeId - cmRail->Levels[clvl].MinNodeId;
	base     = thisClId * CM_GSTATUS_BITS;
	lstat    = statemap_getbits (cmRail->Levels[clvl].LocalMap,  base, CM_GSTATUS_BITS);
	lgstat   = statemap_getbits (cmRail->Levels[clvl].LastGlobalMap,  base, CM_GSTATUS_BITS) & CM_GSTATUS_STATUS_MASK;

	CPRINTF6 (2, "%s: cm_restart_node: Node %d: clvl %d, lgstat %s, gstat %s, lstat %s\n",
		  cmRail->Rail->Name, nodeId, clvl,
		  GlobalStatusString (cmRail->Levels[clvl].LastGlobalMap, thisClId),
		  GlobalStatusString (cmRail->Levels[clvl].GlobalMap, thisClId),
		  GlobalStatusString (cmRail->Levels[clvl].LocalMap, thisClId));
	
	if (lgstat != CM_GSTATUS_CLOSING)
	    statemap_setbits (cmRail->Levels[clvl].LocalMap, base, lstat | CM_GSTATUS_RESTART, CM_GSTATUS_BITS);
    }
    spin_unlock_irqrestore (&cmRail->Lock, flags);
}

void
cm_force_offline (EP_RAIL *rail, int offline, unsigned int reason)
{
    CM_RAIL *cmRail = rail->ClusterRail;
    unsigned long flags;

    spin_lock_irqsave (&cmRail->Lock, flags);
    if (offline)
	cmRail->OfflineReasons |= reason;
    else
	cmRail->OfflineReasons &= ~reason;
    spin_unlock_irqrestore (&cmRail->Lock, flags);
}

static void
cm_remove_rail (EP_SUBSYS *subsys, EP_SYS *epsys, EP_RAIL *rail)
{
    CM_SUBSYS  *sys    = (CM_SUBSYS *) subsys;
    CM_RAIL    *cmRail = sys->Rails[rail->Number];
    int i, lvl, clvl;

    cm_procfs_rail_fini (cmRail);

    sys->Rails[rail->Number] = NULL;
    rail->ClusterRail        = NULL;

    del_timer_sync (&cmRail->HeartbeatTimer);

    cmRail->NextRunTime      = 0;
    cmRail->NextDiscoverTime = 0;
    cmRail->NextHeartbeatTime = 0;
    
    for (clvl = 0; clvl < cmRail->NumLevels; clvl++)
    {
	for (lvl = 0; lvl <= clvl; lvl++)
	{
	    CM_LEVEL *level = &cmRail->Levels[lvl];
	    
	    statemap_destroy (level->SubordinateMap[clvl]);
	    
	    for (i = 0; i < level->NumSegs; i++)
	    {
		statemap_destroy (level->Sgmts[i].Maps[clvl].CurrentInputMap);
		statemap_destroy (level->Sgmts[i].Maps[clvl].InputMap);
		statemap_destroy (level->Sgmts[i].Maps[clvl].OutputMap);
	    }
	}
	
	cmRail->Levels[clvl].Online = 0;
	
	statemap_destroy (cmRail->Levels[clvl].TmpMap);
	statemap_destroy (cmRail->Levels[clvl].GlobalMap);
	statemap_destroy (cmRail->Levels[clvl].LastGlobalMap);
	statemap_destroy (cmRail->Levels[clvl].SubTreeMap);
	statemap_destroy (cmRail->Levels[clvl].LocalMap);
    }
    
    spin_lock_destroy (&cmRail->Lock);

    ep_free_inputq (cmRail->Rail, cmRail->PolledQueue);
    ep_free_inputq (cmRail->Rail, cmRail->IntrQueue);
    ep_free_outputq (cmRail->Rail, cmRail->MsgQueue);

    KMEM_FREE (cmRail, sizeof (CM_RAIL));
}

static int
cm_add_rail (EP_SUBSYS *subsys, EP_SYS *epsys, EP_RAIL *rail)
{
    CM_SUBSYS     *sys = (CM_SUBSYS *) subsys;
    ELAN_POSITION *pos = &rail->Position;
    CM_RAIL       *cmRail;
    int lvl, n, nn, clvl, span, i;
    unsigned long flags;

    KMEM_ZALLOC (cmRail, CM_RAIL *, sizeof (CM_RAIL), 1);

    if (cmRail == NULL)
	return (ENOMEM);
    
    cmRail->Rail     = rail;
    cmRail->NodeId   = pos->pos_nodeid;
    cmRail->NumNodes = pos->pos_nodes;

    spin_lock_init (&cmRail->Lock);

    if ((cmRail->IntrQueue   = ep_alloc_inputq (rail, EP_SYSTEMQ_INTR,   sizeof (CM_MSG), CM_INPUTQ_ENTRIES, IntrQueueCallback, cmRail)) == NULL ||
	(cmRail->PolledQueue = ep_alloc_inputq (rail, EP_SYSTEMQ_POLLED, sizeof (CM_MSG), CM_INPUTQ_ENTRIES, NULL, 0)) == NULL ||
	(cmRail->MsgQueue    = ep_alloc_outputq (rail, sizeof (CM_MSG), CM_NUM_MSG_BUFFERS)) == NULL)
    {
	goto failed;
    }

    /* point to first "spare" message buffer */
    cmRail->NextSpareMsg = 0;

    /* Compute the branching ratios from the switcy arity */
    for (lvl = 0; lvl < CM_MAX_LEVELS; lvl++)
	BranchingRatios[lvl] = (lvl < pos->pos_levels) ? pos->pos_arity[pos->pos_levels - lvl - 1] : 4;
    
    /* now determine the number of levels of hierachy we have */
    /* and how many nodes per level there are */
    for (lvl = 0, nn = 1, n = pos->pos_nodes; 
	 n > 1; 
	 nn *= BranchingRatios[lvl], n = n / BranchingRatios[lvl], lvl++)
    {
	int       nSegs = (n > BranchingRatios[lvl]) ? BranchingRatios[lvl] : n;
	int       nNodes = nn * nSegs;
	CM_LEVEL *level = &cmRail->Levels[lvl];

	for (clvl = 0, span = pos->pos_arity[pos->pos_levels - clvl - 1]; 
	     span < nNodes && clvl < pos->pos_levels - 1;
	     clvl++, span *= pos->pos_arity[pos->pos_levels - clvl - 1])
	    ;
	
	level->SwitchLevel = clvl;
	level->MinNodeId = (pos->pos_nodeid / nNodes) * nNodes;
	level->NumNodes = nNodes;
	level->NumSegs = nSegs;
    }
    
    cmRail->NumLevels      = lvl;
    cmRail->BroadcastLevel = lvl-1;

    CPRINTF4 (2, "%s: NodeId=%d NumNodes=%d NumLevels=%d\n", 
	      rail->Name, pos->pos_nodeid, pos->pos_nodes, cmRail->NumLevels);

    LoadRouteTable (cmRail);
    
    /* Init SGMT constants */
    for (lvl = 0; lvl < cmRail->NumLevels; lvl++)
    {
	CM_LEVEL *level = &cmRail->Levels[lvl];

	level->MySgmt = SegmentNo (cmRail, cmRail->NodeId, lvl);
       
	for (i = 0; i < CM_SGMTS_PER_LEVEL; i++)
	{
	    CM_SGMT *sgmt = &level->Sgmts[i];
	  
	    sgmt->MsgNumber = lvl * CM_SGMTS_PER_LEVEL + i;
	    sgmt->Level = lvl;
	    sgmt->Sgmt = i;
	}
    }

    /* Init maps for each cluster level */
    for (clvl = 0; clvl < cmRail->NumLevels; clvl++)
    {
	int nNodes = cmRail->Levels[clvl].NumNodes;
	int mapBits = (nNodes * CM_GSTATUS_BITS) + (nNodes * EP_SVC_NUM_INDICATORS);
	int clmin;
	int clmax;
	int clid = ClusterIds (cmRail, clvl, &clmin, &clmax);

	for (lvl = 0; lvl <= clvl; lvl++)
	{
	    CM_LEVEL *level = &cmRail->Levels[lvl];

	    level->SubordinateMap[clvl] = statemap_create (mapBits);

	    for (i = 0; i < level->NumSegs; i++)
	    {
		level->Sgmts[i].Maps[clvl].CurrentInputMap = statemap_create (mapBits);
		level->Sgmts[i].Maps[clvl].InputMap        = statemap_create (mapBits);
		level->Sgmts[i].Maps[clvl].OutputMap       = statemap_create (mapBits);
	    }
	}
	
	cmRail->Levels[clvl].Online = 0;

	cmRail->Levels[clvl].TmpMap        = statemap_create (mapBits);
	cmRail->Levels[clvl].GlobalMap     = statemap_create (mapBits);
	cmRail->Levels[clvl].LastGlobalMap = statemap_create (mapBits);
	cmRail->Levels[clvl].SubTreeMap    = statemap_create (mapBits);
	cmRail->Levels[clvl].LocalMap      = statemap_create (mapBits);

	/* Flag everyone outside my next lower cluster as sensed offline... */
	for (i = 0; i < clmin; i++)
	    statemap_setbits (cmRail->Levels[clvl].LocalMap, i * CM_GSTATUS_BITS, CM_GSTATUS_MAY_START, CM_GSTATUS_BITS);
	
	for (i = clmax + 1; i < nNodes; i++)
	    statemap_setbits (cmRail->Levels[clvl].LocalMap, i * CM_GSTATUS_BITS, CM_GSTATUS_MAY_START, CM_GSTATUS_BITS);
	
	/* ...and set my own state */
	statemap_setbits (cmRail->Levels[clvl].LocalMap, clid * CM_GSTATUS_BITS,
			  CM_GSTATUS_CLOSING | CM_GSTATUS_MAY_START, CM_GSTATUS_BITS);
    }
    
    /* compute parameter hash to add to messages */
    cmRail->ParamHash = EP_PROTOCOL_VERSION;
    cmRail->ParamHash = cmRail->ParamHash * 127 + CM_PERIODIC_DISCOVER_INTERVAL;
    cmRail->ParamHash = cmRail->ParamHash * 127 + CM_URGENT_DISCOVER_INTERVAL;
    cmRail->ParamHash = cmRail->ParamHash * 127 + CM_HEARTBEAT_INTERVAL;
    cmRail->ParamHash = cmRail->ParamHash * 127 + CM_P2P_DMA_RETRIES;
    cmRail->ParamHash = cmRail->ParamHash * 127 + CM_P2P_MSG_RETRIES;
    cmRail->ParamHash = cmRail->ParamHash * 127 + CM_BCAST_MSG_RETRIES;
    cmRail->ParamHash = cmRail->ParamHash * 127 + CM_TIMER_SCHEDULE_TIMEOUT;
    cmRail->ParamHash = cmRail->ParamHash * 127 + CM_HEARTBEAT_TIMEOUT;
    cmRail->ParamHash = cmRail->ParamHash * 127 + CM_DISCOVER_TIMEOUT;
    cmRail->ParamHash = cmRail->ParamHash * 127 + BT_NBIPUL;
    cmRail->ParamHash = cmRail->ParamHash * 127 + CM_GSTATUS_BITS;
    cmRail->ParamHash = cmRail->ParamHash * 127 + EP_SVC_NUM_INDICATORS;
    cmRail->ParamHash = cmRail->ParamHash * 127 + cmRail->NumLevels;
    cmRail->ParamHash = cmRail->ParamHash * 127 + cmRail->NumNodes;
    for (i = 0; i < cmRail->NumLevels; i++)
	cmRail->ParamHash = cmRail->ParamHash * 127 + BranchingRatios[i];
    
    spin_lock_irqsave (&cmRail->Lock, flags);

    /* Initialise the timer, but don't add it yet, since
     * __Schedule_Heartbeat() will do this. */

    init_timer (&cmRail->HeartbeatTimer);
    
    cmRail->HeartbeatTimer.function = cm_heartbeat_timer;
    cmRail->HeartbeatTimer.data     = (unsigned long) cmRail;
    cmRail->HeartbeatTimer.expires  = lbolt + hz;

    /* Indicate that heartbeats should be sent 
     * as soon as the timer is run from inside 
     * LowerTopLevel */
    cmRail->NextHeartbeatTime = lbolt;
    
    /* start discovering who else is out there */
    LowerTopLevel (cmRail, 0);

    /* connect to myself straight away - I know I'm here */
    ep_connect_node (rail, cmRail->NodeId);
    
    /* add to all rails */
    sys->Rails[rail->Number] = cmRail;
    rail->ClusterRail = (void *) cmRail;

    spin_unlock_irqrestore (&cmRail->Lock, flags);

    /* Enable the input queues */
    ep_enable_inputq (rail, cmRail->PolledQueue);
    ep_enable_inputq (rail, cmRail->IntrQueue);

    /* Create the procfs entries */
    cm_procfs_rail_init (cmRail);

    return 0;

 failed:
    cm_remove_rail (subsys, epsys, rail);
    return -ENOMEM;
}

static void
cm_fini (EP_SUBSYS *subsys, EP_SYS *epsys)
{
    CM_SUBSYS *sys = (CM_SUBSYS *) subsys;

    cm_procfs_fini(sys);
    
    KMEM_FREE (sys, sizeof (CM_SUBSYS));
}

int
cm_init (EP_SYS *sys)
{
    CM_SUBSYS *subsys;

    KMEM_ZALLOC (subsys, CM_SUBSYS *, sizeof (CM_SUBSYS), 1);

    if (subsys == NULL)
	return (ENOMEM);

    subsys->Subsys.Sys        = sys;
    subsys->Subsys.Name	      = "cm";
    subsys->Subsys.Destroy    = cm_fini;
    subsys->Subsys.AddRail    = cm_add_rail;
    subsys->Subsys.RemoveRail = cm_remove_rail;

    ep_subsys_add (sys, &subsys->Subsys);

    cm_procfs_init (subsys);

    /*
     * Initialise the machineid if it wasn't specified by
     * the modules.conf file - otherwise truncate it to 
     * 16 bits.
     */
    if (MachineId != -1)
	MachineId = (uint16_t) MachineId;
    else
    {
#if defined(LINUX_ALPHA)
	MachineId = (uint16_t)((5 << 12) | HZ);
#elif defined(LINUX_SPARC)
	MachineId = (uint16_t)((4 << 12) | HZ);
#elif defined(LINUX_I386)
	MachineId = (uint16_t)((3 << 12) | HZ);
#elif defined( LINUX_IA64)
	MachineId = (uint16_t)((2 << 12) | HZ);
#elif defined(LINUX_X86_64)
	MachineId = (uint16_t)((1 << 12) | HZ);
#else
	MachineId = (uint16_t)((0 << 12) | HZ);
#endif
    }

    return (0);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
