/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __ELAN_CM_H
#define __ELAN_CM_H

#ident "@(#)$Id: cm.h,v 1.16 2005/03/30 14:06:34 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/cm.h,v $*/

#include <elan/statemap.h>

#if defined(DIGITAL_UNIX)
/*
 * On Tru64 - SMP doesn't mean Symmetric - cpu 0 is a master cpu and is responsible
 * for handling all PCI interrupts and "funneled" operations.  When a kernel thread
 * is made runnable, the scheduler will choose which cpu it will run on at that time,
 * and will only execute a higher priority thread from another cpu's run queue when 
 * it becomes totally idle (apparently also including user processes).  Also the 
 * assert_wait_mesg_timo function uses a per-cpu timeout - these can only get executed
 * at "preemptable" places - so again have no guarantee on when they will execute if
 * they happen to be queued on a "hogged" cpu. The combination of these mean that the Tru64
 * is incapable of scheduling a high priority kernel  thread within a deterministic time
 * of when it should have become runnable - wonderfull.
 *
 * Hence the solution Compaq have proposed it to schedule a timeout onto all of the
 * cpu's timeouts lists at the maximum frequency that we could want to execute code,
 * then to handle the scheduling of work between these ourselves.  With a bit of luck
 * ..... at least one cpu will be sufficiently unloaded to allow us to get a chance
 * to do our important work.
 *
 * However ..... this still is not reliable, since timeouts under Tru64 are still 
 * only run when the currently running kernel thread "co-operates" by calling one
 * of a number of functions which is permitted to run the "lwc"s AND is not holding
 * any spinlocks AND is running ai IPL 0.   However Compaq are unable to provide
 * any upper limit on the time between the "lwc"'s being run and so it is possible
 * for all 4 cpus to not run them for an unbounded time.
 *
 * The solution proposed is to use the RM_TEMP_BACKDOOR hook which was added to 
 * hardclock() to "solve" this problem for Memory Channel.  However, since it
 * is called within the clock interrupt it is not permissible to aquire any
 * spinlocks, nor to run for "too long".  This means that it is not possible to
 * call the heartbeat algorithm from this hook.  
 *
 * Our solution to these limitations is to use the hook to cause an elan interrupt 
 * to be delivered, by issueing a mis-aligned SetEvent command - this causes the device 
 * to trap and ep_cprocTrap() can then run the heartbeat code.  However there is a lock 
 * order violation between the elan_dev::IntrLock and ep_dev::Lock, so we have to 
 * use a trylock and if we fail, then hope that when the interrupt is delievered again
 * some time later we will succeed.
 *
 * However this only works if the kernel is able to respond to the Elan interrupt,
 * so we panic inside the RM_TEMP_BACKDOOR hook if the SetEvent's interrupt has
 * not been taken for more than an CM_TIMER_SCHEDULE_TIMEOUT interval.
 *
 * In fact this is exactly the mechanism that other operating systems use to
 * execute timeouts, since the hardclock interrupt posts a low priority 
 * "soft interrupt" which "pre-eempts" the currently running thread and then
 * executes the timeouts.To block timeouts you use splsoftclock() the same as 
 * in Tru64.
 */
#define PER_CPU_TIMEOUT			TRUE
#endif


#define CM_SGMTS_PER_LEVEL		8			/* maximum nodes in each segment */
#define CM_MAX_LEVELS			6			/* maximum depth of tree */

/* message buffers/dmas/events etc */
#define CM_NUM_NODE_MSG_BUFFERS		(CM_MAX_LEVELS * CM_SGMTS_PER_LEVEL) /* subordinates and leader */
#define CM_NUM_SPARE_MSG_BUFFERS	8			/* spare msg buffers for non-connected nodes */
#define CM_NUM_MSG_BUFFERS		(CM_NUM_NODE_MSG_BUFFERS + CM_NUM_SPARE_MSG_BUFFERS)

#define CM_INPUTQ_ENTRIES		128			/* # entries in input queue */

#define CM_PERIODIC_DISCOVER_INTERVAL	(5000)		/* 5s (infrequent resolution of established leader conflicts) */
#define CM_URGENT_DISCOVER_INTERVAL	(50)		/* 0.05s (more frequently than heartbeats 'cause they don't retry) */
#define CM_HEARTBEAT_INTERVAL		(125)		/* 0.125s */
#define CM_TIMER_SCHEDULE_TIMEOUT	(4000)		/* 4s     Maximum time before a timer that's secheduled to run gets to run (eg blocked in interrupt handlers etc) */
#define CM_THREAD_SCHEDULE_TIMEOUT	(30000)		/* 30s    Maximum time before a thread that's scheduled to run gets to run */
#define CM_THREAD_RUNNING_TIMEOUT	(30000)		/* 30s    Don't expect the manager thread to be running longer than this */

#ifdef PER_CPU_TIMEOUT
#define CM_PERCPU_TIMEOUT_INTERVAL	(50)		/* 0.05s (must be less than all above intervals) */
#define CM_PACEMAKER_INTERVAL		(500)		/* 0.05s */

#define CM_HEARTBEAT_OVERDUE		(250)		/* 0.25s Maximum time a timeout can be overdue before taking extreme action */
#endif

#define CM_P2P_DMA_RETRIES		31

/* We expect at least 1 point-to-point message in CM_P2P_MSG_RETRIES
 * attempts to send one to be successfully received */
#define CM_P2P_MSG_RETRIES		8

/* We expect at least 1 broadcast message in CM_BCAST_MSG_RETRIES attempts
 * to send one to be successfully received. */
#define CM_BCAST_MSG_RETRIES		40

/* Heartbeat timeout allows for a node stalling and still getting its
 * heartbeat. The 2 is to allow for unsynchronised polling times. */
#define CM_HEARTBEAT_TIMEOUT		(CM_TIMER_SCHEDULE_TIMEOUT + (2 + CM_P2P_MSG_RETRIES) * CM_HEARTBEAT_INTERVAL)

/* Discover timeout must be > CM_HEARTBEAT_TIMEOUT to guarantee that people
 * who don't see discovery are considered dead by their leader.  This
 * ensures that by the time a node "discovers" it is a leader of a segment,
 * the previous leader of that segment will have been deemed to be dead by
 * its the parent segment's leader */
#define CM_DISCOVER_TIMEOUT		(CM_TIMER_SCHEDULE_TIMEOUT + (2 + CM_BCAST_MSG_RETRIES) * CM_URGENT_DISCOVER_INTERVAL)

#define CM_WAITING_TIMEOUT		(CM_DISCOVER_TIMEOUT * 100)

/*
 * Convert all timeouts specified in mS into "ticks"
 */
#define MSEC2TICKS(MSEC)		(((MSEC)*HZ)/1000)


/* statemap entry */
typedef struct cm_state_entry
{
    int16_t           level;			/* cluster level to apply to */
    int16_t	      offset;			/* from statemap_findchange() */
    uint16_t          seg[BT_NBIPUL/16]; 	/* ditto */
} CM_STATEMAP_ENTRY;

/* offset is >= 0 for a change to apply and */
#define STATEMAP_NOMORECHANGES	(-1)		/* end of a set of updates */
#define STATEMAP_RESET		(-2)		/* reset the target map */
#define STATEMAP_NOOP		(-3)		/* null token */

/* CM message format */
typedef int8_t	CM_SEQ;				/* heartbeat sequence numbers; at least 2 bits, signed */

/*
 * The message header is received into the last 64 byte block of 
 * the input queue and the Version *MUST* be the last word of the 
 * block to ensure that we can see that the whole of the message
 * has reached main memory after we've seen the input queue pointer
 * have been updated.
 */
typedef struct ep_cm_hdr
{
    uint32_t	       Pad0;
    uint32_t	       Pad1;

    uint8_t	       Type;
    uint8_t	       Level;
    CM_SEQ	       Seq;			/* precision at least 2 bits each*/
    CM_SEQ	       AckSeq;
    
    uint16_t	       NumMaps;
    uint16_t	       MachineId;

    uint16_t	       NodeId;
    uint16_t	       Checksum;

    uint32_t           Timestamp;
    uint32_t           ParamHash;
    uint32_t	       Version;
} CM_HDR;

#define CM_HDR_SIZE	    sizeof (CM_HDR)

typedef struct cm_msg
{
    union {
	CM_STATEMAP_ENTRY   Statemaps[1];		/* piggy-backed statemap updates start here */
	uint8_t             Space[EP_SYSTEMQ_MSG_MAX - CM_HDR_SIZE];
    } Payload;
    
    CM_HDR		    Hdr;
} CM_MSG;

/* The maximum number of statemap entries that can fit within an EP_CM_MSG_BUFFER */
#define CM_MSG_MAXMAPS		(offsetof (CM_MSG, Hdr) / sizeof (CM_STATEMAP_ENTRY))
#define CM_MSG_MAP(mapno)	(CM_MSG_MAXMAPS - (mapno) - 1)

/* The actual special message base & size, including 'nmaps' piggy-backed statemap entries */
#define CM_MSG_BASE(nmaps)	(nmaps == 0 ? offsetof (CM_MSG, Hdr) : offsetof (CM_MSG, Payload.Statemaps[CM_MSG_MAXMAPS - nmaps]))
#define CM_MSG_SIZE(nmaps)	(sizeof (CM_MSG) - CM_MSG_BASE(nmaps))

#define CM_MSG_VERSION				0xcad00005
#define CM_MSG_TYPE_RESOLVE_LEADER		0
#define CM_MSG_TYPE_DISCOVER_LEADER		1
#define CM_MSG_TYPE_NOTIFY			2
#define CM_MSG_TYPE_DISCOVER_SUBORDINATE	3
#define CM_MSG_TYPE_IMCOMING			4
#define CM_MSG_TYPE_HEARTBEAT			5
#define CM_MSG_TYPE_REJOIN			6

/* CM machine segment */
typedef struct cm_sgmtMaps
{
    u_char       InputMapValid;			/* Input map has been set */
    u_char       OutputMapValid;		/* Output map has been set */
    u_char       SentChanges;			/* got an outstanding STATEMAP_NOMORECHANGES to send */
    statemap_t  *OutputMap;			/* state to send */
    statemap_t  *InputMap;			/* state received */
    statemap_t  *CurrentInputMap;		/* state being received */
} CM_SGMTMAPS;

typedef struct cm_sgmt
{
   u_char       State;
   u_char       SendMaps;
   u_char       MsgAcked;
   CM_SEQ	MsgSeq;
   CM_SEQ	AckSeq;
   u_int	NodeId;
   long		UpdateTick;
   long		WaitingTick;
   uint32_t	Timestamp;
   CM_SGMTMAPS  Maps[CM_MAX_LEVELS];		/* Maps[i] == state for cluster level i */
   u_short      MsgNumber;			/* msg buffer to use */
   u_short	NumMaps;			/* # maps in message buffer */
   u_short      Level;
   u_short      Sgmt;
} CM_SGMT;

#define CM_SGMT_ABSENT		0		/* no one there at all */
#define CM_SGMT_WAITING		1		/* waiting for subtree to connect */
#define CM_SGMT_COMING		2		/* expecting a subtree to reconnect */
#define CM_SGMT_PRESENT		3		/* connected */

typedef struct cm_level
{
    int 	       SwitchLevel;
    u_int	       MinNodeId;
    u_int              NumNodes;
    u_int              NumSegs;
    u_int              MySgmt;
   
    /* SubordinateMap[i] == OR of all subordinate maps on this level and down for cluster level i */
    u_char             SubordinateMapValid[CM_MAX_LEVELS];
    statemap_t        *SubordinateMap[CM_MAX_LEVELS];

    /* maps/flags for this cluster level */
    u_int              Online:1;				/* I've gone online (seen myself running) */
    u_int	       Restarting:1;				/* driving my owm restart bit */
    u_char	       OfflineReasons;				/* forced offline by broadcast */

    u_char             GlobalMapValid;
    u_char             SubTreeMapValid;
    u_long	       Connected;

    statemap_t        *LocalMap;		/* state bits I drive */
    statemap_t        *SubTreeMap;		/* OR of my and my subtree states */
    statemap_t        *GlobalMap;		/* OR of all node states */
    statemap_t        *LastGlobalMap;		/* last map I saw */
    statemap_t        *TmpMap;			/* scratchpad */

    CM_SGMT	       Sgmts[CM_SGMTS_PER_LEVEL];
} CM_LEVEL;

#define CM_ROLE_LEADER_CANDIDATE	0
#define CM_ROLE_LEADER			1
#define CM_ROLE_SUBORDINATE		2

/* global status bits */
#define CM_GSTATUS_STATUS_MASK		0x03	/* bits nodes drive to broadcast their status */
#define CM_GSTATUS_ABSENT		0x00	/* Off the network */
#define CM_GSTATUS_STARTING		0x01	/* I'm waiting for everyone to see me online */
#define CM_GSTATUS_RUNNING              0x03	/* up and running */
#define CM_GSTATUS_CLOSING		0x02	/* I'm waiting for everyone to see me offline */

#define CM_GSTATUS_ACK_MASK		0x0c	/* bits node drive to ack other status */
#define CM_GSTATUS_MAY_START		0x04	/* Everyone thinks I may not start */
#define CM_GSTATUS_MAY_RUN		0x08	/* Everyone thinks I may not run */

#define CM_GSTATUS_RESTART		0x10	/* Someone thinks I should restart */
#define CM_GSTATUS_BITS			5

#define CM_GSTATUS_BASE(node)		((node) * CM_GSTATUS_BITS)

#if defined(PER_CPU_TIMEOUT)
typedef struct cm_timeout_data
{
    long		ScheduledAt;				/* lbolt timeout was scheduled to run at */

    unsigned long       EarlyCount;				/* # times run early than NextRun */
    unsigned long	MissedCount;				/* # times run on time - but someone else was running it */
    unsigned long       WastedCount;				/* # times we failed to get the spinlock */
    unsigned long	WorkCount;				/* # times we're the one running */

    unsigned long	WorstDelay;				/* worst scheduling delay */
    unsigned long	BestDelay;				/* best scheduling delay */

    unsigned long	WorstLockDelay;				/* worst delay before getting rail->Lock */

    unsigned long	WorstHearbeatDelay;			/* worst delay before calling DoHeartbeatWork */
} CM_TIMEOUT_DATA;
#endif

typedef struct cm_rail
{
    EP_RAIL	      *Rail;					/* rail we're associated with */
    struct list_head   Link;					/*   and linked on the CM_SUBSYS */

    uint32_t	       ParamHash;				/* hash of critical parameters */
    uint32_t           Timestamp;
    long	       DiscoverStartTick;			/* when discovery start */

    unsigned int       NodeId;					/* my node id */
    unsigned int       NumNodes;				/*   and number of nodes */
    unsigned int       NumLevels;				/* number of levels computed from machine size */
    int		       BroadcastLevel;
    long	       BroadcastLevelTick;
    unsigned int       TopLevel;				/* level at which I'm not a leader */
    unsigned char      Role;					/* state at TopLevel */

    EP_INPUTQ	      *PolledQueue;				/* polled input queue */
    EP_INPUTQ	      *IntrQueue;				/* intr input queue */
    EP_OUTPUTQ	      *MsgQueue;				/* message  */
    unsigned int       NextSpareMsg;				/* next "spare" message buffer to use */

    EP_CM_RAIL_STATS   Stats;					/* statistics */

    kmutex_t	       Mutex;
    spinlock_t	       Lock;
    
    long	       NextHeartbeatTime;			/* next time to check/send heartbeats */
    long	       NextDiscoverTime;			/* next time to progress discovery  */
    long	       NextRunTime;				/* the earlier of the above two or intr requires inputq poll*/

    unsigned int       OfflineReasons;				/* forced offline by procfs/manager thread stuck */

#if defined(PER_CPU_TIMEOUT)
    spinlock_t	       HeartbeatTimeoutsLock;			/* spinlock to sequentialise per-cpu timeouts */
    long	       HeartbeatTimeoutsStarted;		/* bitmap of which timeouts have started */
    long	       HeartbeatTimeoutsStopped;		/* bitmap of which timeouts have stopped */
    long	       HeartbeatTimeoutsShouldStop;		/* flag to indicate timeouts should stop */
    kcondvar_t	       HeartbeatTimeoutsWait;			/* place to sleep waiting for timeouts to stop */
    long	       HeartbeatTimeoutRunning;			/* someone is running the timeout - don't try for the lock */

    long	       HeartbeatTimeoutOverdue;			/* heartbeat seen as overdue - interrupt requested */

    CM_TIMEOUT_DATA   *HeartbeatTimeoutsData;			/* per timeout data */
#else
    struct timer_list  HeartbeatTimer;				/* timer for heartbeat/discovery */
#endif

    CM_LEVEL           Levels[CM_MAX_LEVELS];
} CM_RAIL;

/* OfflineReasons (both per-rail and  */
#define CM_OFFLINE_BROADCAST		(1 << 0)
#define CM_OFFLINE_PROCFS		(1 << 1)
#define CM_OFFLINE_MANAGER		(1 << 2)

typedef struct cm_subsys
{
    EP_SUBSYS		Subsys;
    CM_RAIL            *Rails[EP_MAX_RAILS];
} CM_SUBSYS;

extern int  MachineId;

extern void cm_node_disconnected (EP_RAIL *rail, unsigned nodeId);
extern void cm_restart_node (EP_RAIL *rail, unsigned nodeId);
extern void cm_restart_comms (CM_RAIL *cmRail);
extern int  cm_init (EP_SYS *sys);

extern void DisplayRail(EP_RAIL *rail);
extern void DisplaySegs (EP_RAIL *rail);
extern void DisplayStatus (EP_RAIL *rail);

extern void DisplayNodeMaps  (DisplayInfo *di, CM_RAIL *cmRail);
extern void DisplayNodeSgmts (DisplayInfo *di, CM_RAIL *cmRail);
extern void DisplayRailDo    (DisplayInfo *di, EP_RAIL *rail);

extern int    cm_read_cluster(EP_RAIL *rail,char *page);
extern void   cm_force_offline (EP_RAIL *rail, int offline, unsigned int reason);

extern int    cm_svc_indicator_set      (EP_RAIL *rail, int svc_indicator);
extern int    cm_svc_indicator_clear    (EP_RAIL *rail, int svc_indicator);
extern int    cm_svc_indicator_is_set   (EP_RAIL *rail, int svc_indicator, int nodeId);
extern int    cm_svc_indicator_bitmap   (EP_RAIL *rail, int svc_indicator, bitmap_t * bitmap, int low, int nnodes);

/* cm_procfs.c */
extern void   cm_procfs_init (CM_SUBSYS *subsys);
extern void   cm_procfs_fini (CM_SUBSYS *subsys);
extern void   cm_procfs_rail_init (CM_RAIL *rail);
extern void   cm_procfs_rail_fini (CM_RAIL *rail);

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
#endif /* __ELAN_CM_H */

