/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: epcomms.c,v 1.77 2004/11/30 12:02:06 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/epcomms.c,v $ */

#include <qsnet/kernel.h>
#include <qsnet/kthread.h>
#include <qsnet/autoconf.h>

#include <elan/kcomm.h>
#include <elan/epsvc.h>
#include <elan/epcomms.h>
#include "cm.h"
#include "debug.h"

static void
ep_comms_thread (void *arg)
{
    EP_COMMS_SUBSYS  *subsys = (EP_COMMS_SUBSYS *) arg;
    struct list_head *el;

    kernel_thread_init ("ep_comms");

    /* since ep_alloc_xmtr() has incremented the module use count,
     * we would be preventing the module from being unloaded, so
     * we decrement the use count since this thread must terminate
     * during unload of the module.
     */
    ep_mod_dec_usecount();

    for (;;)
    {
	long nextRunTime = 0;

	/* NOTE - subsys->Lock serializes us against flush/relocations
	 *        caused by rail nodeset transitions.
	 */
	kmutex_lock (&subsys->Lock);
	list_for_each (el, &subsys->Transmitters) {
	    nextRunTime = ep_check_xmtr (list_entry (el, EP_XMTR, Link), nextRunTime);
	}

	list_for_each (el, &subsys->Receivers) {
	    nextRunTime = ep_check_rcvr (list_entry (el, EP_RCVR, Link), nextRunTime);
	}
	kmutex_unlock (&subsys->Lock);

#if ! defined(CONFIG_EP_NO_CHECK_SUM)
	ep_csum_rxds (subsys);	
#endif
	nextRunTime = ep_forward_rxds (subsys, nextRunTime);

	if (ep_kthread_sleep (&subsys->Thread, nextRunTime) < 0)
	    break;
    }

    ep_mod_inc_usecount();

    ep_kthread_stopped (&subsys->Thread);
    kernel_thread_exit();
}

int
ep_comms_add_rail (EP_SUBSYS *s, EP_SYS *sys, EP_RAIL *rail)
{
    EP_COMMS_SUBSYS  *subsys = (EP_COMMS_SUBSYS *) s;
    EP_COMMS_RAIL    *commsRail;
    struct list_head *el;

    printk ("%s: vendorid=%x deviceid=%x\n", rail->Name, rail->Devinfo.dev_vendor_id, rail->Devinfo.dev_device_id);

    switch (rail->Devinfo.dev_device_id)
    {
#if defined(CONFIG_ELAN3) || defined(CONFIG_ELAN3_MODULE)
    case PCI_DEVICE_ID_ELAN3:
	commsRail = ep3comms_add_rail (s, sys, rail);
	break;
#endif
#if defined(CONFIG_ELAN4) || defined(CONFIG_ELAN4_MODULE)
    case PCI_DEVICE_ID_ELAN4:
	commsRail = ep4comms_add_rail (s, sys, rail);
	break;
#endif
    default:
	return 0;
    }

    if (commsRail == NULL)
	return 1;

    commsRail->Rail   = rail;
    commsRail->Subsys = subsys;

    kmutex_lock (&subsys->Lock);
    list_add_tail (&commsRail->Link, &subsys->Rails);
    
    list_for_each (el, &subsys->Receivers) {
	EP_RCVR *rcvr = list_entry (el, EP_RCVR, Link);

	EP_RAIL_OP (commsRail, Rcvr.AddRail) (rcvr, commsRail);
    }
	
    list_for_each (el, &subsys->Transmitters) {
	EP_XMTR *xmtr = list_entry (el, EP_XMTR, Link);

	EP_RAIL_OP (commsRail, Xmtr.AddRail) (xmtr, commsRail);
    }

    kmutex_unlock (&subsys->Lock);

    return 0;
}

void
ep_comms_del_rail (EP_SUBSYS *s, EP_SYS *sys, EP_RAIL *rail)
{
    EP_COMMS_SUBSYS  *subsys    = (EP_COMMS_SUBSYS *) s;
    EP_COMMS_RAIL    *commsRail = NULL;
    struct list_head *el;

    kmutex_lock (&subsys->Lock);
    /* find out rail entry and remove from system list */
    list_for_each (el, &subsys->Rails) {
	if ((commsRail = list_entry (el, EP_COMMS_RAIL, Link))->Rail == rail)
	    break;
    }

    list_del (&commsRail->Link);
    
    list_for_each (el, &subsys->Receivers) {
	EP_RCVR *rcvr = list_entry (el, EP_RCVR, Link);

	EP_RAIL_OP(commsRail, Rcvr.DelRail) (rcvr, commsRail);
    }
	
    list_for_each (el, &subsys->Transmitters) {
	EP_XMTR *xmtr = list_entry (el, EP_XMTR, Link);

	EP_RAIL_OP(commsRail,Xmtr.DelRail) (xmtr, commsRail);
    }

    kmutex_unlock (&subsys->Lock);

    EP_RAIL_OP (commsRail, DelRail) (commsRail);
}

void
ep_comms_fini (EP_SUBSYS *s, EP_SYS *sys)
{
    EP_COMMS_SUBSYS *subsys = (EP_COMMS_SUBSYS *) s;

    ep_kthread_stop (&subsys->Thread);
    ep_kthread_destroy (&subsys->Thread);

    if (subsys->ForwardXmtr)
	ep_free_xmtr (subsys->ForwardXmtr);

    spin_lock_destroy (&subsys->ForwardDescLock);

#if ! defined(CONFIG_EP_NO_CHECK_SUM)
    spin_lock_destroy (&subsys->CheckSumDescLock);
#endif

    kmutex_destroy (&subsys->Lock);

    KMEM_FREE (subsys, sizeof (EP_COMMS_SUBSYS));
}

int
ep_comms_init (EP_SYS *sys)
{
    EP_COMMS_SUBSYS *subsys;

    KMEM_ZALLOC (subsys, EP_COMMS_SUBSYS *, sizeof (EP_COMMS_SUBSYS), 1);

    if (subsys == NULL)
	return (ENOMEM);

    INIT_LIST_HEAD (&subsys->Rails);
    INIT_LIST_HEAD (&subsys->Receivers);
    INIT_LIST_HEAD (&subsys->Transmitters);
    INIT_LIST_HEAD (&subsys->ForwardDescList);

    kmutex_init (&subsys->Lock);
    spin_lock_init (&subsys->ForwardDescLock);

#if ! defined(CONFIG_EP_NO_CHECK_SUM)
    INIT_LIST_HEAD (&subsys->CheckSumDescList);
    spin_lock_init (&subsys->CheckSumDescLock);
#endif

    subsys->Subsys.Sys        = sys;
    subsys->Subsys.Name	      = "epcomms";
    subsys->Subsys.Destroy    = ep_comms_fini;
    subsys->Subsys.AddRail    = ep_comms_add_rail;
    subsys->Subsys.RemoveRail = ep_comms_del_rail;

    ep_subsys_add (sys, &subsys->Subsys);
    ep_kthread_init (&subsys->Thread);

    if ((subsys->ForwardXmtr = ep_alloc_xmtr (subsys->Subsys.Sys)) == NULL)
	goto failed;

    if (kernel_thread_create (ep_comms_thread, subsys) == NULL)
	goto failed;
    ep_kthread_started (&subsys->Thread);

    return (0);

 failed:
    ep_subsys_del (sys, &subsys->Subsys);
    ep_comms_fini (&subsys->Subsys, sys);

    return (ENOMEM);
}

void
ep_comms_display (EP_SYS *sys, char *how)
{
    EP_COMMS_SUBSYS  *subsys = (EP_COMMS_SUBSYS *) ep_subsys_find (sys, EPCOMMS_SUBSYS_NAME);
    struct list_head *el;

    if (how == NULL || !strncmp (how, "rail", 4))
    {
	kmutex_lock (&subsys->Lock);
	list_for_each (el, &subsys->Rails) {
	    EP_COMMS_RAIL *commsRail = list_entry (el, EP_COMMS_RAIL, Link);

	    EP_RAIL_OP(commsRail, DisplayRail) (commsRail);
	}
	kmutex_unlock (&subsys->Lock);
    }
	    
    if (how == NULL || !strncmp (how, "xmtr", 4))
	list_for_each (el, &subsys->Transmitters)
	    ep_display_xmtr (&di_ep_debug, list_entry (el, EP_XMTR, Link));

    if (how == NULL || !strncmp (how, "rcvr", 4)) 
	list_for_each (el, &subsys->Receivers)
	    ep_display_rcvr (&di_ep_debug, list_entry (el, EP_RCVR, Link), (how && how[4] == ',') ? 1 : 0);
}

int
ep_svc_indicator_set (EP_SYS *epsys, int svc_indicator) 
{
    EP_COMMS_SUBSYS  *subsys;
    struct list_head *el;

    EPRINTF1 (DBG_SVC,"ep_svc_indicator_set: %d \n",svc_indicator);

    if (svc_indicator < 0 || svc_indicator > EP_SVC_NUM_INDICATORS)
	return (EP_EINVAL);

    if ((subsys = (EP_COMMS_SUBSYS *) ep_subsys_find (epsys, "epcomms")) == NULL) {
	EPRINTF0 (DBG_SVC,"ep_svc_indicator_set: ep_subsys_find failed\n");
	return (EP_EINVAL);
    }


    kmutex_lock (&subsys->Lock); /* walking rails list and setting info on Rail */
    list_for_each (el, &subsys->Rails) { 
	EP_COMMS_RAIL *commsRail = list_entry (el, EP_COMMS_RAIL, Link);

	cm_svc_indicator_set(commsRail->Rail, svc_indicator);
    }
    kmutex_unlock (&subsys->Lock);

    EPRINTF1 (DBG_SVC,"ep_svc_indicator_set: %d success\n",svc_indicator);
    return (EP_SUCCESS);
}

int
ep_svc_indicator_clear (EP_SYS *epsys, int svc_indicator) 
{
    EP_COMMS_SUBSYS  *subsys;
    struct list_head *el;

    EPRINTF1 (DBG_SVC,"ep_svc_indicator_clear: %d \n",svc_indicator);

    if (svc_indicator < 0 || svc_indicator >= EP_SVC_NUM_INDICATORS)
	return (EP_EINVAL);

    if ((subsys = (EP_COMMS_SUBSYS *) ep_subsys_find (epsys, "epcomms")) == NULL) {
	EPRINTF0 (DBG_SVC,"ep_svc_indicator_clear: ep_subsys_find failed\n");
	return (EP_EINVAL);
    }

    kmutex_lock (&subsys->Lock); /* walking rails list and setting info on Rail */
    list_for_each (el, &subsys->Rails) { 
	EP_COMMS_RAIL *commsRail = list_entry (el, EP_COMMS_RAIL, Link);

	cm_svc_indicator_clear(commsRail->Rail, svc_indicator);
    }
    kmutex_unlock (&subsys->Lock);

    EPRINTF1 (DBG_SVC,"ep_svc_indicator_clear: %d success\n",svc_indicator);
    return (EP_SUCCESS);
}

int 
ep_svc_indicator_is_set (EP_SYS *epsys, int svc_indicator, int nodeId) 
{
    EP_COMMS_SUBSYS  *subsys;
    struct list_head *el;
    int               set = 0;

    EPRINTF2 (DBG_SVC,"ep_svc_indicator_is_set: svc %d node %d \n", svc_indicator, nodeId);

    if ((subsys = (EP_COMMS_SUBSYS *) ep_subsys_find (epsys, "epcomms")) == NULL) {
	EPRINTF0 (DBG_SVC,"ep_svc_indicator_is_set: ep_subsys_find failed\n");
	return (0);
    }

    kmutex_lock (&subsys->Lock); /* walking rails list and setting info on Rail */
    list_for_each (el, &subsys->Rails) { 
	EP_COMMS_RAIL *commsRail = list_entry (el, EP_COMMS_RAIL, Link);

	set |= cm_svc_indicator_is_set(commsRail->Rail, svc_indicator, nodeId);
    }
    kmutex_unlock (&subsys->Lock);

    EPRINTF3 (DBG_SVC,"ep_svc_indicator_is_set: svc %d node %d returning %d\n", svc_indicator, nodeId, set);
    return set;
}

int
ep_svc_indicator_bitmap (EP_SYS *epsys, int svc_indicator, bitmap_t * bitmap, int low, int nnodes) 
{
    EP_COMMS_SUBSYS  *subsys;
    struct list_head *el;

    EPRINTF1 (DBG_SVC,"ep_svc_indicator_bitmap: svc %d\n", svc_indicator);

    if (svc_indicator < 0 || svc_indicator >= EP_SVC_NUM_INDICATORS)
	return (-1);

    if ((subsys = (EP_COMMS_SUBSYS *) ep_subsys_find (epsys, "epcomms")) == NULL) {
	EPRINTF0 (DBG_SVC,"ep_svc_indicator_bitmap: ep_subsys_find failed\n");
	return (-2);
    }

    /* clear bitmap */
    bt_zero (bitmap, nnodes);

    kmutex_lock (&subsys->Lock); /* walking rails list and setting info on Rail */
    list_for_each (el, &subsys->Rails) { 
	EP_COMMS_RAIL *commsRail = list_entry (el, EP_COMMS_RAIL, Link);

	/* this will or in each bit map */
	cm_svc_indicator_bitmap (commsRail->Rail, svc_indicator, bitmap, low, nnodes);
    }
    kmutex_unlock (&subsys->Lock);

    return (0);
}

int
ep_xmtr_svc_indicator_bitmap (EP_XMTR *xmtr, int svc_indicator, bitmap_t * bitmap, int low, int nnodes) 
{
    int i;

    EPRINTF1 (DBG_SVC,"ep_xmtr_svc_indicator_bitmap: svc %d\n", svc_indicator);

    if (svc_indicator < 0 || svc_indicator >= EP_SVC_NUM_INDICATORS)
	return (-1);

    /* clear bitmap */
    bt_zero (bitmap, nnodes);

    for (i = 0; i < EP_MAX_RAILS; i++)
    {
	if (xmtr->RailMask & (1 << i) )
	{
	    /* this will or in each bit map */
	    cm_svc_indicator_bitmap (xmtr->Rails[i]->CommsRail->Rail, svc_indicator, bitmap, low, nnodes);
	}
    }

    return (0);
}

EP_RAILMASK
ep_svc_indicator_railmask (EP_SYS *epsys, int svc_indicator, int nodeId)
{
    EP_COMMS_SUBSYS  *subsys;
    struct list_head *el;
    EP_RAILMASK       rmask=0;

    if ((subsys = (EP_COMMS_SUBSYS *) ep_subsys_find (epsys, "epcomms")) == NULL)
	return (rmask);

    kmutex_lock (&subsys->Lock); /* walking rails list and reading info from Rail */
    list_for_each (el, &subsys->Rails) { 
	EP_COMMS_RAIL *commsRail = list_entry (el, EP_COMMS_RAIL, Link);

	if ( cm_svc_indicator_is_set(commsRail->Rail, svc_indicator,nodeId))
	     rmask |= EP_RAIL2RAILMASK(commsRail->Rail->Number);
    }
    kmutex_unlock (&subsys->Lock);

    return (rmask);
}

EP_RAILMASK
ep_xmtr_svc_indicator_railmask (EP_XMTR *xmtr, int svc_indicator, int nodeId)
{
    EP_RAILMASK    rmask=0;
    EP_COMMS_RAIL *commsRail;
    int            i;

    for (i = 0; i < EP_MAX_RAILS; i++)
    {
	if (xmtr->RailMask & (1 << i) )
	{
	    commsRail = xmtr->Rails[i]->CommsRail;

	    if ( cm_svc_indicator_is_set(commsRail->Rail, svc_indicator,nodeId))
		rmask |= EP_RAIL2RAILMASK(commsRail->Rail->Number);
	}
    }   
    
    EPRINTF3 (DBG_SVC, "ep_xmtr_svc_indicator_railmask: svc %d node %d mask 0x%x\n",  svc_indicator, nodeId, rmask);

    return (rmask);
}

EP_RAILMASK
ep_rcvr_railmask (EP_SYS *epsys, EP_SERVICE service)
{
    EP_COMMS_SUBSYS  *subsys;
    EP_RAILMASK       rmask=0;
    struct list_head *el;
    
    if ((subsys = (EP_COMMS_SUBSYS *) ep_subsys_find (epsys, "epcomms")) == NULL)
	return (rmask);
    
    kmutex_lock (&subsys->Lock);
    list_for_each (el, &subsys->Receivers) {
	EP_RCVR *rcvr = list_entry (el, EP_RCVR, Link);
 
	if (rcvr->Service == service)
	    rmask |= rcvr->RailMask; 
    }
    kmutex_unlock(&subsys->Lock);

    return (rmask);
}

#if ! defined(CONFIG_EP_NO_CHECK_SUM)
uint32_t
ep_calc_check_sum (EP_SYS *sys, EP_ENVELOPE *env, EP_NMD *nmd, int nFrags)
{
    EP_NMH   *nmh;
    int       i;
    uint16_t  check_data = 0;
    uint16_t  check_env  = 0;

    for (i = 0; i < nFrags; i++) {
	/* find the nmh for this frag */
	nmh = ep_nmh_find (&sys->MappingTable, &nmd[i]);

	ASSERT( nmh != NULL);

	/* add the next frag to the check sum */
	check_data = nmh->nmh_ops->op_calc_check_sum (sys, nmh, &nmd[i], check_data);
    }

    check_env = rolling_check_sum ((char *) env, offsetof(EP_ENVELOPE, CheckSum), 0);

    return (EP_ENVELOPE_CHECK_SUM | ( (check_env & 0x7FFF) << 16) | (check_data & 0xFFFF));
}
#endif

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
