/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: kcomm.c,v 1.61.2.2 2005/07/20 12:01:34 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/kcomm.c,v $ */

#include <qsnet/kernel.h>
#include <qsnet/kthread.h>

#include <elan/kcomm.h>
#include <elan/epsvc.h>
#include <elan/epcomms.h>

#include "cm.h"
#include "debug.h"

int MaxSwitchLevels = 5;				/* Max 1024 sized machine */

static char *NodeStateNames[EP_NODE_NUM_STATES] = 
{
    "Disconnected",
    "Connecting",
    "Connnected",
    "LeavingConnected",
    "LocalPassivate",
    "RemotePassivate",
    "Passivated",
    "Disconnecting",
};

static void
ep_xid_cache_fill (EP_SYS *sys, EP_XID_CACHE *cache)
{
    unsigned long flags;

    spin_lock_irqsave (&sys->XidLock, flags);

    cache->Current = sys->XidNext;
    cache->Last    = cache->Current + EP_XID_CACHE_CHUNKS-1;

    sys->XidNext += EP_XID_CACHE_CHUNKS;

    spin_unlock_irqrestore (&sys->XidLock, flags);
}

EP_XID
ep_xid_cache_alloc (EP_SYS *sys, EP_XID_CACHE *cache)
{
    EP_XID xid;
    
    if (cache->Current == cache->Last)
	ep_xid_cache_fill (sys, cache);

    xid.Generation = sys->XidGeneration;
    xid.Handle     = cache->Handle;
    xid.Unique     = cache->Current++;

    return (xid);
}

void
ep_xid_cache_init (EP_SYS *sys, EP_XID_CACHE *cache)
{
    /* Stall manager thread - it doesn't lock the XidCacheList */
    ep_kthread_stall (&sys->ManagerThread);

    cache->Handle = ++sys->XidHandle;

    list_add_tail (&cache->Link, &sys->XidCacheList);

    ep_kthread_resume (&sys->ManagerThread);
}

void
ep_xid_cache_destroy (EP_SYS *sys, EP_XID_CACHE *cache)
{
    /* Stall manager thread - it doesn't lock the XidCacheList */
    ep_kthread_stall (&sys->ManagerThread);

    list_del (&cache->Link);

    ep_kthread_resume (&sys->ManagerThread);
}

EP_XID_CACHE *
ep_xid_cache_find (EP_SYS *sys, EP_XID xid)
{
    struct list_head *el;

    list_for_each (el, &sys->XidCacheList) {
	EP_XID_CACHE *cache = list_entry (el, EP_XID_CACHE, Link);

	if (sys->XidGeneration == xid.Generation && cache->Handle == xid.Handle)
	    return (cache);
    }

    return (NULL);
}

static int
MsgBusy (EP_RAIL *rail, EP_OUTPUTQ *outputq, int slotNum)
{
    switch (rail->Operations.OutputQState (rail, outputq, slotNum))
    {
    case EP_OUTPUTQ_BUSY:			/* still busy */
	return 1;
	
    case EP_OUTPUTQ_FAILED:			/* NACKed */
    {
#if defined(DEBUG_PRINTF)
	EP_MANAGER_MSG *msg = rail->Operations.OutputQMsg (rail, outputq, slotNum);

	EPRINTF4 (DBG_MANAGER, "%s: kcomm msg %d type %d to %d failed\n", rail->Name, slotNum, msg->Hdr.Type, msg->Hdr.DestId);
#endif
	break;
    }
    
    case EP_OUTPUTQ_FINISHED:			/* anything else is finished */
	break;
    }

    return 0;
}

int
ep_send_message (EP_RAIL *rail, int nodeId, int type, EP_XID xid, EP_MANAGER_MSG_BODY *body)
{
    EP_SYS         *sys  = rail->System;
    EP_NODE        *node = &sys->Nodes[nodeId];
    int             n    = EP_MANAGER_OUTPUTQ_SLOTS;
    int             slotNum;
    int             rnum;
    EP_RAIL        *msgRail;
    EP_MANAGER_MSG *msg;
    unsigned long   flags;

    ASSERT (! EP_XID_INVALID (xid));

    if ((rnum = ep_pickRail (node->ConnectedRails)) >= 0)
	msgRail = sys->Rails[rnum];
    else
    {
	if (EP_MANAGER_MSG_TYPE_CONNECTED(type))
	{
	    ep_debugf (DBG_MANAGER, "%s: no rails available, trying to send type %d to %d\n", rail->Name, type, nodeId);
	    return -EHOSTDOWN;
	}

	ep_debugf (DBG_MANAGER, "%s: no rails connected to %d - using receiving rail\n", rail->Name, nodeId);

	msgRail = rail;
    }
    

    spin_lock_irqsave (&msgRail->ManagerOutputQLock, flags);

    slotNum = msgRail->ManagerOutputQNextSlot;

    while (n-- > 0 && MsgBusy (msgRail, msgRail->ManagerOutputQ, slotNum))		/* search for idle message buffer */
    {
	if (++(msgRail->ManagerOutputQNextSlot) == EP_MANAGER_OUTPUTQ_SLOTS)
	    msgRail->ManagerOutputQNextSlot = 0;
      
	slotNum = msgRail->ManagerOutputQNextSlot;
    }

    if (n == 0)							/* all message buffers busy */
    {
	spin_unlock_irqrestore (&msgRail->ManagerOutputQLock, flags);

	ep_debugf (DBG_MANAGER, "%s: all message buffers busy: trying to send type %d to %d\n", msgRail->Name, type, nodeId);
	return -EBUSY;
    }

    msg = msgRail->Operations.OutputQMsg (msgRail, msgRail->ManagerOutputQ, slotNum);
    
    EPRINTF7 (DBG_MANAGER, "%s: ep_send_message: type=%d nodeId=%d rail=%d xid=%08x.%08x.%016llx\n", 
	      msgRail->Name, type, nodeId, rail->Number, xid.Generation, xid.Handle, (long long) xid.Unique);

    msg->Hdr.Version    = EP_MANAGER_MSG_VERSION;
    msg->Hdr.Type       = type;
    msg->Hdr.Rail       = rail->Number;
    msg->Hdr.NodeId     = msgRail->Position.pos_nodeid;
    msg->Hdr.DestId     = nodeId;
    msg->Hdr.Xid        = xid;
    msg->Hdr.Checksum   = 0;

    if (body) bcopy (body, &msg->Body, sizeof (EP_MANAGER_MSG_BODY));

    msg->Hdr.Checksum = CheckSum ((char *) msg, EP_MANAGER_MSG_SIZE);

    if (msgRail->Operations.OutputQSend (msgRail, msgRail->ManagerOutputQ, slotNum, EP_MANAGER_MSG_SIZE,
				      nodeId, EP_SYSTEMQ_MANAGER, EP_MANAGER_OUTPUTQ_RETRIES) < 0)
	IncrStat (msgRail, SendMessageFailed);
    
    if (++(msgRail->ManagerOutputQNextSlot) == EP_MANAGER_OUTPUTQ_SLOTS) /* check this one last next time */
	msgRail->ManagerOutputQNextSlot = 0;

    spin_unlock_irqrestore (&msgRail->ManagerOutputQLock, flags);

    return 0;
}

void
ep_panic_node (EP_SYS *sys, int nodeId, unsigned char *reason)
{
    EP_NODE            *node = &sys->Nodes[nodeId];
    EP_MANAGER_MSG_BODY body;
    EP_XID              xid;
    kcondvar_t          sleep;
    int                 rnum;
    unsigned long       flags;

    if (nodeId > sys->Position.pos_nodes)
	return;

    strncpy (body.PanicReason, reason, sizeof (body.PanicReason));

    kcondvar_init (&sleep);
    spin_lock_irqsave (&sys->NodeLock, flags);
    for (;;)
    {
	if (node->ConnectedRails == 0)
	    break;

	for (rnum = 0; rnum < EP_MAX_RAILS; rnum++)
	    if (node->ConnectedRails & (1 << rnum))
		break;

	xid = ep_xid_cache_alloc(sys, &sys->Rails[rnum]->XidCache);
	
	if (ep_send_message (sys->Rails[rnum], nodeId, EP_MANAGER_MSG_TYPE_REMOTE_PANIC, xid, &body) == 0)
	    break;

	if (kcondvar_timedwaitsig (&sleep, &sys->NodeLock, &flags, lbolt + hz) == CV_RET_SIGPENDING)
	    break;
    }
    spin_unlock_irqrestore (&sys->NodeLock, flags);
    kcondvar_destroy (&sleep);
}

static void
ProcessNeterrRequest (EP_RAIL *msgRail, EP_RAIL *rail, EP_MANAGER_MSG *msg)
{
    EPRINTF4 (DBG_NETWORK_ERROR, "%s: process neterr request - node %d cookies %llx %llx\n", rail->Name, msg->Hdr.NodeId, (long long)msg->Body.Cookies[0], (long long)msg->Body.Cookies[1]);

    rail->Operations.NeterrFixup (rail, msg->Hdr.NodeId, msg->Body.Cookies);
    
    ep_send_message (rail, msg->Hdr.NodeId, EP_MANAGER_MSG_TYPE_NETERR_RESPONSE, msg->Hdr.Xid, &msg->Body);
}


static void
ProcessNeterrResponse (EP_RAIL *msgRail, EP_RAIL *rail, EP_MANAGER_MSG *msg)
{
    EP_SYS       *sys      = rail->System;
    EP_NODE_RAIL *nodeRail = &rail->Nodes[msg->Hdr.NodeId];
    unsigned long flags;

    EPRINTF4 (DBG_NETWORK_ERROR, "%s: process neterr response - node %d cookies %llx %llx\n", rail->Name, msg->Hdr.NodeId, (long long)msg->Body.Cookies[0], (long long)msg->Body.Cookies[1]);

    spin_lock_irqsave (&sys->NodeLock, flags);
    if (EP_XIDS_MATCH (nodeRail->MsgXid, msg->Hdr.Xid))
    {
	EP_INVALIDATE_XID (nodeRail->MsgXid);

	if (nodeRail->NetworkErrorCookies[0] != 0 && nodeRail->NetworkErrorCookies[0] == msg->Body.Cookies[0])
	    nodeRail->NetworkErrorCookies[0] = 0;

	if (nodeRail->NetworkErrorCookies[1] != 0 && nodeRail->NetworkErrorCookies[1] == msg->Body.Cookies[1])
	    nodeRail->NetworkErrorCookies[1] = 0;
	
	if (nodeRail->NetworkErrorCookies[0] == 0 && nodeRail->NetworkErrorCookies[1] == 0)
	    nodeRail->NetworkErrorState &= ~EP_NODE_NETERR_ATOMIC_PACKET;
    }
    spin_unlock_irqrestore (&sys->NodeLock, flags);
}


static void
ProcessGetNodeState (EP_RAIL *msgRail, EP_RAIL *rail, EP_MANAGER_MSG *msg)
{
    EP_NODE_RAIL *nodeRail = &rail->Nodes[msg->Hdr.NodeId];
    unsigned int service = msg->Body.Service;

    EPRINTF5 (DBG_MANAGER, "%s: ProcessGetNodeState: %s - %d %s%s\n",  msgRail->Name, rail->Name, msg->Hdr.NodeId,
	      NodeStateNames[nodeRail->State], nodeRail->NetworkErrorState ? " (NetworkError)" : "");

    msg->Body.NodeState.State             = nodeRail->State;
    msg->Body.NodeState.NetworkErrorState = nodeRail->NetworkErrorState;
    msg->Body.NodeState.Railmask          = ep_rcvr_railmask (rail->System, service);

    if (ep_send_message (rail, msg->Hdr.NodeId, EP_MANAGER_MSG_TYPE_GET_NODE_STATE_RESPONSE, msg->Hdr.Xid, &msg->Body) < 0)
	printk ("%s: get node state for %s[%d] - failed to send response\n", msgRail->Name, rail->Name, msg->Hdr.NodeId);
}

static void
ProcessFlushRequest (EP_RAIL *msgRail, EP_RAIL *rail, EP_MANAGER_MSG *msg)
{
    EP_NODE_RAIL *nodeRail = &rail->Nodes[msg->Hdr.NodeId];

    EPRINTF5 (DBG_MANAGER, "%s: ProcessFlushRequest: %s - %d %s%s\n",  msgRail->Name, rail->Name, msg->Hdr.NodeId,
	      NodeStateNames[nodeRail->State], nodeRail->NetworkErrorState ? " (NetworkError)" : "");

    switch (nodeRail->State)
    {
    case EP_NODE_REMOTE_PASSIVATE:
	nodeRail->NextRunTime = lbolt + MSGBUSY_RETRY_TIME;	/* retransmit our flush request quickly */
	EPRINTF3 (DBG_MANAGER, "%s: ProcessFlushRequest: NextRunTime -> %lx (%lx)\n", rail->Name, nodeRail->NextRunTime, lbolt);
	/* DROPTHROUGH */

    case EP_NODE_PASSIVATED:
    case EP_NODE_DISCONNECTED:
	if (nodeRail->NetworkErrorState != 0)
	    break;

	if (ep_send_message (rail, msg->Hdr.NodeId, EP_MANAGER_MSG_TYPE_FLUSH_RESPONSE, msg->Hdr.Xid, NULL) < 0)
	    printk ("%s: flush request for %s[%d] - failed to send response\n", msgRail->Name, rail->Name, msg->Hdr.NodeId);
	break;
	
    default:
	EPRINTF4 (DBG_MANAGER, "%s: flush request for %s[%d] - node not in approriate state - %s\n", msgRail->Name, rail->Name, msg->Hdr.NodeId, NodeStateNames[nodeRail->State]);
	break;
    }
}

static void
ProcessFlushResponse (EP_RAIL *msgRail, EP_RAIL *rail, EP_MANAGER_MSG *msg)
{
    EP_NODE_RAIL *nodeRail= &rail->Nodes[msg->Hdr.NodeId];

    EPRINTF5 (DBG_MANAGER, "%s: ProcessFlushResponse: %s - %d %s%s\n",  msgRail->Name, rail->Name, msg->Hdr.NodeId,
	      NodeStateNames[nodeRail->State], EP_XIDS_MATCH (nodeRail->MsgXid, msg->Hdr.Xid) ? " (XIDS match)" : "");

    if (nodeRail->State == EP_NODE_REMOTE_PASSIVATE && EP_XIDS_MATCH(nodeRail->MsgXid, msg->Hdr.Xid))
    {
	EP_INVALIDATE_XID (nodeRail->MsgXid);

	printk ("%s: flush response from %d - move to passivated list\n", rail->Name, msg->Hdr.NodeId);
	list_del (&nodeRail->Link);

	/* Node is now passivated - attempt to failover  messages */
	list_add_tail (&nodeRail->Link, &rail->PassivatedList);
	nodeRail->State = EP_NODE_PASSIVATED;
    }
    else
    {
	printk ("%s: flush response from %d - not passivating (%s) or XIDs mismatch (%llx %llx)\n", rail->Name, 
		msg->Hdr.NodeId, NodeStateNames[nodeRail->State], (long long) nodeRail->MsgXid.Unique, (long long) msg->Hdr.Xid.Unique);
    }
}

static void
ProcessMapNmdRequest (EP_RAIL *msgRail, EP_RAIL *rail, EP_MANAGER_MSG *msg)
{
    EP_SYS          *sys     = rail->System;
    EP_MAP_NMD_BODY *msgBody = &msg->Body.MapNmd;
    int              i;

    EPRINTF4 (DBG_MANAGER, "%s: Map NMD request from %d for %d NMDs to railmask %x\n", rail->Name, msg->Hdr.NodeId, msgBody->nFrags, msgBody->Railmask);
    
    for (i = 0; i < msgBody->nFrags; i++)
	ep_nmd_map_rails (sys, &msgBody->Nmd[i], msgBody->Railmask);
    
    /* Must flush TLBs before responding */
    for (i = 0; i < EP_MAX_RAILS; i++)
	if (sys->Rails[i] && sys->Rails[i]->TlbFlushRequired)
	    ep_perrail_dvma_sync (sys->Rails[i]);

    if (ep_send_message (rail, msg->Hdr.NodeId, EP_MANAGER_MSG_TYPE_MAP_NMD_RESPONSE, msg->Hdr.Xid, &msg->Body) < 0)
	printk ("%s: map nmd request for %s[%d] - failed to send response\n", msgRail->Name, rail->Name, msg->Hdr.NodeId);
}

static void
ProcessXidMessage (EP_RAIL *msgRail, EP_MANAGER_MSG *msg, EP_XID xid)
{
    EP_XID_CACHE *xidCache = ep_xid_cache_find (msgRail->System, xid);

    EPRINTF6 (DBG_MANAGER, "%s: ProcessXidMessage: XID=%08x.%0x8.%016llx -> %p(%p)\n",
	      msgRail->Name, xid.Generation, xid.Handle, (long long) xid.Unique,
	      xidCache  ? xidCache->MessageHandler : 0, xidCache  ? xidCache->Arg : 0);
    
    if (xidCache != NULL)
	xidCache->MessageHandler (xidCache->Arg, msg);
}

static void
ProcessMessage (EP_RAIL *msgRail, void *arg, void *msgbuf)
{
    EP_SYS         *sys  = msgRail->System;
    EP_MANAGER_MSG *msg  = (EP_MANAGER_MSG *) msgbuf;
    uint16_t        csum = msg->Hdr.Checksum;
    EP_RAIL        *rail;

    if (msg->Hdr.Version != EP_MANAGER_MSG_VERSION)
	return;

    msg->Hdr.Checksum= 0;
    if (CheckSum ((char *) msg, EP_MANAGER_MSG_SIZE) != csum)
    {
	printk ("%s: checksum failed on msg from %d (%d) (%x != %x) ?\n", msgRail->Name, msg->Hdr.NodeId, msg->Hdr.Type, csum, CheckSum ((char *) msg, EP_MANAGER_MSG_SIZE));
	return;
    }

    if ((rail = sys->Rails[msg->Hdr.Rail]) == NULL)
    {
	printk ("%s: rail no longer exists for msg from %d?\n", msgRail->Name, msg->Hdr.NodeId);
	return;
    }

    EPRINTF7 (DBG_MANAGER, "%s: ProcessMessage (%s) type=%d node=%d XID=%08x.%0x8.%016llx\n", 
	      msgRail->Name, rail->Name, msg->Hdr.Type, msg->Hdr.NodeId,
	      msg->Hdr.Xid.Generation, msg->Hdr.Xid.Handle, (long long)msg->Hdr.Xid.Unique);

    switch (msg->Hdr.Type)
    {
    case EP_MANAGER_MSG_TYPE_REMOTE_PANIC:
	msg->Body.PanicReason[EP_PANIC_STRLEN] = '\0';		/* ensure string terminated */

	printk ("%s: remote panic call from elan node %d - %s\n", msgRail->Name, msg->Hdr.NodeId, msg->Body.PanicReason);
	panic ("ep: remote panic request\n");
	break;

    case EP_MANAGER_MSG_TYPE_NETERR_REQUEST:
	ProcessNeterrRequest (msgRail, rail, msg);
	break;

    case EP_MANAGER_MSG_TYPE_NETERR_RESPONSE:
	ProcessNeterrResponse (msgRail, rail, msg);
	break;

    case EP_MANAGER_MSG_TYPE_FLUSH_REQUEST:
	ProcessFlushRequest (msgRail, rail, msg);
	break;

    case EP_MANAGER_MSG_TYPE_FLUSH_RESPONSE:
	ProcessFlushResponse (msgRail, rail, msg);
	break;

    case EP_MANAGER_MSG_TYPE_MAP_NMD_REQUEST:
	ProcessMapNmdRequest (msgRail, rail, msg);
	break;

    case EP_MANAGER_MSG_TYPE_MAP_NMD_RESPONSE:
	ProcessXidMessage (msgRail, msg, msg->Hdr.Xid);
	break;

    case EP_MANAGER_MSG_TYPE_FAILOVER_REQUEST:
	ProcessXidMessage (msgRail, msg, msg->Body.Failover.Xid);
	break;

    case EP_MANAGER_MSG_TYPE_FAILOVER_RESPONSE:
	ProcessXidMessage (msgRail, msg, msg->Hdr.Xid);
	break;
       
    case EP_MANAGER_MSG_TYPE_GET_NODE_STATE:
	ProcessGetNodeState (msgRail, rail, msg);
	break;

    case EP_MANAGER_MSG_TYPE_GET_NODE_STATE_RESPONSE: 
	ProcessXidMessage (msgRail, msg, msg->Hdr.Xid);
	break;

    default:
	printk ("%s: Unknown message type %d from %d\n", msgRail->Name, msg->Hdr.Type, msg->Hdr.NodeId);
	break;
    }
}


static void
ManagerQueueEvent (EP_RAIL *rail, void *arg)
{
    ep_kthread_schedule ((EP_KTHREAD *) arg, lbolt);
}

void
UpdateConnectionState (EP_RAIL *rail, statemap_t *map)
{
    EP_SYS *sys = rail->System;
    bitmap_t seg;
    int offset, nodeId;
    unsigned long flags;
    
    while ((offset = statemap_findchange (map, &seg, 1)) >= 0)
    {
	for (nodeId = offset; nodeId < (offset + BT_NBIPUL) && nodeId < rail->Position.pos_nodes; nodeId++)
	{
	    EP_NODE      *node     = &sys->Nodes[nodeId];
	    EP_NODE_RAIL *nodeRail = &rail->Nodes[nodeId];

	    if (statemap_getbits (map, nodeId, 1))
	    {
		spin_lock_irqsave (&sys->NodeLock, flags);

		switch (nodeRail->State)
		{
		case EP_NODE_DISCONNECTED:
		    EPRINTF2 (DBG_MANAGER, "%s: Node %d -> Disconnected \n", rail->Name, nodeId);
		    break;

		case EP_NODE_CONNECTING:
		    EPRINTF2 (DBG_MANAGER, "%s: Node %d -> Connect\n", rail->Name, nodeId);
		    
		    /* load the route table entry *before*  setting the state
		     * to connected, since DMA's can be initiated as soon as
		     * the node is marked as connected */
		    rail->Operations.LoadNodeRoute (rail, nodeId);
		    
		    nodeRail->State = EP_NODE_CONNECTED;
		    
		    statemap_setbits (rail->NodeSet, nodeId, 1, 1);
		    if (statemap_getbits (sys->NodeSet, nodeId, 1) == 0)
			statemap_setbits (sys->NodeSet, nodeId, 1, 1);

		    /* Add to rails connected to this node */
		    node->ConnectedRails |= (1 << rail->Number);

		    /* Finally lower the per-node context filter */
		    rail->Operations.LowerFilter (rail, nodeId);
		    break;
		    
		case EP_NODE_LEAVING_CONNECTED:
		    EPRINTF2 (DBG_MANAGER, "%s: Node %d -> Local Passivate\n", rail->Name, nodeId);
		    
		    /* Raise the per-node context filter */
		    rail->Operations.RaiseFilter (rail, nodeId);

		    /* If it's resolving network errors it will be on the NodeNeterrList,
		     * remove if from this list before placing it on the LocalPassivateList
		     * as we'll resolve the network error later in RemotePassivate */
		    if (nodeRail->NetworkErrorState)
			list_del (&nodeRail->Link);

		    list_add_tail (&nodeRail->Link, &rail->LocalPassivateList);
		    nodeRail->State = EP_NODE_LOCAL_PASSIVATE;

		    /* Remove from rails connected to this node */
		    node->ConnectedRails &= ~(1 << rail->Number);
		    break;

		default:
		    printk ("%s: Node %d - in NodeChangeMap with state %d\n", rail->Name, nodeId, nodeRail->State);
		    panic ("Node in NodeChangeMap with invalid state\n");
		    break;
		}
		spin_unlock_irqrestore (&sys->NodeLock, flags);
	    }
	}
    }
}

void
ProgressNetworkError (EP_RAIL *rail, EP_NODE_RAIL *nodeRail)
{
    EP_SYS             *sys    = rail->System;
    int                 nodeId = nodeRail - rail->Nodes;
    EP_MANAGER_MSG_BODY msg;

    ASSERT (nodeRail->State >= EP_NODE_CONNECTED && nodeRail->State <= EP_NODE_REMOTE_PASSIVATE);

    if (BEFORE (lbolt, nodeRail->NextRunTime))
	return;

    if (nodeRail->NetworkErrorState & EP_NODE_NETERR_DMA_PACKET)
	nodeRail->NetworkErrorState &= ~EP_NODE_NETERR_DMA_PACKET;
    
    if (nodeRail->NetworkErrorState & EP_NODE_NETERR_ATOMIC_PACKET)
    {
	if (EP_XID_INVALID (nodeRail->MsgXid))
	    nodeRail->MsgXid = ep_xid_cache_alloc (sys, &rail->XidCache);

	msg.Cookies[0] = nodeRail->NetworkErrorCookies[0];
	msg.Cookies[1] = nodeRail->NetworkErrorCookies[1];
	
	EPRINTF4 (DBG_NETWORK_ERROR, "%s: progress neterr - node %d cookies %llx %llx\n", rail->Name, nodeId, (long long)msg.Cookies[0], (long long)msg.Cookies[1]);

	if (ep_send_message (rail, nodeId, EP_MANAGER_MSG_TYPE_NETERR_REQUEST, nodeRail->MsgXid, &msg) == 0)
	    nodeRail->NextRunTime = lbolt + MESSAGE_RETRY_TIME;
	else
	    nodeRail->NextRunTime = lbolt + MSGBUSY_RETRY_TIME;
    }
}

long
ProgressNodeLists (EP_RAIL *rail, long nextRunTime)
{
    EP_SYS           *sys = rail->System;
    struct list_head *el, *nel;
    unsigned long flags;

    spin_lock_irqsave (&sys->NodeLock, flags);
    list_for_each_safe (el, nel, &rail->NetworkErrorList) {
	EP_NODE_RAIL *nodeRail = list_entry (el, EP_NODE_RAIL, Link);
	int           nodeId   = nodeRail - rail->Nodes;

	ProgressNetworkError (rail, nodeRail);

	if (nodeRail->NetworkErrorState == 0)
	{
	    EPRINTF2 (DBG_NETWORK_ERROR, "%s: lower context filter for node %d due to network error\n", rail->Name, nodeId);

	    rail->Operations.LowerFilter (rail, nodeId);

	    list_del (&nodeRail->Link);
	    continue;
	}
	
	if (nextRunTime == 0 || AFTER (nextRunTime, nodeRail->NextRunTime))
	    nextRunTime = nodeRail->NextRunTime;
    }
    spin_unlock_irqrestore (&sys->NodeLock, flags);

    if (! list_empty (&rail->LocalPassivateList))
    {
	EPRINTF1 (DBG_MANAGER, "%s: Locally Passivating Nodes\n", rail->Name);
	
	/* We have disconnected from some nodes or have left ourselves
	 * flush through all communications and determine whether we
	 * need to perform rail failover */
	rail->Operations.FlushFilters (rail);
	
	ep_call_callbacks (rail, EP_CB_FLUSH_FILTERING, rail->NodeSet);

	rail->Operations.FlushQueues (rail);

	ep_call_callbacks (rail, EP_CB_FLUSH_FLUSHING, rail->NodeSet);

	while (! list_empty (&rail->LocalPassivateList))
	{
	    EP_NODE_RAIL *nodeRail = list_entry (rail->LocalPassivateList.next, EP_NODE_RAIL, Link);
	    int           nodeId   = nodeRail - rail->Nodes;

	    list_del (&nodeRail->Link);

	    rail->Operations.UnloadNodeRoute (rail, nodeId);
	    
	    if (nodeRail->NetworkErrorState == 0 && nodeRail->MessageState == 0)
	    {
		EPRINTF2 (DBG_MANAGER, "%s: Node %d -> Disconnecting\n", rail->Name, nodeId);

		list_add_tail (&nodeRail->Link, &rail->DisconnectingList);
		nodeRail->State = EP_NODE_DISCONNECTING;
	    }
	    else
	    {
		EPRINTF2 (DBG_MANAGER, "%s: Node %d -> Remote Passivate\n", rail->Name, nodeId);

		list_add_tail (&nodeRail->Link, &rail->RemotePassivateList);
		nodeRail->State = EP_NODE_REMOTE_PASSIVATE;

		if (nodeRail->NetworkErrorState == 0)
		    nodeRail->NextRunTime = lbolt;
	    }
	}

	ep_call_callbacks (rail, EP_CB_PASSIVATED, rail->NodeSet);
    }

    list_for_each_safe (el, nel, &rail->RemotePassivateList) {
	EP_NODE_RAIL *nodeRail = list_entry (el, EP_NODE_RAIL, Link);
	int	      nodeId   = nodeRail - rail->Nodes;
	EP_NODE      *node     = &sys->Nodes[nodeId];

	if (node->ConnectedRails == 0)				/* no rails connected to this node (anymore) */
	{
	    /*  Remove from this list */
	    list_del (&nodeRail->Link);

	    EPRINTF2 (DBG_MANAGER, "%s: Node %d, no rails, Remote Passivate -> Disconnecting\n", rail->Name, nodeId);

	    /* transition towards disconnected */
	    list_add_tail (&nodeRail->Link, &rail->DisconnectingList);
	    nodeRail->State = EP_NODE_DISCONNECTING;
	    continue;
	}

	EPRINTF6 (DBG_MANAGER, "%s: Node %d - %s NetworkErrorState=%x NextRunTime=%lx (%lx)\n",
		  rail->Name, nodeId, NodeStateNames[nodeRail->State], nodeRail->NetworkErrorState,
		  nodeRail->NextRunTime, nextRunTime);

	if (nodeRail->NetworkErrorState)
	{
	    ProgressNetworkError (rail, nodeRail);
	}
	else if (! BEFORE (lbolt, nodeRail->NextRunTime))
	{
	    if (EP_XID_INVALID (nodeRail->MsgXid))
		nodeRail->MsgXid = ep_xid_cache_alloc (sys, &rail->XidCache);

	    if (ep_send_message (rail, nodeId, EP_MANAGER_MSG_TYPE_FLUSH_REQUEST, nodeRail->MsgXid, NULL) == 0)
		nodeRail->NextRunTime = lbolt + MESSAGE_RETRY_TIME;
	    else
		nodeRail->NextRunTime = lbolt + MSGBUSY_RETRY_TIME;
	}

	if (nextRunTime == 0 || AFTER (nextRunTime, nodeRail->NextRunTime))
	    nextRunTime = nodeRail->NextRunTime;
    }
    
    if (! list_empty (&rail->PassivatedList)) 
    {
	ep_call_callbacks (rail, EP_CB_FAILOVER, rail->NodeSet);

	list_for_each_safe (el, nel, &rail->PassivatedList) {
	    EP_NODE_RAIL *nodeRail = list_entry (rail->PassivatedList.next, EP_NODE_RAIL, Link);
	    int           nodeId   = nodeRail - rail->Nodes;
	    EP_NODE      *node     = &sys->Nodes[nodeId];

	    ASSERT (nodeRail->NetworkErrorState == 0);

	    if (node->ConnectedRails == 0)
	    {
		/*  Remove from this list */
		list_del (&nodeRail->Link);

		EPRINTF2 (DBG_MANAGER, "%s: Node %d, no rails, Passivated -> Disconnecting\n", rail->Name, nodeId);

		/* transition towards disconnected */
		list_add_tail (&nodeRail->Link, &rail->DisconnectingList);
		nodeRail->State = EP_NODE_DISCONNECTING;
		continue;
	    }
	    
	    EPRINTF6 (DBG_MANAGER, "%s: Node %d - %s NetworkErrorState=%x NextRunTime=%lx (%lx)\n",
		      rail->Name, nodeId, NodeStateNames[nodeRail->State], nodeRail->NetworkErrorState,
		      nodeRail->NextRunTime, nextRunTime);

	    if (nodeRail->MessageState == 0)
	    {
		EPRINTF2 (DBG_MANAGER, "%s: Node %d, no messages, Passivated -> Disconnecting\n", rail->Name,nodeId);

		list_del (&nodeRail->Link);
		list_add_tail (&nodeRail->Link, &rail->DisconnectingList);
		nodeRail->State = EP_NODE_DISCONNECTING;
		continue;
	    }

	    nodeRail->MessageState = 0;
	    nodeRail->NextRunTime  = lbolt + FAILOVER_RETRY_TIME;

	    if (nextRunTime == 0 || AFTER (nextRunTime, nodeRail->NextRunTime))
		nextRunTime = nodeRail->NextRunTime;
	}
    }

    if (! list_empty (&rail->DisconnectingList))
    {
	ep_call_callbacks (rail, EP_CB_DISCONNECTING, rail->NodeSet);

	while (! list_empty (&rail->DisconnectingList))
	{
	    EP_NODE_RAIL *nodeRail = list_entry (rail->DisconnectingList.next, EP_NODE_RAIL, Link);
	    int           nodeId   = nodeRail - rail->Nodes;
	    EP_NODE      *node     = &sys->Nodes[nodeId];

	    EPRINTF2 (DBG_MANAGER, "%s: Node %d, Disconnecting -> Disconnected\n", rail->Name, nodeId);

	    list_del (&nodeRail->Link);

	    rail->Operations.NodeDisconnected (rail, nodeId);

	    /* Clear the network error state */
	    nodeRail->NextRunTime            = 0;
	    nodeRail->NetworkErrorState      = 0;
	    nodeRail->NetworkErrorCookies[0] = 0;
	    nodeRail->NetworkErrorCookies[1] = 0;

	    /* Clear the message state */
	    nodeRail->MessageState = 0;

	    cm_node_disconnected (rail, nodeId);

	    nodeRail->State = EP_NODE_DISCONNECTED;
	    
	    statemap_setbits (rail->NodeSet, nodeId, 0, 1);

	    if (node->ConnectedRails == 0)
		statemap_setbits (sys->NodeSet, nodeId, 0, 1);
	}

	ep_call_callbacks (rail, EP_CB_DISCONNECTED, rail->NodeSet);
    }

    return (nextRunTime);
}

void
DisplayNodes (EP_RAIL *rail)
{
    EP_SYS *sys = rail->System;
    int i, state, count;
    unsigned long flags;

    spin_lock_irqsave (&sys->NodeLock, flags);

    for (state = 0; state < EP_NODE_NUM_STATES; state++)
    {
	for (count = i = 0; i < rail->Position.pos_nodes; i++)
	{
	    ASSERT (rail->Nodes[i].State < EP_NODE_NUM_STATES);

	    if (rail->Nodes[i].State == state)
		if (state != EP_NODE_DISCONNECTED)
		    printk ("%s %d", !count++ ? NodeStateNames[state] : "", i);
	}
	if (count)
	    printk ("%s (%d total)\n", state == EP_NODE_DISCONNECTED ? NodeStateNames[state] : "", count);
    }
    spin_unlock_irqrestore (&sys->NodeLock, flags);
}

static void
PositionFound (EP_RAIL *rail, ELAN_POSITION *pos)
{
    EP_SYS           *sys = rail->System;
    struct list_head *el;
    int i;

    /* only called from the ep_managage whilst rail->State == EP_RAIL_STATE_STARTED */
    ASSERT ( rail->State == EP_RAIL_STATE_STARTED );

#if defined(PER_CPU_TIMEOUT)
    /*
     * On Tru64 - if we're running in a "funnelled" thread, then we will be 
     * unable to start the per-cpu timeouts, so if we return then eventually
     * the ep_manager() thread will find the network position and we're
     * in control of our own destiny.
     */
    if (THREAD_IS_FUNNELED(current_thread()))
    {
	ep_kthread_schedule (&sys->ManagerThread, lbolt);
	return;
    }
#endif

    sprintf (rail->Name, "ep%d[%d]", rail->Number, pos->pos_nodeid);

    if (pos->pos_levels > MaxSwitchLevels)
    {
	for (i = 0; i < (pos->pos_levels - MaxSwitchLevels); i++)
	    pos->pos_nodes /= pos->pos_arity[i];

	for (i = 0; i < MaxSwitchLevels; i++)
	    pos->pos_arity[i] = pos->pos_arity[i + (pos->pos_levels - MaxSwitchLevels)];

	pos->pos_levels = MaxSwitchLevels;
	pos->pos_nodeid = pos->pos_nodeid % pos->pos_nodes;
				
	printk ("%s: limiting switch levels to %d\n", rail->Name, MaxSwitchLevels);
	printk ("%s: nodeid=%d level=%d numnodes=%d\n", rail->Name, pos->pos_nodeid, pos->pos_levels, pos->pos_nodes);

	sprintf (rail->Name, "ep%d[%d]", rail->Number, pos->pos_nodeid);
    }

    if (rail->Position.pos_mode != ELAN_POS_UNKNOWN && rail->Position.pos_nodeid != pos->pos_nodeid)
    {
	printk ("%s: NodeId has changed from %d to %d\n", rail->Name, rail->Position.pos_nodeid, pos->pos_nodeid);
	panic ("ep: PositionFound: NodeId has changed\n");
    }

    if (sys->Position.pos_mode != ELAN_POS_UNKNOWN && (sys->Position.pos_nodeid != pos->pos_nodeid || sys->Position.pos_nodes != pos->pos_nodes))
    {
	printk ("%s: position incompatible - disabling rail\n", rail->Name);
	rail->State = EP_RAIL_STATE_INCOMPATIBLE;
	return;
    }

    if (sys->Position.pos_mode == ELAN_POS_UNKNOWN)
    {
	sys->Position = *pos;
	sys->NodeSet  = statemap_create (pos->pos_nodes);
	KMEM_ZALLOC (sys->Nodes, EP_NODE *, pos->pos_nodes * sizeof (EP_NODE), 1);
    }

    rail->Position             = *pos;
    rail->SwitchBroadcastLevel = pos->pos_levels - 1;
    rail->State                = EP_RAIL_STATE_RUNNING;

    for (i = 0; i < pos->pos_levels; i++)
    {
	rail->SwitchProbeTick[i]   = lbolt;
	rail->SwitchLast[i].uplink = 4;
    }

    rail->Operations.PositionFound (rail, pos);

    INIT_LIST_HEAD (&rail->NetworkErrorList);
    INIT_LIST_HEAD (&rail->LocalPassivateList);
    INIT_LIST_HEAD (&rail->RemotePassivateList);
    INIT_LIST_HEAD (&rail->PassivatedList);
    INIT_LIST_HEAD (&rail->DisconnectingList);

    rail->NodeSet       = statemap_create (rail->Position.pos_nodes);
    rail->NodeChangeMap = statemap_create (rail->Position.pos_nodes);
    rail->NodeChangeTmp = statemap_create (rail->Position.pos_nodes);

    KMEM_ZALLOC (rail->Nodes, EP_NODE_RAIL *, rail->Position.pos_nodes * sizeof (EP_NODE_RAIL), 1);

    for (i = 0; i < rail->Position.pos_nodes; i++)
    {
	spin_lock_init (&rail->Nodes[i].CookieLock);

	INIT_LIST_HEAD (&rail->Nodes[i].StalledDmas);

	rail->Nodes[i].State = EP_NODE_DISCONNECTED;
    }

    /* Notify all subsystems that a new rail has been enabled */
    kmutex_lock (&sys->SubsysLock);
    list_for_each (el, &sys->Subsystems) { 
	EP_SUBSYS *subsys = list_entry (el, EP_SUBSYS, Link);

	if (subsys->AddRail)
	    subsys->AddRail (subsys, sys, rail);

	/* XXXX: what to do if the subsystem refused to add the rail ? */
    }
    kmutex_unlock (&sys->SubsysLock);

    /* Now enable the manager input queue */
    ep_enable_inputq (rail, rail->ManagerInputQ);
}

static void
ep_manager (void *arg)
{
    EP_SYS            *sys = (EP_SYS *) arg;
    struct list_head *el;
    ELAN_POSITION     pos;
    unsigned long     flags;

    kernel_thread_init ("ep_manager");
    kernel_thread_become_highpri();

    for (;;)
    {
	long nextRunTime = lbolt + MSEC2TICKS(CM_THREAD_SCHEDULE_TIMEOUT);

	list_for_each (el, &sys->ManagedRails) {
	    EP_RAIL *rail = list_entry (el, EP_RAIL, ManagerLink);

	    switch (rail->State)
	    {
	    case EP_RAIL_STATE_STARTED:
		if (ProbeNetwork (rail, &pos) == 0)
		{
		    PositionFound (rail, &pos);
		    break;
		}

		if (nextRunTime == 0 || AFTER (nextRunTime, lbolt + HZ))
		    nextRunTime = lbolt + HZ;
		break;

	    case EP_RAIL_STATE_RUNNING:
		if (ep_poll_inputq (rail, rail->ManagerInputQ, 100, ProcessMessage, rail) >= 100)
		    nextRunTime = lbolt;
		
		/* Handle any nodes which the cluster membership subsystem
		 * has indicated are to begin connecting or disconnecting */
		spin_lock_irqsave (&sys->NodeLock, flags);
		if (! statemap_changed (rail->NodeChangeMap))
		    spin_unlock_irqrestore (&sys->NodeLock, flags);
		else
		{
		    /*
		     * Take a copy of the statemap, and zero all entries so
		     * we only see new requests next time
		     */
		    statemap_copy (rail->NodeChangeTmp, rail->NodeChangeMap);
		    statemap_zero (rail->NodeChangeMap);
		    spin_unlock_irqrestore (&sys->NodeLock, flags);
		    
		    UpdateConnectionState (rail, rail->NodeChangeTmp);
		}

		nextRunTime = ProgressNodeLists (rail, nextRunTime);

		if (statemap_changed (rail->NodeSet))
		{
		    ep_call_callbacks (rail, EP_CB_NODESET, rail->NodeSet);

		    statemap_clearchanges (rail->NodeSet);
		}
		break;

	    case EP_RAIL_STATE_INCOMPATIBLE:
		break;
	    }
	}


	EPRINTF5 (DBG_MANAGER, "ep_manager: sleep now=%lx nextRunTime=%lx (%ld) [%lx (%ld)]\n",
		  lbolt, nextRunTime, nextRunTime ? nextRunTime - lbolt : 0, sys->ManagerThread.next_run,
		  sys->ManagerThread.next_run ? sys->ManagerThread.next_run - lbolt : 0);

	if (ep_kthread_sleep (&sys->ManagerThread, nextRunTime) < 0)
	    break;
    }

    ep_kthread_stopped (&sys->ManagerThread);
    kernel_thread_exit();
}

void
ep_connect_node (EP_RAIL *rail, int nodeId)
{
    EP_SYS       *sys  = rail->System;
    EP_NODE_RAIL *node = &rail->Nodes[nodeId];
    unsigned long flags;
  
    spin_lock_irqsave (&sys->NodeLock, flags);

    EPRINTF2 (DBG_MANAGER, "%s: ep_connect_node: nodeId %d\n", rail->Name, nodeId);

    ASSERT (node->State == EP_NODE_DISCONNECTED && statemap_getbits (rail->NodeChangeMap, nodeId, 1) == 0);
    
    node->State = EP_NODE_CONNECTING;

    statemap_setbits (rail->NodeChangeMap, nodeId, 1, 1);

    spin_unlock_irqrestore (&sys->NodeLock, flags);

    ep_kthread_schedule (&sys->ManagerThread, lbolt);
}

int
ep_disconnect_node (EP_RAIL *rail, int nodeId)
{
    EP_SYS       *sys  = rail->System;
    EP_NODE_RAIL *node = &rail->Nodes[nodeId];
    int	          state;
    unsigned long flags;
  
    spin_lock_irqsave (&sys->NodeLock, flags);
    
    EPRINTF3 (DBG_MANAGER, "%s: ep_disconnect_node: nodeId %d - %s\n", rail->Name, nodeId, NodeStateNames[node->State]);

    switch (state = node->State)
    {
    case EP_NODE_CONNECTING:
	statemap_setbits (rail->NodeChangeMap, nodeId, 0, 1);

	node->State = EP_NODE_DISCONNECTED;
	break;
	
    case EP_NODE_CONNECTED:
	statemap_setbits (rail->NodeChangeMap, nodeId, 1, 1);

	node->State = EP_NODE_LEAVING_CONNECTED;
	break;

    case EP_NODE_LEAVING_CONNECTED:
	/* no assert on NodeChangeMap as the map could have been taken but not acted on */
	break;
	
    default:
	ASSERT (statemap_getbits (rail->NodeChangeMap, nodeId, 1) == 0);
	break;
    }
    spin_unlock_irqrestore (&sys->NodeLock, flags);

    if (state == EP_NODE_CONNECTED)
	ep_kthread_schedule (&sys->ManagerThread, lbolt);

    return state;
}

int
ep_manager_add_rail (EP_SYS *sys, EP_RAIL *rail)
{
    if ((rail->ManagerOutputQ = ep_alloc_outputq (rail, EP_MANAGER_MSG_SIZE, EP_MANAGER_OUTPUTQ_SLOTS)) == NULL)
	return -ENOMEM;

    if ((rail->ManagerInputQ = ep_alloc_inputq (rail, EP_SYSTEMQ_MANAGER, EP_MANAGER_MSG_SIZE, EP_MANAGER_INPUTQ_SLOTS,
						 ManagerQueueEvent, &sys->ManagerThread)) == NULL)
    {
	ep_free_outputq (rail, rail->ManagerOutputQ);
	return -ENOMEM;
    }

    spin_lock_init (&rail->ManagerOutputQLock);

    ep_xid_cache_init (sys, &rail->XidCache);

    ep_kthread_stall (&sys->ManagerThread);
    list_add_tail (&rail->ManagerLink, &sys->ManagedRails);
    ep_kthread_resume (&sys->ManagerThread);

    return (0);
}

void
ep_manager_remove_rail (EP_SYS *sys, EP_RAIL *rail)
{
    if (rail->ManagerInputQ != NULL)
    {
	ep_kthread_stall (&sys->ManagerThread);
	list_del (&rail->ManagerLink);
	ep_kthread_resume (&sys->ManagerThread);

	ep_xid_cache_destroy (sys, &rail->XidCache);

	spin_lock_destroy (&rail->ManagerOutputQLock);

	ep_disable_inputq (rail, rail->ManagerInputQ);
	ep_free_inputq (rail, rail->ManagerInputQ);
	ep_free_outputq (rail, rail->ManagerOutputQ);
    }
}

int
ep_manager_init (EP_SYS *sys)
{
    INIT_LIST_HEAD (&sys->ManagedRails);

    ep_kthread_init (&sys->ManagerThread);

    if (kernel_thread_create (ep_manager, (void *) sys) == 0)
	return (ENOMEM);
    
    ep_kthread_started (&sys->ManagerThread);
    
    return (0);
}

void
ep_manager_fini (EP_SYS *sys)
{
    ep_kthread_stop (&sys->ManagerThread);
    ep_kthread_destroy (&sys->ManagerThread);
}

int
ep_sys_init (EP_SYS *sys)
{
    kmutex_init (&sys->SubsysLock);   
    kmutex_init (&sys->StartStopLock);
    spin_lock_init (&sys->NodeLock);

    INIT_LIST_HEAD (&sys->Subsystems);

    /* initialise the xid allocators */
    spin_lock_init (&sys->XidLock);
    INIT_LIST_HEAD (&sys->XidCacheList);

    /* initially don't know where we are in the network */
    sys->Position.pos_mode = ELAN_POS_UNKNOWN;

    /* initialise the network mapping descriptor hash tables */
    ep_nmh_init (&sys->MappingTable);

    /* intialise the shared allocators */
    ep_shared_alloc_init (sys);

    /* initialise the dvma space */
    ep_dvma_init (sys);

    /* intiialise the rail manager */
    ep_manager_init (sys);

    /* initialise all subsystems */
    cm_init (sys);
    ep_comms_init (sys);
    //ep_msgsys_init (sys);

    return (0);
}

void
ep_sys_fini (EP_SYS *sys)
{
    /* Destroy the subsystems in the reverse order to their creation */
    while (! list_empty (&sys->Subsystems))
    {
	EP_SUBSYS *subsys = list_entry (sys->Subsystems.prev, EP_SUBSYS, Link);

	list_del (&subsys->Link);
	
	subsys->Destroy (subsys, sys);
    }

    ep_manager_fini(sys);
    ep_dvma_fini (sys);
    ep_shared_alloc_fini (sys);

    ep_nmh_fini (&sys->MappingTable);

    if (sys->Position.pos_mode != ELAN_POS_UNKNOWN) {
	statemap_destroy (sys->NodeSet);
	KMEM_FREE(sys->Nodes, sys->Position.pos_nodes * sizeof (EP_NODE));
    }

    spin_lock_destroy (&sys->XidLock);

    spin_lock_destroy (&sys->NodeLock);
    kmutex_destroy (&sys->SubsysLock);
    kmutex_destroy (&sys->StartStopLock);
}

void
ep_shutdown (EP_SYS *sys)
{
    sys->Shutdown = 1;
}

int
ep_init_rail (EP_SYS *sys, EP_RAIL *rail)
{
    static int rnum;

    rail->System              = sys;
    rail->State               = EP_RAIL_STATE_UNINITIALISED;
    rail->Number              = rnum++;
    rail->Position.pos_mode   = ELAN_POS_UNKNOWN;
    rail->Position.pos_nodeid = ELAN_INVALID_NODE;

    rail->CallbackRegistered  = 0;

    sprintf (rail->Name, "ep%d", rail->Number);

    /* Initialise externally visible locks */
    kmutex_init (&rail->CallbackLock);

    ep_alloc_init (rail);

    sys->Rails[rail->Number] = rail;

    return 0;
}

void
ep_destroy_rail (EP_RAIL *rail)
{
    ASSERT (rail->State == EP_RAIL_STATE_UNINITIALISED);

    ep_alloc_fini (rail);

    kmutex_destroy (&rail->CallbackLock);

    rail->System->Rails[rail->Number] = NULL;

    rail->Operations.DestroyRail (rail);
}

/* We need to traverse the Subsystems lists backwards
 * but it's not defined in <linux/list.h> */
#define list_for_each_backwards(pos,list) \
	for (pos = (list)->prev; pos != (list); \
	     pos = (pos)->prev)

void
__ep_stop_rail (EP_RAIL *rail)
{
    /* called holding the sys->Lock */
    EP_SYS           *sys = rail->System;
    struct list_head *el;

    rail->Operations.StallRail (rail);

    /* Notify all subsystems that this rail is being stopped */
    if (rail->State == EP_RAIL_STATE_RUNNING)
    {
	kmutex_lock (&sys->SubsysLock);
	list_for_each_backwards (el, &sys->Subsystems) { 
	    EP_SUBSYS *subsys = list_entry (el, EP_SUBSYS, Link);
	    
	    if (subsys->RemoveRail)
		subsys->RemoveRail (subsys, sys, rail);
	}
	kmutex_unlock (&sys->SubsysLock);

	ep_manager_remove_rail (sys, rail);

	KMEM_FREE (rail->Nodes, rail->Position.pos_nodes * sizeof (EP_NODE_RAIL));

	statemap_destroy (rail->NodeChangeTmp);
	statemap_destroy (rail->NodeChangeMap);
	statemap_destroy (rail->NodeSet);
    }

    ep_dvma_remove_rail (sys, rail);
    ep_shared_alloc_remove_rail (sys, rail);

    rail->Operations.StopRail (rail);

    rail->State = EP_RAIL_STATE_UNINITIALISED;
}

void
ep_stop_rail (EP_RAIL *rail)
{
    EP_SYS *sys = rail->System;

    /* stall ep_manager                      */
    /* and remove the rail from the manaager */

    ep_kthread_stall (&sys->ManagerThread);
    if ( rail->State == EP_RAIL_STATE_STARTED ) 
        ep_manager_remove_rail (sys, rail);
    ep_kthread_resume (&sys->ManagerThread);

    __ep_stop_rail (rail);
}

int
ep_start_rail (EP_RAIL *rail)
{
    EP_SYS *sys = rail->System;

    ASSERT (rail->State == EP_RAIL_STATE_UNINITIALISED);

    if (rail->Operations.StartRail (rail) < 0)
	return -ENXIO;
    
    kmutex_lock (&sys->StartStopLock);
    /* Add this rail to the shared allocator */
    if (ep_shared_alloc_add_rail (rail->System, rail))
	goto failed;

    /* Add this rail to dvma kmap */
    if (ep_dvma_add_rail (rail->System, rail))
	goto failed;

    /* rail is now started */
    rail->State = EP_RAIL_STATE_STARTED;

    /* notify the rail manager of the new rail */
    if (ep_manager_add_rail (rail->System, rail))
	goto failed;

    kmutex_unlock (&sys->StartStopLock);
    return (ESUCCESS);

 failed:
    printk ("%s: start failed\n", rail->Name);
    kmutex_unlock (&sys->StartStopLock);
    __ep_stop_rail (rail);

    return (ENOMEM);   
}

void
ep_subsys_add (EP_SYS *sys, EP_SUBSYS *subsys)
{
    kmutex_lock (&sys->SubsysLock);
    list_add_tail (&subsys->Link, &sys->Subsystems);
    kmutex_unlock (&sys->SubsysLock);
}

void
ep_subsys_del (EP_SYS *sys, EP_SUBSYS *subsys)
{
    kmutex_lock (&sys->SubsysLock);
    list_del (&subsys->Link);
    kmutex_unlock (&sys->SubsysLock);
}

EP_SUBSYS *
ep_subsys_find (EP_SYS *sys, char *name)
{
    struct list_head *el;

    ASSERT ( !in_interrupt());

    kmutex_lock (&sys->SubsysLock); 
    list_for_each (el, &sys->Subsystems) {
	EP_SUBSYS *subsys = list_entry (el, EP_SUBSYS, Link);

	if (! strcmp (subsys->Name, name))
	{
	    kmutex_unlock (&sys->SubsysLock);
	    return (subsys);
	}
    }

    kmutex_unlock (&sys->SubsysLock);
    return (NULL);
}

int
ep_waitfor_nodeid (EP_SYS *sys)
{
    int i, printed = 0;
    kcondvar_t Sleep;
    spinlock_t Lock;

    kcondvar_init (&Sleep);
    spin_lock_init (&Lock);

#define TICKS_TO_WAIT	(10*hz)
#define TICKS_PER_LOOP	(hz/10)
    for (i = 0; sys->Position.pos_mode == ELAN_POS_UNKNOWN && i < TICKS_TO_WAIT; i += TICKS_PER_LOOP)
    {
	if (! printed++)
	    printk ("ep: waiting for network position to be found\n");

	spin_lock (&Lock);
	kcondvar_timedwait (&Sleep, &Lock, NULL, lbolt + TICKS_PER_LOOP);
	spin_unlock (&Lock);
    }

    if (sys->Position.pos_mode == ELAN_POS_UNKNOWN)
	printk ("ep: network position not found after waiting\n");
    else if (printed)
	printk ("ep: network position found at nodeid %d\n", sys->Position.pos_nodeid);

    spin_lock_destroy (&Lock);
    kcondvar_destroy (&Sleep);

    return (sys->Position.pos_mode == ELAN_POS_UNKNOWN ? ELAN_INVALID_NODE : sys->Position.pos_nodeid);
}

int
ep_nodeid (EP_SYS *sys)
{
    return (sys->Position.pos_mode == ELAN_POS_UNKNOWN ? ELAN_INVALID_NODE : sys->Position.pos_nodeid);
}

int
ep_numnodes (EP_SYS *sys)
{
    return (sys->Position.pos_nodes);
}

void
ep_fillout_stats(EP_RAIL *r, char *str) 
{
    sprintf(str+strlen(str),"SendMessageFailed %lu NeterrAtomicPacket %lu NeterrDmaPacket %lu \n", r->Stats.SendMessageFailed, r->Stats.NeterrAtomicPacket, r->Stats.NeterrDmaPacket);
    sprintf(str+strlen(str),"Rx %lu  %lu /sec\n",   GET_STAT_TOTAL(r->Stats,rx), GET_STAT_PER_SEC(r->Stats,rx) ); 
    sprintf(str+strlen(str),"MBytes %lu  %lu MB/sec\n", GET_STAT_TOTAL(r->Stats,rx_len)/ (1024*1024), GET_STAT_PER_SEC(r->Stats,rx_len) / (1024*1024)); 
    sprintf(str+strlen(str),"Tx %lu  %lu /sec\n",   GET_STAT_TOTAL(r->Stats,tx), GET_STAT_PER_SEC(r->Stats,tx) ); 
    sprintf(str+strlen(str),"MBytes %lu  %lu MB/sec\n", GET_STAT_TOTAL(r->Stats,tx_len)/ (1024*1024), GET_STAT_PER_SEC(r->Stats,tx_len) / (1024*1024)); 
}


/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
