/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: kmap.c,v 1.12 2004/12/14 10:19:23 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/kmap.c,v $*/

#include <qsnet/kernel.h>
#include <qsnet/kpte.h>

#include <elan/kcomm.h>

#include "debug.h"

#if defined(DIGITAL_UNIX)
#  define kernel_map			(first_task->map)
#  define vaddr_to_phys(map, addr)	(pmap_extract (vm_map_pmap ((vm_map_t) map), (unsigned long) addr))
#elif defined(LINUX)
#  define kernel_map			get_kern_mm()
#  define vaddr_to_phys(map, addr)	(kmem_to_phys(addr))
#elif defined(SOLARIS)
#  define kernel_map			&kas
#  define vaddr_to_phys(map,addr)	ptob(hat_getpfnum (((struct as *) map)->a_hat, (caddr_t) addr))
#endif

void
ep_perrail_kaddr_map (EP_RAIL *rail, EP_ADDR eaddr, virtaddr_t kaddr, unsigned long len,  unsigned int perm, int ep_attr)
{
    rail->Operations.KaddrMap (rail, eaddr, kaddr, len, perm, ep_attr);
}

void
ep_perrail_sdram_map (EP_RAIL *rail, EP_ADDR eaddr, sdramaddr_t saddr, unsigned long len, unsigned int perm, int ep_attr)
{
    rail->Operations.SdramMap (rail, eaddr, saddr, len, perm, ep_attr);
}

void
ep_perrail_unmap (EP_RAIL *rail, EP_ADDR eaddr, unsigned long len)
{
    rail->Operations.Unmap (rail, eaddr, len);
}

void
ep_perrail_dvma_sync (EP_RAIL *rail)
{
    if (rail->TlbFlushRequired)
    {
	rail->TlbFlushRequired = 0;

	rail->Operations.FlushTlb (rail);
    }
}


static int ep_dvma_map_rails (EP_SYS *sys, EP_NMH *nmh, EP_NMD *nmd, EP_RAILMASK mask);

#if ! defined(CONFIG_EP_NO_CHECK_SUM)
static uint16_t ep_dvma_calc_check_sum (EP_SYS *sys, EP_NMH *nmh, EP_NMD *nmd, uint16_t check_sum);
#endif

EP_NMH_OPS ep_dvma_nmh_ops = 
{
    ep_dvma_map_rails,
#if ! defined(CONFIG_EP_NO_CHECK_SUM)
    ep_dvma_calc_check_sum
#endif
};

extern void
ep_dvma_init (EP_SYS *sys)
{
    EP_DVMA_STATE *d = &sys->DvmaState;

    kmutex_init (&d->dvma_lock);
    
    INIT_LIST_HEAD (&d->dvma_handles);
    INIT_LIST_HEAD (&d->dvma_rails);

    d->dvma_rmap = ep_rmallocmap (EP_DVMA_RMAP_SIZE, "dvma_rmap", 1);

    ep_rmfree (d->dvma_rmap, EP_DVMA_TOP - EP_DVMA_BASE, EP_DVMA_BASE);
}

extern void
ep_dvma_fini (EP_SYS *sys)
{
    EP_DVMA_STATE *d = &sys->DvmaState;

    ep_rmfreemap (d->dvma_rmap);

    kmutex_destroy (&d->dvma_lock);
}
    
extern int
ep_dvma_add_rail (EP_SYS *sys, EP_RAIL *rail)
{
    EP_DVMA_STATE    *d = &sys->DvmaState;
    EP_RAIL_ENTRY    *l;
    struct list_head *el;

    KMEM_ZALLOC (l, EP_RAIL_ENTRY *, sizeof (EP_RAIL_ENTRY), 1);

    if (l == NULL)
	return (ENOMEM);

    kmutex_lock (&d->dvma_lock);

    l->Rail = rail;

    list_add_tail (&l->Link, &d->dvma_rails);

    list_for_each (el, &d->dvma_handles) {
	EP_DVMA_NMH *desc   = list_entry (el, EP_DVMA_NMH, dvma_link);
	int          npages = desc->dvma_nmh.nmh_nmd.nmd_len >> PAGESHIFT;

	desc->dvma_rails[rail->Number] = rail;
	desc->dvma_railmask |= ( 1 << rail->Number);

	desc->dvma_private[rail->Number] = rail->Operations.DvmaReserve (rail, desc->dvma_nmh.nmh_nmd.nmd_addr, npages);
    }

    kmutex_unlock (&d->dvma_lock);
    return (0);
}

extern void
ep_dvma_remove_rail (EP_SYS *sys, EP_RAIL *rail)
{
    EP_DVMA_STATE    *d = &sys->DvmaState;
    struct list_head *el;

    kmutex_lock (&d->dvma_lock);
    
    list_for_each (el, &d->dvma_handles) {
	EP_DVMA_NMH *desc   = list_entry (el, EP_DVMA_NMH, dvma_link);
	int          npages = desc->dvma_nmh.nmh_nmd.nmd_len >> PAGESHIFT;

	desc->dvma_rails[rail->Number] = NULL;
	desc->dvma_railmask &= ~(1 << rail->Number);

	rail->Operations.DvmaRelease (rail, desc->dvma_nmh.nmh_nmd.nmd_addr, npages, desc->dvma_private[rail->Number]);
    }

    list_for_each (el, &d->dvma_rails) {
	EP_RAIL_ENTRY *tmp = list_entry (el, EP_RAIL_ENTRY, Link);

	if (tmp->Rail == rail)
	{
	    list_del (el);

	    KMEM_FREE (tmp, sizeof (EP_RAIL_ENTRY));
	    break;
	}
    }
    kmutex_unlock (&d->dvma_lock);
}

EP_NMH *
ep_dvma_reserve (EP_SYS *sys, unsigned npages, unsigned perm)
{
    EP_DVMA_STATE    *d = &sys->DvmaState;
    EP_DVMA_NMH      *desc;
    EP_ADDR	      addr;
    struct list_head *el;
    int               i;

    KMEM_ZALLOC (desc, EP_DVMA_NMH *, offsetof (EP_DVMA_NMH, dvma_attrs[npages]), 1);
    
    if (desc == NULL)
	return (NULL);

    if ((addr = ep_rmalloc (d->dvma_rmap, npages << PAGESHIFT, 0)) == 0)
    {

	KMEM_FREE (desc, sizeof (EP_DVMA_NMH));
	return (NULL);
    }

    spin_lock_init (&desc->dvma_lock);

    desc->dvma_perm = perm;

    kmutex_lock (&d->dvma_lock);
    /* reserve the mapping resource */
    list_for_each (el, &d->dvma_rails) {
	EP_RAIL *rail = list_entry (el, EP_RAIL_ENTRY, Link)->Rail;

	EPRINTF4 (DBG_KMAP, "%s: ep_dvma_reserve desc=%p npages=%d rail=%p\n", rail->Name, desc, npages, rail);

	if ((desc->dvma_private[rail->Number] = rail->Operations.DvmaReserve (rail, addr, npages)) == NULL)
	{
	    printk ("%s: !!ep_dvma_reserve - rail->DvmaReserve failed\n", rail->Name);
	    goto failed;
	}

	desc->dvma_rails[rail->Number] = rail;
	desc->dvma_railmask |= (1 << rail->Number);
    }

    /* insert into the network mapping handle table */
    desc->dvma_nmh.nmh_nmd.nmd_addr = addr;
    desc->dvma_nmh.nmh_nmd.nmd_len  = npages << PAGESHIFT;
    desc->dvma_nmh.nmh_nmd.nmd_attr = EP_NMD_ATTR (sys->Position.pos_nodeid, 0);
    desc->dvma_nmh.nmh_ops          = &ep_dvma_nmh_ops;

    ep_nmh_insert (&sys->MappingTable, &desc->dvma_nmh);

    list_add (&desc->dvma_link, &d->dvma_handles);

    kmutex_unlock (&d->dvma_lock);

    return (&desc->dvma_nmh);

 failed:

    kmutex_unlock (&d->dvma_lock);

    for (i = 0; i < EP_MAX_RAILS; i++)
	if (desc->dvma_rails[i] != NULL)
	    desc->dvma_rails[i]->Operations.DvmaRelease (desc->dvma_rails[i], addr, npages, desc->dvma_private[i]);

    ep_rmfree (d->dvma_rmap, npages << PAGESHIFT, addr);

    KMEM_FREE (desc, sizeof (EP_DVMA_NMH));
    return (NULL);
}

void
ep_dvma_release (EP_SYS *sys, EP_NMH *nmh)
{
    EP_DVMA_STATE *d      = &sys->DvmaState;
    EP_DVMA_NMH   *desc   = (EP_DVMA_NMH *) nmh;
    EP_ADDR        addr   = nmh->nmh_nmd.nmd_addr;
    int            npages = nmh->nmh_nmd.nmd_len >> PAGESHIFT;
    EP_RAIL       *rail;
    int            i;

    kmutex_lock (&d->dvma_lock);

    list_del (&desc->dvma_link);
    
    ep_nmh_remove (&sys->MappingTable, nmh);

    for (i = 0; i < EP_MAX_RAILS; i++)
	if ((rail = desc->dvma_rails[i]) != NULL)
	    rail->Operations.DvmaRelease (rail, addr, npages, desc->dvma_private[i]);

    ep_rmfree (d->dvma_rmap, npages << PAGESHIFT, addr);

    KMEM_FREE (desc, offsetof (EP_DVMA_NMH, dvma_attrs[npages]));

    kmutex_unlock (&d->dvma_lock);
}

void
ep_dvma_load (EP_SYS *sys, void *map, caddr_t vaddr, unsigned len, EP_NMH *nmh, unsigned index, EP_RAILMASK *hints, EP_NMD *subset)
{
    EP_DVMA_NMH *desc = (EP_DVMA_NMH *) nmh;
    unsigned     offset = (unsigned long) vaddr & PAGEOFFSET;
    unsigned     npages = btopr (len + offset);
    EP_ADDR      addr   = nmh->nmh_nmd.nmd_addr + (index << PAGESHIFT);
    int		 rmask  = *hints;
    EP_RAIL     *rail;
    register int i, rnum;
    unsigned long flags;

    EPRINTF7 (DBG_KMAP, "ep_dvma_load: map=%p vaddr=%p len=%x nmh=%p(%x,%x) index=%d\n",
	      map, vaddr, len, nmh, nmh->nmh_nmd.nmd_addr, nmh->nmh_nmd.nmd_len, index);

    /* If no rail specified, then map into all rails */
    if (rmask == 0)
	rmask = desc->dvma_railmask;

    ASSERT ((index + npages) <= (nmh->nmh_nmd.nmd_len >> PAGESHIFT));

    /* If not map specified then use the kernel map */
    if (map == NULL)
	map = kernel_map;

    spin_lock_irqsave (&desc->dvma_lock, flags);
    /* Now map each of the specified pages (backwards) */

    vaddr = (vaddr - offset) + (npages-1)*PAGESIZE;
    for (i = npages-1; i >= 0; i--, vaddr -= PAGESIZE)
    {
	physaddr_t paddr = vaddr_to_phys (map, vaddr);
	
	for (rnum = 0; rnum < EP_MAX_RAILS; rnum++)
	{
	    if (! (rmask & (1 << rnum)) || (rail = desc->dvma_rails[rnum]) == NULL)
		rmask &= ~(1 << rnum);
	    else
	    {
		rail->Operations.DvmaSetPte (rail, desc->dvma_private[rnum], index + i, paddr, desc->dvma_perm);

		desc->dvma_attrs[index + i] |= (1 << rnum);
	    }
	}
    }

    for (rnum = 0; rnum < EP_MAX_RAILS; rnum++)
	if ((rmask & (1 << rnum)) && (rail = desc->dvma_rails[rnum]) != NULL)
	    rail->TlbFlushRequired = 1;

    spin_unlock_irqrestore (&desc->dvma_lock, flags);

    /* Construct the network mapping handle to be returned. */
    subset->nmd_addr = addr + offset;
    subset->nmd_len  = len;
    subset->nmd_attr = EP_NMD_ATTR(sys->Position.pos_nodeid, rmask);
}

void
ep_dvma_unload (EP_SYS *sys, EP_NMH *nmh, EP_NMD *nmd)
{
    EP_DVMA_NMH *desc = (EP_DVMA_NMH *) nmh;
    unsigned     offset = nmd->nmd_addr & PAGEOFFSET;
    unsigned     npages = btopr (nmd->nmd_len + offset);
    unsigned     index  = (nmd->nmd_addr - nmh->nmh_nmd.nmd_addr) >> PAGESHIFT;
    EP_RAIL     *rail;
    int          rnum;
    int          rmask;
    register int i;
    unsigned long flags;
    
    spin_lock_irqsave (&desc->dvma_lock, flags);

    /* compute which rails we need to unload on */
    for (rmask = 0, i = 0; i < npages; i++)
    {
	rmask |= desc->dvma_attrs[index + i];
	
	desc->dvma_attrs[index + i] = 0;
    }
    
    for (rnum = 0; rnum < EP_MAX_RAILS; rnum++)
	if ((rmask & (1 << rnum)) && (rail = desc->dvma_rails[rnum]) != NULL)
	    rail->Operations.DvmaUnload (rail, desc->dvma_private[rnum], index, npages);

    spin_unlock_irqrestore (&desc->dvma_lock, flags);
}

int
ep_dvma_map_rails (EP_SYS *sys, EP_NMH *nmh, EP_NMD *nmd, EP_RAILMASK mask)
{
    EP_DVMA_NMH *desc = (EP_DVMA_NMH *) nmh;
    unsigned     offset = nmd->nmd_addr & PAGEOFFSET;
    unsigned     npages = btopr (nmd->nmd_len + offset);
    unsigned     index  = (nmd->nmd_addr - nmh->nmh_nmd.nmd_addr) >> PAGESHIFT;
    int          r, rnum;
    register int i;
    unsigned long flags;

    spin_lock_irqsave (&desc->dvma_lock, flags);

    EPRINTF4 (DBG_KMAP, "ep_dvma_map_rails: nmd=%08x.%08x.%08x mask=%04x\n", nmd->nmd_addr, nmd->nmd_len, nmd->nmd_attr, mask);

    if ((mask &= desc->dvma_railmask) == 0)
    {
	printk ("ep_dvma_map_rails: no intersecting rails %04x.%04x\n", mask, desc->dvma_railmask);
	spin_unlock_irqrestore (&desc->dvma_lock, flags);
	return (-1);
    }

    for (i = npages-1; i >= 0; i--)
    {
	int pgidx = (index + i);

	for (rnum = 0; rnum < EP_MAX_RAILS; rnum++)
	    if (desc->dvma_attrs[pgidx] & (1 << rnum))
		break;
	
	if (rnum == EP_MAX_RAILS)
	{
	    EPRINTF3 (DBG_KMAP, "ep_dvma_map_rails: nmh=%p idx=%x [%08x] not ptes valid\n", nmh, pgidx, 
		      nmh->nmh_nmd.nmd_addr + ((pgidx) << PAGESHIFT));
	    mask = 0;
	}
	else
	{
	    EP_RAIL   *rail  = desc->dvma_rails[rnum];
	    physaddr_t paddr = rail->Operations.DvmaReadPte (rail, desc->dvma_private[rnum], pgidx);
	    
	    EPRINTF5 (DBG_KMAP, "%s: ep_dvma_map_rails: nmh=%p idx=%x [%08x] paddr %llx\n", rail->Name, nmh, pgidx,
		      nmh->nmh_nmd.nmd_addr + (pgidx << PAGESHIFT), (long long) paddr);
	    
	    for (r = 0; r < EP_MAX_RAILS; r++)
	    {
		if ((mask & (1 << r)) == 0)
		    continue;
		
		if ((desc->dvma_attrs[pgidx] & (1 << r)) == 0)
		{
		    EPRINTF5 (DBG_KMAP, "%s: ep_dvma_map_rails: nmh=%p idx=%x [%08x] paddr=%llx\n",
			      desc->dvma_rails[rnum]->Name, nmh, pgidx, nmh->nmh_nmd.nmd_addr + (pgidx << PAGESHIFT), 
			      (long long) paddr);
		    
		    rail->Operations.DvmaSetPte (rail, desc->dvma_private[rnum], pgidx, paddr, desc->dvma_perm);
		    
		    desc->dvma_attrs[pgidx] |= (1 << r);
		}
	    }
	}
    }

    for (rnum = 0; rnum < EP_MAX_RAILS; rnum++)
	if ((mask & (1 << rnum)) != 0)
	    desc->dvma_rails[rnum]->TlbFlushRequired = 1;

    EPRINTF4 (DBG_KMAP, "ep_dvma_map_rails: nmd=%08x.%08x.%08x|%04x\n", nmd->nmd_addr, nmd->nmd_len, nmd->nmd_attr, mask);

    /* Finally update the network memory descriptor */
    nmd->nmd_attr |= mask;

    spin_unlock_irqrestore (&desc->dvma_lock, flags);

    return (0);
}
#if ! defined(CONFIG_EP_NO_CHECK_SUM)
#include <linux/highmem.h>

/* Generic rolling checksum algorithm */
uint16_t
rolling_check_sum (char *msg, int nob, uint16_t sum)
{
    while (nob-- > 0)
	sum = sum * 13 + *msg++;

    return (sum);
}

#if ! defined(NO_RMAP)
void  
unmap_phys_address(unsigned long phys_addr)
{
    unsigned long pfn = (phys_addr >> PAGE_SHIFT);
    
    if (pfn_valid(pfn)) 
	kunmap(pfn_to_page(pfn));
}

void * 
map_phys_address(unsigned long phys_addr)
{
    unsigned long pfn = (phys_addr >> PAGE_SHIFT);
    
    if (pfn_valid(pfn)) 
	return  kmap(pfn_to_page(pfn));
    
    return NULL;
}
#else
void  
unmap_phys_address(unsigned long phys_addr)
{
    struct page *p = virt_to_page(__va(phys_addr));
    
    if (VALID_PAGE(p)) 
	kunmap(p);
}

void * 
map_phys_address(unsigned long phys_addr)
{
    struct page *p = virt_to_page(__va(phys_addr));
				
    if (VALID_PAGE(p)) 
	return  kmap(p);
    
    return NULL;
}
#endif

uint16_t
ep_dvma_calc_check_sum (EP_SYS *sys, EP_NMH *nmh, EP_NMD *nmd, uint16_t check_sum)
{
    /* cant be called from an interupt */

    EP_DVMA_NMH *desc = (EP_DVMA_NMH *) nmh;
    unsigned     offset = nmd->nmd_addr & PAGEOFFSET;
    unsigned     npages = btopr (nmd->nmd_len + offset);
    unsigned     index  = (nmd->nmd_addr - nmh->nmh_nmd.nmd_addr) >> PAGESHIFT;
    unsigned     start, len;
    int          rnum;
    register int i;
    unsigned long flags;
    EP_RAIL      *rail;


    spin_lock_irqsave (&desc->dvma_lock, flags);

    EPRINTF3 (DBG_KMAP, "ep_dvma_calc_check_sum: nmd=%08x.%08x.%08x \n", nmd->nmd_addr, nmd->nmd_len, nmd->nmd_attr);
 
    /* find a rail */
    for (rnum = 0; rnum < EP_MAX_RAILS; rnum++)
	if (desc->dvma_attrs[index] & (1 << rnum))
	    break;
	
    ASSERT (rnum != EP_MAX_RAILS);
 
    rail = desc->dvma_rails[rnum];

    for (i = 0; i <= (npages-1); i++)
    {
	int        pgidx = (index + i);
	physaddr_t paddr = rail->Operations.DvmaReadPte (rail, desc->dvma_private[rnum], pgidx);
	void *     virt;

	spin_unlock_irqrestore (&desc->dvma_lock, flags); /* unlock for check sum calc */

	virt = map_phys_address(paddr);

	if (!virt)
	    printk("ep_dvma_calc_check_sum: virt = NULL ! \n");
	else {
	    if ( i == 0 ) {
		/* last bit of the first page */
		start =  (nmd->nmd_addr & (PAGESIZE - 1)) ;
		len   =  PAGESIZE - start;
		if ( len > nmd->nmd_len) /* less than the remaining page */ 
		    len =  nmd->nmd_len;
	    } else {
		if ( i != (npages-1)) {
		    /* all of the middle pages    */
		    start = 0;
		    len   = PAGESIZE;
		} else {
		    /* first bit of the last page */
		    start = 0;
		    len   = ((nmd->nmd_addr + nmd->nmd_len -1) & (PAGESIZE -1)) +1;
		}
	    }

	    check_sum = rolling_check_sum (((char *)virt)+start, len, check_sum);
	    unmap_phys_address(paddr);
   
	    /* re aquire the lock */
	    spin_lock_irqsave (&desc->dvma_lock, flags);
	}
	
	EPRINTF5 (DBG_KMAP, "%s: ep_dvma_calc_check_sum: nmh=%p idx=%x [%08x] paddr %llx\n", rail->Name, nmh, pgidx,
		  nmh->nmh_nmd.nmd_addr + (pgidx << PAGESHIFT), (long long) paddr);    	
    }

    EPRINTF4 (DBG_KMAP, "ep_dvma_calc_check_sum: nmd=%08x.%08x.%08x = %d\n", nmd->nmd_addr, nmd->nmd_len, nmd->nmd_attr, check_sum);

    spin_unlock_irqrestore (&desc->dvma_lock, flags);

    return (check_sum);
}
#endif
/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
