/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: kmsg_elan4.c,v 1.10 2005/02/28 14:05:38 david Exp $"
/*      $Source: /cvs/master/quadrics/epmod/kmsg_elan4.c,v $ */

#include <qsnet/kernel.h>

#include <elan/kcomm.h>

#include "debug.h"
#include "kcomm_vp.h"
#include "kcomm_elan4.h"

#include <elan4/trtype.h>

static void
ep4_inputq_interrupt (EP4_RAIL *rail, void *arg)
{
    EP4_INPUTQ *inputq = (EP4_INPUTQ *) arg;

    /* mark the queue as "fired" to cause a single waitevent
     * to be issued next time the queue is polled */
    atomic_inc (&inputq->q_fired);
    
    (*inputq->q_callback)(&rail->r_generic, inputq->q_arg);
}

EP_INPUTQ *
ep4_alloc_inputq (EP_RAIL *r, unsigned qnum, unsigned slotSize, unsigned slotCount,
		  EP_INPUTQ_CALLBACK *callback, void *arg)
{
    EP4_RAIL     *rail = (EP4_RAIL *) r;
    EP4_INPUTQ   *inputq;
    E4_Event32    qevent;
    void         *slots;
    int           i;

    ASSERT ((slotSize & (EP_SYSTEMQ_MSG_ALIGN-1)) == 0);

    KMEM_ALLOC (inputq, EP4_INPUTQ *, sizeof (EP4_INPUTQ), 1);

    if (inputq == NULL)
	return (EP_INPUTQ *) NULL;
    
    if ((slots = ep_alloc_main (&rail->r_generic, slotSize * slotCount, 0, &inputq->q_slotsAddr)) == NULL)
    {
	KMEM_FREE (inputq, sizeof (EP4_INPUTQ));
	return (EP_INPUTQ *) NULL;
    }

    inputq->q_slotSize  = slotSize;
    inputq->q_slotCount = slotCount;
    inputq->q_callback  = callback;
    inputq->q_arg       = arg;
    inputq->q_slots     = slots;

    /* Initialise all the slots to be "unreceived" */
    for (i = 0; i < slotCount; i++)
	((uint32_t *) ((unsigned long) slots + (i+1) * slotSize))[-1] = EP_SYSTEMQ_UNRECEIVED;
    
    inputq->q_last      = inputq->q_slotsAddr + (slotCount-1) * slotSize;
    inputq->q_fptr      = inputq->q_slotsAddr;
    inputq->q_desc      = EP_SYSTEMQ_DESC (rail->r_queuedescs, qnum);
    inputq->q_descAddr  = EP_SYSTEMQ_ADDR (qnum);
    inputq->q_eventAddr = rail->r_elan_addr + offsetof (EP4_RAIL_ELAN, r_qevents[qnum]);

    if (callback)
    {
	if ((inputq->q_ecq = ep4_get_ecq (rail, EP4_ECQ_EVENT, 1)) == 0)
	{
	    ep_free_main (&rail->r_generic, inputq->q_slotsAddr, inputq->q_slotSize * inputq->q_slotCount);

	    KMEM_FREE (inputq, sizeof (EP4_INPUTQ));
	    return (EP_INPUTQ *) NULL;
	}

	if ((inputq->q_wcq = ep4_get_ecq (rail, EP4_ECQ_MAIN, 4)) == 0)
	{
	    ep4_put_ecq (rail, inputq->q_ecq, 1);
	    ep_free_main (&rail->r_generic, inputq->q_slotsAddr, inputq->q_slotSize * inputq->q_slotCount);

	    KMEM_FREE (inputq, sizeof (EP4_INPUTQ));
	    return (EP_INPUTQ *) NULL;
	}

	ep4_register_intcookie (rail, &inputq->q_intcookie, inputq->q_descAddr, ep4_inputq_interrupt, inputq);

	inputq->q_count = 0;

	atomic_set (&inputq->q_fired, 0);

	/* Initialise the queue event */
	qevent.ev_CountAndType = E4_EVENT_INIT_VALUE (callback ? -32 : 0, E4_EVENT_WRITE, E4_EVENT_DTYPE_LONG, 0);
	qevent.ev_WritePtr     = inputq->q_ecq->ecq_addr;
	qevent.ev_WriteValue   = (inputq->q_intcookie.int_val << E4_MAIN_INT_SHIFT) | INTERRUPT_CMD;
    }

    /* copy the event down to sdram */
    elan4_sdram_copyq_to_sdram (rail->r_ctxt.ctxt_dev, &qevent, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_qevents[qnum]), sizeof (E4_Event32));

    return (EP_INPUTQ *) inputq;
}

void
ep4_free_inputq (EP_RAIL *r, EP_INPUTQ *q)
{
    EP4_RAIL   *rail   = (EP4_RAIL *) r;
    EP4_INPUTQ *inputq = (EP4_INPUTQ *) q;

    ep_free_main (&rail->r_generic, inputq->q_slotsAddr, inputq->q_slotSize * inputq->q_slotCount);

    if (inputq->q_callback)
    {
	ep4_deregister_intcookie (rail, &inputq->q_intcookie);
	ep4_put_ecq (rail, inputq->q_ecq, 1);
	ep4_put_ecq (rail, inputq->q_wcq, 4);
    }

    KMEM_FREE (inputq, sizeof (EP4_INPUTQ));
}

void
ep4_enable_inputq (EP_RAIL *r, EP_INPUTQ *q)
{
    EP4_RAIL     *rail     = (EP4_RAIL *) r;
    EP4_INPUTQ   *inputq   = (EP4_INPUTQ *) q;
    EP_ADDR       lastSlot = inputq->q_slotsAddr + (inputq->q_slotCount-1) * inputq->q_slotSize;
    E4_InputQueue qdesc;

    qdesc.q_bptr    = inputq->q_slotsAddr;
    qdesc.q_fptr    = inputq->q_slotsAddr;
    qdesc.q_control = E4_InputQueueControl (inputq->q_slotsAddr, lastSlot, inputq->q_slotSize);
    qdesc.q_event   = inputq->q_callback ? inputq->q_eventAddr : 0;

    /* copy the queue descriptor down to sdram */
    ep4_write_qdesc (rail, inputq->q_desc, &qdesc);

    EPRINTF5 (DBG_KMSG,  "ep_enable_inputq: %x - %016llx %016llx %016llx %016llx\n", (int) inputq->q_descAddr,
	    elan4_sdram_readq (rail->r_ctxt.ctxt_dev, inputq->q_desc + 0),
	    elan4_sdram_readq (rail->r_ctxt.ctxt_dev, inputq->q_desc + 8),
	    elan4_sdram_readq (rail->r_ctxt.ctxt_dev, inputq->q_desc + 16),
	    elan4_sdram_readq (rail->r_ctxt.ctxt_dev, inputq->q_desc + 24));
}

void
ep4_disable_inputq (EP_RAIL *r, EP_INPUTQ *q)
{
    EP4_RAIL     *rail   = (EP4_RAIL *) r;
    EP4_INPUTQ   *inputq = (EP4_INPUTQ *) q;
    E4_InputQueue qdesc;

    /* Initialise the input queue descriptor as "full" with no event */
    qdesc.q_bptr    = 0;
    qdesc.q_fptr    = 8;
    qdesc.q_control = E4_InputQueueControl(qdesc.q_bptr, qdesc.q_fptr, 8);
    qdesc.q_event   = 0;

    /* copy the queue descriptor down to sdram */
    ep4_write_qdesc (rail, inputq->q_desc, &qdesc);
}

int
ep4_poll_inputq (EP_RAIL *r, EP_INPUTQ *q, int maxCount, EP_INPUTQ_HANDLER *handler, void *arg)
{
    EP4_RAIL   *rail   = (EP4_RAIL *) r;
    ELAN4_DEV  *dev    = rail->r_ctxt.ctxt_dev; 
    EP4_INPUTQ *inputq = (EP4_INPUTQ *) q;
    sdramaddr_t qdesc = inputq->q_desc;
    E4_Addr     fptr  = inputq->q_fptr;
    E4_Addr     bptr  = elan4_sdram_readl (dev, qdesc + offsetof (E4_InputQueue, q_bptr));
    int		count = 0;
    int         delay;

    while (bptr != 0 && fptr != bptr)
    {
	while (fptr != bptr)
	{
	    unsigned long slot = (unsigned long) inputq->q_slots + (fptr - inputq->q_slotsAddr);
	    
	    /* Poll the final word of the message until the message has completely
	     * arrived in main memory. */
	    for (delay = 1; ((uint32_t *) (slot + inputq->q_slotSize))[-1] == EP_SYSTEMQ_UNRECEIVED && delay < EP_SYSTEMQ_UNRECEIVED_TLIMIT; delay <<= 1)
		DELAY (delay);
	    
	    EPRINTF4(DBG_KMSG, "ep4_poll_inputq: %x slot %d of %d [%08x]\n", (int)inputq->q_descAddr,
		     ((int)(fptr - inputq->q_slotsAddr))/inputq->q_slotSize, 
		     inputq->q_slotCount, ((uint32_t *) (slot + inputq->q_slotSize))[-1]);
	    
	    /* Call the message handler */
	    (*handler) (r, arg, (void *) slot);
	    
	    /* reset the last word of the slot to "unreceived" */
	    ((uint32_t *) (slot + inputq->q_slotSize))[-1] = EP_SYSTEMQ_UNRECEIVED;
	    
	    /* move on the front pointer */
	    fptr = (fptr == inputq->q_last) ? inputq->q_slotsAddr : fptr + inputq->q_slotSize;
	    
	    elan4_sdram_writel (dev, qdesc + offsetof (E4_InputQueue, q_fptr), fptr);
	    
	    inputq->q_count++;
	    
	    if (++count >= maxCount && maxCount)
	    {
		inputq->q_fptr = fptr;

		return count;
	    }
	}

	bptr = elan4_sdram_readl (dev, qdesc + offsetof (E4_InputQueue, q_bptr));
    }

    inputq->q_fptr = fptr;

    /* Only insert a single wait event command if the callback has
     * occured, otherwise just acrue the count as we've just periodically
     * polled it.
     */
    if (inputq->q_callback && atomic_read (&inputq->q_fired))
    {
	atomic_dec (&inputq->q_fired);

	ep4_wait_event_cmd (inputq->q_wcq, inputq->q_eventAddr,
			    E4_EVENT_INIT_VALUE (-inputq->q_count << 5, E4_EVENT_WRITE, E4_EVENT_DTYPE_LONG, 0),
			    inputq->q_ecq->ecq_addr,
			    (inputq->q_intcookie.int_val << E4_MAIN_INT_SHIFT) | INTERRUPT_CMD);

	inputq->q_count = 0;
    }

    return count;
}

#define Q_MSG(q,slotNum)         (unsigned long)((q)->q_main      + (slotNum) * (q)->q_slotSize)
#define Q_MSG_ADDR(q,slotNum)			((q)->q_mainAddr  + (slotNum) * (q)->q_slotSize)
#define Q_DONE(q,slotNum)        *((E4_uint64 *)((q)->q_main      + (q)->q_slotCount * (q)->q_slotSize + (slotNum) * sizeof (E4_uint64)))
#define Q_DONE_ADDR(q,slotNum)			((q)->q_mainAddr  + (q)->q_slotCount * (q)->q_slotSize + (slotNum) * sizeof (E4_uint64))

#define Q_MAIN_SIZE(q)			((q)->q_slotCount * ((q)->q_slotSize + sizeof (E4_uint64)))

#define Q_DONE_VAL(val,cnt)		((cnt) << 16 | (val))
#define Q_DONE_RET(done)		((int) ((done) & 0xffff))
#define Q_DONE_CNT(done)		((int) ((done) >> 16))

EP_OUTPUTQ *
ep4_alloc_outputq (EP_RAIL *r, unsigned slotSize, unsigned slotCount)
{
    EP4_RAIL    *rail = (EP4_RAIL *) r;
    EP4_OUTPUTQ *outputq;
    int          i;

    ASSERT ((slotSize & (EP_SYSTEMQ_MSG_ALIGN-1)) == 0);

    KMEM_ALLOC (outputq, EP4_OUTPUTQ *, sizeof (EP4_OUTPUTQ), 1);

    if (outputq == NULL)
	return NULL;

    spin_lock_init (&outputq->q_lock);

    outputq->q_slotCount = slotCount;
    outputq->q_slotSize  = slotSize;
    outputq->q_main      = ep_alloc_main (r, Q_MAIN_SIZE(outputq), 0, &outputq->q_mainAddr);

    if (outputq->q_main == (E4_uint64 *) NULL)
    {
	KMEM_FREE (outputq, sizeof (EP_OUTPUTQ));
	return NULL;
    }

    outputq->q_cq = elan4_alloccq (&rail->r_ctxt, CQ_Size64K, CQ_STENEnableBit | CQ_WriteEnableBit, CQ_Priority);

    if (outputq->q_cq == (ELAN4_CQ *) NULL)
    {
	ep_free_main (&rail->r_generic, outputq->q_mainAddr, Q_MAIN_SIZE(outputq));

	KMEM_FREE (outputq, sizeof (EP_OUTPUTQ));
    }

    outputq->q_dwords = CQ_Size (outputq->q_cq->cq_size) >> 3;

    /* mark all the queue slots as finished */
    for (i = 0; i < slotCount; i++)
	Q_DONE(outputq, i) = Q_DONE_VAL (EP_OUTPUTQ_FINISHED, 0);

    return (EP_OUTPUTQ *) outputq;
}

void
ep4_free_outputq (EP_RAIL *r, EP_OUTPUTQ *q)
{
    EP4_RAIL    *rail    = (EP4_RAIL *) r;
    EP4_OUTPUTQ *outputq = (EP4_OUTPUTQ *) q;

    elan4_freecq (&rail->r_ctxt, outputq->q_cq);

    ep_free_main (&rail->r_generic, outputq->q_mainAddr, Q_MAIN_SIZE(outputq));

    spin_lock_destroy (&outputq->q_lock);

    KMEM_FREE (outputq, sizeof (EP4_OUTPUTQ));
}

void *
ep4_outputq_msg (EP_RAIL *r, EP_OUTPUTQ *q, unsigned slotNum)
{
    return (void *) Q_MSG ((EP4_OUTPUTQ *) q, slotNum);
}

int
ep4_outputq_state (EP_RAIL *r, EP_OUTPUTQ *q, unsigned slotNum)
{
    EPRINTF2 (DBG_KMSG, "ep4_outputq_state: slotNum %d state %x\n", slotNum, (int)Q_DONE((EP4_OUTPUTQ *) q, slotNum));

    return Q_DONE_RET(Q_DONE((EP4_OUTPUTQ *)q, slotNum));
}

int
ep4_outputq_send (EP_RAIL *r, EP_OUTPUTQ *q, unsigned slotNum, unsigned size, 
		  unsigned vp, unsigned qnum, unsigned retries)
{
    EP4_OUTPUTQ *outputq = (EP4_OUTPUTQ *) q;
    unsigned int nbytes  = roundup (size, 32);
    unsigned int base    = outputq->q_slotSize - nbytes;
    unsigned int i, dwords;
    unsigned long flags;
    E4_uint64 val;

    spin_lock_irqsave (&outputq->q_lock, flags);

    EPRINTF4 (DBG_KMSG, "ep4_outputq_send: slotNum=%d size=%d vp=%d qnum=%d\n", slotNum, size, vp, qnum);

    /* compute command queue size as follows - each slot uses
     *     overhead:	14 dwords +
     *	   data > 128 ?	36 dwords
     *     data > 64  ? 18 dwords
     *     data > 32  ? 10 dwords
     *     else         6  dwords
     */
    dwords = 14 + (size > 128 ? 36 :
		   size > 64  ? 18 :
		   size  ? 10 : 6);

    outputq->q_dwords += Q_DONE_CNT (Q_DONE(outputq, slotNum));

    if (dwords > outputq->q_dwords)
    {
	/* attempt to reclaim command queue space from other slots */
	i = slotNum;
	do {
	    if (++i == outputq->q_slotCount)
		i = 0;
	    
	    val = Q_DONE(outputq, i);

	    if ((Q_DONE_RET (val) == EP_OUTPUTQ_FINISHED || Q_DONE_RET (val) == EP_OUTPUTQ_FAILED) && Q_DONE_CNT(val) > 0)
	    {
		outputq->q_dwords += Q_DONE_CNT (val);

		Q_DONE(outputq, i) = Q_DONE_VAL(Q_DONE_RET(val), 0);
	    }
	} while (i != slotNum && dwords > outputq->q_dwords);
    }

    if (dwords > outputq->q_dwords)
    {
	spin_unlock_irqrestore (&outputq->q_lock, flags);
	
	EPRINTF0 (DBG_KMSG, "ep4_outputq_state: no command queue space\n");
	return 0;
    }

    outputq->q_dwords -= dwords;

    Q_DONE(outputq, slotNum) = Q_DONE_VAL (EP_OUTPUTQ_BUSY, dwords);

    if (outputq->q_retries != retries)
    {
	outputq->q_retries = retries;

	elan4_guard (outputq->q_cq, GUARD_CHANNEL(1) | GUARD_RESET(retries));
	elan4_nop_cmd (outputq->q_cq, 0);
    }

    /* transfer the top "size" bytes from message buffer to top of input queue */
    elan4_open_packet (outputq->q_cq, OPEN_PACKET (0, PACK_OK | RESTART_COUNT_ZERO, vp));
    elan4_sendtrans0 (outputq->q_cq, TR_INPUT_Q_GETINDEX, EP_SYSTEMQ_ADDR(qnum));

    /* send upto EP_SYSTEMQ_MSG_MAX (256) bytes of message to the top of the slot */
    if (size > 128)
    {
	elan4_sendtransp (outputq->q_cq, TR_WRITE (128 >> 3, 0, TR_DATATYPE_DWORD), base + 0,   (void *) (Q_MSG(outputq, slotNum) + base + 0));
	elan4_sendtransp (outputq->q_cq, TR_WRITE (128 >> 3, 0, TR_DATATYPE_DWORD), base + 128, (void *) (Q_MSG(outputq, slotNum) + base + 128));
    }
    else if (size > 64)
	elan4_sendtransp (outputq->q_cq, TR_WRITE (128 >> 3, 0, TR_DATATYPE_DWORD), base, (void *) (Q_MSG(outputq, slotNum) + base));
    else if (size > 32)
	elan4_sendtransp (outputq->q_cq, TR_WRITE (64 >> 3, 0, TR_DATATYPE_DWORD),  base, (void *) (Q_MSG(outputq, slotNum) + base));
    else
	elan4_sendtransp (outputq->q_cq, TR_WRITE (32 >> 3, 0, TR_DATATYPE_DWORD),  base, (void *) (Q_MSG(outputq, slotNum) + base));
    elan4_sendtrans1 (outputq->q_cq, TR_INPUT_Q_COMMIT, EP_SYSTEMQ_ADDR(qnum), 0 /* no cookie */);

    elan4_guard (outputq->q_cq, GUARD_CHANNEL (1) | GUARD_TEST(0, PACK_OK) | GUARD_RESET (outputq->q_retries));
    elan4_write_dword_cmd (outputq->q_cq, Q_DONE_ADDR(outputq, slotNum), Q_DONE_VAL (EP_OUTPUTQ_FINISHED, dwords));

    elan4_guard (outputq->q_cq, GUARD_CHANNEL (1) | GUARD_TEST(0, RESTART_COUNT_ZERO) | GUARD_RESET (outputq->q_retries));
    elan4_write_dword_cmd (outputq->q_cq, Q_DONE_ADDR(outputq, slotNum), Q_DONE_VAL (EP_OUTPUTQ_FAILED, dwords));

    spin_unlock_irqrestore (&outputq->q_lock, flags);

    return 1;
}
