/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: kthread.c,v 1.5 2004/05/19 08:54:57 david Exp $ $Name: QSNETMODULES-5-11-3_20050907 $"
/*      $Source: /cvs/master/quadrics/epmod/kthread.c,v $*/

#include <qsnet/kernel.h>

#include <elan/kthread.h>

void
ep_kthread_init (EP_KTHREAD *kt)
{
	spin_lock_init (&kt->lock);
	kcondvar_init (&kt->wait);
	
	kt->next_run     = 0;
	kt->should_stall = 0;
	kt->started      = 0;
	kt->should_stop  = 0;
	kt->stopped      = 0;
	kt->state        = KT_STATE_RUNNING;
}

void
ep_kthread_destroy (EP_KTHREAD *kt)
{
	spin_lock_destroy (&kt->lock);
	kcondvar_destroy (&kt->wait);
}

void
ep_kthread_started (EP_KTHREAD *kt)
{
	unsigned long flags;
	
	spin_lock_irqsave (&kt->lock, flags);
	kt->started = 1;
	spin_unlock_irqrestore(&kt->lock, flags);
}

void
ep_kthread_stopped (EP_KTHREAD *kt)
{
	unsigned long flags;
	
	spin_lock_irqsave (&kt->lock, flags);
	kt->stopped = 1;
	kcondvar_wakeupall (&kt->wait, &kt->lock);
	spin_unlock_irqrestore(&kt->lock, flags);
}

int
ep_kthread_should_stall (EP_KTHREAD *kth)
{
	return (kth->should_stall);
}

int
ep_kthread_sleep (EP_KTHREAD *kt, long next_run)
{
	unsigned long flags;

	spin_lock_irqsave (&kt->lock, flags);
	if (next_run && (kt->next_run == 0 || BEFORE (next_run, kt->next_run)))
		kt->next_run = next_run;

	if (kt->should_stop)
	{
		spin_unlock_irqrestore (&kt->lock, flags);
		return (-1);
	}
	
	do {
		if (kt->should_stall)
			kcondvar_wakeupall (&kt->wait, &kt->lock);

		kt->state = KT_STATE_SLEEPING;
		kt->running = 0;
		if (kt->should_stall || kt->next_run == 0)
			kcondvar_wait (&kt->wait, &kt->lock, &flags);
		else
			kcondvar_timedwait (&kt->wait,&kt->lock, &flags, kt->next_run);
		kt->state = KT_STATE_RUNNING;
		kt->running = lbolt;
	} while (kt->should_stall);
	kt->next_run = 0;
	spin_unlock_irqrestore (&kt->lock, flags);
	
	return (0);
}

void
ep_kthread_schedule (EP_KTHREAD *kt, long tick)
{
	unsigned long flags;
	
	spin_lock_irqsave (&kt->lock, flags);
	if (kt->next_run == 0 || BEFORE (tick, kt->next_run))
	{
		kt->next_run = tick;
		if (!kt->should_stall && kt->state == KT_STATE_SLEEPING)
		{
			kt->state = KT_STATE_SCHEDULED;
			kcondvar_wakeupone (&kt->wait, &kt->lock);
		}
	}
	spin_unlock_irqrestore (&kt->lock, flags);
}

void
ep_kthread_stall (EP_KTHREAD *kt)
{
	unsigned long flags;
	
	spin_lock_irqsave (&kt->lock, flags);
	if (kt->should_stall++ == 0)
		kcondvar_wakeupall (&kt->wait, &kt->lock);

	while (kt->state != KT_STATE_SLEEPING)
		kcondvar_wait (&kt->wait, &kt->lock, &flags);
	spin_unlock_irqrestore (&kt->lock, flags);
}

void
ep_kthread_resume (EP_KTHREAD *kt)
{
	unsigned long flags;

	spin_lock_irqsave (&kt->lock, flags);
	if (--kt->should_stall == 0)
	{
		kt->state = KT_STATE_SCHEDULED;
		kcondvar_wakeupone (&kt->wait, &kt->lock);
	}
	spin_unlock_irqrestore (&kt->lock, flags);
}

void
ep_kthread_stop (EP_KTHREAD *kt)
{
	unsigned long flags;
	
	spin_lock_irqsave (&kt->lock, flags);
	kt->should_stop = 1;
	while (kt->started && !kt->stopped)
	{
		kcondvar_wakeupall (&kt->wait, &kt->lock);
		kcondvar_wait (&kt->wait, &kt->lock, &flags);
	}
	spin_unlock_irqrestore (&kt->lock, flags);
}

int
ep_kthread_state (EP_KTHREAD *kt, long *time)
{
	unsigned long flags;
	int res = KT_STATE_SLEEPING;

	spin_lock_irqsave (&kt->lock, flags);

	if (kt->next_run) {
		*time = kt->next_run;
		res   = kt->should_stall ? KT_STATE_STALLED : KT_STATE_SCHEDULED;
	}

	if (kt->running) {
		*time = kt->running;
		res   = KT_STATE_RUNNING;
	}

	spin_unlock_irqrestore (&kt->lock, flags);
	
	return res;
}

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
