/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: neterr_elan4.c,v 1.3.2.1 2005/07/20 11:35:37 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/neterr_elan4.c,v $ */

#include <qsnet/kernel.h>

#include <elan/kcomm.h>

#include "kcomm_vp.h"
#include "kcomm_elan4.h"
#include "debug.h"

struct neterr_desc
{
    EP4_RAIL         *rail;
    unsigned int      nodeid;
    EP_NETERR_COOKIE *cookies;
    int		      done;
} ;

static int
dma_matches_cookie (EP4_RAIL *rail, E4_uint64 vproc, E4_uint64 cookie, unsigned int nodeId, EP_NETERR_COOKIE *cookies, const char *where)
{
    if ((EP_VP_ISDATA (vproc) && EP_VP_TO_NODE (vproc) == nodeId) && (cookie == cookies[0] || cookie == cookies[1]))
    {
	EPRINTF3 (DBG_NETWORK_ERROR, "%s: match cookie %016llx on %s\n", rail->r_generic.Name, (long long)cookie, where);

	return 1;
    }
    return 0;
}

static void
ep4_neterr_dma_flushop (ELAN4_DEV *dev, void *arg, int qfull)
{
    struct neterr_desc *desc  = (struct neterr_desc *) arg;
    EP4_RAIL           *rail  = desc->rail;
    E4_uint64           qptrs = read_reg64 (dev, DProcHighPriPtrs);
    E4_uint32           qsize = E4_QueueSize (E4_QueueSizeValue (qptrs));
    E4_uint32           qfptr = E4_QueueFrontPointer (qptrs);
    E4_uint32           qbptr = E4_QueueBackPointer (qptrs);
    E4_DProcQueueEntry  qentry;
    unsigned long       flags;

    while ((qfptr != qbptr) || qfull)
    {
	E4_uint64 cookie = elan4_sdram_readq (dev, qfptr + offsetof (E4_DProcQueueEntry, Desc.dma_cookie));
	E4_uint64 vproc  = elan4_sdram_readq (dev, qfptr + offsetof (E4_DProcQueueEntry, Desc.dma_vproc));

	if (dma_matches_cookie (rail, vproc, cookie, desc->nodeid, desc->cookies, "runq "))
	{
	    elan4_sdram_copyq_from_sdram (dev, qfptr, &qentry, sizeof (E4_DProcQueueEntry));

	    ep4_queue_dma_stalled (rail, &qentry.Desc);

	    /* Replace the dma with one which will "disappear" */
	    qentry.Desc.dma_typeSize = DMA_ShMemWrite | dev->dev_ctxt.ctxt_num;
	    qentry.Desc.dma_cookie   = 0;
	    qentry.Desc.dma_vproc    = 0;
	    qentry.Desc.dma_srcAddr  = 0;
	    qentry.Desc.dma_dstAddr  = 0;
	    qentry.Desc.dma_srcEvent = 0;
	    qentry.Desc.dma_dstEvent = 0;

	    elan4_sdram_copyq_to_sdram (dev, &qentry, qfptr, sizeof (E4_DProcQueueEntry));
	}
	
	qfptr = (qfptr & ~(qsize-1)) | ((qfptr + sizeof (E4_DProcQueueEntry)) & (qsize-1));
	qfull = 0;
    }

    spin_lock_irqsave (&rail->r_haltop_lock, flags);
    desc->done = 1;
    kcondvar_wakeupall (&rail->r_haltop_sleep, &rail->r_haltop_lock);
    spin_unlock_irqrestore (&rail->r_haltop_lock, flags);
}

static void
ep4_neterr_dma_haltop (ELAN4_DEV *dev, void *arg)
{
    struct neterr_desc *desc = (struct neterr_desc *) arg;

    elan4_queue_dma_flushop (dev, &desc->rail->r_flushop, 1);
}

void
ep4_neterr_fixup_dmas (EP4_RAIL *rail, unsigned int nodeId, EP_NETERR_COOKIE *cookies)
{
    EP_NODE_RAIL      *nodeRail = &rail->r_generic.Nodes[nodeId];
    struct neterr_desc desc;
    struct list_head   matchedList;
    struct list_head  *el, *nel;
    unsigned long      flags;
    register int       i;

    desc.rail    = rail;
    desc.nodeid  = nodeId;
    desc.cookies = cookies;
    desc.done    = 0;

    INIT_LIST_HEAD (&matchedList);

    /* First -  stall the retry thread, so that it will no longer restart
     *          any dma's from the retry list */
    ep_kthread_stall (&rail->r_retry_thread);
    
    /* Second - flush through all command queues targetted by events, thread etc */
    ep4_flush_ecqs (rail);
    
    /* Third - queue a halt operation to flush through all DMA's which are executing
     *         or on the run queues */
    kmutex_lock (&rail->r_haltop_mutex);
    
    rail->r_haltop.op_mask      = INT_DProcHalted;
    rail->r_haltop.op_function  = ep4_neterr_dma_haltop;
    rail->r_haltop.op_arg       = &desc;

    rail->r_flushop.op_function = ep4_neterr_dma_flushop;
    rail->r_flushop.op_arg      = &desc;
    
    elan4_queue_haltop (rail->r_ctxt.ctxt_dev, &rail->r_haltop);

    spin_lock_irqsave (&rail->r_haltop_lock, flags);
    while (! desc.done)
	kcondvar_wait (&rail->r_haltop_sleep, &rail->r_haltop_lock, &flags);
    spin_unlock_irqrestore (&rail->r_haltop_lock, flags);
    kmutex_unlock (&rail->r_haltop_mutex);

    /* Fourth - run down the dma retry lists and move all entries to the cancelled
     *          list.  Any dma's which were on the run queues have already been
     *          moved there */
    spin_lock_irqsave (&rail->r_dma_lock, flags);
    for (i = EP_RETRY_BASE; i < EP_NUM_RETRIES; i++)
    {
	list_for_each_safe (el,nel, &rail->r_dma_retrylist[i]) {
	    EP4_DMA_RETRY *retry    = list_entry (el, EP4_DMA_RETRY, retry_link);
	    
	    if (dma_matches_cookie (rail, retry->retry_dma.dma_vproc, retry->retry_dma.dma_cookie, nodeId, cookies, "retry"))
	    {
		/* remove from retry list */
		list_del (&retry->retry_link);
		
		/* add to list of dmas which matched */
		list_add_tail (&retry->retry_link, &matchedList);
	    }
	}
    }
    
    list_for_each_safe (el, nel, &nodeRail->StalledDmas) {
	EP4_DMA_RETRY *retry = list_entry (el, EP4_DMA_RETRY, retry_link);
	
	if (dma_matches_cookie (rail, retry->retry_dma.dma_vproc, retry->retry_dma.dma_cookie, nodeId, cookies, "stalled"))
	{
	    /* remove from retry list */
	    list_del (&retry->retry_link);
	    
	    /* add to list of dmas which matched */
	    list_add_tail (&retry->retry_link, &matchedList);
	}
    }
    spin_unlock_irqrestore (&rail->r_dma_lock, flags);
    
    /* Now "set" the source event of any put DMA#'s we can use the dma 
     * retry command queue as the retry thread is stalled */
    while (! list_empty (&matchedList))
    {
	EP4_DMA_RETRY *retry = list_entry (matchedList.next, EP4_DMA_RETRY, retry_link);
	
	list_del (&retry->retry_link);

	elan4_set_event_cmd (rail->r_dma_ecq->ecq_cq, retry->retry_dma.dma_srcEvent);

	spin_lock_irqsave (&rail->r_dma_lock, flags);
	list_add (&retry->retry_link, &rail->r_dma_freelist);
	spin_unlock_irqrestore (&rail->r_dma_lock, flags);
    }

    /* Flush through the command queues to ensure that all the setevents have executed */
    ep4_flush_ecqs (rail);

    /* Finally - allow the retry thread to run again */
    ep_kthread_resume (&rail->r_retry_thread);
}

void
ep4_add_neterr_ops (EP4_RAIL *rail, EP4_NETERR_OPS *ops)
{
    /* we're called from the ManagerThread, so no need to stall it */
    list_add_tail (&ops->op_link, &rail->r_neterr_ops);
}
void
ep4_remove_neterr_ops (EP4_RAIL *rail, EP4_NETERR_OPS *ops)
{
    EP_SYS *sys = rail->r_generic.System;

    ep_kthread_stall (&sys->ManagerThread);
    list_del (&ops->op_link);
    ep_kthread_resume (&sys->ManagerThread);
}

void
ep4_neterr_fixup_sten (EP4_RAIL *rail, unsigned int nodeId, EP_NETERR_COOKIE *cookies)
{
    struct list_head *el;

    /* First -  stall the retry thread, so that it will no longer restart
     *          any sten packets from the retry list */
    ep_kthread_stall (&rail->r_retry_thread);

    /* Second - flush through all command queues targetted by events, thread etc */
    ep4_flush_ecqs (rail);

    list_for_each (el, &rail->r_neterr_ops) {
	EP4_NETERR_OPS *op = list_entry (el, EP4_NETERR_OPS, op_link);

	(op->op_func) (rail, op->op_arg, nodeId, cookies);
    }

    /* Flush through the command queues to ensure that all the setevents have executed */
    ep4_flush_ecqs (rail);

    /* Finally - allow the retry thread to run again */
    ep_kthread_resume (&rail->r_retry_thread);
}

void
ep4_neterr_fixup (EP_RAIL *r, unsigned int nodeId, EP_NETERR_COOKIE *cookies)
{
    EP4_RAIL *rail = (EP4_RAIL *) r;

    /* network error cookies can come from the following :
     *
     *   DMA  engine
     *     if a DMA matches a network error cookie, then we just need to 
     *     execute the local setevent *before* returning.
     *
     *   STEN packet
     *     if the STEN packet was generated with as a WAIT_FOR_EOP
     *     and it's not present on the retry lists, then re-create
     *     it.
     *
     */
    EPRINTF4 (DBG_NETWORK_ERROR, "%s: ep4_neterr_fixup: node %d cookies <%lld%s%s%s%s> <%lld%s%s%s%s>\n",
	      rail->r_generic.Name, nodeId, EP4_COOKIE_STRING(cookies[0]), EP4_COOKIE_STRING(cookies[1]));

    if ((cookies[0] & EP4_COOKIE_DMA) || (cookies[1] & EP4_COOKIE_DMA))
	ep4_neterr_fixup_dmas (rail, nodeId, cookies);

    if ((cookies[0] & EP4_COOKIE_STEN) || (cookies[1] & EP4_COOKIE_STEN))
	ep4_neterr_fixup_sten (rail, nodeId, cookies);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */

