/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: procfs_linux.c,v 1.60.2.3 2005/09/07 14:35:03 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/procfs_linux.c,v $*/

#include <qsnet/kernel.h>
#include <qsnet/autoconf.h>

#include <elan/kcomm.h>
#include <elan/epsvc.h>
#include <elan/epcomms.h>

#include "cm.h"
#include "debug.h"
#include "conf_linux.h"
#include <qsnet/module.h>
#include <linux/wait.h>
#include <linux/poll.h>

#include <qsnet/procfs_linux.h>

struct proc_dir_entry *ep_procfs_root;
struct proc_dir_entry *ep_config_root;

/*
 * We provide a slightly "special" interface for /proc/elan/device%d/nodeset,
 * so that it can be included in a "poll" system call.  On each "read" on the
 * file, we generate a new nodeset if a) the previous one has been completely
 * read and b) if it has changed since it was generated.
 *
 * Unfortunately ... this doesn't allow "tail -f" to work, since this uses
 * fstat() on the fd, as we only hold the last nodeset string, we could not
 * handle the case where two processes were reading a different rates.
 * We could maybe have implemented this as a "sliding window", so that we 
 * add a new nodeset string, when it has changed and someone reads past 
 * end of the last one.   Then if someone read from before out "window"
 * we would produce "padding" data.  The problem with this, is that a 
 * simple "cat" on /proc/elan/device%d/nodeset will read the whole "file"
 * which will be mostly padding !
 *
 * Just to not that the purpose of this interface is:
 *    1) to allow cat /proc/elan/device%d/nodeset to show the current
 *       nodeset.
 *    2) to allow rms (or similar) to poll() on the file, and when the
 *       nodeset changes read a new one.
 *
 * so ... we don't bother solving the troublesome "tail -f" problem.
 */

typedef struct nodeset_private
{
    struct nodeset_private *pr_next;
    EP_RAIL                *pr_rail;
    unsigned		    pr_changed;
    char		   *pr_page;
    unsigned		    pr_off;
    unsigned		    pr_len;
} NODESET_PRIVATE;

NODESET_PRIVATE   *ep_nodeset_list;
wait_queue_head_t  ep_nodeset_wait;
spinlock_t         ep_nodeset_lock;

static int
proc_write_state(struct file *file, const char *buffer,
		 unsigned long count, void *data)
{
    EP_RAIL *rail = (EP_RAIL *) data;
    char    tmpbuf[128];
    int     res;

    if (count > sizeof (tmpbuf)-1)
	return (-EINVAL);
    
    MOD_INC_USE_COUNT;
    
    if (copy_from_user (tmpbuf, buffer, count))
	res = -EFAULT;
    else 
    {
	tmpbuf[count] = '\0'; 	

	if (tmpbuf[count-1] == '\n')
	    tmpbuf[count-1] = '\0';

	if (! strcmp (tmpbuf, "start") && rail->State == EP_RAIL_STATE_UNINITIALISED)
	    ep_start_rail (rail);
	
	if (! strcmp (tmpbuf, "stop") && rail->State > EP_RAIL_STATE_UNINITIALISED)
	    ep_stop_rail (rail);
	
	if (! strcmp (tmpbuf, "offline") && rail->State > EP_RAIL_STATE_UNINITIALISED)
	    cm_force_offline (rail, 1, CM_OFFLINE_PROCFS);

	if (! strcmp (tmpbuf, "online") && rail->State > EP_RAIL_STATE_UNINITIALISED)
	    cm_force_offline (rail, 0, CM_OFFLINE_PROCFS);

	if (! strncmp (tmpbuf, "restart=", 8) && rail->State == EP_RAIL_STATE_RUNNING)
	    cm_restart_node (rail, simple_strtol (tmpbuf + 8, NULL, 0));

	if (! strncmp (tmpbuf, "panic=", 6))
	    ep_panic_node (rail->System, simple_strtol(tmpbuf + 6, NULL, 0),
			   strchr (tmpbuf, ',') ? strchr(tmpbuf, ',') + 1 : "remote panic request");

	if (! strncmp (tmpbuf, "raise=", 6) && rail->State > EP_RAIL_STATE_UNINITIALISED)
	    rail->Operations.RaiseFilter (rail, simple_strtol (tmpbuf + 6, NULL, 0));

	if (! strncmp (tmpbuf, "lower=", 6) && rail->State > EP_RAIL_STATE_UNINITIALISED)
	    rail->Operations.LowerFilter (rail, simple_strtol (tmpbuf + 6, NULL, 0));
	
	res = count;
    }

    MOD_DEC_USE_COUNT;

    return (res);
}

static int
proc_read_state(char *page, char **start, off_t off,
		int count, int *eof, void *data)
{
    EP_RAIL *rail = (EP_RAIL *) data;
    int     len;

    switch (rail->State)
    {
    case EP_RAIL_STATE_UNINITIALISED:
	len = sprintf (page, "uninitialised\n");
	break;
    case EP_RAIL_STATE_STARTED:
	len = sprintf (page, "started\n");
	break;
    case EP_RAIL_STATE_RUNNING:
	len = sprintf (page, "running NodeId=%d NumNodes=%d\n", rail->Position.pos_nodeid, rail->Position.pos_nodes);
	break;
    case EP_RAIL_STATE_INCOMPATIBLE:
	len = sprintf (page, "incompatible NodeId=%d NumNodes=%d\n", rail->Position.pos_nodeid, rail->Position.pos_nodes);
	break;
    default:
	len = sprintf (page, "<unknown>\n");
	break;
    }

    return (qsnet_proc_calc_metrics (page, start, off, count, eof, len));
}

static int
proc_write_display(struct file *file, const char *buffer,
		   unsigned long count, void *data)
{
    EP_RAIL *rail = (EP_RAIL *) data;
    char    tmpbuf[128];
    int     res;

    if (count > sizeof (tmpbuf)-1)
	return (-EINVAL);
    
    MOD_INC_USE_COUNT;
    
    if (copy_from_user (tmpbuf, buffer, count))
	res = -EFAULT;
    else 
    {
	tmpbuf[count] = '\0'; 	

	if (tmpbuf[count-1] == '\n')
	    tmpbuf[count-1] = '\0';

	if (! strcmp (tmpbuf, "rail"))
	    DisplayRail (rail);
	if (! strcmp (tmpbuf, "segs"))
	    DisplaySegs (rail);
	if (! strcmp (tmpbuf, "nodes"))
	    DisplayNodes (rail);
	if (! strcmp (tmpbuf, "status"))
	    DisplayStatus (rail);
	if (! strcmp (tmpbuf, "debug") && rail->Operations.Debug)
	    rail->Operations.Debug (rail);
	if (! strncmp (tmpbuf, "epcomms", 7))
	    ep_comms_display (rail->System, tmpbuf[7] == '=' ? tmpbuf + 8 : NULL);
	res = count;
    }

    MOD_DEC_USE_COUNT;

    return (res);
}

static int
proc_read_display(char *page, char **start, off_t off,
		  int count, int *eof, void *data)
{
    int len = sprintf (page, "<unreadable>\n");
    
    return (qsnet_proc_calc_metrics (page, start, off, count, eof, len));
}


static int
proc_read_stats(char *page, char **start, off_t off,
		int count, int *eof, void *data)
{
    EP_RAIL *rail = (EP_RAIL *) data;

    if ( rail == NULL ) {
	strcpy(page,"proc_read_stats rail=NULL\n");
    } else {
	page[0] = 0;
	ep_fillout_stats(rail, page);
	rail->Operations.FillOutStats (rail, page);
    }
    return (qsnet_proc_calc_metrics (page, start, off, count, eof, strlen(page)));
}

static int
proc_read_devinfo(char *page, char **start, off_t off,
		  int count, int *eof, void *data)
{
    EP_RAIL       *rail    = (EP_RAIL *) data;
    ELAN_DEVINFO  *devinfo = &rail->Devinfo;
    ELAN_POSITION *pos     = &rail->Position;
    char          *p       = page;
    
    switch (devinfo->dev_device_id)
    {
    case PCI_DEVICE_ID_ELAN3:
	p += sprintf (p, "ep%d is elan3 %d rev %c\n", rail->Number, 
		      devinfo->dev_instance, 'a' + devinfo->dev_revision_id);
	break;
	
    case PCI_DEVICE_ID_ELAN4:
	p += sprintf (p, "ep%d is elan4 %d rev %c\n", rail->Number, 
		      devinfo->dev_instance, 'a' + devinfo->dev_revision_id);
	break;
    default:
	p += sprintf (p, "ep%d is unknown %x/%x\n", rail->Number, devinfo->dev_vendor_id, devinfo->dev_device_id);
	break;
    }

    if (rail->State == EP_RAIL_STATE_RUNNING)
	p += sprintf (p, "ep%d nodeid %d numnodes %d\n", rail->Number, pos->pos_nodeid, pos->pos_nodes);

    return (qsnet_proc_calc_metrics (page, start, off, count, eof, p - page));
}

static struct rail_info
{
    char *name;
    int (*read_func) (char *page, char **start, off_t off, int count, int *eof, void *data);
    int (*write_func) (struct file *file, const char *buf, unsigned long count, void *data);
} rail_info[] = {
    {"state",   proc_read_state,   proc_write_state},
    {"display", proc_read_display, proc_write_display},
    {"stats",   proc_read_stats,   NULL},
    {"devinfo", proc_read_devinfo, NULL},
};

static int
nodeset_open (struct inode *inode, struct file *file)
{
    NODESET_PRIVATE *pr;

    if ((pr = kmalloc (sizeof (NODESET_PRIVATE), GFP_KERNEL)) == NULL)
	return (-ENOMEM);
    
    pr->pr_changed = 1;
    pr->pr_off     = 0;
    pr->pr_len     = 0;
    pr->pr_page    = NULL;
    pr->pr_rail    = (EP_RAIL *)( PDE(inode)->data );

    spin_lock (&ep_nodeset_lock);
    pr->pr_next = ep_nodeset_list;
    ep_nodeset_list = pr;
    spin_unlock (&ep_nodeset_lock);

    file->private_data = (void *) pr;

    MOD_INC_USE_COUNT;
    return (0);
}

static int
nodeset_release (struct inode *inode, struct file *file)
{
    NODESET_PRIVATE *pr = (NODESET_PRIVATE *) file->private_data;
    NODESET_PRIVATE **ppr;

    spin_lock (&ep_nodeset_lock);
    for (ppr = &ep_nodeset_list; (*ppr) != pr; ppr = &(*ppr)->pr_next)
	;
    (*ppr) = pr->pr_next;
    spin_unlock (&ep_nodeset_lock);

    if (pr->pr_page)
	free_page ((unsigned long) pr->pr_page);
    kfree (pr);
    
    MOD_DEC_USE_COUNT;
    return (0);
}

static ssize_t
nodeset_read (struct file *file, char *buf, size_t count, loff_t *ppos)
{
    NODESET_PRIVATE *pr  = (NODESET_PRIVATE *) file->private_data;
    EP_RAIL          *rail = pr->pr_rail;
    int              error;
    unsigned long    flags;

    if (!pr->pr_changed && pr->pr_off >= pr->pr_len)
	return (0);

    if ((error = verify_area (VERIFY_WRITE, buf, count)) != 0)
	return (error);

    if (pr->pr_page == NULL && (pr->pr_page = (char *) __get_free_page (GFP_KERNEL)) == NULL)
	return (-ENOMEM);

    if (pr->pr_off >= pr->pr_len)
    {
	kmutex_lock (&rail->CallbackLock);
	if (rail->State == EP_RAIL_STATE_RUNNING)
	{
	    spin_lock_irqsave (&rail->System->NodeLock, flags);
	    ep_sprintf_bitmap (pr->pr_page, PAGESIZE, statemap_tobitmap(rail->NodeSet), 0, 0, rail->Position.pos_nodes);
	    spin_unlock_irqrestore (&rail->System->NodeLock, flags);

	    if (rail->SwitchBroadcastLevel == -1)
		strcat (pr->pr_page, "<disconnected>");
	    else if (rail->SwitchBroadcastLevel < (rail->Position.pos_levels-1))
		sprintf (pr->pr_page + strlen (pr->pr_page), "<%d>", rail->SwitchBroadcastLevel);
	    strcat (pr->pr_page, "\n");
	}
	else
	    strcpy (pr->pr_page, "<not running>\n");
	kmutex_unlock (&rail->CallbackLock);

	pr->pr_len     = strlen (pr->pr_page);
	pr->pr_off     = 0;
	pr->pr_changed = 0;
    }

    if (count >= (pr->pr_len - pr->pr_off))
	count = pr->pr_len - pr->pr_off;

    copy_to_user (buf, pr->pr_page + pr->pr_off, count);

    pr->pr_off += count;
    *ppos      += count;

    if (pr->pr_off >= pr->pr_len)
    {
	free_page ((unsigned long) pr->pr_page);
	pr->pr_page = NULL;
    }

    return (count);
}

static unsigned int
nodeset_poll (struct file *file, poll_table *wait)
{
    NODESET_PRIVATE *pr = (NODESET_PRIVATE *) file->private_data;

    poll_wait (file, &ep_nodeset_wait, wait);
    if (pr->pr_changed || pr->pr_off < pr->pr_len)
	return (POLLIN | POLLRDNORM);
    return (0);
}

static void 
nodeset_callback (void *arg, statemap_t *map)
{
    EP_RAIL         *rail = (EP_RAIL *) arg;
    NODESET_PRIVATE *pr;

    ep_display_bitmap (rail->Name, "Nodeset", statemap_tobitmap(map), 0, ep_numnodes(rail->System));

    spin_lock (&ep_nodeset_lock);
    for (pr = ep_nodeset_list; pr; pr = pr->pr_next)
	if (pr->pr_rail == rail)
	    pr->pr_changed = 1;
    spin_unlock (&ep_nodeset_lock);

    wake_up_interruptible (&ep_nodeset_wait);
}

static int
proc_open (struct inode *inode, struct file *file)
{
    QSNET_PROC_PRIVATE *pr;
    CM_RAIL            *cmRail;
    EP_RAIL            *epRail;
    int                 pages = 4;
    unsigned long       flags;

    if ((pr = kmalloc (sizeof (QSNET_PROC_PRIVATE), GFP_KERNEL)) == NULL)
	return (-ENOMEM);
    
    epRail = (EP_RAIL *)(PDE(inode)->data);
	
    do {	
	pr->pr_data_len = PAGESIZE * pages;

	KMEM_ZALLOC (pr->pr_data, char *, pr->pr_data_len, 1);
	if (pr->pr_data == NULL) 
	{ 
	    pr->pr_len  = sprintf (pr->pr_data, "Out of Memory\n");
	    break;
	} 
	
	pr->pr_off     = 0;
	pr->pr_len     = 0;
	pr->pr_data[0] = 0;
	
	if (epRail->State != EP_RAIL_STATE_RUNNING) 
	{ 
	    pr->pr_len  = sprintf (pr->pr_data, "Rail not Running\n");
	    break;
	} 
	else 
	{
	    pr->pr_di.func  = qsnet_proc_character_fill;
	    pr->pr_di.arg   = (long)pr;

	    if (!strcmp("maps", file->f_dentry->d_iname)) 
	    {
		cmRail = epRail->ClusterRail;

		spin_lock_irqsave (&cmRail->Lock, flags);
		DisplayNodeMaps (&pr->pr_di, cmRail);	
		spin_unlock_irqrestore (&cmRail->Lock, flags);	
	    }

	    if (!strcmp("segs", file->f_dentry->d_iname)) 
	    {
		cmRail = epRail->ClusterRail;
		
		spin_lock_irqsave (&cmRail->Lock, flags);	
		DisplayNodeSgmts (&pr->pr_di, cmRail);
		spin_unlock_irqrestore (&cmRail->Lock, flags);
	    }

	    if (!strcmp("tree", file->f_dentry->d_iname)) 
		DisplayRailDo (&pr->pr_di, epRail);
	}

	if ( pr->pr_len < pr->pr_data_len) 
	    break; /* we managed to get all the output into the buffer */

	pages++;
	KMEM_FREE ( pr->pr_data,  pr->pr_data_len);
    } while (1);
	

    file->private_data = (void *) pr;

    MOD_INC_USE_COUNT;
    return (0);
}

struct file_operations proc_nodeset_operations = 
{
    read:	nodeset_read,
    poll:	nodeset_poll,
    open:	nodeset_open,
    release:	nodeset_release,
};

struct file_operations proc_operations = 
{
    read:	qsnet_proc_read,
    open:	proc_open,
    release:	qsnet_proc_release,
};

void
ep_procfs_rail_init (EP_RAIL *rail)
{
    struct proc_dir_entry *dir;
    struct proc_dir_entry *p;
    char                   name[10];
    int                    i;

    sprintf (name, "rail%d", rail->Number);

    if ((dir = rail->ProcDir = proc_mkdir (name, ep_procfs_root)) == NULL)
	return;

    for (i = 0; i < sizeof (rail_info)/sizeof (rail_info[0]); i++)
    {
	if ((p = create_proc_entry (rail_info[i].name, 0, dir)) != NULL)
	{
	    p->read_proc  = rail_info[i].read_func;
	    p->write_proc = rail_info[i].write_func;
	    p->data       = rail;
	    p->owner      = THIS_MODULE;
	}
    }

    if ((p = create_proc_entry ("nodeset", 0, dir)) != NULL)
    {
	p->proc_fops = &proc_nodeset_operations;
	p->owner     = THIS_MODULE;
	p->data      = rail;

	rail->CallbackRegistered = 1;
	ep_register_callback (rail, EP_CB_NODESET, nodeset_callback, rail);
    }
     
    if ((p = create_proc_entry ("maps", 0, dir)) != NULL)
    {
	p->proc_fops = &proc_operations;
	p->owner     = THIS_MODULE;
	p->data      = rail;	
    }
    
    if ((p = create_proc_entry ("segs", 0, dir)) != NULL)
    {
	p->proc_fops = &proc_operations;
	p->owner     = THIS_MODULE;
	p->data      = rail;
    }
    
    if ((p = create_proc_entry ("tree", 0, dir)) != NULL)
    {
	p->proc_fops = &proc_operations;
	p->owner     = THIS_MODULE;
	p->data      = rail;
    }

}

void
ep_procfs_rail_fini (EP_RAIL *rail)
{
    struct proc_dir_entry *dir = rail->ProcDir;
    char name[10];
    int  i;

    if (dir == NULL)
	return;

    if (rail->CallbackRegistered)
    {
	ep_remove_callback (rail, EP_CB_NODESET, nodeset_callback, rail);

	remove_proc_entry ("nodeset", dir);
    }

    remove_proc_entry ("maps",    dir);
    remove_proc_entry ("segs",    dir);
    remove_proc_entry ("tree",    dir);

    for (i = 0; i < sizeof (rail_info)/sizeof (rail_info[0]); i++)
	remove_proc_entry (rail_info[i].name, dir);

    sprintf (name, "rail%d", rail->Number);
    remove_proc_entry (name, ep_procfs_root);
}

#include "quadrics_version.h"
static char     quadrics_version[] = QUADRICS_VERSION;

void
ep_procfs_init()
{
    extern int txd_stabilise;
    extern int MaxSwitchLevels;

    printk ("ep Module (version %s)\n", quadrics_version);

    spin_lock_init (&ep_nodeset_lock);
    init_waitqueue_head (&ep_nodeset_wait);

    ep_procfs_root = proc_mkdir ("ep", qsnet_procfs_root);
    ep_config_root = proc_mkdir ("config", ep_procfs_root);

    qsnet_proc_register_str (ep_procfs_root, "version", quadrics_version, 1);

    qsnet_proc_register_hex (ep_config_root, "epdebug",               &epdebug,               0);
    qsnet_proc_register_hex (ep_config_root, "epdebug_console",       &epdebug_console,       0);
    qsnet_proc_register_hex (ep_config_root, "epdebug_cmlevel",       &epdebug_cmlevel,       0);
#if ! defined(CONFIG_EP_NO_CHECK_SUM)
    qsnet_proc_register_hex (ep_config_root, "epdebug_check_sum",     &epdebug_check_sum,     0);
#endif
    qsnet_proc_register_hex (ep_config_root, "epcomms_forward_limit", &epcomms_forward_limit, 0);
    qsnet_proc_register_int (ep_config_root, "txd_stabilise",         &txd_stabilise,         0);
    qsnet_proc_register_int (ep_config_root, "assfail_mode",          &assfail_mode,          0);
    qsnet_proc_register_int (ep_config_root, "max_switch_levels",     &MaxSwitchLevels,       1);

    ep_procfs_rcvr_xmtr_init();
}

void
ep_procfs_fini(void)
{
    ep_procfs_rcvr_xmtr_fini();

    remove_proc_entry ("max_switch_levels",     ep_config_root);
    remove_proc_entry ("assfail_mode",          ep_config_root);
    remove_proc_entry ("txd_stabilise",         ep_config_root);
    remove_proc_entry ("epcomms_forward_limit", ep_config_root);

#if ! defined(CONFIG_EP_NO_CHECK_SUM)
    remove_proc_entry ("epdebug_check_sum",     ep_config_root);
#endif
    remove_proc_entry ("epdebug_cmlevel",       ep_config_root);
    remove_proc_entry ("epdebug_console",       ep_config_root);
    remove_proc_entry ("epdebug",               ep_config_root);

    remove_proc_entry ("version", ep_procfs_root);
    
    remove_proc_entry ("config", ep_procfs_root);
    remove_proc_entry ("ep", qsnet_procfs_root);

    spin_lock_destroy (&ep_nodeset_lock);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
