/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: rmap.c,v 1.15 2004/05/19 10:24:38 david Exp $"
/*      $Source: /cvs/master/quadrics/epmod/rmap.c,v $ */

#include <qsnet/kernel.h>
#include <elan/rmap.h>

#include "debug.h"

void
ep_display_rmap (EP_RMAP *mp)
{
    EP_RMAP_ENTRY *bp;
    unsigned long flags;
    
    spin_lock_irqsave (&mp->m_lock, flags);
    ep_debugf (DBG_DEBUG, "map: %s size %d free %d\n", mp->m_name, mp->m_size, mp->m_free);
    for (bp = &mp->m_map[0]; bp->m_size; bp++)
	ep_debugf (DBG_DEBUG, "   [%lx - %lx]\n", bp->m_addr, bp->m_addr+bp->m_size-1);
    spin_unlock_irqrestore (&mp->m_lock, flags);
}

void
ep_mapinit (EP_RMAP *mp, char *name, u_int mapsize)
{
    spin_lock_init (&mp->m_lock);
    kcondvar_init (&mp->m_wait);
    
    /* The final segment in the array has size 0 and acts as a delimiter
     * we insure that we never use segments past the end of the array by
     * maintaining a free segment count in m_free.  When excess segments
     * occur we discard some resources */
    
    mp->m_size = mapsize;
    mp->m_free = mapsize;
    mp->m_name = name;
    
    bzero (mp->m_map, sizeof (EP_RMAP_ENTRY) * (mapsize+1));
}

EP_RMAP *
ep_rmallocmap (size_t mapsize, char *name, int cansleep)
{
    EP_RMAP *mp;

    KMEM_ZALLOC (mp, EP_RMAP *, sizeof (EP_RMAP) + mapsize*sizeof (EP_RMAP_ENTRY), cansleep);

    if (mp != NULL)
	ep_mapinit (mp, name, mapsize);

    return (mp);
}

void
ep_rmfreemap (EP_RMAP *mp)
{
    spin_lock_destroy (&mp->m_lock);
    kcondvar_destroy (&mp->m_wait);
    
    KMEM_FREE (mp, sizeof (EP_RMAP) + mp->m_size * sizeof (EP_RMAP_ENTRY));
}

static u_long
ep_rmalloc_locked (EP_RMAP *mp, size_t size)
{
    EP_RMAP_ENTRY *bp;
    u_long            addr;
    
    ASSERT (size > 0);
    ASSERT (SPINLOCK_HELD (&mp->m_lock));

    for (bp = &mp->m_map[0]; bp->m_size; bp++)
    {
	if (bp->m_size >= size)
	{
	    addr = bp->m_addr;
	    bp->m_addr += size;
	    
	    if ((bp->m_size -= size) == 0)
	    {
		/* taken all of this slot - so shift the map down */
		do {
		    bp++;
		    (bp-1)->m_addr = bp->m_addr;
		} while (((bp-1)->m_size = bp->m_size) != 0);

		mp->m_free++;
	    }
	    return (addr);
	}
    }

    return (0);
}

u_long
ep_rmalloc (EP_RMAP *mp, size_t size, int cansleep)
{
    unsigned long addr;
    unsigned long flags;

    spin_lock_irqsave (&mp->m_lock, flags);
    while ((addr = ep_rmalloc_locked (mp, size)) == 0 && cansleep)
    {
	mp->m_want = 1;
	kcondvar_wait (&mp->m_wait, &mp->m_lock, &flags);
    }

    spin_unlock_irqrestore (&mp->m_lock, flags);

    return (addr);
}



u_long
ep_rmalloc_constrained (EP_RMAP *mp, size_t size, u_long alo, u_long ahi, u_long align, int cansleep)
{
    EP_RMAP_ENTRY *bp, *bp2, *lbp;
    unsigned long addr=0;
    size_t        delta;
    int           ok;
    unsigned long flags;

    spin_lock_irqsave (&mp->m_lock, flags);
 again:
    for (bp = &mp->m_map[0]; bp->m_size; bp++)
    {
	delta = 0;
	
	if (alo < bp->m_addr)
	{
	    addr = bp->m_addr;
	    
	    if (addr & (align-1))
		addr = (addr + (align-1)) & ~(align-1);
	    
	    delta = addr - bp->m_addr;
	    
	    if (ahi >= bp->m_addr + bp->m_size)
		ok = (bp->m_size >= (size + delta));
	    else
		ok = ((bp->m_addr + size + delta) <= ahi);
	}
	else
	{
	    addr = alo;
	    if (addr & (align-1))
		addr = (addr + (align-1)) & ~(align-1);
	    delta = addr - bp->m_addr;
	    
	    if (ahi >= bp->m_addr + bp->m_size)
		ok = ((alo + size + delta) <= (bp->m_addr + bp->m_size));
	    else
		ok = ((alo + size + delta) <= ahi);
	}

	if (ok)
	    break;
    }	
    
    if (bp->m_size == 0)
    {
	if (cansleep)
	{
	    mp->m_want = 1;
	    kcondvar_wait (&mp->m_wait, &mp->m_lock, &flags);
	    goto again;
	}
	spin_unlock_irqrestore (&mp->m_lock, flags);
	return (0);
    }

    /* found an approriate map entry - so take the bit out which we want */
    if (bp->m_addr == addr) 
    {
	if (bp->m_size == size) 
	{
	    /* allocate entire segment and compress map */
	    bp2 = bp;
	    while (bp2->m_size) 
	    {
		bp2++;
		(bp2-1)->m_addr = bp2->m_addr;
		(bp2-1)->m_size = bp2->m_size;
	    }
	    mp->m_free++;
	}
	else 
	{
	    /* take from start of segment */
	    bp->m_addr += size;
	    bp->m_size -= size;
	}
    }
    else 
    {
	if (bp->m_addr + bp->m_size == addr + size) 
	{
	    /* take from end of segment */
	    bp->m_size -= size;
	}
	else 
	{
	    /* split the segment loosing the last entry if there's no space */
	    if (mp->m_free == 0) 
	    {
		/* find last map entry */
		for (lbp = bp; lbp->m_size != 0; lbp++)
		    ;
		lbp--;
		
		if (lbp->m_size > (lbp-1)->m_size)
		    lbp--;
		
		printk ("%s: lost resource map entry [%lx, %lx]\n",
			mp->m_name, lbp->m_addr, lbp->m_addr + lbp->m_size);
		
		*lbp = *(lbp+1);
		(lbp+1)->m_size = 0;
		
		mp->m_free++;
	    }
	    
	    for (bp2 = bp; bp2->m_size != 0; bp2++)
		continue;
	    
	    for (bp2--; bp2 > bp; bp2--)
	    {
		(bp2+1)->m_addr = bp2->m_addr;
		(bp2+1)->m_size = bp2->m_size;
	    }

	    mp->m_free--;
	    
	    (bp+1)->m_addr = addr + size;
	    (bp+1)->m_size = bp->m_addr + bp->m_size - (addr + size);
	    bp->m_size = addr - bp->m_addr;
	}
    }

    spin_unlock_irqrestore (&mp->m_lock, flags);
    return (addr);
}

void
ep_rmfree (EP_RMAP *mp, size_t size, u_long addr)
{
    EP_RMAP_ENTRY *bp;
    unsigned long t;
    unsigned long flags;

    spin_lock_irqsave (&mp->m_lock, flags);

    ASSERT (addr != 0 && size > 0);
	
again:
    /* find the piece of the map which starts after the returned space
     * or the end of the map */
    for (bp = &mp->m_map[0]; bp->m_addr <= addr && bp->m_size != 0; bp++)
	;

    /* bp points to the piece to the right of where we want to go */
    
    if (bp > &mp->m_map[0] && (bp-1)->m_addr + (bp-1)->m_size >= addr) 
    {
	/* merge with piece on the left */
	
	ASSERT ((bp-1)->m_addr + (bp-1)->m_size <= addr);
	
	(bp-1)->m_size += size;
	
	ASSERT (bp->m_size == 0 || addr+size <= bp->m_addr);
	
	if (bp->m_size && (addr + size) == bp->m_addr)
	{
	    /* merge witht he piece on the right by 
	     * growing the piece on the left and shifting
	     * the map down */
	    
	    ASSERT ((addr + size) <= bp->m_addr);
	    
	    (bp-1)->m_size += bp->m_size;
	    while (bp->m_size) 
	    {
		bp++;
		(bp-1)->m_addr = bp->m_addr;
		(bp-1)->m_size = bp->m_size;
	    }
	    
	    mp->m_free++;
	}
    }
    else if (addr + size >= bp->m_addr && bp->m_size)
    {
	/* merge with piece to the right */
	
	ASSERT ((addr + size) <= bp->m_addr);
	
	bp->m_addr -= size;
	bp->m_size += size;
    }
    else
    {
	/* doesn't join with left or right - check for map
	   overflow and discard the smallest of the last or
	   next to last entries */

	if (mp->m_free == 0)
	{
	    EP_RMAP_ENTRY *lbp;
	    
	    /* find last map entry */
	    for (lbp = bp; lbp->m_size != 0; lbp++)
		;
	    lbp--;
	    
	    if (lbp->m_size > (lbp-1)->m_size)
		lbp--;
	    
	    printk ("%s: lost resource map entry [%lx, %lx]\n", 
		    mp->m_name, lbp->m_addr, lbp->m_addr + lbp->m_size);
	    
	    *lbp = *(lbp+1);
	    (lbp+1)->m_size = 0;

	    mp->m_free++;
	    goto again;
	}

	/* make a new entry and push the remaining ones up */
	do {
	    t = bp->m_addr;
	    bp->m_addr = addr;
	    addr = t;
	    t = bp->m_size;
	    bp->m_size = size;
	    bp++;
	} while ((size = t) != 0);

	mp->m_free--;
    }
    
    /* if anyone blocked on rmalloc failure, wake 'em up */
    if (mp->m_want)
    {
	mp->m_want = 0;
	kcondvar_wakeupall (&mp->m_wait, &mp->m_lock);
    }

    spin_unlock_irqrestore (&mp->m_lock, flags);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
