/*
 *    Copyright (c) 2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: debug.c,v 1.22 2005/03/23 11:04:54 david Exp $"
/*      $Source: /cvs/master/quadrics/qsnet/debug.c,v $ */

#include <qsnet/kernel.h>
#include <qsnet/debug.h>
#include <qsnet/procfs_linux.h>

caddr_t        qsnet_debug_buffer_ptr = NULL;
int	       qsnet_debug_front      = 0;
int	       qsnet_debug_back       = 0;
int            qsnet_debug_lost_lines = 0;
int	       qsnet_debug_disabled   = 0;

int	       qsnet_debug_line_size  = 256;
int	       qsnet_debug_num_lines  = 8192;

int	       qsnet_assfail_mode     = 1;			/* default to BUG() */

int            qsnet_debug_running    = 0;
int            kqsnet_debug_running   = 0;

static spinlock_t qsnet_debug_lock;
static kcondvar_t qsnet_debug_wait;
static char       qsnet_debug_buffer_space[8192];

#define QSNET_DEBUG_PREFIX_MAX_SIZE	32
#define QSNET_DEBUG_MAX_WORDWRAP	15

/* must be larger than  QSNET_DEBUG_PREFIX_MAX_SIZE +  QSNET_DEBUG_MAX_WORDWRAP + 2 */
#if defined(DIGITAL_UNIX) 
#define QSNET_DEBUG_CONSOLE_WIDTH 80
#elif defined(LINUX)
#define QSNET_DEBUG_CONSOLE_WIDTH 128
#endif

#define isspace(CH)    ((CH==' ') | (CH=='\t') | (CH=='\n'))

#ifdef LINUX
#define ALLOC_DEBUG_BUFFER(ptr)		do { (ptr) = (void *)__get_free_pages (GFP_KERNEL, get_order (qsnet_debug_num_lines * qsnet_debug_line_size)); } while (0)
#define FREE_DEBUG_BUFFER(ptr)		free_pages ((unsigned long) ptr, get_order (qsnet_debug_num_lines * qsnet_debug_line_size))
#else
#define ALLOC_DEBUG_BUFFER(ptr)		KMEM_ALLOC (ptr, caddr_t, qsnet_debug_num_lines * qsnet_debug_line_size, 1)
#define FREE_DEBUG_BUFFER(ptr)		KMEM_FREE (ptr, qsnet_debug_num_lines * qsnet_debug_line_size)
#endif

void
qsnet_debug_init ()
{
	spin_lock_init (&qsnet_debug_lock);
	kcondvar_init  (&qsnet_debug_wait);

	qsnet_debug_front      = 0;
	qsnet_debug_back       = 0;
	qsnet_debug_lost_lines = 0;

	if (qsnet_debug_line_size < (QSNET_DEBUG_PREFIX_MAX_SIZE + QSNET_DEBUG_MAX_WORDWRAP + 2))
		qsnet_debug_line_size = 256;

	qsnet_debug_running    = 1;

	qsnet_proc_register_int (qsnet_procfs_config, "assfail_mode", &qsnet_assfail_mode, 0);
}

void
qsnet_debug_fini()
{
	if (!qsnet_debug_running) return;

	remove_proc_entry ("assfail_mode", qsnet_procfs_config);

	spin_lock_destroy (&qsnet_debug_lock);
	kcondvar_destroy  (&qsnet_debug_wait);
	
	if (qsnet_debug_buffer_ptr)
		FREE_DEBUG_BUFFER (qsnet_debug_buffer_ptr);

	qsnet_debug_buffer_ptr     = NULL;
	qsnet_debug_lost_lines = 0;	
	qsnet_debug_running    = 0;	
}

void
qsnet_debug_disable(int val)
{
	qsnet_debug_disabled = val;
}

void
qsnet_debug_alloc()
{
	caddr_t ptr;
	unsigned long flags;

	if (!qsnet_debug_running) return;

	if (qsnet_debug_buffer_ptr == NULL)
	{
		ALLOC_DEBUG_BUFFER (ptr);

		if (ptr != NULL)
		{
			spin_lock_irqsave (&qsnet_debug_lock, flags);
			if (qsnet_debug_buffer_ptr == NULL)
			{
				qsnet_debug_buffer_ptr = ptr;
				spin_unlock_irqrestore (&qsnet_debug_lock, flags);
			}
			else
			{
				spin_unlock_irqrestore (&qsnet_debug_lock, flags);

				FREE_DEBUG_BUFFER (ptr);
			}
		}
	}
	
}

static void 
qsnet_prefix_debug(unsigned int mode, const char *prefix, char *buffer) 
{
	/* assumes caller has lock */

	int  prefixlen = strlen(prefix);
	char pref[QSNET_DEBUG_PREFIX_MAX_SIZE];
	int  prefix_done = 0;

	if (!qsnet_debug_running) return;

	if (qsnet_debug_disabled)
		return;

	if (prefixlen >= QSNET_DEBUG_PREFIX_MAX_SIZE) 
	{
		strncpy(pref,prefix,QSNET_DEBUG_PREFIX_MAX_SIZE -2);
		strcpy (&pref[QSNET_DEBUG_PREFIX_MAX_SIZE-5],"... ");

		prefix = pref;
                prefixlen = strlen(prefix);
	}

#ifdef CONFIG_MPSAS
	{
		char *p;
#define TRAP_PUTCHAR_B			(0x17a - 256)
#define SAS_PUTCHAR(c)			do {\
			register int o0 asm ("o0") = (c);\
\
			asm volatile ("ta %0; nop" \
				      : /* no outputs */\
				      : /* inputs */ "i" (TRAP_PUTCHAR_B), "r" (o0)\
				      : /* clobbered */ "o0");\
\
			if (o0 == '\n') {\
				o0 = '\r';\
\
				asm volatile ("ta %0; nop" \
					      : /* no outputs */\
					      : /* inputs */ "i" (TRAP_PUTCHAR_B), "r" (o0)\
					      : /* clobbered */ "o0");\
			}\
		} while(0)

		for (p = prefix; *p; p++)
			SAS_PUTCHAR (*p);

		for (p = buffer; *p; p++)
			SAS_PUTCHAR (*p);
	}
#else
	if (mode & QSNET_DEBUG_BUFFER)
	{
		if (qsnet_debug_buffer_ptr == NULL)
			qsnet_debug_lost_lines++;
		else
		{		    
			caddr_t base = &qsnet_debug_buffer_ptr[qsnet_debug_line_size * qsnet_debug_back];
			caddr_t lim  = base + qsnet_debug_line_size - 2;
			caddr_t p;
		
			p = buffer; 
			prefix_done = 0;
			while (*p) 
			{
				/* sort out prefix */
				if ( prefix_done++ ) 
				{
					int i;
					for(i=0;i<prefixlen;i++)
						base[i] = ' ';
					/* memset(base,' ',prefixlen); */
				}
				else
					strcpy(base,prefix);
				base += prefixlen; /* move the base on */

				/* copy data */
				for ( ; *p && (base < lim); )
					*base++ = *p++;

				/* if line split then add \n */
				if ((base == lim) && (*base != '\n'))
				{
					char *ptr;
					int   count;

					*base = '\n';
					/* we added a \n cos it was end of line put next char was \n */
					if (*p == '\n') 
						p++;
					else
					{
						/* lets see if we can back track and find a white space to break on */
						ptr = base-1;
						count = 1;
						while ( ( !isspace(*ptr) ) && ( count < QSNET_DEBUG_MAX_WORDWRAP ))
						{
							count++;
							ptr--;
						}

						if ( isspace(*ptr) ) 
						{
							/* found somewhere to wrap to */
							p -= (count-1); /* need to loose the white space */
							base = ptr;
							*base = '\n';
						}
					}
					base++;
				}
				*base = '\0';

				/* move on pointers */
				qsnet_debug_back = (++qsnet_debug_back == qsnet_debug_num_lines) ? 0 : qsnet_debug_back;	    
				if (qsnet_debug_back == qsnet_debug_front)
				{
					qsnet_debug_lost_lines++;
					qsnet_debug_front = (++qsnet_debug_front == qsnet_debug_num_lines) ? 0 : qsnet_debug_front;
				}
				base  = &qsnet_debug_buffer_ptr[qsnet_debug_line_size * qsnet_debug_back];
				lim  =  base + qsnet_debug_line_size - 2;
			}
			kcondvar_wakeupone (&qsnet_debug_wait, &qsnet_debug_lock);
		}
	}

	if (mode & QSNET_DEBUG_CONSOLE)
	{
		int     remaining = QSNET_DEBUG_CONSOLE_WIDTH - prefixlen;
		caddr_t p;
		char    line[QSNET_DEBUG_CONSOLE_WIDTH +2];
		int     len;
	    
		strcpy (pref,prefix);
		prefix_done = 0;

		p = buffer;
		while ( *p )
		{
			/* use the prefix only once */
			if  ( prefix_done++ > 0 ) 
				{
					int i;
					for(i=0;i<prefixlen;i++)
						pref[i] = ' ';
					/* memset(perf,' ',prefixlen); */
				}	

			len=strlen(p);
			if (len > remaining) len = remaining;
		  
			strncpy(line, p, len);
			line[len] = 0;
			p += len;
		    
			/* word wrap */
			if ((len == remaining) && *p && !isspace(*p))
			{
				/* lets see if we can back track and find a white space to break on */
				char * ptr = &line[len-1];
				int    count = 1;

				while ( ( !isspace(*ptr) ) && ( count < QSNET_DEBUG_MAX_WORDWRAP ))
				{
					count++;
					ptr--;
				}

				if ( isspace(*ptr) ) 
				{
					/* found somewhere to wrap to */
					p -= (count-1); /* need to loose the white space */
					len -= count;
				}		
			}

			if (line[len-1] != '\n' ) 
			{
				line[len] = '\n';
				line[len+1] = 0;
			}

			/* we put a \n in so dont need another one next */
			if ( *p == '\n')
				p++;

#if defined(DIGITAL_UNIX)
			{
				char *pr;

				for (pr = pref; *pr; pr++)
					cnputc (*pr);

				for (pr = line; *pr; pr++)
					cnputc (*pr); 
			}
#elif defined(LINUX)
			printk("%s%s",pref,line);
#endif
		}
	}
#endif /* CONFIG_MPSAS */
}

void
qsnet_vdebugf (unsigned int mode, const char *prefix, const char *fmt, va_list ap)
{
	unsigned long flags;

	if (!qsnet_debug_running) return;

	spin_lock_irqsave (&qsnet_debug_lock, flags);

	qsnet_debug_buffer_space[0] = '\0';

#if defined(DIGITAL_UNIX)
	prf (qsnet_debug_buffer_space+strlen(qsnet_debug_buffer_space), NULL, fmt, ap);
#elif defined(LINUX)
	vsprintf (qsnet_debug_buffer_space+strlen(qsnet_debug_buffer_space), fmt, ap);
#endif

	if (prefix == NULL)
		printk ("qsnet_vdebugf: prefix==NULL\n");
	else
		qsnet_prefix_debug(mode, prefix, qsnet_debug_buffer_space);

	spin_unlock_irqrestore (&qsnet_debug_lock, flags);
}

void kqsnet_debugf(char *fmt,...)
{
	if ( kqsnet_debug_running ) {
		va_list ap;
		char string[20];
		
		sprintf (string, "mm=%p:", current->mm);
		va_start(ap, fmt);
		qsnet_vdebugf(QSNET_DEBUG_BUFFER, string, fmt, ap);
		va_end(ap);
	}	
}
void 
qsnet_debugf(unsigned int mode, const char *fmt,...)
{
	va_list       ap;
	unsigned long flags;

	if (!qsnet_debug_running) return;

	spin_lock_irqsave (&qsnet_debug_lock, flags);

	qsnet_debug_buffer_space[0] = '\0';

	va_start (ap, fmt);
#if defined(DIGITAL_UNIX)
	prf (qsnet_debug_buffer_space+strlen(qsnet_debug_buffer_space), NULL, fmt, ap);
#elif defined(LINUX)
	vsprintf (qsnet_debug_buffer_space+strlen(qsnet_debug_buffer_space), fmt, ap);
#endif
	va_end (ap);

	qsnet_prefix_debug(mode, "", qsnet_debug_buffer_space); 

	spin_unlock_irqrestore (&qsnet_debug_lock, flags);
}

int
qsnet_debug_buffer (caddr_t ubuffer, int len)
{
	caddr_t buffer, ptr, base;
	int     remain, len1;
	unsigned long flags;
	static  char qsnet_space[65536];

	if (!qsnet_debug_running) return (0);

	if (len < qsnet_debug_line_size)
		return (-1);

	if (len > (qsnet_debug_line_size * qsnet_debug_num_lines))
		len = qsnet_debug_line_size * qsnet_debug_num_lines;
    
	if ( len > 65536 ) {
		KMEM_ZALLOC (buffer, caddr_t, len, 1);
	} else 
		buffer = qsnet_space;

	if (buffer == NULL)
		return (-1);

	if (qsnet_debug_buffer_ptr == NULL)
		qsnet_debug_alloc();

	if (qsnet_debug_buffer_ptr == NULL)
	{
		if ( len > 65536 )
			KMEM_FREE (buffer, len);
		return (-1);
	}

	spin_lock_irqsave (&qsnet_debug_lock, flags);
    
	while (!qsnet_debug_lost_lines && (qsnet_debug_back == qsnet_debug_front))
		if (kcondvar_waitsig (&qsnet_debug_wait, &qsnet_debug_lock, &flags) == 0)
			break;
    
	ptr    = buffer;
	remain = len;

	if (qsnet_debug_lost_lines)
	{
		qsnet_debug_lost_lines = 0;
		strcpy (ptr, "Debug Buffer has overflowed!!\n");
		len1 = strlen (ptr);

		remain -= len1;
		ptr    += len1;
	}

	while (qsnet_debug_front != qsnet_debug_back)
	{
		/* copy the line from DebugFront */
		base = &qsnet_debug_buffer_ptr[qsnet_debug_front*qsnet_debug_line_size];

		len1 = strlen (base);

		if (len1 > remain)
			break;
	
		bcopy (base, ptr, len1);
	
		ptr += len1;
		remain -= len1;

		qsnet_debug_front = (++qsnet_debug_front == qsnet_debug_num_lines) ? 0 : qsnet_debug_front;
	}

	spin_unlock_irqrestore (&qsnet_debug_lock, flags);

	len1 = ptr - buffer;

	if (len1 != 0 && copyout (buffer, ubuffer, len1))
		len1 = -1;

	if ( len > 65536 )
		KMEM_FREE (buffer, len);
   
	return (len1);
}

void
qsnet_debug_buffer_on() 
{
	if (qsnet_debug_buffer_ptr == NULL)
		qsnet_debug_alloc();
}

void 
qsnet_debug_buffer_clear()
{
	unsigned long flags;

	qsnet_debug_buffer_on();
	
	if (qsnet_debug_buffer_ptr != NULL){
		spin_lock_irqsave (&qsnet_debug_lock, flags);
		qsnet_debug_front      = 0;
		qsnet_debug_back       = 0;
		qsnet_prefix_debug(QSNET_DEBUG_BUFFER,"Clear","");
		spin_unlock_irqrestore (&qsnet_debug_lock, flags);	
	}
}

void 
qsnet_debug_buffer_mark(char *str)
{
	unsigned long flags;	

	qsnet_debug_buffer_on();

	if (qsnet_debug_buffer_ptr != NULL) {
		spin_lock_irqsave (&qsnet_debug_lock, flags);
		qsnet_prefix_debug(QSNET_DEBUG_BUFFER,"Mark",str);
		spin_unlock_irqrestore (&qsnet_debug_lock, flags);
	}
}
int
qsnet_debug_dump ()
{
	unsigned long flags;

	if (!qsnet_debug_running) return (0);

	if (qsnet_debug_buffer_ptr == NULL)
		qsnet_debug_alloc();

	if (qsnet_debug_buffer_ptr == NULL)
		return (-1);

	spin_lock_irqsave (&qsnet_debug_lock, flags);

	while (qsnet_debug_front != qsnet_debug_back)
	{
		printk ("%s", &qsnet_debug_buffer_ptr[qsnet_debug_front*qsnet_debug_line_size]);

		qsnet_debug_front = (++qsnet_debug_front == qsnet_debug_num_lines) ? 0 : qsnet_debug_front;
	}

	if (qsnet_debug_lost_lines)
		printk ("\n**** Debug buffer has lost %d lines\n****\n",qsnet_debug_lost_lines);

	spin_unlock_irqrestore (&qsnet_debug_lock, flags);

	return (0);
}

int
qsnet_debug_kmem (void *handle)
{
	if (!qsnet_debug_running) return (0);

#ifdef KMEM_DEBUG
	qsnet_kmem_display(handle);
#endif
	return (0);
}

int
qsnet_assfail (char *ex, const char *func, char *file, int line)
{
	qsnet_debugf (QSNET_DEBUG_BUFFER, "qsnet: assertion failure: %s, function: %s, file %s, line: %d\n", ex, func, file, line);

	printk (KERN_EMERG "qsnet: assertion failure: %s, function: %s, file %s, line: %d\n", ex, func, file, line);

	if (panicstr)
		return (0);

	if (qsnet_assfail_mode & 1)				/* return to BUG() */
		return 1;

	if (qsnet_assfail_mode & 2)
		panic ("qsnet: assertion failure: %s, function: %s, file %s, line: %d\n", ex, func, file, line);
	if (qsnet_assfail_mode & 4)
		qsnet_debug_disable (1);

	return 0;

}


/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
