/*
 * Copyright (c) 1996-2003 by Quadrics Supercomputers World Ltd.
 * Copyright (c) 2004-2006 by Quadrics Ltd
 *
 * For licensing information please see the supplied COPYING file
 *
 * rms_kern.c:    RMS kernel module
 *
 * $Source: /cvs/master/quadrics/rmsmod/rms_kern.c,v $
 */

#ident "@(#)$Id: rms_kern.c,v 1.77.2.8 2005/09/07 14:35:04 mike Exp $"

#include <stddef.h>
#include <qsnet/kernel.h>
#include <qsnet/autoconf.h>
#include <rms/rmscall.h>


/*
 * extended stats added in version 5
 * patch free kernel support added in version 6
 */
#define RMS_MODVERSION 6

#ifdef PROCESS_ACCT 
#define TIMEVAL_TO_MSEC(tv) ((tv)->tv_sec * 1000 + (tv)->tv_usec / 1000)
#define TIMEVAL_TO_CT(tv)   ((tv)->tv_sec * HZ + (tv)->tv_usec / (1000000L / HZ))
#endif

#ifdef get_mm_counter
#define PROC_RSS(proc)	((proc)->mm ? get_mm_counter(proc->mm, rss) : 0)
#else
#ifdef RSS_ATOMIC
#define PROC_RSS(proc)	((proc)->mm ? atomic_read(&(proc)->mm->rss) : 0)
#else
#define PROC_RSS(proc)	((proc)->mm ? (proc)->mm->rss : 0)
#endif
#endif

/*
 * 2.6 kernels don't consistently export put_task_struct
 */

#ifdef free_task_struct
#undef NO_PUTGET_TASK
#else
#define NO_PUTGET_TASK
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
#	define 	RMS_NCPUS()	smp_num_cpus
#else
#	define RMS_NCPUS()      num_online_cpus()
#endif

#define CURUID()	CURPROC()->uid
#define p_pid		pid


/* care needed with conversion to millisecs on 32-bit Linux */
#ifdef LINUX_I386
#define CT_TO_MSEC(x)	ct_to_msec(x)

uint64_t ct_to_msec(clock_t t)
{
    uint64_t msecs;
    if (t < 2000000)
    {
	t = (1000 * t)/HZ;
	msecs = t;
    }
    else
    {
	t = t / HZ;
	msecs = t * 1000;
    }
    return(msecs);
}
#else
#define CT_TO_MSEC(x)	(((x) * 1000)/HZ)
#endif

#ifndef FALSE
#define FALSE		(0)
#define TRUE		(!FALSE)
#endif

#include <linux/time.h>
#include <linux/proc_fs.h>
#ifndef NO_PTRACK
#include <linux/ptrack.h>
#endif
#include <qsnet/module.h>
#ifndef NO_SHM_CLEANUP
extern int shm_cleanup(void);
#endif

struct cap_desc {

    struct cap_desc *next;
    int              index;	/* index of capability in program */
    ELAN_CAPABILITY  cap;	/* elan capability */

};

struct proc_desc {
    
    struct proc_desc    *next;
    struct task_struct  *task;
    struct prg_desc     *program;	/* controlling program         */
    int                  mycap;	        /* index of my capability      */
    int                  myctx;	        /* context number for process  */
    int                  flags;
    int                  vp;	        /* elan virtual process number */
    int                  pid;	        /* process id                  */

    /* last set of stats sampled */
    uint64_t             utime;
    uint64_t             stime;
    uint64_t             majflt;
    int                  maxrss;

};

struct prg_desc {
    
    struct prg_desc  *next;		
    int               id;	/* program id                          */
    int               flags;	/* program status flags                */
    uid_t             uid;	/* user id                             */
    int               ncpus;	/* number of cpus allocated to program */
    int               nprocs;	/* number of processes in program      */
    struct proc_desc *pdescs;	/* processes in this program           */
    int               ncaps;	/* number of capabilities              */
    struct cap_desc  *caps;	/* elan capabilities                   */
    char             *corepath;	/* core path for parallel program      */
    int               psid;	/* processor set id                    */

    uint64_t       cutime;	/* user time accumulated by children   */
    uint64_t       cstime;	/* system time accumulated by children */
    uint64_t       start_time;	/* time program created                */
    uint64_t       end_time;	/* time last process exited            */
    uint64_t       sched_time;	/* last time job was scheduled         */
    uint64_t       accum_atime;	/* allocated time last deschedule      */
    uint64_t       memint;	/* accumulated memory integral         */
    uint64_t       ebytes;	/* data transferred by the Elan(s)     */
    uint64_t       exfers;	/* number of Elan data transfers       */
    uint64_t       majflt;	/* number of major faults              */
    int            maxrss;	/* maximum size to date                */
    
    struct proc_dir_entry *proc_entry;

};

static int rms_ptrack_callback (void *arg, int phase, struct task_struct *child);

static void prgsignal(struct prg_desc *program, int signo);
static uint64_t gettime(void);
static void freeProgram(struct prg_desc *program);

static struct prg_desc *programs = 0;

kmutex_t rms_lock;

#ifdef NO_PTRACK
int      ptrack_enabled = 0;
#else
int      ptrack_enabled = 1;
#endif

int rms_init(void)
{
    kmutex_init (&rms_lock);

    DBG(printk("rms: initialising ptrack %d\n", ptrack_enabled));

    return(ESUCCESS);
}

int rms_reconfigure(void)
{
    return(ESUCCESS);
}

int rms_programs_registered(void)
{
    /*
    ** Called when trying to unload rms.mod will not succeed
    ** if programs registered
    */
 
   struct prg_desc *program, **pp;

   kmutex_lock(&rms_lock);

   for (program = programs; program; program = program->next)
   {
	if (program->nprocs != 0)
	{
            kmutex_unlock(&rms_lock);
            return(EBUSY);
	}
   }

   /*
   ** We have traversed the programs list and no processes registered
   ** Now free the memory
   */
	
    pp = &programs;
    while ((program = *pp) != NULL)
    {
        *pp = program->next;
        freeProgram(program);
    }
    kmutex_unlock(&rms_lock);
   
    return(ESUCCESS);

}

int rms_fini(void)
{
    /*
     * don't allow an unload if there are programs registered
     */
    if (rms_programs_registered())
        return(EBUSY);

    kmutex_destroy (&rms_lock);

    DBG(printk("rms: removed\n"));

    return(ESUCCESS);
}

extern struct proc_dir_entry *rms_procfs_programs;

/*
 * display one pid per line if there isn't enough space 
 * for another pid then add "...\n" and stop 
 */
int pids_callback(char* page, char** start, off_t off, int count, int* eof, void* data)
{
    struct prg_desc *program = (struct prg_desc *)data;
    struct proc_desc *pdesc;
    char *ptr = page;
    int bytes = 0, nb;

    kmutex_lock(&rms_lock);
    
    for (pdesc = program->pdescs; pdesc; pdesc = pdesc->next)
    {
	if (bytes > count - 15)
	{
	    bytes += sprintf(ptr,"...\n");
	    break;
	}
	nb = sprintf(ptr, "%d %d\n", pdesc->pid, pdesc->vp);
	bytes += nb;
	ptr += nb;
    }
    kmutex_unlock(&rms_lock);
    
    return(bytes);
}

int status_callback(char* page, char** start, off_t off, int count, int* eof, void* data)
{
    struct prg_desc *program = (struct prg_desc *)data;
    int bytes;
    if (program->flags & PRG_KILLED)
	bytes = sprintf(page, "killed\n");
    else if (program->flags & PRG_SUSPEND)
	bytes = sprintf(page, "suspended\n");
    else
	bytes = sprintf(page, "running\n");
    return(bytes);
}

void rms_create_proc_entry(struct prg_desc *program)
{
    struct proc_dir_entry *p;
    char name[32];

    if (rms_procfs_programs)
    {
	sprintf(name,"%d", program->id);
	if ((program->proc_entry = proc_mkdir(name, rms_procfs_programs)) != NULL)
	{
	    if ((p = create_proc_entry ("pids", S_IRUGO, program->proc_entry)) != NULL)
	    {
		p->owner = THIS_MODULE;
		p->data = program;
		p->read_proc = pids_callback;
	    }
	    if ((p = create_proc_entry ("status", S_IRUGO, program->proc_entry)) != NULL)
	    {
		p->owner = THIS_MODULE;
		p->data = program;
		p->read_proc = status_callback;
	    }
	}
    }
}

void rms_remove_proc_entry(struct prg_desc *program)
{
    char name[32];
    if (rms_procfs_programs)
    {
	if (program->proc_entry)
	{
	    remove_proc_entry ("pids", program->proc_entry);
	    remove_proc_entry ("status", program->proc_entry);
	}
	sprintf(name,"%d", program->id);
	remove_proc_entry (name, rms_procfs_programs);
    }
}

/*
 * find a program from its index/pid
 */
static struct prg_desc *findProgram(const int id)
{
    struct prg_desc *program;
    for (program = programs; program; program = program->next)
	if (program->id == id)
	    return(program);
    return(0);
}

static struct proc_desc *findProcess(const int pid)
{
    struct prg_desc *program;
    struct proc_desc *pdesc;

    for (program = programs; program; program = program->next)
	for (pdesc = program->pdescs; pdesc; pdesc = pdesc->next)
	    if (pdesc->pid == pid)
		return(pdesc);
    
    return(0);
}

static void freeProgram(struct prg_desc *program)
{
    struct proc_desc *pdesc;
    struct cap_desc *cdesc;

    rms_remove_proc_entry(program);

    while ((pdesc = program->pdescs) != NULL)
    {
	program->pdescs = pdesc->next;
	KMEM_FREE(pdesc, sizeof(struct proc_desc));
    }

    while ((cdesc = program->caps) != NULL)
    {
	program->caps = cdesc->next;
	KMEM_FREE(cdesc, sizeof(struct cap_desc));
    }

    if (program->corepath)
	KMEM_FREE(program->corepath, MAXCOREPATHLEN + 1);

    KMEM_FREE(program, sizeof(struct prg_desc));

    MOD_DEC_USE_COUNT;
}

/*
 * rms_prgcreate
 *
 * create a new program description
 */
int rms_prgcreate(int id, uid_t uid, int cpus)
{
    struct prg_desc *program;
    struct proc_desc *pdesc;
    
    DBG(printk("rms_prgcreate :: program %d pid %d uid %d cpus %d\n", id, CURPROC()->p_pid, uid, cpus));
    
    /*
     * parallel programs are created as root by the rmsd as it forks the loader
     */
    if (CURUID())
	return(EACCES);
    
    /*
     * program ids must be unique
     */
    kmutex_lock(&rms_lock);
    program = findProgram(id);
    kmutex_unlock(&rms_lock);
    if (program)
	return(EINVAL);

    /*
     * create a new program description
     */
    KMEM_ALLOC(program, struct prg_desc *, sizeof(struct prg_desc), TRUE);
    if (!program)
	return(ENOMEM);

    program->id = id;
    program->flags = PRG_RUNNING;
    program->ncpus = cpus;
    program->nprocs = 1;
    program->uid = uid;
    program->ncaps = 0;
    program->caps = 0;
    program->corepath = 0;
    program->psid = 0;
    program->start_time = program->sched_time = gettime();
    program->end_time = 0;
    program->accum_atime = 0;
    program->cutime = 0;
    program->cstime = 0;
    program->maxrss = 0;
    program->memint = 0;
    program->majflt = 0;
    program->ebytes = 0;
    program->exfers = 0;

    KMEM_ALLOC(pdesc, struct proc_desc *, sizeof(struct proc_desc), TRUE);
    if (!pdesc)
    {
	KMEM_FREE (program,sizeof (struct prg_desc));
	return(ENOMEM);
    }

    /* bump the reference count on the caller */
    pdesc->task = ptrack_enabled ? CURPROC() : NULL;

    pdesc->next = 0;
    pdesc->mycap = ELAN_CAP_UNINITIALISED;
    pdesc->myctx = ELAN_CAP_UNINITIALISED;
    pdesc->vp = -1;		/* rmsloader */
    pdesc->program = program;
    pdesc->pid = CURPROC()->p_pid;
    program->pdescs = pdesc;
    
    rms_create_proc_entry(program);
    
    kmutex_lock(&rms_lock);

#ifndef NO_PTRACK

    if (ptrack_enabled)
    {
	DBG(printk("rms_prgcreate :: ptrack enabled - fork callback\n"));
	if (ptrack_register (rms_ptrack_callback, NULL) != 0)
	{
	    kmutex_unlock(&rms_lock);
	    KMEM_FREE(pdesc,sizeof(struct proc_desc));
	    KMEM_FREE(program,sizeof(struct prg_desc));
	    return(ENOMEM);
	}
    }
    else
    {
	DBG(printk("rms_prgcreate :: ptrack disabled - no fork callback\n"));
    }
#endif

    program->next = programs;
    programs = program;
    
    MOD_INC_USE_COUNT;
    
    kmutex_unlock(&rms_lock);
    return(ESUCCESS);
}


/*
 * can't rely on put_task_struct being exported
 * so we need to make sure that a proc is valid
 * before extracting the stats
 */

int getProcessStats(struct proc_desc *pdesc)
{
    struct task_struct *task = 0;
    
    if (ptrack_enabled)
	task = pdesc->task;
    else
    {
	read_lock(&tasklist_lock);
	task = find_task_by_pid(pdesc->pid);
    }
    
    if (task)
    {
#ifdef PROCESS_ACCT
	pdesc->utime = TIMEVAL_TO_MSEC(&task->utime);
	pdesc->stime = TIMEVAL_TO_MSEC(&task->stime);      
	
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)                           
	pdesc->utime = CT_TO_MSEC(task->times.tms_utime);
	pdesc->stime = CT_TO_MSEC(task->times.tms_stime);
	
#else
	pdesc->utime = CT_TO_MSEC(task->utime);
	pdesc->stime = CT_TO_MSEC(task->stime);
#endif
	pdesc->majflt = task->maj_flt;

	/*
	 * the ptrack exit callbacks occur before exit_mm
	 * but if ptrack is not present we can get called
	 * with task->mm = 0
	 */
	pdesc->maxrss = PROC_RSS(task) >> (20 - PAGE_SHIFT);
    }
    
    if (!ptrack_enabled)
	read_unlock(&tasklist_lock);
    
    return(task ? 0 : -1);
}




/*
 * remove a process descriptor keeping track of the
 * accumulated resource usage
 */

static void removeProcDesc(struct prg_desc *program, struct proc_desc *pdesc)
{
#ifndef NO_PTRACK
    struct proc_desc *p;
#endif
    int maxrss;

    /*
     * keep track of the resources used by processes that have 
     * exited, if ptrack is enabled then we will be called 
     * as the process exists, otherwise we will have the last
     * sample
     */
    getProcessStats(pdesc);
    
    program->cutime += pdesc->utime;
    program->cstime += pdesc->stime;
    program->majflt += pdesc->majflt;
    maxrss = pdesc->maxrss;

    /*
     * process specific shared memory cleanup requires the shm_cleanup
     * patch, otherwise the run time system is left to do the job with
     * a blunt axe 
     */
#ifndef NO_SHM_CLEANUP
    shm_cleanup();
#endif
    
    /* 
     * remove process from program 
     */
    KMEM_FREE(pdesc, sizeof(struct proc_desc));
    program->nprocs--;
    
    /*
     * update the memory high water mark for the program
     *
     * safe to access the task structures if we have incremented
     * their reference counts as they were added to the program 
     * however, the mm can be zero
     */
#ifndef NO_PTRACK
    for (p = program->pdescs; p; p = p->next)
	maxrss += PROC_RSS(p->task) >> (20 - PAGE_SHIFT);
    
    if (maxrss > program->maxrss)
	program->maxrss = maxrss;
#endif
    
    DBG(printk("rms_removproc :: program %d procs %d cutime %lld cstime %lld mem %d\n", 
	       program->id, program->nprocs, 
	       (long long)program->cutime, (long long)program->cstime, 
	       program->maxrss));
    
    /*
     * final update to the program if this is the last process
     */
    if (program->nprocs == 0)
    {
	program->end_time = gettime();
	program->flags &= ~PRG_RUNNING;
	program->accum_atime += program->ncpus * (program->end_time - program->sched_time);
	DBG(printk("rms_removproc :: last process has gone\n"));
    }
}


/*
 * rms_prgdestroy
 *
 * destroy a program description
 */
int rms_prgdestroy(int id)
{
    struct prg_desc *program, **pp;
    struct proc_desc *pdesc;

    int status = ESRCH;

    /*
     * parallel programs are created and destroyed by the rmsd
     */
    if (CURUID())
	return(EACCES);

    kmutex_lock(&rms_lock);
    
    pp = &programs;
    while ((program = *pp) != NULL)
    {
	if (program->id == id)
	{
	    /*
	     * with ptrack disabled there won't be any exit callbacks
	     */
	    if (!ptrack_enabled)
	    {
		while ((pdesc = program->pdescs))
		{
		    program->pdescs = pdesc->next;
		    removeProcDesc(program, pdesc);
		}
	    }

	    if (program->nprocs == 0)
	    {
		DBG(printk("rms_prgdestro :: removing program %d\n", program->id));
		*pp = program->next;
		freeProgram(program);
		status = ESUCCESS;
	    }
	    else
	    {
		DBG(printk("rms_prgdestro :: failed to remove program %d: %d\n", program->id, program->nprocs));
		status = ECHILD;
		pp = &program->next;
	    }
	}
	else
	    pp = &program->next;
    }
    
    kmutex_unlock(&rms_lock);
    return(status);
}

/*
 * rms_prgids
 */
int rms_prgids(int maxids, int *prgids, int *nprgs)
{
    struct prg_desc *program;
    int count = 0, *buf, *bufp;
    int status = ESUCCESS;

    if (maxids < 1)
        return(EINVAL);

    kmutex_lock(&rms_lock);

    for (program = programs; program; program = program->next)
        count++;
    count = MIN(count, maxids);

    if (count > 0)
    {
        KMEM_ALLOC(buf, int *, count * sizeof(int), TRUE);
 	if (buf)
	{		   
	    for (program = programs, bufp=buf; bufp < buf + count; 
	         program = program->next)
	        *bufp++ = program->id;
	
	    if (copyout(buf, prgids, sizeof(int) * count))
	        status = EFAULT;

	    KMEM_FREE(buf, count * sizeof(int));
   	}
	else
	    status = ENOMEM;
    }
    
    if (copyout(&count, nprgs, sizeof(int)))
	status = EFAULT;

    kmutex_unlock(&rms_lock);
    
    return(status);
}

/*
 * rms_prginfo
 */
int rms_prginfo(int id, int maxpids, pid_t *pids, int *nprocs)
{
    struct prg_desc *program;
    struct proc_desc *pdesc;
    pid_t *pidp, *buf;
    int status = ESUCCESS;

    kmutex_lock(&rms_lock);

    if ((program = findProgram(id)) != NULL)
    {
	if (program->nprocs > 0)
	{
	    KMEM_ALLOC(buf, pid_t *, program->nprocs * sizeof(pid_t), TRUE);
	    if (buf)
	    {
		for (pidp = buf, pdesc = program->pdescs; pdesc; pdesc = pdesc->next)
		    *pidp++ = pdesc->pid;
		
		if (copyout(buf, pids, sizeof(pid_t) * MIN(program->nprocs, maxpids)))
		    status = EFAULT;
		
		KMEM_FREE(buf, program->nprocs * sizeof(pid_t));
	    }
	    else
		status = ENOMEM;
	}
	
	if (copyout(&program->nprocs, nprocs, sizeof(int)))
	    status = EFAULT;
    }
    else
	status = ESRCH;

    kmutex_unlock(&rms_lock);
    
    return(status);
}

/*
 * Deliver a signal to all the processes in a program
 */
static void prgsignal(struct prg_desc *program, int signo)
{
    struct proc_desc *pdesc;
    DBG(printk("rms_prgsignal :: program %d signal %d\n", program->id, signo));
    for (pdesc = program->pdescs; pdesc; pdesc = pdesc->next)
	kill_proc(pdesc->pid, signo, 1);
}

int rms_prgsignal(int id, int signo)
{
    struct prg_desc *program;
    int status = ESUCCESS;
    
    kmutex_lock(&rms_lock);
    
    if ((program = findProgram(id)) != NULL)
    {
	if (CURUID() == 0 || CURUID() == program->uid)
	{
	    prgsignal(program, signo);
	    if (signo == SIGKILL)
		program->flags |= PRG_KILLED;
	}
	else
	    status = EACCES;
    }
    else
	status = ESRCH;
    
    kmutex_unlock(&rms_lock);
    
    return(status);
}

int rms_prgaddcap(int id, int index, ELAN_CAPABILITY *cap)
{
    struct prg_desc *program;
    struct cap_desc *cdesc;
    int status = ESUCCESS;

    if (cap == NULL)
        return(EINVAL);

    kmutex_lock(&rms_lock);
    if ((program = findProgram(id)) != NULL)
    {
	KMEM_ALLOC(cdesc, struct cap_desc *, sizeof(struct cap_desc), TRUE);
	if (cdesc)
	{
	    cdesc->index = index;
	    if (copyin(cap, &cdesc->cap, sizeof(ELAN_CAPABILITY)))
	    {
		KMEM_FREE(cdesc, sizeof(struct cap_desc));
		status = EFAULT;
	    }
	    else
	    {
		DBG(printk("rms_prgaddcap :: program %d index %d context %d<-->%d\n",
			   program->id, index, cdesc->cap.cap_lowcontext, cdesc->cap.cap_highcontext));
		cdesc->next = program->caps;
		program->caps = cdesc;
		program->ncaps++;
	    }
	}
	else
	    status = ENOMEM;
    }
    else
	status = ESRCH;

    kmutex_unlock(&rms_lock);
    return(status);
}

static uint64_t gettime(void)
{
    uint64_t now;

    struct timeval tv;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,17)
    get_fast_time(&tv);
#else
    do_gettimeofday(&tv);
#endif
    now = tv.tv_sec * 1000 + tv.tv_usec / 1000;
    return(now);
}


/*
 * new stats collection interface, 64-bit with addition of Elan stats
 */
int rms_prggetstats(int id, prgstats_t *stats)
{
    struct prg_desc *program = 0;
    struct proc_desc *pdesc;

    int maxrss, status = ESUCCESS;
    prgstats_t totals;
    uint64_t now = gettime();

    kmutex_lock(&rms_lock);
    
    if (id < 0)
    {
	if ((pdesc = findProcess(CURPROC()->p_pid)) != NULL)
	    program = pdesc->program;
    }
    else
	program = findProgram(id);

    if (program)
    {
	if (CURUID() == 0 || CURUID() == program->uid)
	{
	    totals.flags = program->flags;
	    totals.ncpus = program->ncpus;
	    maxrss = 0;
	    
	    if (program->nprocs > 0)
		totals.etime = now - program->start_time;
	    else
		totals.etime = program->end_time - program->start_time;
	    
	    totals.atime = program->accum_atime;
	    if (program->flags & PRG_RUNNING)
		totals.atime += program->ncpus * (now - program->sched_time);
	    
	    totals.utime = program->cutime;
	    totals.stime = program->cstime;
	    totals.pageflts = program->majflt;
	    totals.memint = program->memint;
	    
 	    for (pdesc = program->pdescs; pdesc; pdesc = pdesc->next)
	    {
		getProcessStats(pdesc);
		totals.utime += pdesc->utime;
		totals.stime += pdesc->stime;
		totals.pageflts += pdesc->majflt;
		maxrss += pdesc->maxrss;
	    }
	    
	    if (maxrss > program->maxrss)
		program->maxrss = maxrss;
	    
	    totals.mem = program->maxrss;
	    totals.ebytes = program->ebytes;
	    totals.exfers = program->exfers;
	    
	    DBG(printk("rms_prggetsta :: program %d utime %lld stime %lld mem %d flt %lld\n", 
		       program->id, (long long)totals.utime, (long long)totals.stime, 
		       totals.mem, (long long)totals.pageflts));
	    
	    if (copyout(&totals, stats, sizeof(prgstats_t)))
		status = EFAULT;
	}
	else
	    status = EACCES;
    }
    else
	status = ESRCH;
    
    kmutex_unlock(&rms_lock);
    return(status);
}

int rms_prgsuspend(int id)
{
    struct prg_desc *program;
    int status = ESUCCESS;

    kmutex_lock(&rms_lock);
    
    if ((program = findProgram(id)) != NULL)
    {
	if (CURUID() == 0 || CURUID() == program->uid)
	{
	    program->flags &= ~PRG_RUNNING;
	    program->flags |=  PRG_SUSPEND;
	    program->accum_atime += program->ncpus * (gettime() - program->sched_time);

	    /* suspend/resume just use signals for now */
	    prgsignal(program, SIGSTOP);
	}
	else
	    status = EACCES;
    }
    else
	status = ESRCH;

    kmutex_unlock(&rms_lock);
    return(status);
}

int rms_prgresume(int id)
{
    struct prg_desc *program;
    int status = ESUCCESS;

    kmutex_lock(&rms_lock);
    
    if ((program = findProgram(id)) != NULL)
    {
	if (CURUID() == 0 || CURUID() == program->uid)
	{
	    program->flags &= ~PRG_SUSPEND;
	    program->flags |=  PRG_RUNNING;
	    program->sched_time = gettime();
	    prgsignal(program, SIGCONT);
	}
	else
	    status = EACCES;
    }
    else
	status = ESRCH;

    kmutex_unlock(&rms_lock);
    return(status);
}


int rms_ncaps(int *ncaps)
{
    struct proc_desc *pdesc;
    int status = ESUCCESS;
    
    kmutex_lock(&rms_lock);
    if ((pdesc = findProcess(CURPROC()->p_pid)) != NULL)
    {
	if (copyout(&pdesc->program->ncaps, ncaps, sizeof(int)))
	    status = EFAULT;
    }
    else
	status = ESRCH;

    kmutex_unlock(&rms_lock);
    return(status);
}

int rms_getprgid(pid_t pid, int *id)
{
    struct proc_desc *pdesc;
    int status = ESUCCESS;
    
    if (pid == 0)
	pid = CURPROC()->p_pid;
    
    kmutex_lock(&rms_lock);
    if ((pdesc = findProcess(pid)) != NULL)
    {
	if (copyout(&pdesc->program->id, id, sizeof(int)))
	    status = EFAULT;
    }
    else
	status = ESRCH;

    kmutex_unlock(&rms_lock);
    return(status);
}

int rms_setcap(int index, int ctx)
{
    struct proc_desc *pdesc;
    struct cap_desc *cdesc;
    int status = EINVAL;
    
    DBG(printk("rms_setcap    :: process %d cap %d ctx %d\n",CURPROC()->p_pid,index,ctx));

    kmutex_lock(&rms_lock);
    if ((pdesc = findProcess(CURPROC()->p_pid)) != NULL)
    {
	for (cdesc = pdesc->program->caps; cdesc; cdesc = cdesc->next)
	    if (cdesc->index == index && 0 <= ctx && ctx <= (cdesc->cap.cap_highcontext - cdesc->cap.cap_lowcontext + 1))
	    {
		pdesc->mycap = index;
		pdesc->myctx = cdesc->cap.cap_lowcontext + ctx;
		status = ESUCCESS;
	    }
    }
    else
	status = ESRCH;

    kmutex_unlock(&rms_lock);
    return(status);
}


int rms_mycap(int *index)
{
    struct proc_desc *pdesc;
    int status = ESUCCESS;
    
    DBG(printk("rms_mycap :: process %d\n", CURPROC()->p_pid));
    
    kmutex_lock(&rms_lock);
    if ((pdesc = findProcess(CURPROC()->p_pid)) != NULL)
    {
	DBG(printk("rms_mycap :: found process %d mycap = %d\n", CURPROC()->p_pid, pdesc->mycap));
	if (copyout(&pdesc->mycap, index, sizeof(int)))
	    status = EFAULT;
    }
    else
	status = ESRCH;

    kmutex_unlock(&rms_lock);
    return(status);
}

int rms_getcap(int index, ELAN_CAPABILITY *cap)
{
    struct proc_desc *pdesc;
    struct cap_desc *cdesc;
    int status = ESUCCESS;
    
    kmutex_lock(&rms_lock);
    if ((pdesc = findProcess(CURPROC()->p_pid)) != NULL)
    {
	for (cdesc = pdesc->program->caps; cdesc; cdesc = cdesc->next)
	    if (cdesc->index == index)
		break;
	
	if (cdesc)
	{
	    /* tell each process about its own context */
	    cdesc->cap.cap_mycontext = pdesc->myctx;
	    
	    if (copyout(&cdesc->cap, cap, ELAN_CAP_SIZE(&cdesc->cap)))
		status = EFAULT;
	    
	    DBG(printk("rms_getcap    :: program %d index %d context %d<-->%d\n", pdesc->program->id, 
		       cdesc->index, cdesc->cap.cap_lowcontext, cdesc->cap.cap_highcontext));
	}
	else
	    status = EINVAL;
    }
    else
	status = ESRCH;
    
    kmutex_unlock(&rms_lock);
    return(status);
}

static void
addProcDesc (struct proc_desc *pdesc,
	     struct prg_desc *program, 
	     struct proc_desc *parent, 
	     int               pid)
{
	pdesc->mycap = (parent ? parent->mycap : ELAN_CAP_UNINITIALISED);
	pdesc->myctx = (parent ? parent->myctx : ELAN_CAP_UNINITIALISED);
	pdesc->program = program;
	pdesc->vp = -1;	             /* assigned by elaninitdone */
	pdesc->pid = pid;

	pdesc->next = program->pdescs;
	program->pdescs = pdesc;
	program->nprocs++;
}

static int
rms_fork_callback (struct task_struct *curproc, struct task_struct *child)
{
    struct prg_desc *program;
    struct proc_desc *parent;
    struct proc_desc *pdesc = NULL;

    kmutex_lock(&rms_lock);
    
    DBG(printk("rms_fork_func :: phase is fork pid %d child %d\n", curproc->p_pid, child->p_pid));

    /*
     * find the process that forked
     */
    if ((parent = findProcess(curproc->p_pid)) != NULL)
    {
	program = parent->program;
	
	DBG(printk("rms_fork_func :: program is %d flags %d\n", program->id, program->flags));
	
	/*
	 * processes can be blocked in fork while prgsignal is in progress
	 * so check to see if the PRG_KILLED flag is set
	 */
	if (program->flags & PRG_KILLED)
	    DBG(printk("rms_fork_func :: fork handler called after program killed\n"));
	else
	{
	    /*
	     * create a new process description and add to program
	     */
	    KMEM_ALLOC(pdesc, struct proc_desc *, sizeof(struct proc_desc), TRUE);
	    if (pdesc == NULL)
		printk("rms_creatproc :: memory allocation failed\n");
	    else
	    {
		addProcDesc (pdesc, program, parent, child->p_pid);
		
		pdesc->task = child;
		
	    }
	}
    }
    else
	DBG(printk("rms_fork_func :: no program\n"));
    
    kmutex_unlock (&rms_lock);
    
    return(pdesc == NULL);
}

static void
rms_exit_callback (struct task_struct *curproc)
{
    struct prg_desc *program;
    struct proc_desc *pdesc, **pdescp;

    kmutex_lock(&rms_lock);
    
    DBG(printk("rms_exit_func :: process %d exiting\n", curproc->p_pid));

    /*
     * find the process that exited and accumulate 
     * resource usage in its parent program
     */
    for (program = programs, pdesc = 0; program && !pdesc; program = program->next)
    {
	pdescp = &program->pdescs;
	while ((pdesc = *pdescp) != NULL)
	{
	    if (pdesc->pid == curproc->p_pid)
	    {
		*pdescp = pdesc->next;
		removeProcDesc(program, pdesc);
		break;
	    }
	    else
		pdescp = &pdesc->next;
	}
    }
    kmutex_unlock  (&rms_lock);
}

#ifndef NO_PTRACK

static int
rms_ptrack_callback (void *arg, int phase, struct task_struct *child)
{
    switch (phase)
    {
    case PTRACK_PHASE_CLONE:
	if (rms_fork_callback (current, child))
	    return PTRACK_DENIED;
	else
	    return PTRACK_INNHERIT;

    case PTRACK_PHASE_CLONE_FAIL:
	DBG(printk("rms_fork_func :: fork failed pid %d child %d\n", current->p_pid, child->p_pid));
	rms_exit_callback(child);
	break;

    case PTRACK_PHASE_EXIT:
	rms_exit_callback(current);
	break;
    }
    return PTRACK_FINISHED;
}

#endif

/*
 * rms_elaninitdone - mark a process as having successfully completed elan initialisation
 */
int rms_elaninitdone(int vp)
{
    int status = ESUCCESS;
    struct proc_desc *pdesc;
    
    DBG(printk("rms_elaninit  :: process %d vp %d\n", CURPROC()->p_pid, vp));
    
    kmutex_lock(&rms_lock);
    if ((pdesc = findProcess(CURPROC()->p_pid)) != NULL)
	pdesc->vp = vp;
    else
	status = ESRCH;
    kmutex_unlock(&rms_lock);
    return(status);
}


/*
 * rms_prgelanpids - return the ids of processes that have completed elan initialisation
 */
int rms_prgelanpids(int id, int maxpids, int *vps, pid_t *pids, int *npids)
{
    struct prg_desc *program;
    struct proc_desc *pdesc;
    pid_t *pidbuf;
    int status = ESUCCESS, count = 0, *vpbuf;
    
    DBG(printk("rms_elanpids  :: process %d id %d\n", CURPROC()->p_pid, id));
    
    kmutex_lock(&rms_lock);
    
    if ((program = findProgram(id)) != NULL)
    {
	if (program->nprocs > 0)
	{
	    KMEM_ALLOC(pidbuf, pid_t *, program->nprocs * sizeof(pid_t), TRUE);
	    KMEM_ALLOC(vpbuf, int *, program->nprocs * sizeof(int), TRUE);
	    if (pidbuf && vpbuf)
	    {
		for (pdesc = program->pdescs; pdesc; pdesc = pdesc->next)
		    if (pdesc->vp >= 0)
		    {
			pidbuf[count] = pdesc->pid;
			vpbuf[count] = pdesc->vp;
			count++;
		    }
	    
		if (count > 0 && (copyout(pidbuf, pids, sizeof(pid_t) * MIN(count, maxpids)) ||
				  copyout(vpbuf, vps, sizeof(int) * MIN(count, maxpids))))
		    status = EFAULT;
		
		KMEM_FREE(pidbuf, program->nprocs * sizeof(pid_t));
		KMEM_FREE(vpbuf, program->nprocs * sizeof(int));
	    }
	    else
		status = ENOMEM;
	}

	if (copyout(&count, npids, sizeof(int)))
	    status = EFAULT;
    }
    else
	status = ESRCH;

    kmutex_unlock(&rms_lock);
    
    return(status);

}

int rms_setpset(int psid)
{
    struct prg_desc *program;
    struct proc_desc *pdesc;
    int status = ESUCCESS;

    if (CURUID())
	return(EACCES);

    kmutex_lock(&rms_lock);
    
    if ((pdesc = findProcess(CURPROC()->p_pid)) != NULL)
    {
	program = pdesc->program;
	program->psid = psid;
    }
    else
	status = ESRCH;

    kmutex_unlock(&rms_lock);
    return(status);
}


int rms_getpset(int id, int *psid)
{
    struct prg_desc *program;
    int status = ESUCCESS;
    
    kmutex_lock(&rms_lock);
    if ((program = findProgram(id)) != NULL)
    {
	if (copyout(&program->psid, psid, sizeof(int)))
	    status = EFAULT;
    }
    else
	status = ESRCH;
    
    kmutex_unlock(&rms_lock);
    return(status);
}

int
rms_setelanstats(int id, uint64_t ebytes, uint64_t exfers)
{
    struct prg_desc *program;
    int status = ESUCCESS;
    
    DBG(printk("rms_setelanst :: process %d id %d\n", CURPROC()->p_pid, id));

    kmutex_lock(&rms_lock);
    if ((program = findProgram(id)) != NULL)
    {
	if (CURUID() == 0 || CURUID() == program->uid)
	{
	    program->ebytes = ebytes;
	    program->exfers = exfers;
	}
	else
	    status = EACCES;
    }
    else
	status = ESRCH;
    
    kmutex_unlock(&rms_lock);
    return(status);
}

int
rms_modversion(void)
{
    return(RMS_MODVERSION);
}

int
rms_addproc(int id, pid_t pid)
{
    struct prg_desc *program;
    struct task_struct *task;
    struct proc_desc *parent;
    struct proc_desc *pdesc;
    int status;
    
    DBG(printk("rms_addproc   :: program %d proc %d pid %d\n", id, CURPROC()->p_pid, pid));
    
    kmutex_lock(&rms_lock);
    if ((program = findProgram(id)) != NULL)
    {
	if (CURUID() == 0 || CURUID() == program->uid)
	{
	    if (findProcess(pid))
		status = ESRCH;
	    else
	    {
		KMEM_ALLOC(pdesc, struct proc_desc *, sizeof(struct proc_desc), TRUE);

		if (pdesc == NULL)
		    status = ENOMEM;
		else
		{
		    read_lock(&tasklist_lock);

		    if ((task = find_task_by_pid(pid)) == NULL)
			status = ESRCH;
		    else
		    {
#ifdef NO_NPTL
			pid_t ppid = task->p_pptr->pid;
#else
			pid_t ppid = task->parent->pid;
#endif
			for (parent = program->pdescs; parent; parent = parent->next)
			    if (parent->pid == ppid)
				break;
		    
			addProcDesc (pdesc, program, parent, pid);
			status = ESUCCESS;
		    }

		    read_unlock (&tasklist_lock);
		    
		    if (status != ESUCCESS)
			KMEM_FREE (pdesc, sizeof (struct proc_desc));
		}
	    }
	}
	else
	    status = EACCES;
    }
    else
	status = ESRCH;
    
    kmutex_unlock(&rms_lock);
    return(status);
}


int
rms_removeproc(int id, pid_t pid)
{
    struct prg_desc *program;
    struct proc_desc *pdesc, **pdescp;
    int status;
    
    DBG(printk("rms_removproc :: program %d proc %d pid %d\n", id, CURPROC()->p_pid, pid));

    kmutex_lock(&rms_lock);
    if ((program = findProgram(id)) != NULL)
    {
	if (CURUID() == 0 || CURUID() == program->uid)
	{
	    status = ESRCH;
	    pdescp = &program->pdescs;
	    while ((pdesc = *pdescp) != NULL)
	    {
		if (pdesc->pid == pid)
		{
		    
		    *pdescp = pdesc->next;
		    removeProcDesc(program, pdesc);
		    status = ESUCCESS;
		    break;
		}
		else
		    pdescp = &pdesc->next;
	    }
	}
	else
	    status = EACCES;
    }
    else
	status = ESRCH;
    
    kmutex_unlock(&rms_lock);
    return(status);
}



/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */







