/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __ELAN3_DMA_H
#define __ELAN3_DMA_H

#ident "$Id: dma.h,v 1.38 2002/08/21 12:43:27 david Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/elan3/dma.h,v $ */

#include <elan3/e3types.h>
#include <elan3/events.h>

/* Alignment for a DMA descriptor */
#define E3_DMA_ALIGN		(32)

/* The maximum size a DMA can be (i.e. < 2GB) */
#define E3_MAX_DMA_SIZE		0x7fffffff

/* This macro returns TRUE if a fixup for the ELAN_REVB_BUG_2 problem is required 
 * i.e. if the DMA begins in the last 64-bytes of a page and its size causes it to enter the
 * next page, hence causing the Elan to issue 2 (64-byte) block reads to different pages.
 * See GNAT hw-elan3/3263
 */
#define E3_DMA_REVB_BUG_2(SIZE, ADDR, PAGESIZE)	\
	( (((int) (ADDR) & (PAGESIZE-64)) == (PAGESIZE-64)) && (-(((int) (ADDR) | ~(PAGESIZE-1))) < (SIZE)) )

/* There is a point where a dma runs quicker from main memory than
 * when running from sdram and having to copy all the data down
 * first.
 */
#define E3_DMA_SDRAM_CUTOFF	128

typedef union _e3_DmaType
{
    E3_uint32 type;
    struct
    {
#if defined(__LITTLE_ENDIAN__)
	E3_uint32 dataType:2;	/* Bits 0 to 1   */
	E3_uint32 direction:3;	/* Bit  4 to 2   */
	E3_uint32 opCode:4;	/* Bits 5 to 8   */
	E3_uint32 failCount:6;	/* Bits 9 to 14  */
	E3_uint32 isRemote:1;	/* Bit  15       */
	E3_uint32 Context:13;	/* Bits 16 to 28 */
	E3_uint32 :3;		/* Bits 29 to 31 */
#else
	E3_uint32 :3;		/* Bits 29 to 31 */
	E3_uint32 Context:13;	/* Bits 16 to 28 */
	E3_uint32 isRemote:1;	/* Bit  15       */
	E3_uint32 failCount:6;	/* Bits 9 to 14  */
	E3_uint32 opCode:4;	/* Bits 5 to 8   */
	E3_uint32 direction:3;	/* Bit  4 to 2   */
	E3_uint32 dataType:2;	/* Bits 0 to 1   */
#endif
    } s;
} E3_DmaType;

#define E3_DMA_CONTEXT_MASK	(ALL_CONTEXT_BITS << 16)

#define E3_DMA_CONTEXT(type)	(((type) >> 16) & ALL_CONTEXT_BITS)
#define E3_DMA_ISREMOTE(type)	(((type) >> 15) & 1)
#define E3_DMA_FAILCOUNT(type)	(((type) >> 9) & 0x3F)
#define E3_DMA_OPCODE(type)	(((type) >> 5) & 0xF)
#define E3_DMA_DIRECTION(type)	(((type) >> 2) & 0x7)
#define EP_DMA_DATATYPE(type)	(((type) >> 0) & 0x3)

#define E3_DMA_TYPE(dataType, direction, opCode, failCount) \
    (((dataType) & 0x3) | (((direction) & 7) << 2) | (((opCode) & 0xF) << 5) | (((failCount) & 0x3F) << 9))


typedef union _e3_CookieVProc
{
    E3_uint32 cookie_vproc;
    struct
    {
#if defined(__LITTLE_ENDIAN__)
	E3_uint32 vproc:16;			/* Bit  15 to 0  */
	E3_uint32 cookie:16;			/* Bits 31 to 16 */
#else
	E3_uint32 cookie:16;			/* Bits 31 to 16 */
	E3_uint32 vproc:16;			/* Bit  15 to 0  */
#endif
    } s;
} E3_CookieVProc;

#define E3_DMA_COOKIE_PROC(Cookie, VProc)  (((VProc) & 0xffff) | (((Cookie) << 16)))

#define DMA_COOKIE_MASK		(0xffff0000)
#define DMA_PROCESS_MASK	(0x0000ffff)

/* We use the bottom bit of the cookie to
 * distinguish main/thread generated cookies
 */
#define DMA_COOKIE_THREAD	(0x01 << 16)

/* We use the next bit of the cookie to
 * distinguish locally/remotely generated cookies 
 */
#define DMA_COOKIE_REMOTE	(0x02 << 16)

/* Assign and increment cookie (NB: we have reserved the bottom two bits)
 */
#define DMA_COOKIE(COOKIE, VPROC) 	 ((((COOKIE) += (0x4 << 16)) & DMA_COOKIE_MASK) | VPROC)
#define DMA_REMOTE_COOKIE(COOKIE, VPROC) ((((COOKIE) += (0x4 << 16)) & DMA_COOKIE_MASK) | DMA_COOKIE_REMOTE | VPROC)

#define DMA_COOKIE_REFRESH(COOKIEVP, COOKIE)				\
do { 									\
	COOKIEVP &= ~DMA_COOKIE_MASK;	 	/* Clear cookie */	\
	COOKIEVP |= DMA_COOKIE(COOKIE,0);	/* Assign new cookie */	\
} while (0)

typedef struct e3_dma
{
    E3_DmaType		dma_u;
    E3_uint32		dma_size;
    E3_Addr		dma_source;
    E3_Addr		dma_dest;
    E3_Addr		dma_destEvent;
    E3_CookieVProc	dma_destCookieProc;
    E3_Addr		dma_srcEvent;
    E3_CookieVProc	dma_srcCookieProc;
} E3_DMA;


/*
 * Word-swapped version of DMA descriptor.
 * This is used by the UltraSPARC code to format the descriptor
 * in main memory before block-copying it down to Elan SDRAM.
 * In the process it does a dword (64-bit) conversion and so swaps
 * the word order on a double-word pair basis
 */
typedef struct e3_dma_swapped
{
    E3_uint32		dma_size;
    E3_DmaType		dma_u;
    E3_Addr		dma_dest;
    E3_Addr		dma_source;
    E3_CookieVProc	dma_destCookieProc;
    E3_Addr		dma_destEvent;
    E3_CookieVProc	dma_srcCookieProc;
    E3_Addr		dma_srcEvent;
} E3_DMA_SWAPPED;

/* Define a Main memory structure for DMA desc based on Endianess of machine */
#if defined(__LITTLE_ENDIAN__)
#define E3_DMA_MAIN E3_DMA
#else
#define E3_DMA_MAIN E3_DMA_SWAPPED;
#endif

#define dma_type	 dma_u.type
#define dma_failCount    dma_u.s.failCount
#define dma_isRemote     dma_u.s.isRemote
#define dma_opCode       dma_u.s.opCode
#define dma_direction    dma_u.s.direction
#define dma_dataType     dma_u.s.dataType
#define dma_queueContext dma_u.s.Context

#define dma_destCookieVProc   dma_destCookieProc.cookie_vproc
#define dma_destVProc	      dma_destCookieProc.s.vproc
#define dma_destCookie	      dma_destCookieProc.s.cookie
#define dma_srcCookieVProc    dma_srcCookieProc.cookie_vproc
#define dma_srcVProc	      dma_srcCookieProc.s.vproc
#define dma_srcCookie	      dma_srcCookieProc.s.cookie

/*
 * Values for dma_opCode
 */
#define DMA_NORMAL			0
#define DMA_QUEUED			1
#define DMA_NORMAL_BROADCAST		2
#define DMA_QUEUED_BROADCAST		3
#define DMA_NORMAL_UNSAFE		4
#define DMA_QUEUED_UNSAFE		5
#define DMA_NORMAL_BROADCAST_UNSAFE	6
#define DMA_QUEUED_BROADCAST_UNSAFE	7

/*
 * Values for dma_direction
 */
#define DMA_WRITE		0
#define DMA_READ_REQUEUE	1
#define DMA_READ		3
#define DMA_READ_BROADCAST	7

/*
 * Values for dma_dataType
 */
#define DMA_BYTE		0
#define DMA_HALFWORD		1
#define DMA_WORD		2
#define DMA_DOUBLE		3

/* OUT OF DATE ?
  #define DMA_OPCODE_SHIFT	3
  #define DMA_FAILCOUNT_SHIFT	9
*/
#define DMA_TYPE_ISREMOTE	(1 << 15)
#define DMA_TYPE_READ		(3 << 2)
#define DMA_TYPE_READ_REQUEUE	(1 << 2)
#define DMA_TYPE_DIRECTION_MASK	(3 << 2)

#endif /* __ELAN3_DMA_H */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
