/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef _ELAN3_SPINLOCK_
#define _ELAN3_SPINLOCK_

#ident "$Id: spinlock.h,v 1.31 2003/09/24 13:57:24 david Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/elan3/spinlock.h,v $*/

/*
 * This spinlock is designed for main/elan processor interactions.
 * The lock is split over Elan/Main memory in such a way that
 * we don't end up busy-polling over the PCI.
 * In the Elan memory we have two words; one is a sequence number
 * and the other is a lock word for main.
 * In main memory we have a copy of the sequence number which main polls when it is
 * waiting for the Elan to drop the lock. Main polls this word until it becomes
 * equal to the sequence number it sampled.
 * The Elan drops the lock by writing the current sequence number to main memory.
 * It is coded to always give priority to the Elan thread, and so when both go for the
 * lock, main will back off first.
 *
 * 18/3/98
 * This has been extended to avoid a starvation case where both the main and thread claim the
 * lock and so both backoff (thread does a break). So now, main attempts to claim the
 * lock by writing 'mainLock' then samples the 'sl_seq' and if it has the lock
 * it sets 'mainGotLock'. The thread will now see the 'sl_mainLock' set, but will only
 * backoff with a c_break_busywait() if 'mainGotLock' is set too.
 */
typedef struct elan3_spinlock_elan {
    union {
	volatile E3_uint64	mainLocks;		/* main writes this dble word */
	struct {
	    volatile E3_uint32	mainLock;		/* main wants a lock */
	    volatile E3_uint32	mainGotLock;		/* main has the lock */
	} s;
    } sl_u;
    volatile E3_uint32		sl_seq;			/* thread owns this word */
    volatile E3_uint32       	sl_mainWait;		/* performance counter */
    volatile E3_uint32		sl_elanWait;		/* performance counter */
    volatile E3_uint32		sl_elanBusyWait;	/* performance counter */
    /* NOTE: The lock/seq words must be within the same 32-byte Elan cache-line */
    E3_uint64                   sl_pad[5];		/* pad to 64-bytes */
} ELAN3_SPINLOCK_ELAN;

#define sl_mainLocks sl_u.mainLocks
#define sl_mainLock  sl_u.s.mainLock
#define sl_mainGotLock sl_u.s.mainGotLock

#define SL_MAIN_RECESSIVE	1
#define SL_MAIN_DOMINANT	2

/* Declare this as a main memory cache block for efficiency */
typedef union elan3_spinlock_main {
    volatile E3_uint32		sl_seq;			/* copy of seq number updated by Elan */
    volatile E3_uint32		sl_Int32[E3_BLK_SIZE/sizeof (E3_uint32)];
} ELAN3_SPINLOCK_MAIN;

/* Main/Main or Elan/Elan lock word */
typedef volatile int	ELAN3_SPINLOCK;

#ifdef __ELAN3__

/* Main/Elan interlock */

#define ELAN3_ME_SPINENTER(SLE,SL) do {\
      			asm volatile ("! elan3_spinlock store barrier");\
			(SLE)->sl_seq++; \
			if ((SLE)->sl_mainLock) \
			  elan3_me_spinblock(SLE, SL);\
			asm volatile ("! elan3_spinlock store barrier");\
		} while (0)
#define ELAN3_ME_SPINEXIT(SLE,SL) do {\
			asm volatile ("! elan3_spinlock store barrier");\
			(SL)->sl_seq = (SLE)->sl_seq;\
			asm volatile ("! elan3_spinlock store barrier");\
		} while (0)


/* Elan/Elan interlock */
#define ELAN3_SPINENTER(L)	do {\
			   asm volatile ("! store barrier");\
			   if (c_swap ((L), 1)) elan3_spinenter(L);\
			   asm volatile ("! store barrier");\
			} while (0)
#define ELAN3_SPINEXIT(L)	do {\
			   asm volatile ("! store barrier");\
			   c_swap((L), 0);\
			   asm volatile ("! store barrier");\
			} while (0)

extern void elan3_me_spinblock (ELAN3_SPINLOCK_ELAN *sle, ELAN3_SPINLOCK_MAIN *sl);
extern void elan3_spinenter (ELAN3_SPINLOCK *l);

#else			   

/* Main/Elan interlock */
#ifdef DEBUG
#define ELAN3_ME_SPINENTER(SDRAM,SLE,SL) do {\
		        register E3_int32 maxLoops = 0x7fffffff;	\
			register E3_uint32 seq;\
			elan3_write32_sdram(SDRAM, (SLE) + offsetof(ELAN3_SPINLOCK_ELAN, sl_mainLock), SL_MAIN_RECESSIVE); \
			MEMBAR_STORELOAD(); \
			seq = elan3_read32_sdram(SDRAM, (SLE) + offsetof(ELAN3_SPINLOCK_ELAN, sl_seq)); \
			while (seq != (SL)->sl_seq) {\
			    elan3_write32_sdram(SDRAM, (SLE) + offsetof(ELAN3_SPINLOCK_ELAN, sl_mainLock), 0); \
			    while ((SL)->sl_seq == (seq-1) && maxLoops--) ; \
			    if (maxLoops < 0) {	\
			        printf("Failed to get ME lock %lx/%lx seq %d sle_seq %d sl_seq %d\n", \
				       SL, SLE, seq, \
				       elan3_read32_sdram(SDRAM, (SLE) + offsetof(ELAN3_SPINLOCK_ELAN, sl_seq)), \
				       (SL)->sl_seq); \
			    } \
			    elan3_write32_sdram(SDRAM, (SLE) + offsetof(ELAN3_SPINLOCK_ELAN, sl_mainLock), SL_MAIN_RECESSIVE); \
			    MEMBAR_STORELOAD(); \
			    seq = elan3_read32_sdram(SDRAM, (SLE) + offsetof(ELAN3_SPINLOCK_ELAN, sl_seq)); \
			}\
			elan3_write32_sdram(SDRAM, (SLE) + offsetof(ELAN3_SPINLOCK_ELAN, sl_mainGotLock), 1); \
			MEMBAR_LOADLOAD();\
		} while (0)
#else
#define ELAN3_ME_SPINENTER(SDRAM,SLE,SL) do {\
			register E3_uint32 seq;\
			elan3_write32_sdram(SDRAM, SLE + offsetof(ELAN3_SPINLOCK_ELAN, sl_mainLock), SL_MAIN_RECESSIVE); \
			MEMBAR_STORELOAD(); \
			seq = elan3_read32_sdram(SDRAM, SLE + offsetof(ELAN3_SPINLOCK_ELAN, sl_seq)); \
			while (seq != (SL)->sl_seq) {\
			    elan3_write32_sdram(SDRAM, SLE + offsetof(ELAN3_SPINLOCK_ELAN, sl_mainLock), 0); \
			    while ((SL)->sl_seq == (seq-1)) ; \
			    elan3_write32_sdram(SDRAM, SLE + offsetof(ELAN3_SPINLOCK_ELAN, sl_mainLock), SL_MAIN_RECESSIVE); \
			    MEMBAR_STORELOAD(); \
			    seq = elan3_read32_sdram(SDRAM, SLE + offsetof(ELAN3_SPINLOCK_ELAN, sl_seq)); \
			}\
			elan3_write32_sdram(SDRAM, SLE + offsetof(ELAN3_SPINLOCK_ELAN, sl_mainGotLock), 1); \
			MEMBAR_LOADLOAD();\
		} while (0)
#endif
#define ELAN3_ME_FORCEENTER(SDRAM,SLE,SL) do { \
	register E3_uint32 seq; \
	MEMBAR_STORELOAD(); \
	elan3_write32_sdram(SDRAM, SLE + offsetof(ELAN3_SPINLOCK_ELAN, sl_mainLock), SL_MAIN_DOMINANT); \
	MEMBAR_STORELOAD(); \
	seq = elan3_read32_sdram(SDRAM, SLE + offsetof(ELAN3_SPINLOCK_ELAN, sl_seq)); \
	while (seq != (SL)->sl_seq) \
	{ \
		/* NOTE: we MUST call elan3_usecspin here for kernel comms */\
		while ((SL)->sl_seq == (seq)-1) \
			elan3_usecspin (1); \
		seq = elan3_read32_sdram(SDRAM, SLE + offsetof(ELAN3_SPINLOCK_ELAN, sl_seq)); \
	} \
	elan3_write32_sdram(SDRAM, SLE + offsetof(ELAN3_SPINLOCK_ELAN, sl_mainGotLock), 1); \
	MEMBAR_LOADLOAD(); \
} while (0)

#define ELAN3_ME_TRYENTER(SDRAM,SLE,SL,SEQ) do { \
    elan3_write32_sdram(SDRAM, SLE + offsetof(ELAN3_SPINLOCK_ELAN, sl_mainLock), SL_MAIN_RECESSIVE); \
    MEMBAR_STORELOAD(); \
    SEQ = elan3_read32_sdram(SDRAM, SLE + offsetof(ELAN3_SPINLOCK_ELAN, sl_seq)); \
} while (0)

#define ELAN3_ME_CHECKENTER(SDRAM,SLE,SL,SEQ) do { \
    if ((SEQ) == ((SL)->sl_seq)) { \
	 elan3_write32_sdram(SDRAM, SLE + offsetof(ELAN3_SPINLOCK_ELAN, sl_mainGotLock), 1); \
   	 MEMBAR_LOADLOAD();\
    } \
    else ELAN3_ME_SPINENTER(SLE,SL); \
} while (0)
	
#define ELAN3_ME_SPINEXIT(SDRAM,SLE,SL) do {\
			MEMBAR_STORESTORE(); \
			elan3_write64_sdram(SDRAM, SLE + offsetof(ELAN3_SPINLOCK_ELAN, sl_mainLocks), 0); \
			MEMBAR_STORESTORE(); \
		} while (0)


/* Main/Main */
#define ELAN3_SPINENTER(L)	do {\
			   while (c_swap ((L), 1)) ; \
			} while (0)
#define ELAN3_SPINEXIT(L)	do {\
			   c_swap((L), 0);\
			} while (0)
#endif /* _ELAN3_ */

#endif /* _ELAN3_SPINLOCK_H */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
