/* -*- linux-c -*-
 *
 *    Copyright (C) 2002-2004 Quadrics Ltd.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 */

/*
 * Callbacks for IO processor page table updates.
 */

#ifndef __LINUX_IOPROC_H__
#define __LINUX_IOPROC_H__

#include <linux/sched.h>
#include <linux/mm.h>

typedef struct ioproc_ops {
	struct ioproc_ops *next;
	void *arg;

	void (*release)(void *arg, struct mm_struct *mm);
	void (*sync_range)(void *arg, struct vm_area_struct *vma, unsigned long start, unsigned long end);
	void (*invalidate_range)(void *arg, struct vm_area_struct *vma, unsigned long start, unsigned long end);
	void (*update_range)(void *arg, struct vm_area_struct *vma, unsigned long start, unsigned long end);

	void (*change_protection)(void *arg, struct vm_area_struct *vma, unsigned long start, unsigned long end, pgprot_t newprot);

	void (*sync_page)(void *arg, struct vm_area_struct *vma, unsigned long address);
	void (*invalidate_page)(void *arg, struct vm_area_struct *vma, unsigned long address);
	void (*update_page)(void *arg, struct vm_area_struct *vma, unsigned long address);

} ioproc_ops_t;

/* IOPROC Registration
 * 
 * Called by the IOPROC device driver to register its interest in page table
 * changes for the process associated with the supplied mm_struct
 *
 * The caller should first allocate and fill out an ioproc_ops structure with 
 * the function pointers initialised to the device driver specific code for
 * each callback. If the device driver doesn't have code for a particular 
 * callback then it should set the function pointer to be NULL.
 * The ioproc_ops arg parameter will be passed unchanged as the first argument
 * to each callback function invocation.
 *
 * The ioproc registration is not inherited across fork() and should be called
 * once for each process that the IOPROC device driver is interested in.
 *
 * Must be called holding the mm->page_table_lock
 */
extern int ioproc_register_ops(struct mm_struct *mm, struct ioproc_ops *ip);


/* IOPROC De-registration
 * 
 * Called by the IOPROC device driver when it is no longer interested in page 
 * table changes for the process associated with the supplied mm_struct
 *
 * Normally this is not needed to be called as the ioproc_release() code will
 * automatically unlink the ioproc_ops struct from the mm_struct as the
 * process exits
 *
 * Must be called holding the mm->page_table_lock
 */
extern int ioproc_unregister_ops(struct mm_struct *mm, struct ioproc_ops *ip);

#ifdef CONFIG_IOPROC

/* IOPROC Release
 *
 * Called during exit_mmap() as all vmas are torn down and unmapped.
 *
 * Also unlinks the ioproc_ops structure from the mm list as it goes.
 *
 * No need for locks as the mm can no longer be accessed at this point
 *
 */
static inline void 
ioproc_release(struct mm_struct *mm)
{
	struct ioproc_ops *cp;

	while ((cp = mm->ioproc_ops) != NULL) {
		mm->ioproc_ops = cp->next;
        
		if (cp->release)
			cp->release(cp->arg, mm);
	}
}

/* IOPROC SYNC RANGE
 *
 * Called when a memory map is synchronised with its disk image i.e. when the 
 * msync() syscall is invoked. Any future read or write to the associated 
 * pages by the IOPROC should cause the page to be marked as referenced or 
 * modified.
 *
 * Called holding the mm->page_table_lock
 */
static inline void 
ioproc_sync_range(struct vm_area_struct *vma, unsigned long start, unsigned long end)
{
	struct ioproc_ops *cp;

	for (cp = vma->vm_mm->ioproc_ops; cp; cp = cp->next)
		if (cp->sync_range)
			cp->sync_range(cp->arg, vma, start, end);
}

/* IOPROC INVALIDATE RANGE
 *
 * Called whenever a valid PTE is unloaded e.g. when a page is unmapped by the
 * user or paged out by the kernel. 
 *
 * After this call the IOPROC must not access the physical memory again unless
 * a new translation is loaded.
 *
 * Called holding the mm->page_table_lock
 */
static inline void 
ioproc_invalidate_range(struct vm_area_struct *vma, unsigned long start, unsigned long end)
{
	struct ioproc_ops *cp;
	
	for (cp = vma->vm_mm->ioproc_ops; cp; cp = cp->next)
		if (cp->invalidate_range)
			cp->invalidate_range(cp->arg, vma, start, end);
}

/* IOPROC UPDATE RANGE
 *
 * Called whenever a valid PTE is loaded e.g. mmaping memory, moving the brk 
 * up, when breaking COW or faulting in an anonymous page of memory.
 *
 * These give the IOPROC device driver the opportunity to load translations 
 * speculatively, which can improve performance by avoiding device translation
 * faults.
 *
 * Called holding the mm->page_table_lock
 */
static inline void 
ioproc_update_range(struct vm_area_struct *vma, unsigned long start, unsigned long end)
{
	struct ioproc_ops *cp;

	for (cp = vma->vm_mm->ioproc_ops; cp; cp = cp->next)
		if (cp->update_range)
			cp->update_range(cp->arg, vma, start, end);
}


/* IOPROC CHANGE PROTECTION
 *
 * Called when the protection on a region of memory is changed i.e. when the 
 * mprotect() syscall is invoked.
 *
 * The IOPROC must not be able to write to a read-only page, so if the 
 * permissions are downgraded then it must honour them. If they are upgraded 
 * it can treat this in the same way as the ioproc_update_[range|sync]() calls
 *
 * Called holding the mm->page_table_lock
 */
static inline void 
ioproc_change_protection(struct vm_area_struct *vma, unsigned long start, unsigned long end, pgprot_t newprot)
{
	struct ioproc_ops *cp;

	for (cp = vma->vm_mm->ioproc_ops; cp; cp = cp->next)
		if (cp->change_protection)
			cp->change_protection(cp->arg, vma, start, end, newprot);
}

/* IOPROC SYNC PAGE
 *
 * Called when a memory map is synchronised with its disk image i.e. when the 
 * msync() syscall is invoked. Any future read or write to the associated page
 * by the IOPROC should cause the page to be marked as referenced or modified.
 *
 * Not currently called as msync() calls ioproc_sync_range() instead
 *
 * Called holding the mm->page_table_lock
 */
static inline void 
ioproc_sync_page(struct vm_area_struct *vma, unsigned long addr)
{
	struct ioproc_ops *cp;

	for (cp = vma->vm_mm->ioproc_ops; cp; cp = cp->next)
		if (cp->sync_page)
			cp->sync_page(cp->arg, vma, addr);
}

/* IOPROC INVALIDATE PAGE
 *
 * Called whenever a valid PTE is unloaded e.g. when a page is unmapped by the
 * user or paged out by the kernel. 
 *
 * After this call the IOPROC must not access the physical memory again unless
 * a new translation is loaded.
 *
 * Called holding the mm->page_table_lock
 */
static inline void 
ioproc_invalidate_page(struct vm_area_struct *vma, unsigned long addr)
{
	struct ioproc_ops *cp;

	for (cp = vma->vm_mm->ioproc_ops; cp; cp = cp->next)
		if (cp->invalidate_page)
			cp->invalidate_page(cp->arg, vma, addr);
}

/* IOPROC UPDATE PAGE
 *
 * Called whenever a valid PTE is loaded e.g. mmaping memory, moving the brk 
 * up, when breaking COW or faulting in an anoymous page of memory.
 *
 * These give the IOPROC device the opportunity to load translations 
 * speculatively, which can improve performance by avoiding device translation
 * faults.
 *
 * Called holding the mm->page_table_lock
 */
static inline void 
ioproc_update_page(struct vm_area_struct *vma, unsigned long addr)
{
	struct ioproc_ops *cp;

	for (cp = vma->vm_mm->ioproc_ops; cp; cp = cp->next)
		if (cp->update_page)
			cp->update_page(cp->arg, vma, addr);
}

#else

/* ! CONFIG_IOPROC so make all hooks empty */

#define ioproc_release(mm)			do { } while (0)

#define ioproc_sync_range(vma,start,end)		do { } while (0)

#define ioproc_invalidate_range(vma, start,end)	do { } while (0)

#define ioproc_update_range(vma, start, end)	do { } while (0)

#define ioproc_change_protection(vma, start, end, prot)	do { } while (0)

#define ioproc_sync_page(vma, addr)		do { } while (0)

#define ioproc_invalidate_page(vma, addr)	do { } while (0)

#define ioproc_update_page(vma, addr)		do { } while (0)

#endif /* CONFIG_IOPROC */

#endif /* __LINUX_IOPROC_H__ */
