/*
 *    Copyright (C) 2000  Regents of the University of California
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#if	!defined(_LINUX_CONDVAR_H)
#define	_LINUX_CONDVAR_H

#if 	defined(__KERNEL__)

#include <linux/list.h>
#include <qsnet/debug.h>

#define CV_RET_SIGPENDING	0
#define CV_RET_TIMEOUT		(-1)
#define CV_RET_NORMAL		1

struct kcondvar_task {
	struct task_struct	*task;		/* need to wrap task in this */
	struct list_head	list;		/*   to thread as a list */
	int 			blocked;
};

typedef struct {
	struct list_head	task_list; 	/* list of kcondvar_task's */
} kcondvar_t;

#define kcondvar_wait(c,l,fl)			debug_kcondvar_wait(c, l, fl, 0,  TASK_UNINTERRUPTIBLE)
#define kcondvar_waitsig(c,l,fl)		debug_kcondvar_wait(c, l, fl, 0,  TASK_INTERRUPTIBLE)
#define kcondvar_timedwait(c,l,fl,to)		debug_kcondvar_wait(c, l, fl, to, TASK_UNINTERRUPTIBLE)
#define kcondvar_timedwaitsig(c,l,fl,to)	debug_kcondvar_wait(c, l, fl, to, TASK_INTERRUPTIBLE)
#define kcondvar_wakeupone(c,l)			kcondvar_wakeup(c, l, 0)
#define kcondvar_wakeupall(c,l)			kcondvar_wakeup(c, l, 1)
 
extern __inline__ void
kcondvar_init(kcondvar_t *c)
{
	INIT_LIST_HEAD(&c->task_list);
}

extern __inline__ void
kcondvar_destroy(kcondvar_t *c)
{
	ASSERT(list_empty(&c->task_list));
}

/*
 * We thread a struct kcondvar_task, allocated on the stack, onto the kcondvar_t's
 * task_list, and take it off again when we wake up.
 */
extern __inline__ int
debug_kcondvar_wait(kcondvar_t *c, spinlock_t *l, unsigned long *fl, long tmo, int state)
{
	struct kcondvar_task cvt;
	int ret = CV_RET_NORMAL;

	ASSERT(!in_interrupt());		/* we can block */
	ASSERT(SPINLOCK_HELD(l));		/* enter holding lock */

	cvt.task = current;
	cvt.blocked = 1;
	list_add(&cvt.list, &c->task_list);
	do {
	       /* Note: we avoid using TASK_UNINTERRUPTIBLE here because avenrun()
		* (linux/kernel/timer.c:calc_load())
		* computation treats it like TASK_RUNNABLE hence creates false high
		* load averages when we create kernel threads.
		* The cvt.blocked flag distinguishes a signal wakeup from a kcondvar_wakeup.
		*
		* However, if we do take a signal we could end up busily spinning here, if
		* we ignore it (state == TASK_UNINTERRUPTIBLE) so once we see a signal
		* pending we do sleep TASK_UNINTERRUPTIBLE to stop a busy spin.
		* I have now blocked all signals for kernel threads to prevent this
		* happening but other users of kcondvar_wait may still hit this spin.
		*/
		set_current_state (signal_pending(current) ? state : TASK_INTERRUPTIBLE);

		if (fl)
		    spin_unlock_irqrestore(l, *fl);
		else
		    spin_unlock(l);
		if (tmo) {
			if (tmo <= jiffies || !schedule_timeout(tmo - jiffies))
				ret = CV_RET_TIMEOUT;
		} else
			schedule();
		if (fl)
		    spin_lock_irqsave (l, *fl);
		else
		    spin_lock(l);
		
		/* signal_pending - Only exit the loop if the user was waiting TASK_INTERRUPTIBLE */
		if ((state == TASK_INTERRUPTIBLE) && signal_pending(current))
			ret = CV_RET_SIGPENDING;

	} while (cvt.blocked && ret == CV_RET_NORMAL);
	list_del(&cvt.list);

	/* Reset task state in case we didn't sleep above */
	set_current_state (TASK_RUNNING);

	return ret;				/* return holding lock */
}

extern __inline__ void
kcondvar_wakeup(kcondvar_t *c, spinlock_t *l, int wakeall)
{
	struct list_head *lp;
	struct kcondvar_task *cvtp;

	ASSERT(SPINLOCK_HELD(l)); 			/* already holding lock */
	for (lp = c->task_list.next; lp != &c->task_list; lp = lp->next) {
		cvtp = list_entry(lp, struct kcondvar_task, list);
		if (cvtp->blocked) {
			cvtp->blocked = 0;
			/* wake_up_process added to kernel/ksyms.c */
			wake_up_process(cvtp->task); 
			if (!wakeall)
				break;
		}
	}
}						/* return still holding lock */


#endif /* __KERNEL__ */
#endif /* _LINUX_CONDVAR_H */
