/* 
 *    Copyright (C) 2000  Regents of the University of California
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 *    Complex - Reader/Writer locks
 *    Ref: "UNIX Systems for Modern Architectures", by Curt Schimmel, 
 *    sec 11.6.3. 
 *
 *    This implementation is based on semaphores and may not be called from 
 *    interrupt handlers.
 *
 */

#if	!defined(_LINUX_RWLOCK_H)
#define	_LINUX_RWLOCK_H

#if	defined(__KERNEL__)

typedef enum { RD, WRT, ANY } crwlock_type_t;

#define crwlock_write_held(l) debug_crwlock_held(l, WRT, __BASE_FILE__,__LINE__)
#define crwlock_read_held(l) debug_crwlock_held(l, RD, __BASE_FILE__, __LINE__)
#define crwlock_held(l)      debug_crwlock_held(l, ANY, __BASE_FILE__, __LINE__)

#define crwlock_read(l)	     debug_crwlock_read(l, __BASE_FILE__, __LINE__)
#define crwlock_write(l)     debug_crwlock_write(l, __BASE_FILE__, __LINE__)
#define crwlock_done(l)      debug_crwlock_done(l, __BASE_FILE__, __LINE__)

#if     defined(DEBUG_RWLOCK) && defined(__alpha__) && !defined(DEBUG_SPINLOCK)
#define DEBUG_SPINLOCK
#endif

#include <linux/spinlock.h>
#include <asm/semaphore.h>
#include <qsnet/debug.h>
#include <qsnet/mutex.h>
#include <linux/version.h>

#if	!defined(DEBUG_SPINLOCK)
#define debug_spin_lock(lock, file, line)       spin_lock(lock)
#endif

typedef struct {
        spinlock_t		m_lock;         /* protects cnt fields below */
        int                     m_rdcnt;        /* # of rdrs in crit section */
        int                     m_wrcnt;        /* # of wrtrs in crit section */
        int                     m_rdwcnt;       /* # of waiting readers */
        int                     m_wrwcnt;       /* # of waiting writers */
        struct semaphore        m_rdwait;       /* sema where readers wait */
        struct semaphore        m_wrwait;       /* sema where writers wait */
        pid_t			m_wrholder;	/* task holding write lock */
} crwlock_t;
 
extern __inline__ void 
crwlock_init(crwlock_t *l)
{
	l->m_lock = SPIN_LOCK_UNLOCKED;
#if	LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	l->m_rdwait = MUTEX_LOCKED;
	l->m_wrwait = MUTEX_LOCKED;
#else
	sema_init(&l->m_rdwait,0);
	sema_init(&l->m_wrwait,0);
#endif
	l->m_rdcnt = l->m_wrcnt = l->m_rdwcnt = l->m_wrwcnt = 0;
	l->m_wrholder = PID_NONE;
}

extern __inline__ void 
crwlock_destroy(crwlock_t *l)
{
	ASSERT(l->m_rdcnt == 0 && l->m_wrcnt == 0);
}

/*
 * If a writer has the lock presently or there are writers waiting,
 * then we have to wait.
 */
extern __inline__ void 
debug_crwlock_read(crwlock_t *l, char *file, int line)
{
	ASSERT(!in_interrupt());
	spin_lock(&l->m_lock);
	if (l->m_wrcnt || l->m_wrwcnt) {
		l->m_rdwcnt++;
		spin_unlock(&l->m_lock);
		down(&l->m_rdwait); /* P */
	} else {
		l->m_rdcnt++;
		spin_unlock(&l->m_lock);
	}
}

/*
 * If we're the last reader, and a writer is waiting,
 * then let the writer go now.
 */
/* private */
extern __inline__ void 
debug_crwlock_read_done(crwlock_t *l, char *file, int line)
{
	spin_lock(&l->m_lock);
	l->m_rdcnt--;
	if (l->m_wrwcnt && l->m_rdcnt == 0) {
		l->m_wrcnt = 1;
		l->m_wrwcnt--;
		spin_unlock(&l->m_lock);
		up(&l->m_wrwait); /* V */	
		return;
	}
	spin_unlock(&l->m_lock);
}

extern __inline__ void 
debug_crwlock_write(crwlock_t *l, char *file, int line)
{
	ASSERT(!in_interrupt());
	spin_lock(&l->m_lock);
	if (l->m_wrcnt || l->m_rdcnt) {		/* block if lock is in use */
		l->m_wrwcnt++;
		spin_unlock(&l->m_lock);
		down(&l->m_wrwait); /* P */
	} else {				/* lock is not in use */
		l->m_wrcnt = 1;
		spin_unlock(&l->m_lock);
	}
	l->m_wrholder = current->pid;
}

/* private */
extern __inline__ void
debug_crwlock_write_done(crwlock_t *l, char *file, int line)
{
	int rdrs;

	spin_lock(&l->m_lock);
	l->m_wrholder = PID_NONE;
	if (l->m_rdwcnt) { 			/* let any readers go first */
		l->m_wrcnt = 0;
		rdrs = l->m_rdwcnt;
		l->m_rdcnt = rdrs;
		l->m_rdwcnt = 0;
		spin_unlock(&l->m_lock);
		while (rdrs--)
			up(&l->m_rdwait); /* V */
	} else if (l->m_wrwcnt) {		/* or let any writer go */
		l->m_wrwcnt--;
		spin_unlock(&l->m_lock);
		up(&l->m_wrwait); /* V */
	} else {				/* nobody waiting, unlock */
		l->m_wrcnt = 0;
		spin_unlock(&l->m_lock);
	}
}

extern __inline__ void
debug_crwlock_done(crwlock_t *l, char *file, int line)
{
	if (l->m_wrholder == current->pid)
		debug_crwlock_write_done(l, file, line);
	else
		debug_crwlock_read_done(l, file, line);
}

/*
 * Return nonzero if lock is held
 */
extern __inline__ int  
debug_crwlock_held(crwlock_t *l, crwlock_type_t t, char *file, int line)
{
	int res;

	spin_lock(&l->m_lock);
	switch(t) {
		case RD:
			res = l->m_rdcnt;
			break;
		case WRT:
			res = l->m_wrcnt;
			break;
		case ANY:
			res = l->m_wrcnt + l->m_rdcnt;
			break;
	}
	spin_unlock(&l->m_lock);

	return res;
}

#endif /* __KERNEL__ */
#endif /* _LINUX_RWLOCK_H */
