/*
 *    Copyright (C) 2000  Regents of the University of California
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Derived from exit_actn.c by
 *    Copyright (C) 2003 Quadrics Ltd.
 */


#include <linux/module.h>
#include <linux/spinlock.h>
#include <linux/sched.h>
#include <linux/ptrack.h>
#include <linux/slab.h>
#include <linux/list.h>

#include <asm/errno.h>

int
ptrack_register (ptrack_callback_t callback, void *arg)
{
       struct ptrack_desc *desc = kmalloc (sizeof (struct ptrack_desc), GFP_KERNEL);
       
       if (desc == NULL)
               return -ENOMEM;

	desc->callback = callback;
	desc->arg      = arg;
       
       list_add_tail (&desc->link, &current->ptrack_list);
       
       return 0;
}

void
ptrack_deregister (ptrack_callback_t callback, void *arg)
{      
       struct list_head *el, *nel;
       
       list_for_each_safe (el, nel, &current->ptrack_list) {
               struct ptrack_desc *desc = list_entry (el, struct ptrack_desc, link);
               
               if (desc->callback == callback && desc->arg == arg) {
                       list_del (&desc->link);
                       kfree (desc);
               }
       }
}

int
ptrack_registered (ptrack_callback_t callback, void *arg)
{
       struct list_head *el;
       
       list_for_each (el, &current->ptrack_list) {
               struct ptrack_desc *desc = list_entry (el, struct ptrack_desc, link);
               
               if (desc->callback == callback && desc->arg == arg)
                       return 1;
       }
       return 0;
}      
        
int
ptrack_call_callbacks (int phase, struct task_struct *child)
{
       struct list_head *el, *nel;
       struct ptrack_desc *new;
	int res;

       if (phase == PTRACK_PHASE_CLONE)
               INIT_LIST_HEAD (&child->ptrack_list);

       list_for_each_safe (el, nel, &current->ptrack_list) {
               struct ptrack_desc *desc = list_entry (el, struct ptrack_desc, link);
               
	       res = desc->callback (desc->arg, phase, child);
               
               switch (phase)
               {
               case PTRACK_PHASE_EXIT:
                       list_del (&desc->link);
                       kfree (desc);
                       break;
                       
               case PTRACK_PHASE_CLONE:
		       switch (res)
		       {
		       case PTRACK_FINISHED:
			       break;

		       case PTRACK_INNHERIT:
			       if ((new = kmalloc (sizeof (struct ptrack_desc), GFP_ATOMIC)) == NULL)
			       {
				       /* allocation failed - notify that this process is not going
					* to be started by signalling clone failure.
					*/
				       desc->callback (desc->arg, PTRACK_PHASE_CLONE_FAIL, child);
				       
				       goto failed;
			       }

			       	new->callback = desc->callback;
				new->arg      = desc->arg;
                               
                               list_add_tail (&new->link, &child->ptrack_list);
			       break;

		       case PTRACK_DENIED:
			       goto failed;
                       }
		       break;
               }
       }

       return 0;

 failed:
       while (! list_empty (&child->ptrack_list))
       {
	       struct ptrack_desc *desc = list_entry (child->ptrack_list.next, struct ptrack_desc, link);
	       
	       desc->callback (desc->arg, PTRACK_PHASE_CLONE_FAIL, child);

	       list_del (&desc->link);
	       kfree (desc);
       }
       return 1;
}
EXPORT_SYMBOL(ptrack_register);
EXPORT_SYMBOL(ptrack_deregister);
EXPORT_SYMBOL(ptrack_registered);
